Require Import committable.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import syntax.
Require Import trails.

Definition coherent ts tr : Prop :=
  forall deps depr ids idr ps pr,
    ts = thread_trail ids ps ->
    tr = thread_trail idr pr ->
    deps = dep ts -> 
    depr = dep tr ->
    ((ids <> idr)
      /\ 
      (forall p, 
        In trail depr (thread_trail ids p) -> extends ps p)
      /\
      (forall p, 
        In trail deps (thread_trail idr p) -> extends pr p)
      /\
      (forall id p1 p2, 
        In trail deps (thread_trail id p1) ->
        In trail depr (thread_trail id p2) ->
        (extends p1 p2 \/ extends p2 p1))
      /\
      (forall hp1 hp2,
        In trail deps (heap_trail hp1) ->
        In trail depr (heap_trail hp2) ->
        (heap_extends hp1 hp2 \/ heap_extends hp2 hp1))
).

Definition heap_coherent t ht : Prop :=
  forall dep_t dep_ht id p hp,
    t = thread_trail id p ->
    ht = heap_trail hp ->
    dep_t = dep t -> 
    dep_ht = dep ht ->
    (forall p', 
      In trail dep_ht (thread_trail id p') -> extends p p')
    /\
    (forall hp', In trail dep_t (heap_trail hp') -> heap_extends hp hp')
    /\
    (forall id1 p1 p2, 
      In trail dep_t (thread_trail id1 p1) ->
      In trail dep_ht (thread_trail id1 p2) ->
      (extends p1 p2 \/ extends p2 p1))
    /\
    (forall hp1 hp2,
      In trail dep_t (heap_trail hp1) ->
      In trail dep_ht (heap_trail hp2) ->
      (heap_extends hp1 hp2 \/ heap_extends hp2 hp1))
.

Lemma committable_implies_coherent : 
  forall c id1 id2 p1 p2, 
    committable c -> 
    id1 <> id2 ->
    In trail c (thread_trail id1 p1) -> 
    In trail c (thread_trail id2 p2) -> 
    coherent (thread_trail id1 p1) (thread_trail id2 p2).

intros c id1 id2 p1 p2 comm id1_ne_id2 t1_in_c t2_in_c.

unfold coherent. 
intros deps depr ids idr ps pr eq1 eq2 dep1 dep2. 
inversion eq1. inversion eq2. clear eq1 eq2.
subst deps depr id1 id2 p1 p2.
destruct comm as [ uniq_heap [ uniq_thread [ parent_thread parent_heap ] ] ].

split. auto.

split. Focus.
intros p in_dep.
assert (exists ps',
  In trail c (thread_trail ids ps') /\
  extends ps' p) as [ ps' [ ps'_in_c ps'_extends ] ].
eapply parent_thread. apply t2_in_c. auto.
assert (ps' = ps). eapply uniq_thread. apply ps'_in_c.
auto. subst ps'. auto.
Unfocus.

split. Focus.
intros p in_dep.
assert (exists pr',
  In trail c (thread_trail idr pr') /\
  extends pr' p) as [ pr' [ pr'_in_c pr'_extends ] ].
eapply parent_thread. apply t1_in_c. auto.
assert (pr' = pr). eapply uniq_thread. apply pr'_in_c.
auto. subst pr'. auto.
Unfocus.

split. Focus.
intros id p1 p2 p1_in_dep p2_in_dep.
assert (exists ps',
  In trail c (thread_trail id ps') /\
  extends ps' p1) as [ ps' [ ps'_in_c ps'_extends ] ].
eapply parent_thread. apply t1_in_c. auto.
assert (exists pr',
  In trail c (thread_trail id pr') /\
  extends pr' p2) as [ pr' [ pr'_in_c pr'_extends ] ].
eapply parent_thread. apply t2_in_c. auto.
assert (ps' = pr'). eapply uniq_thread.
eauto. auto.
eapply common_parent_implies_extends.
eauto. subst pr'. auto.
Unfocus.

intros hp1 hp2 hp1_in_dep hp2_in_dep.
assert (exists hp1',
  In trail c (heap_trail hp1') /\ heap_extends hp1' hp1)
as [ hp1' [ hp1'_in_c hp1'_extends ] ].
eapply parent_heap. apply t1_in_c. auto.
assert (exists hp2',
  In trail c (heap_trail hp2') /\ heap_extends hp2' hp2)
as [ hp2' [ hp2'_in_c hp2'_extends ] ].
eapply parent_heap. apply t2_in_c. auto.
unfold unique in uniq_heap.
destruct uniq_heap as [ hp [ hp_in_c hp_unique ] ].
assert (hp = hp1'). apply hp_unique. auto.
assert (hp = hp2'). apply hp_unique. auto.
subst hp1' hp2'.
eapply common_parent_implies_heap_extends.
eauto. auto.

Qed.

Lemma committable_implies_heap_coherent : 
  forall c id p hp, 
    committable c -> 
    In trail c (thread_trail id p) -> 
    In trail c (heap_trail hp) -> 
    heap_coherent (thread_trail id p) (heap_trail hp).

intros c id1 p1 hp2 comm t1_in_c t2_in_c.

unfold heap_coherent. 
intros dep_t dep_ht id p hp eq1 eq2 dep1 dep2. 
inversion eq1. inversion eq2. clear eq1 eq2.
subst dep_t dep_ht id p hp.
destruct comm as [ uniq_heap [ uniq_thread [ parent_thread parent_heap ] ] ].

split. Focus.
intros p in_dep.
assert (exists p',
  In trail c (thread_trail id1 p') /\
  extends p' p) as [ p' [ p'_in_c p'_extends ] ].
eapply parent_thread. apply t2_in_c. auto.
assert (p' = p1). eapply uniq_thread. apply p'_in_c.
auto. subst p'. auto.
Unfocus.

split. Focus.
intros p in_dep.
assert (exists hp',
  In trail c (heap_trail hp') /\
  heap_extends hp' p) as [ hp' [ hp'_in_c hp'_extends ] ].
eapply parent_heap. apply t1_in_c. auto.
unfold unique in uniq_heap.
destruct uniq_heap as [ hp [ hp_in_c hp_unique ] ].
assert (hp = hp'). apply hp_unique. auto.
assert (hp = hp2). apply hp_unique. auto.
subst hp' hp2. auto.
Unfocus.

split. Focus.
intros id2 p2 p3 p2_in_dep p3_in_dep.
assert (exists ps',
  In trail c (thread_trail id2 ps') /\
  extends ps' p2) as [ ps' [ ps'_in_c ps'_extends ] ].
eapply parent_thread. apply t1_in_c. auto.
assert (exists pr',
  In trail c (thread_trail id2 pr') /\
  extends pr' p3) as [ pr' [ pr'_in_c pr'_extends ] ].
eapply parent_thread. apply t2_in_c. auto.
assert (ps' = pr'). eapply uniq_thread.
eauto. auto.
eapply common_parent_implies_extends.
eauto. subst pr'. auto.
Unfocus.

intros hp3 hp4 hp3_in_dep hp4_in_dep.
assert (exists hp1',
  In trail c (heap_trail hp1') /\ heap_extends hp1' hp3)
as [ hp1' [ hp1'_in_c hp1'_extends ] ].
eapply parent_heap. apply t1_in_c. auto.
assert (exists hp2',
  In trail c (heap_trail hp2') /\ heap_extends hp2' hp4)
as [ hp2' [ hp2'_in_c hp2'_extends ] ].
eapply parent_heap. apply t2_in_c. auto.
unfold unique in uniq_heap.
destruct uniq_heap as [ hp [ hp_in_c hp_unique ] ].
assert (hp = hp1'). apply hp_unique. auto.
assert (hp = hp2'). apply hp_unique. auto.
subst hp1' hp2'.
eapply common_parent_implies_heap_extends.
eauto. auto.

Qed.
