Require Import consistent.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import trails.

Definition committable (c : Ensemble trail) : Prop :=
  (exists! hp, In trail c (heap_trail hp))
  /\
  (forall id p1 p2,
    In trail c (thread_trail id p1) ->
    In trail c (thread_trail id p2) ->
    p1 = p2)
  /\
  (forall t id p,
    In trail c t ->
    In trail (dep t) (thread_trail id p) ->
    exists p', In trail c (thread_trail id p') /\ extends p' p)
  /\
  (forall t hp,
    In trail c t ->
    In trail (dep t) (heap_trail hp) ->
    exists hp', In trail c (heap_trail hp') /\ heap_extends hp' hp).

Lemma committable_implies_consistent :
  forall c t,
    committable c ->
    In trail c t ->
    consistent t.

intros c t comm in_c.
unfold committable in comm.
destruct comm as [ uniq_heap [ unique_thread [ parent_thread parent_heap ] ] ].

unfold consistent.
split.

  (* Subgoal 1: each pair of matching thread trails have a common suffix *)
  clear uniq_heap parent_heap.
  intros id p1 p2 in_dep1 in_dep2.
  
  assert (exists p' : path,
    In trail c (thread_trail id p') /\ extends p' p1) 
  as [ p1' [ in_c_p1' p1'_extends_p1 ] ]; eauto.
  assert (exists p' : path,
    In trail c (thread_trail id p') /\ extends p' p2) 
  as [ p2' [ in_c_p2' p2'_extends_p2 ] ]; eauto.

  assert (p1' = p2'). eauto. subst p2'.
  
  eapply common_parent_implies_extends. eauto. auto.

  (* Subgoal 2: each pair of heap trails have a matching suffix *)
  clear unique_thread parent_thread.

  intros p1 p2 in_dep1 in_dep2.
  assert (exists p' : heap_path,
    In trail c (heap_trail p') /\ heap_extends p' p1)
  as [p1' [ in_c_p1' p1'_extends_p1 ] ]; eauto.
  assert (exists p' : heap_path,
    In trail c (heap_trail p') /\ heap_extends p' p2)
  as [p2' [ in_c_p2' p2'_extends_p2 ] ]; eauto.

  destruct uniq_heap as [ hp uniq_heap ].
  unfold unique in uniq_heap.
  destruct uniq_heap as [ hp_in_c uniq_heap ].

  assert (hp = p1'). eauto. subst p1'.
  assert (hp = p2'). eauto. subst p2'.
  
  eapply common_parent_implies_heap_extends. eauto. auto.

Qed.
