Require Import Coq.Lists.List.
Require Import Coq.Sets.Ensembles.
Require Import trails.
Require Import dep.
Require Import committable.
Require Import syntax.

Lemma committable_after_end_lockdown :
  forall c n id_ld p_ld hp_ld,
    ~ In trail c (thread_trail id_ld p_ld) ->
    ~ In trail c (heap_trail hp_ld) ->
    committable (Union trail c
      (Couple trail
        (thread_trail id_ld p_ld)
        (heap_trail hp_ld))) ->
    committable (Union trail c
      (Couple trail
        (thread_trail id_ld (path_cons (path_heap n hp_ld) p_ld))
        (heap_trail (heap_path_cons (heap_path_thread n id_ld p_ld) hp_ld)))).

intros c n id_ld p_ld hp_ld t_ld_not_in_c ht_ld_not_in_c comm_before.
inversion comm_before as
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
unfold unique in unique_heap.
destruct unique_heap as [ hp [ hp_in_before hp_unique ] ].
assert (forall t hp',
  In trail (Union trail c
    (Couple trail (thread_trail id_ld p_ld) (heap_trail hp_ld))) t ->
  In trail (dep t) (heap_trail hp') ->
  heap_extends hp hp') as hp_parent.
intros t hp' in_before in_dep.
assert (exists hp'' : heap_path,
  In trail (Union trail c
    (Couple trail (thread_trail id_ld p_ld) (heap_trail hp_ld))) 
  (heap_trail hp'') /\ heap_extends hp'' hp')
as [ hp'' [ hp''_in_before hp''_extends ] ].
eapply parent_heap. apply in_before. auto.
assert (hp = hp''). apply hp_unique. auto. subst hp''. auto.
assert (hp = hp_ld). apply hp_unique. 
right. right. subst hp.

split.

(* unique heap *)
Focus.
unfold unique. 
exists (heap_path_cons (heap_path_thread n id_ld p_ld) hp_ld).
split. right. right.
intros hp' in_after.
inversion in_after as [ ht' ht'_in_c | ht' [ | ] ].
assert (hp_ld = hp'). apply hp_unique. left. auto. subst hp'.
tauto.

reflexivity.
Unfocus.

split.

(* unique thread *)
Focus.
intros id p1 p2 t1_in_after t2_in_after.
assert (forall p, ~ In trail c (thread_trail id_ld p)) as id_ld_not_in_c.
intros p in_c. assert (p_ld = p). eapply unique_thread.
right. left. left. auto. subst p. tauto.
inversion t1_in_after as [ t1 t1_in_c | t1 [ | ] ];
inversion t2_in_after as [ t2 t2_in_c | t2 [ | ] ].
eapply unique_thread;
eauto using Union_intror, Union_introl, Couple_l, Couple_r.
subst id. assert (~ In trail c (thread_trail id_ld p1)).
apply id_ld_not_in_c. tauto.
subst id. assert (~ In trail c (thread_trail id_ld p2)).
apply id_ld_not_in_c. tauto.
reflexivity.
Unfocus.

split.

(* parent thread *)
Focus.

intros t id p in_after in_dep.
inversion in_after as [ t' in_c | t' [ | ] ].

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_ld p_ld)
      (heap_trail hp_ld))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. left. apply in_c. auto.

inversion in_before as [ t2 t2_in_c | t2 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_heap n hp_ld) p').
split. right. left. apply extends_cons. auto.

unfold dep in in_dep. simpl in in_dep.
inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 false ] | t2 t2_is_ld ].

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_ld p_ld)
      (heap_trail hp_ld))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. left. left. auto.

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_heap n hp_ld) p').
split. right. left. apply extends_cons. auto.

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_ld p_ld)
      (heap_trail hp_ld))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. right. left. auto.

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_heap n hp_ld) p').
split. right. left. apply extends_cons. auto.

inversion false.

inversion t2_is_ld.
exists (path_cons (path_heap n hp_ld) p_ld).
split. right. left. apply extends_refl.

unfold dep in in_dep. simpl in in_dep.
inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 t2_is_ld ] | t2 false ].

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_ld p_ld)
      (heap_trail hp_ld))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. left. left. auto. 

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_heap n hp_ld) p').
split. right. left. apply extends_cons. auto.

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_ld p_ld)
      (heap_trail hp_ld))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. right. left. auto.

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_heap n hp_ld) p').
split. right. left. apply extends_cons. auto.

inversion t2_is_ld.
exists (path_cons (path_heap n hp_ld) p_ld).
split. right. left. apply extends_refl.

inversion false.

Unfocus.

(* parent heap *)
Focus.

intros t1 hp2 in_after in_dep.
exists (heap_path_cons (heap_path_thread n id_ld p_ld) hp_ld).
split. right. right.

inversion in_after as [ t' in_c | t' [ | ] ].

eapply heap_extends_cons. eapply hp_parent. 
left. apply in_c. auto.

unfold dep in in_dep. simpl in in_dep.

inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 t2_is_ld ] | t2 false ].

eapply heap_extends_cons. eapply hp_parent.
right. left. left. simpl. auto.
eapply heap_extends_cons. eapply hp_parent.
right. right. left. simpl. auto.

inversion t2_is_ld. apply heap_extends_refl.

inversion false.

unfold dep in in_dep. simpl in in_dep.
inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 false ] | t2 t2_is_ld ].

eapply heap_extends_cons. eapply hp_parent.
right. left. left. simpl. auto.
eapply heap_extends_cons. eapply hp_parent.
right. right. left. simpl. auto.

inversion false.

inversion t2_is_ld. apply heap_extends_refl.

Unfocus.

Qed.
