Require Import Coq.Lists.List.
Require Import Coq.Sets.Ensembles.
Require Import trails.
Require Import dep.
Require Import committable.
Require Import syntax.

Lemma committable_after_send_recv :
  forall c n id_s p_s id_r p_r,
    ~ In trail c (thread_trail id_s p_s) ->
    ~ In trail c (thread_trail id_r p_r) ->
    id_s <> id_r ->
    committable (Union trail c
      (Couple trail
        (thread_trail id_s p_s)
        (thread_trail id_r p_r))) ->
    committable (Union trail c
      (Couple trail
        (thread_trail id_s (path_cons (path_send n id_r p_r) p_s))
        (thread_trail id_r (path_cons (path_recv n id_s p_s) p_r)))).

intros c n id_s p_s id_r p_r t_r_not_in_c t_s_not_in_c id_neq comm_before.
inversion comm_before as
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
unfold unique in unique_heap.
destruct unique_heap as [ hp [ hp_in_before hp_unique ] ].
inversion hp_in_before as [ ht ht_in_c | ht [ | ] ]; subst ht.
assert (forall t hp',
  In trail (Union trail c
    (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r))) t ->
  In trail (dep t) (heap_trail hp') ->
  heap_extends hp hp') as hp_parent.
intros t hp' in_before in_dep.
assert (exists hp'' : heap_path,
  In trail (Union trail c
    (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r))) 
  (heap_trail hp'') /\ heap_extends hp'' hp')
as [ hp'' [ hp''_in_before hp''_extends ] ].
eapply parent_heap. apply in_before. auto.
assert (hp = hp''). apply hp_unique. auto. subst hp''. auto.

split.

(* unique heap *)
Focus.
unfold unique. exists hp. split. left. auto.
intros hp' in_after.
inversion in_after as [ ht' ht'_in_c | ht' [ | ] ].
apply hp_unique. left. auto.
Unfocus.

split.

(* unique thread *)
Focus.
intros id p1 p2 t1_in_after t2_in_after.
assert (forall p, ~ In trail c (thread_trail id_s p)) as id_s_not_in_c.
intros p in_c. assert (p_s = p). eapply unique_thread.
right. left. left. auto. subst p. tauto.
assert (forall p, ~ In trail c (thread_trail id_r p)) as id_r_not_in_c.
intros p in_c. assert (p_r = p). eapply unique_thread.
right. right. left. auto. subst p. tauto.
inversion t1_in_after as [ t1 t1_in_c | t1 [ | ] ];
inversion t2_in_after as [ t2 t2_in_c | t2 [ | ] ].
eapply unique_thread;
eauto using Union_intror, Union_introl, Couple_l, Couple_r.
subst id. assert (~ In trail c (thread_trail id_s p1)).
apply id_s_not_in_c. tauto.
subst id. assert (~ In trail c (thread_trail id_r p1)).
apply id_r_not_in_c. tauto.
subst id. assert (~ In trail c (thread_trail id_s p2)).
apply id_s_not_in_c. tauto.
subst id. assert (~ In trail c (thread_trail id_r p2)).
apply id_r_not_in_c. tauto.
subst id_s id_r. tauto.
subst id. assert (~ In trail c (thread_trail id_r p2)).
apply id_r_not_in_c. tauto.
subst id_s id_r. tauto.
reflexivity.
Unfocus.

split.

(* parent thread *)
Focus.

intros t id p in_after in_dep.
inversion in_after as [ t' in_c | t' [ | ] ].

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_s p_s)
      (thread_trail id_r p_r))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. left. apply in_c. auto.

inversion in_before as [ t2 t2_in_c | t2 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_send n id_r p_r) p').
split. right. left. apply extends_cons. auto.
exists (path_cons (path_recv n id_s p_s) p').
split. right. right. apply extends_cons. auto.

unfold dep in in_dep. simpl in in_dep.
inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 t2_is_r ] | t2 t2_is_s ].

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_s p_s)
      (thread_trail id_r p_r))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. left. left. auto.

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_send n id_r p_r) p').
split. right. left. apply extends_cons. auto.
exists (path_cons (path_recv n id_s p_s) p').
split. right. right. apply extends_cons. auto.

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_s p_s)
      (thread_trail id_r p_r))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. right. left. auto.

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_send n id_r p_r) p').
split. right. left. apply extends_cons. auto.
exists (path_cons (path_recv n id_s p_s) p').
split. right. right. apply extends_cons. auto.

inversion t2_is_r.
exists (path_cons (path_recv n id_s p_s) p_r).
split. right. right. apply extends_refl.

inversion t2_is_s.
exists (path_cons (path_send n id_r p_r) p_s).
split. right. left. apply extends_refl.

unfold dep in in_dep. simpl in in_dep.
inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 t2_is_s ] | t2 t2_is_r ].

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_s p_s)
      (thread_trail id_r p_r))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. right. left. auto.

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_send n id_r p_r) p').
split. right. left. apply extends_cons. auto.
exists (path_cons (path_recv n id_s p_s) p').
split. right. right. apply extends_cons. auto.

assert (exists p',
  In trail
  (Union trail c
    (Couple trail (thread_trail id_s p_s)
      (thread_trail id_r p_r))) 
  (thread_trail id p') /\ extends p' p)
as [ p' [ in_before p'_extends ] ].
eapply parent_thread. right. left. left. auto.

inversion in_before as [ t3 t3_in_c | t3 [ | ] ].
exists p'. split. left. auto. auto.
exists (path_cons (path_send n id_r p_r) p').
split. right. left. apply extends_cons. auto.
exists (path_cons (path_recv n id_s p_s) p').
split. right. right. apply extends_cons. auto.

inversion t2_is_s.
exists (path_cons (path_send n id_r p_r) p_s).
split. right. left. apply extends_refl.

inversion t2_is_r.
exists (path_cons (path_recv n id_s p_s) p_r).
split. right. right. apply extends_refl.

Unfocus.

(* parent heap *)
Focus.

intros t1 hp2 in_after in_dep.
exists hp. split.
left. auto.

inversion in_after as [ t' in_c | t' [ | ] ].
eapply hp_parent. left. apply in_c. auto.

unfold dep in in_dep. simpl in in_dep.

inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 t2_is_r ] | t2 t2_is_s ].

eapply hp_parent. right. left. left. auto.
eapply hp_parent. right. right. left. auto.
inversion t2_is_r. inversion t2_is_s.

inversion in_dep 
  as [ t2'' [ t2' [ t2 t2_in_DA | t2 t2_in_DA ] | t2 t2_is_s ] | t2 t2_is_r ].

eapply hp_parent. right. right. left. auto.
eapply hp_parent. right. left. left. auto.
inversion t2_is_s. inversion t2_is_r.

Unfocus.

Qed.
