Require Import committable.
Require Import committable_before_choose_right_helper.
Require Import consistent.
Require Import Coq.Logic.Classical.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import trails.

Lemma committable_before_choose_right :
  forall c id p,
    ~ In trail c (thread_trail id (path_cons path_right p)) ->
    committable (Add trail c 
      (thread_trail id (path_cons path_right p))) ->
    committable (Add trail c (thread_trail id p)).

intros c id p t'_not_in_c comm_after.

inversion comm_after as
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
unfold unique in unique_heap.
destruct unique_heap as [ hp [ hp_in_after hp_unique ] ].
assert (forall t hp',
  In trail (Add trail c (thread_trail id (path_cons path_right p))) t ->
  In trail (dep t) (heap_trail hp') ->
  heap_extends hp hp') as hp_parent.
intros t hp' in_after in_dep.
assert (exists hp'',
  In trail (Add trail c (thread_trail id (path_cons path_right p)))
  (heap_trail hp'') /\ heap_extends hp'' hp') as hp''.
eapply parent_heap. apply in_after. auto.
destruct hp'' as [ hp'' [ hp''_in_after hp''_extends ] ].
assert (hp = hp''). apply hp_unique. auto. subst hp''. auto.
assert (In trail c (heap_trail hp)) as hp_in_c.
inversion hp_in_after as [ ht ht_in_c | ht false ]. 
auto. inversion false.

split.

(* unique heap *)
Focus.
unfold unique. exists hp.
split. left. auto.
intros hp' in_before. inversion in_before as [ ht' in_c | ht' false ].
apply hp_unique. left. auto. inversion false.
Unfocus.

split.

(* unique thread *)
Focus.
intros id1 p1 p2 t1_in_before t2_in_before.
assert (id = id1 \/ id <> id1) as classic.
apply classic. destruct classic as [ | id_not_id1 ].
subst id1.
assert (forall p', ~ In trail c (thread_trail id p')) as id_not_in_c.
intros p' t'_in_c.
assert (p' = path_cons path_right p).
eapply unique_thread. left. eauto. right. apply In_singleton.
subst p'. tauto.
inversion t1_in_before as [ t1 t1_in_c | t1 t1_is_t ];
inversion t2_in_before as [ t2 t2_in_c | t2 t2_is_t ];
subst t1 t2.
unfold not in id_not_in_c. assert False. eauto. tauto.
unfold not in id_not_in_c. assert False. eauto. tauto.
unfold not in id_not_in_c. assert False. eauto. tauto.
inversion t1_is_t. inversion t2_is_t. subst p1 p2. reflexivity.
inversion t1_in_before as [ t1 t1_in_c | t1 false1 ];
inversion t2_in_before as [ t2 t2_in_c | t2 false2 ];
subst t1 t2.
eapply unique_thread. left. eauto. left. auto.
inversion false2. tauto. inversion false1. tauto.
inversion false1. tauto.
Unfocus.

split.

(* Parent thread *)
Focus.

(** Import our helper lemmas *)
assert ((forall p_x id_x,
  consistent (thread_trail id_x p_x) ->
  In trail (dep_aux_thread id_x p_x)
  (thread_trail id (path_cons path_right p)) ->
  id_x <> id ->
  exists p'', 
    In trail (dep_aux_thread id_x p_x) (thread_trail id p'')
    /\ extends_strict p'' (path_cons path_right p))
/\ (forall hp_x,
  consistent (heap_trail hp_x) ->
  In trail (dep_aux_heap hp_x)
  (thread_trail id (path_cons path_right p)) ->
  exists p'', 
    In trail (dep_aux_heap hp_x) (thread_trail id p'')
    /\ extends_strict p'' (path_cons path_right p)))
as [ helper_thread helper_heap ].
apply right_helper.

(** Let t1 be in before and <id2, p2> be in dep(t1).
   Want to show there exists <id2, p2'> in before such that
   p2' extends p2.
   *)
intros t1 id2 p2 t1_in_before t2_in_dep_t1.
destruct t1_in_before as [ t1 t1_in_c | t1 t1_is_t ].

(** Let t1 be in c.  Then there exists p2' such that <id2, p2'>
   is in after and p2' extends p2.
   *)
assert (exists p2',
  In trail (Add trail c (thread_trail id (path_cons path_right p)))
  (thread_trail id2 p2') /\
  extends p2' p2) as [ p2' [ t2'_in_after p2'_extends ] ].
eapply parent_thread. left. apply t1_in_c. auto.

(** Let <id2, p2'> in c.  Then <id2, p2'> in before, so take p2'.
   *)
inversion t2'_in_after as [ t2' t2'_in_c | t2' t2'_is_t' ]; subst t2'.
exists p2'. split. left. auto. auto.

 (** Let <id2, p2'> be <id, Left:p>.  Then Left:p extends p2.
    Either p extends p2 or p2 = Left:p.
   *)
inversion t2'_is_t'. subst id2 p2'.
inversion p2'_extends as [ | elt p2'' ]. subst p2.

(** Suppose p2 = Left:p.  This case is vacuous.
   There exists p'' such that p'' extends strictly Left:p.
*)
assert (exists p'',
  In trail (dep_aux t1) (thread_trail id p'') /\
  extends_strict p'' (path_cons path_right p)) 
as [ p'' [ p'_in_dep_t1 p''_extends_strict ] ].

(** if t1 is a thread, we need to show that id1 <> id *)
induction t1 as [ id1 p1 | hp1 ].
assert (id1 <> id). unfold not. intros. subst id1.
destruct t'_not_in_c.
cut (p1 = path_cons path_right p).
intros. subst p1. auto.
eapply unique_thread. left. eauto. right. auto.

(** now apply the helper lemma *)
simpl. apply helper_thread. 
eapply committable_implies_consistent. apply comm_after.
left. auto.
unfold dep in t2_in_dep_t1. inversion t2_in_dep_t1 as [ | t2 eq ].
auto. inversion eq. subst. tauto.
auto.

(** when t1 is a heap, simply apply the helper lemma *)
simpl. apply helper_heap.
eapply committable_implies_consistent. apply comm_after.
left. auto.
unfold dep in t2_in_dep_t1. inversion t2_in_dep_t1 as [ | t2 false ].
auto. inversion false.

(** since p'' extends strictly, we have a contradiction
   Left:p extends p'', because of committability, but clearly doesn't
   because p'' extends it strictly
   *)
assert (extends (path_cons path_right p) p'').
assert (exists p''',
  In trail
  (Add trail c (thread_trail id (path_cons path_right p)))
  (thread_trail id p''') /\ extends p''' p'')
as [ p''' [ p'''_in_after p'''_extends ] ].
eapply parent_thread. left. apply t1_in_c. left. auto.
assert (p''' = path_cons path_right p).
eapply unique_thread. apply p'''_in_after. right. apply In_singleton.
subst p'''. apply p'''_extends.

assert (~ extends (path_cons path_right p) p'').
apply extends_strict_implies_reverse_not_extends. auto.

(** contradiction *)
tauto.

(** Remember the case when p extends p2? *)
subst elt p2''. exists p. split. right. apply In_singleton. auto.

(** Now let t1 be <id, p> 
   Either t2 is in DA t1 or it is t1
   *)
unfold dep in t2_in_dep_t1.
inversion t2_in_dep_t1 as [ t2 t2_in_DA_t1 | t2 t2_is_t1 ].

(** Let t2 in DA t1 
   Then by committability there exists p2'
   *)
inversion t1_is_t. subst t1.
assert (exists p2',
  In trail (Add trail c (thread_trail id (path_cons path_right p)))
  (thread_trail id2 p2')
  /\ extends p2' p2) as [ p2' [ p2'_in_after p2'_extends ] ].
eapply parent_thread. right. apply In_singleton.
left. simpl. auto.

(** is <id2, p2'> in c, or is it t'? *)
inversion p2'_in_after as [ t2' t2'_in_c | t2' t2'_is_t' ].
exists p2'. split. left. auto. auto.
inversion t2'_is_t'. subst id2 p2'.
exists p. split. right. apply In_singleton.
apply extends_strict_implies_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply committable_implies_consistent.
apply comm_after. right. apply In_singleton.
apply extends_cons. apply extends_refl. auto.

(** t1 is t, and t2 is t1 is t *)
inversion t1_is_t. subst t1. inversion t2_is_t1. subst id2 p2.
exists p. split. right. apply In_singleton. apply extends_refl.
Unfocus.

(* heap parent *)
intros t1 hp2 t1_in_before hp2_in_dep_t1.

(** cases: t1 in c, t1 is t *)
destruct t1_in_before as [ t1 t1_in_c | t1 t1_is_t ].
exists hp. split. left. auto. eapply hp_parent. left. apply t1_in_c. auto.
inversion t1_is_t. subst t1.
unfold dep in hp2_in_dep_t1.
inversion hp2_in_dep_t1 as [ ht2 hp2_in_DA | ht2 false ].
exists hp. split. left. auto. eapply hp_parent. right.
apply In_singleton. unfold dep. simpl. left. auto.
inversion false.

Qed.