Require Import consistent.
Require Import Coq.Logic.Classical.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import syntax.
Require Import trails.

Lemma right_helper :
  forall id_r p_r,
    (forall p id,
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons path_right p_r)) ->
      id <> id_r ->
      exists p_r'', 
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'')
        /\ extends_strict p_r'' (path_cons path_right p_r))
    /\ (forall hp,
      consistent (heap_trail hp) ->
      In trail (dep_aux_heap hp)
      (thread_trail id_r (path_cons path_right p_r)) ->
      exists p_r'', 
        In trail (dep_aux_heap hp) (thread_trail id_r p_r'')
        /\ extends_strict p_r'' (path_cons path_right p_r)).

intros id_r p_r.

apply trail_induction with
(P := fun elt =>
  match elt with
    | path_left => True
    | path_right => True
    | (path_send n id p) => 
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons path_right p_r)) ->
      id <> id_r ->
      exists p_r'', 
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'')
        /\ extends_strict p_r'' (path_cons path_right p_r)
    | (path_recv n id p) =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons path_right p_r)) ->
      id <> id_r ->
      exists p_r'', 
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'')
        /\ extends_strict p_r'' (path_cons path_right p_r)
    | (path_heap n hp) => 
      consistent (heap_trail hp) ->
      In trail (dep_aux_heap hp)
      (thread_trail id_r (path_cons path_right p_r)) ->
      exists p_r'', 
        In trail (dep_aux_heap hp) (thread_trail id_r p_r'')
        /\ extends_strict p_r'' (path_cons path_right p_r)
  end)
(P0 := fun elt =>
  match elt with
    | (heap_path_thread n id p) =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons path_right p_r)) ->
      id <> id_r ->
      exists p_r'', 
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'')
        /\ extends_strict p_r'' (path_cons path_right p_r)
  end)
(P1 := fun p =>
  forall id,
    consistent (thread_trail id p) ->
    In trail (dep_aux_thread id p)
    (thread_trail id_r (path_cons path_right p_r)) ->
    id <> id_r ->
    exists p_r'', 
      In trail (dep_aux_thread id p) (thread_trail id_r p_r'')
      /\ extends_strict p_r'' (path_cons path_right p_r))
(P2 := fun hp =>
  consistent (heap_trail hp) ->
  In trail (dep_aux_heap hp)
  (thread_trail id_r (path_cons path_right p_r)) ->
  exists p_r'', 
    In trail (dep_aux_heap hp) (thread_trail id_r p_r'')
    /\ extends_strict p_r'' (path_cons path_right p_r));
auto.

intros id consis false. simpl in false. inversion false.

intros elt IH_elt p IH id consis in_DA not_r.

induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left and Right *)
Focus.
simpl. apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
Unfocus. Focus.
simpl. apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
Unfocus.

(* Send *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_r' [ t_r in_DA_t | t_r in_DA_t_x ] | t_r is_t_x ].

(* t_r in DA(id,p) *)
assert (exists p_r'', 
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'')
  /\ extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_l.
apply classic. destruct cmp_x_l. subst id_x.
(* By Lemma B, px \extends Send(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons path_right p_r))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Recv(<t,p'>):px \extends(strictly) Send(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_recv n_x id p) p_x)
(path_cons path_right p_r)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Recv(<t,p'>):px*)
exists (path_cons (path_recv n_x id p) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert (exists p_r'', 
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'')
  /\ extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH_elt. auto. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
exists p''. split. simpl. left. right. auto. auto.
inversion is_t_x.
Unfocus.

(* Recv *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_r' [ t_r in_DA_t | t_r in_DA_t_x ] | t_r is_t_x ].

(* t_r in DA(id,p) *)
assert (exists p_r'', 
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'')
  /\ extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_l.
apply classic. destruct cmp_x_l. subst id_x.
(* By Lemma B, px \extends Send(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons path_right p_r))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Recv(<t,p'>):px \extends(strictly) Send(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_send n_x id p) p_x)
(path_cons path_right p_r)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Recv(<t,p'>):px*)
exists (path_cons (path_send n_x id p) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert (exists p_r'', 
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'')
  /\ extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH_elt. auto. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
exists p''. split. simpl. left. right. auto. auto.
inversion is_t_x.
Unfocus.

(* Heap *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_r' [ t_r in_DA_t | t_r in_DA_t_x ] | t_r false ].
assert (exists p_r'',
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (heap_trail hp_x)) as consis_x.
eapply consistent_implies_heap_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply heap_extends_cons. apply heap_extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert (exists p_r'',
  In trail (dep_aux_heap hp_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
exists p''. split. simpl. left. right. auto. auto.

inversion false.
Unfocus.

(* heap path nil *)
Focus.
intros consis false. simpl in false. inversion false.
Unfocus.

(* heap path *)
intros elt IH_elt hp IH consis in_DA.

simpl in in_DA. induction elt as [ n_x id_x p_x ].
inversion in_DA as [ t_r' [ t_r in_DA_t_x | t_r in_DA_t ] | t_r is_t_x ].

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_s.
apply classic. destruct cmp_x_s. subst id_x.
(* By Lemma B, px \extends Send(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons path_right p_r))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Recv(<t,p'>):px \extends(strictly) Send(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_heap n_x hp) p_x)
(path_cons path_right p_r)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Recv(<t,p'>):px*)
exists (path_cons (path_heap n_x hp) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert (exists p_r'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH_elt. auto. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
exists p''. split. simpl. left. left. auto. auto.

assert (exists p_r'',
  In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons path_right p_r)) as p''.
apply IH. eapply consistent_implies_heap_extends_consistent.
apply consis. apply heap_extends_cons. apply heap_extends_refl. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
exists p''. split. simpl. left. right. auto. auto.

inversion is_t_x. 

Qed.
