Require Import committable.
Require Import committable_before_end_lockdown_helper.
Require Import consistent.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import heap_in_dep.
Require Import syntax.
Require Import trails.

Lemma committable_before_end_lockdown :
  forall c n id p hp,
    ~ In trail c (thread_trail id (path_cons (path_heap n hp) p)) ->
    ~ In trail c (heap_trail (heap_path_cons (heap_path_thread n id p) hp)) ->
    committable (Union trail c 
      (Couple trail
        (thread_trail id (path_cons (path_heap n hp) p))
        (heap_trail (heap_path_cons (heap_path_thread n id p) hp)))) ->
    committable (Union trail c 
      (Couple trail
        (thread_trail id p)
        (heap_trail hp))).

intros c n1 id1 p1 hp2 t1'_not_in_c ht2'_not_in_c comm_after.
unfold committable. 
inversion comm_after as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
unfold unique in unique_heap.
destruct unique_heap as [ hp [ hp2'_in_after hp2'_unique ] ].
assert (forall t hp',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1))
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2)))) t ->
  In trail (dep t) (heap_trail hp') ->
  heap_extends hp hp') as hp2'_parent.
intros t hp' in_after in_dep.
assert (exists hp'',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1))
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2))))
  (heap_trail hp'') /\ heap_extends hp'' hp') as hp''.
eapply parent_heap. apply in_after. auto.
destruct hp'' as [ hp'' [ hp''_in_after hp''_extends ] ].
assert (hp = hp''). apply hp2'_unique. auto. subst hp''. auto.
assert (hp = heap_path_cons (heap_path_thread n1 id1 p1) hp2).
apply hp2'_unique. right. right. subst hp. clear hp2'_in_after.

split.

(* Unique heap *)
Focus.
unfold unique. exists hp2. split.
right. right.
intros hp in_before.
inversion in_before as [ ht ht_in_c | ht [ | ] ].
assert (heap_path_cons (heap_path_thread n1 id1 p1) hp2 = hp).
apply hp2'_unique. left. auto. subst hp. tauto. reflexivity.
Unfocus.

split.

(* Unique thread *)
Focus.
intros id3 p3 p4 t3_in_after t4_in_after.

inversion t3_in_after as [ t3 t3_in_c | t3 [ | ] ];
inversion t4_in_after as [ t4 t4_in_c | t4 [ | ] ];
clear t3_in_after t4_in_after.

eapply unique_thread.
left. apply t3_in_c. left. apply t4_in_c.

subst id3 p4.
assert (p3 = path_cons (path_heap n1 hp2) p1).
eapply unique_thread. left. apply t3_in_c. right. left.
subst p3. tauto.

subst id3 p3.
assert (p4 = path_cons (path_heap n1 hp2) p1).
eapply unique_thread. left. apply t4_in_c. right. left.
subst p4. tauto.

reflexivity.
Unfocus.

split.

(* Parent thread *)
Focus.
intros t id p t_in_before p_in_dep_t.

inversion t_in_before as [ t_ t_in_c | t_ t_not_in_c ].
subst t_.

(* Let t in c. Then exists p' such that <id, p'> in c_after 
   and p' extends p *)
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1))
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2))))
  (thread_trail id p') /\ extends p' p) as p'.
eapply parent_thread. left. apply t_in_c. apply p_in_dep_t.
destruct p' as [ p' [ t'_in_after p'_extends ] ].

inversion t'_in_after as [ t' t'_in_c | t' [ | ] ]; clear t'_in_after.

(* Case: <id, p'> in c.  Then take p' *)
exists p'. split. left. apply t'_in_c. apply p'_extends.

(* Case: <id, p'> = <id1, Send(<id2, p2>):p1> *)
subst id p'. inversion p'_extends as [ p1' | elt p1_ p_ p1_extends ].

(* Case: p = p' = Heap(hp2):p1 *)
subst p1' p. induction t as [ id_t p_t | hp_t ].
assert (forall p id,
  consistent (thread_trail id p) ->
  In trail (dep_aux_thread id p)
  (thread_trail id1 (path_cons (path_heap n1 hp2) p1)) ->
  id <> id1 ->
  (exists p1'',
    In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
    extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
  (exists hp2'',
    In trail (dep_aux_thread id p) (heap_trail hp2'') /\
    heap_extends_strict hp2'' 
    (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) 
as helper.
eapply proj1. apply heap_helper1.
assert ((exists p1'',
    In trail (dep_aux_thread id_t p_t) (thread_trail id1 p1'') /\
    extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp2'',
  In trail (dep_aux_thread id_t p_t) (heap_trail hp2'') /\
  heap_extends_strict hp2'' 
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
eapply helper. eapply committable_implies_consistent.
apply comm_after. left. auto.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p in_DA | p false ]. 
auto. inversion false. subst. tauto.
unfold not. intros. subst id_t.
assert (p_t = (path_cons (path_heap n1 hp2) p1)).
eapply unique_thread. left. apply t_in_c. right. left. subst p_t. tauto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_heap n1 hp2) p1) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1))
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2))))
  (thread_trail id1 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_heap n1 hp2) p1)).
eapply unique_thread. apply in_after. right. left.
subst p'. auto.
assert (~ extends (path_cons (path_heap n1 hp2) p1) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.
destruct p'' as [ p'' [ in_before strict ] ].
assert (heap_extends (heap_path_cons (heap_path_thread n1 id1 p1) hp2) p'').
eapply hp2'_parent. left. apply t_in_c. unfold dep. left. auto.
assert (~ heap_extends (heap_path_cons (heap_path_thread n1 id1 p1) hp2) p'').
apply heap_extends_strict_implies_reverse_not_extends. auto.
tauto.

assert (heap_path_cons (heap_path_thread n1 id1 p1) hp2 = hp_t).
apply hp2'_unique. left. auto. subst hp_t. tauto.

(* Case: p1 extends p. Take p1. *)
exists p1. split. right. left. auto.

(* Case: t is not in c *)
inversion t_not_in_c.

(* Case: t is <id1, p1> *)
subst t t_. clear t_not_in_c.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p_ in_DA | p_ p_is_t ].

(* Case: <id, p> in DA(<id1, p1> *)
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1))
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2))))
  (thread_trail id p') /\ extends p' p) as p'.
eapply parent_thread. right. left. unfold dep. simpl. left. left. left. auto.
destruct p' as [ p' [ in_after p'_extends ] ].
inversion in_after as [ t' t'_in_c | t' [ | ] ]; clear in_after.

(* Case: <id, p'> in c *)
exists p'. split. left. auto. auto.

(* Case: <id, p'> = <id1, Heap(hp2):p1> *)
subst id p'.
exists p1. split. right. left.
eapply consistent_implies_dep_extends.
eapply consistent_implies_extends_consistent.
eapply committable_implies_consistent.
unfold committable. apply comm_after.
right. left. apply extends_cons. apply extends_refl.
unfold dep. auto.

(* Case: p is p1 *)
inversion p_is_t as [ eq ]. subst id p.
exists p1. split. right. left. apply extends_refl.

(* Case: t is <id2, p2> *)
subst t t_. clear t_not_in_c.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p_ in_DA | p_ p_is_t ].

(* Case: <id, p> in DA(<id2, p2> *)
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1))
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2))))
  (thread_trail id p') /\ extends p' p) as p'.
eapply parent_thread. right. right. unfold dep. simpl. left. left. right. auto.
destruct p' as [ p' [ in_after p'_extends ] ].
inversion in_after as [ t' t'_in_c | t' [ | ] ]; clear in_after.

(* Case: <id, p'> in c *)
exists p'. split. left. auto. auto.

(* Case: <id, p'> = <id2, Send(<id2, p2>):p1> *)
subst id p'. subst p_.
inversion p'_extends as [ p1' | elt p1_ p_ p1_extends ].

(* Case: p = p' = Recv(<id1, p1>):p2 *)
subst p1' p.
assert (~ In trail (dep_aux (heap_trail hp2))
  (thread_trail id1 (path_cons (path_heap n1 hp2) p1))).
apply heap_helper2. tauto.

(* Case: p1 extends p *)
exists p1. split. right. left. auto.

(* Case: p is p2 *)
inversion p_is_t. 
Unfocus.

(* Parent heap *)
intros t hp' t_in_before hp'_in_dep_t.
exists hp2. split. right. right.

(* 
   Let t in before and let <heap, hp'> in dep t.
*)

destruct t_in_before as [ t t_in_c | t [ | ] ].

(*
   Let t in c.  Then hp2' extends hp' by committability.
*)

assert (heap_extends (heap_path_cons (heap_path_thread n1 id1 p1) hp2) hp')
  as hp2'_extends.
eapply hp2'_parent. left. apply t_in_c. auto.

(*
   If hp2' extends strictly hp', then hp2 extends hp'.
*)

inversion hp2'_extends as [ | elt hp2_ ].

(*
   Else, we have <heap, hp2'> in dep t.
*)

subst hp'.

(*
   By heap_helper3 (first need to show that t is a 
   thread and doesn't have thread id id1), there exists either hp2''
   or p1'' extending stuff strictly.  Show contradiction.
*)

induction t as [ id3 p3 | hp3 ].
assert (id3 <> id1). unfold not. intros. subst id3.
assert (p3 = (path_cons (path_heap n1 hp2) p1)).
eapply unique_thread. left. apply t_in_c. right. left. subst p3. tauto.

assert ((forall p id,
  consistent (thread_trail id p) ->
  In trail (dep_aux_thread id p)
  (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2)) ->
  id <> id1 ->
  (exists p1'',
    In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
    extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
  (exists hp2'',
    In trail (dep_aux_thread id p) (heap_trail hp2'') /\
    heap_extends_strict hp2'' (heap_path_cons (heap_path_thread n1 id1 p1) hp2)))) as helper. eapply proj1. apply heap_helper3.
assert ((exists p1'',
  In trail (dep_aux_thread id3 p3) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp2'',
  In trail (dep_aux_thread id3 p3) (heap_trail hp2'') /\
  heap_extends_strict hp2'' 
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply helper.
eapply committable_implies_consistent.
apply comm_after. left. auto.
unfold dep in hp'_in_dep_t. inversion hp'_in_dep_t as [ | ht' false ].
auto. inversion false. auto.

destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
assert (extends (path_cons (path_heap n1 hp2) p1) p'').
assert (exists p''',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1))
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2)))) 
  (thread_trail id1 p''') /\ extends p''' p'')
as [ p''' [ in_after p'''_extends ] ].
eapply parent_thread. left. apply t_in_c.
left. auto.
assert (p''' = (path_cons (path_heap n1 hp2) p1)).
eapply unique_thread. apply in_after. right. left. subst p'''. auto.
assert (~ extends (path_cons (path_heap n1 hp2) p1) p'').
apply extends_strict_implies_reverse_not_extends. auto. 
tauto.
assert (heap_extends (heap_path_cons (heap_path_thread n1 id1 p1) hp2) p'').
eapply hp2'_parent. left. apply t_in_c.
left. auto.
assert (~ heap_extends (heap_path_cons (heap_path_thread n1 id1 p1) hp2) p'').
apply heap_extends_strict_implies_reverse_not_extends. auto. 
tauto.
assert ((heap_path_cons (heap_path_thread n1 id1 p1) hp2) = hp3).
apply hp2'_unique. left. auto. subst hp3. tauto.

(* take hp2 *)
auto.

(*
   Let t be <id1, p1>. Then <heap, hp'> in DA(id1, p1).
   So hp2' extends hp' by committability.
*)

unfold dep in hp'_in_dep_t. inversion hp'_in_dep_t as [ | ht false ].
assert (heap_extends (heap_path_cons (heap_path_thread n1 id1 p1) hp2) hp')
  as hp2'_extends.
eapply hp2'_parent.
right. left. unfold dep. simpl. left. left. left. auto.

(*
   If hp2' extends strictly hp', then hp2 extends hp'.

   Else, we have <heap, hp2'> in DA(id1,p1). Not true by helper4.

*)

inversion hp2'_extends. 
subst hp'.
assert (~ In trail (dep_aux (thread_trail id1 p1))
  (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2))).
apply heap_helper4. tauto.
auto.
inversion false.

(*
   Let t be <heap, hp2>.  Then <heap, hp'> in DA(heap, hp2).
   So hp2' extends hp' by committability.
   *)

unfold dep in hp'_in_dep_t. inversion hp'_in_dep_t as [ | ht is_t ].
assert (heap_extends
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2) hp') as hp2'_extends.
eapply hp2'_parent. right. right. unfold dep. simpl. left. left. right. auto.

(*
   If hp2' extends strictly hp', then hp2 extends hp'.

   Else, we have <heap, hp2'> in DA(hp2). Not true by helper4.
   *)
inversion hp2'_extends. subst hp'.
apply consistent_implies_dep_heap_extends.
eapply consistent_implies_heap_extends_consistent.
eapply committable_implies_consistent.
apply comm_after. right. right. apply heap_extends_cons. 
apply heap_extends_refl. left. auto.

auto. 

inversion is_t. apply heap_extends_refl.

Qed.
