Require Import consistent.
Require Import Coq.Arith.Arith.
Require Import Coq.Logic.Classical.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import heap_in_dep.
Require Import syntax.
Require Import trails.

Lemma heap_helper1 :
  forall n_ld id_ld p_ld hp_ld,
    (forall p id,
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
      id <> id_ld ->
      (exists p_ld'',
        In trail (dep_aux_thread id p) (thread_trail id_ld p_ld'') /\
        extends_strict p_ld'' (path_cons (path_heap n_ld hp_ld) p_ld)) \/
      (exists hp_ld'',
        In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
        heap_extends_strict hp_ld'' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)))
    /\ (forall (hp:heap_path), True).

intros n1 id1 p1 hp2.

apply trail_induction with
  (P := fun elt =>
    match elt with
      | path_left => True
      | path_right => True
      | (path_send n id p) => 
        consistent (thread_trail id p) ->
        In trail (dep_aux_thread id p)
        (thread_trail id1 (path_cons (path_heap n1 hp2) p1)) ->
        id <> id1 ->
        (exists p_ld'' : path,
          In trail (dep_aux_thread id p) (thread_trail id1 p_ld'') /\
          extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
        (exists hp_ld'' : heap_path,
          In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
          heap_extends_strict hp_ld''
          (heap_path_cons (heap_path_thread n1 id1 p1) hp2))
      | (path_recv n id p) => 
        consistent (thread_trail id p) ->
        In trail (dep_aux_thread id p)
        (thread_trail id1 (path_cons (path_heap n1 hp2) p1)) ->
        id <> id1 ->
        (exists p_ld'' : path,
          In trail (dep_aux_thread id p) (thread_trail id1 p_ld'') /\
          extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
        (exists hp_ld'' : heap_path,
          In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
          heap_extends_strict hp_ld''
          (heap_path_cons (heap_path_thread n1 id1 p1) hp2))
      | (path_heap n hp) => True
    end)
(P0 := fun elt => 
  match elt with 
    | heap_path_thread n id p =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id1 (path_cons (path_heap n1 hp2) p1)) ->
      id <> id1 ->
      (exists p_ld'' : path,
        In trail (dep_aux_thread id p) (thread_trail id1 p_ld'') /\
        extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
      (exists hp_ld'' : heap_path,
        In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
        heap_extends_strict hp_ld''
        (heap_path_cons (heap_path_thread n1 id1 p1) hp2))
  end)
(P1 := fun p =>
  forall id,
    consistent (thread_trail id p) ->
    In trail (dep_aux_thread id p)
    (thread_trail id1 (path_cons (path_heap n1 hp2) p1)) ->
    id <> id1 ->
    (exists p_ld'' : path,
      In trail (dep_aux_thread id p) (thread_trail id1 p_ld'') /\
      extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
    (exists hp_ld'' : heap_path,
      In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
      heap_extends_strict hp_ld''
      (heap_path_cons (heap_path_thread n1 id1 p1) hp2)))
(P2 := fun (hp:heap_path) => True).

auto. auto. auto. auto. auto. auto.

intros id consis false. simpl in false. inversion false.

intros elt IH_elt p IH_path id consis in_DA not_id1.

induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left *)
simpl in in_DA. simpl. apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.

(* Right *)
simpl in in_DA. simpl. apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.

(* Send *)
Focus.
simpl in in_DA.

(** either it's in DA p, DA x, or it's the receiver (vacuous) *)
inversion in_DA as [ t1'' [ t1' in_DA_p | t1' in_DA_x ] | t1' false ].

(** in DA p *)
assert ((exists p_ld'' : path,
  In trail (dep_aux_thread id p) (thread_trail id1 p_ld'') /\
  extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp_ld'' : heap_path,
  In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
  heap_extends_strict hp_ld''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

(** in DA x *)
assert (id_x = id1 \/ id_x <> id1) as classic.
apply classic. destruct classic.
subst id_x.
left. 
assert (extends_strict p_x (path_cons (path_heap n1 hp2) p1))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_x.
exists (path_cons (path_recv n_x id p) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity. 
apply extends_strict_implies_extends. auto.

(* id_x <> id1 *)
assert ((exists p_ld'' : path,
  In trail (dep_aux_thread id_x p_x) (thread_trail id1 p_ld'') /\
  extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp_ld'' : heap_path,
  In trail (dep_aux_thread id_x p_x) (heap_trail hp_ld'') /\
  heap_extends_strict hp_ld''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_elt.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

inversion false.
Unfocus.

(* Send *)
Focus.
simpl in in_DA.

(** either it's in DA p, DA x, or it's the receiver (vacuous) *)
inversion in_DA as [ t1'' [ t1' in_DA_p | t1' in_DA_x ] | t1' false ].

(** in DA p *)
assert ((exists p_ld'' : path,
  In trail (dep_aux_thread id p) (thread_trail id1 p_ld'') /\
  extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp_ld'' : heap_path,
  In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
  heap_extends_strict hp_ld''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

(** in DA x *)
assert (id_x = id1 \/ id_x <> id1) as classic.
apply classic. destruct classic.
subst id_x.
left. 
assert (extends_strict p_x (path_cons (path_heap n1 hp2) p1))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_x.
exists (path_cons (path_send n_x id p) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity. 
apply extends_strict_implies_extends. auto.

(* id_x <> id1 *)
assert ((exists p_ld'' : path,
  In trail (dep_aux_thread id_x p_x) (thread_trail id1 p_ld'') /\
  extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp_ld'' : heap_path,
  In trail (dep_aux_thread id_x p_x) (heap_trail hp_ld'') /\
  heap_extends_strict hp_ld''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_elt.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

inversion false.
Unfocus.

(* Heap *)
Focus.
simpl in in_DA.

(** either it's in DA p, DA x, or it's the receiver (vacuous) *)
inversion in_DA as [ t1'' [ t1' in_DA_p | t1' in_DA_x ] | t1' false ].

(** in DA p *)
assert ((exists p_ld'' : path,
  In trail (dep_aux_thread id p) (thread_trail id1 p_ld'') /\
  extends_strict p_ld'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp_ld'' : heap_path,
  In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
  heap_extends_strict hp_ld''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

(** in DA x *)
assert (exists hp_x',
  In trail (dep (heap_trail hp_x)) (heap_trail hp_x')
  /\ heap_extends hp_x' 
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))
as [ hp_x' [ hp_x'_in_dep hp_x'_extends ] ].
apply heap_in_dep1. left. auto.
unfold dep in hp_x'_in_dep.
inversion hp_x'_in_dep as [ ht' ht'_in_DA | ht' ht'_is_x ].
right. esplit. split. simpl. right. apply In_singleton.
eapply heap_extends_strict_cons. reflexivity.
eapply heap_extends_trans.
apply heap_extends_strict_implies_extends.
apply consistent_implies_dep_aux_heap_extends_strict.
eapply consistent_implies_heap_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply heap_extends_cons. apply heap_extends_refl.
apply ht'_in_DA. auto. 

inversion ht'_is_x. subst ht' hp_x'.
right. esplit. split. simpl. right. apply In_singleton.
eapply heap_extends_strict_cons. reflexivity. auto.

inversion false.

Unfocus.

auto. auto.

Qed.

Lemma heap_helper2 :
  forall n_ld id_ld p_ld hp_ld,
    ~ In trail (dep_aux (heap_trail hp_ld))
    (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)).

intros n id p hp in_DA.
assert (trail_elt_count (heap_trail hp) >= 
  (trail_elt_count (thread_trail id (path_cons (path_heap n hp) p))))
as hp_ge.
apply dep_aux_implies_elt_count_ge. auto.
simpl in hp_ge.
unfold ge in hp_ge.
rewrite plus_n_Sm in hp_ge.
assert (S (heap_elt_count hp) <= heap_elt_count hp) as contra.
eapply le_trans. 
instantiate (1 := elt_count p + S (heap_elt_count hp)). auto with arith.
auto.
assert (~ S (heap_elt_count hp) <= heap_elt_count hp).
apply le_Sn_n. tauto.

Qed.

Lemma heap_helper3 :
  forall n_ld id_ld p_ld hp_ld,
    (forall p id,
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
      id <> id_ld ->
      (exists p_ld'',
        In trail (dep_aux_thread id p) (thread_trail id_ld p_ld'') /\
        extends_strict p_ld'' (path_cons (path_heap n_ld hp_ld) p_ld)) \/
      (exists hp_ld'',
        In trail (dep_aux_thread id p) (heap_trail hp_ld'') /\
        heap_extends_strict hp_ld'' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))) /\
    (forall (hp:heap_path), True).

intros n1 id1 p1 hp2.

apply trail_induction with
  (P := fun elt =>
    match elt with
      | path_left => True
      | path_right => True
      | (path_send n id p) => 
        consistent (thread_trail id p) ->
        In trail (dep_aux_thread id p)
        (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2)) ->
        id <> id1 ->
        (exists p1'' : path,
          In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
          extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
        (exists hp2'' : heap_path,
          In trail (dep_aux_thread id p) (heap_trail hp2'') /\
          heap_extends_strict hp2''
          (heap_path_cons (heap_path_thread n1 id1 p1) hp2))
      | (path_recv n id p) => 
        consistent (thread_trail id p) ->
        In trail (dep_aux_thread id p)
        (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2)) ->
        id <> id1 ->
        (exists p1'' : path,
          In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
          extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
        (exists hp2'' : heap_path,
          In trail (dep_aux_thread id p) (heap_trail hp2'') /\
          heap_extends_strict hp2''
          (heap_path_cons (heap_path_thread n1 id1 p1) hp2))
      | (path_heap n hp) => True
    end)
(P0 := fun elt => 
  match elt with 
    | heap_path_thread n id p =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2)) ->
      id <> id1 ->
      (exists p1'' : path,
        In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
        extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
      (exists hp2'' : heap_path,
        In trail (dep_aux_thread id p) (heap_trail hp2'') /\
        heap_extends_strict hp2''
        (heap_path_cons (heap_path_thread n1 id1 p1) hp2))
  end)
(P1 := fun p =>
  forall id,
    consistent (thread_trail id p) ->
    In trail (dep_aux_thread id p)
    (heap_trail (heap_path_cons (heap_path_thread n1 id1 p1) hp2)) ->
    id <> id1 ->
    (exists p1'' : path,
      In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
      extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
    (exists hp2'' : heap_path,
      In trail (dep_aux_thread id p) (heap_trail hp2'') /\
      heap_extends_strict hp2''
      (heap_path_cons (heap_path_thread n1 id1 p1) hp2)))
(P2 := fun (hp:heap_path) => True).

auto. auto. auto. auto. auto. auto.

intros id consis false. simpl in false. inversion false.

intros elt IH_elt p IH_path id consis in_DA not_id1.

induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left *)
simpl in in_DA. simpl. apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.

(* Right *)
simpl in in_DA. simpl. apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.

(* Send *)
Focus.
simpl in in_DA.

(** either it's in DA p, DA x, or it's the receiver (vacuous) *)
inversion in_DA as [ t1'' [ t1' in_DA_p | t1' in_DA_x ] | t1' false ].

(** in DA p *)
assert ((exists p1'' : path,
  In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp2'' : heap_path,
  In trail (dep_aux_thread id p) (heap_trail hp2'') /\
  heap_extends_strict hp2''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

(** in DA x *)
assert (id_x = id1 \/ id_x <> id1) as classic.
apply classic. destruct classic.
subst id_x.
left.  esplit. split. simpl. right. apply In_singleton.

assert (exists p_x',
  In trail (dep (thread_trail id1 p_x)) (thread_trail id1 p_x')
  /\ extends p_x' (path_cons (path_heap n1 hp2) p1))
as [ p_x' [ p_x'_in_dep p_x'_extends ] ].
apply heap_in_dep2. left. auto.
unfold dep in p_x'_in_dep.
inversion p_x'_in_dep as [ t' t'_in_DA | t' t'_is_x ].
eapply extends_strict_cons. reflexivity.
eapply extends_trans.
apply extends_strict_implies_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
apply t'_in_DA. auto. 

inversion t'_is_x. subst t' p_x'.
eapply extends_strict_cons. reflexivity. auto.

(* id_x <> id1 *)
assert ((exists p1'' : path,
  In trail (dep_aux_thread id_x p_x) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp2'' : heap_path,
  In trail (dep_aux_thread id_x p_x) (heap_trail hp2'') /\
  heap_extends_strict hp2''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_elt.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

inversion false.
Unfocus.

(* Recv *)
Focus.
simpl in in_DA.

(** either it's in DA p, DA x, or it's the receiver (vacuous) *)
inversion in_DA as [ t1'' [ t1' in_DA_p | t1' in_DA_x ] | t1' false ].

(** in DA p *)
assert ((exists p1'' : path,
  In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp2'' : heap_path,
  In trail (dep_aux_thread id p) (heap_trail hp2'') /\
  heap_extends_strict hp2''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

(** in DA x *)
assert (id_x = id1 \/ id_x <> id1) as classic.
apply classic. destruct classic.
subst id_x.
left.  esplit. split. simpl. right. apply In_singleton.

assert (exists p_x',
  In trail (dep (thread_trail id1 p_x)) (thread_trail id1 p_x')
  /\ extends p_x' (path_cons (path_heap n1 hp2) p1))
as [ p_x' [ p_x'_in_dep p_x'_extends ] ].
apply heap_in_dep2. left. auto.
unfold dep in p_x'_in_dep.
inversion p_x'_in_dep as [ t' t'_in_DA | t' t'_is_x ].
eapply extends_strict_cons. reflexivity.
eapply extends_trans.
apply extends_strict_implies_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
apply t'_in_DA. auto. 

inversion t'_is_x. subst t' p_x'.
eapply extends_strict_cons. reflexivity. auto.

(* id_x <> id1 *)
assert ((exists p1'' : path,
  In trail (dep_aux_thread id_x p_x) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp2'' : heap_path,
  In trail (dep_aux_thread id_x p_x) (heap_trail hp2'') /\
  heap_extends_strict hp2''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_elt.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

inversion false.
Unfocus.

(* Heap *)
Focus.
simpl in in_DA.

(** either it's in DA p, DA x, or it's the receiver (vacuous) *)
inversion in_DA as [ t1'' [ t1' in_DA_p | t1' in_DA_x ] | t1' false ].

(** in DA p *)
assert ((exists p1'' : path,
  In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_heap n1 hp2) p1)) \/
(exists hp2'' : heap_path,
  In trail (dep_aux_thread id p) (heap_trail hp2'') /\
  heap_extends_strict hp2''
  (heap_path_cons (heap_path_thread n1 id1 p1) hp2))) as p''.
apply IH_path.
eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl.
auto. auto.
simpl. unfold Add.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA strict ] ].
left. exists p''. auto using Union_introl, Union_intror, In_singleton.
right. exists p''. auto using Union_introl, Union_intror, In_singleton.

(** in DA x *)
right. esplit. split. simpl. right. apply In_singleton.
eapply heap_extends_strict_cons. reflexivity.
apply heap_extends_strict_implies_extends.
apply consistent_implies_dep_aux_heap_extends_strict.
eapply consistent_implies_heap_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply heap_extends_cons. apply heap_extends_refl.
apply in_DA_x. 

inversion false. subst n1 id1 p1 hp2. clear false. tauto.

Unfocus.

auto. auto.

Qed.

Lemma heap_helper4 :
  forall n_ld id_ld p_ld hp_ld,
    ~ In trail (dep_aux (thread_trail id_ld p_ld))
    (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)).

intros n id p hp in_DA.
assert (trail_elt_count (thread_trail id p) >=
  (trail_elt_count (heap_trail (heap_path_cons (heap_path_thread n id p) hp))))
as p_ge.
apply dep_aux_implies_elt_count_ge. auto.
simpl in p_ge.
unfold ge in p_ge.
rewrite plus_n_Sm in p_ge.
assert (S (elt_count p) <= elt_count p) as contra.
eapply le_trans. 
instantiate (1 := heap_elt_count hp + S (elt_count p)). auto with arith.
auto.
assert (~ S (elt_count p) <= elt_count p).
apply le_Sn_n. tauto.

Qed.

