Require Import committable.
Require Import committable_before_send_recv_helper.
Require Import consistent.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import send_in_dep_implies_recv_in_dep.
Require Import trails.

Lemma committable_before_send_recv :
  forall c n1 id1 p1 id2 p2,
    ~ In trail c (thread_trail id1 (path_cons (path_send n1 id2 p2) p1)) ->
    ~ In trail c (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2)) ->
    committable (Union trail c 
      (Couple trail
        (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
        (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2)))) ->
    committable (Union trail c 
      (Couple trail
        (thread_trail id1 p1)
        (thread_trail id2 p2))).

intros c n1 id1 p1 id2 p2 t1'_not_in_c t2'_not_in_c comm_after.
unfold committable. unfold committable in comm_after.
inversion comm_after as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
unfold unique in unique_heap.
destruct unique_heap as [ hp [ hp_in_after hp_unique ] ].
assert (forall t hp',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2)))) t ->
  In trail (dep t) (heap_trail hp') ->
  heap_extends hp hp') as hp_parent.
intros t hp' in_after in_dep.
assert (exists hp'',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (heap_trail hp'') /\ heap_extends hp'' hp') as hp''.
eapply parent_heap. apply in_after. auto.
destruct hp'' as [ hp'' [ hp''_in_after hp''_extends ] ].
assert (hp = hp''). apply hp_unique. auto. subst hp''. auto.
assert (In trail c (heap_trail hp)) as hp_in_c.
inversion hp_in_after as [ ht ht_in_c | ht [ | ] ]. auto.

split.

(* Unique heap *)
Focus.
unfold unique. exists hp. split.
left. auto.
intros hp' in_before.
inversion in_before as [ ht ht_in_c | ht [ | ] ].
apply hp_unique. left. auto.
Unfocus.

split.

(* Unique thread *)
Focus.
intros id3 p3 p4 t3_in_after t4_in_after.

inversion t3_in_after as [ t3 t3_in_c | t3 [ | ] ];
inversion t4_in_after as [ t4 t4_in_c | t4 [ | ] ];
clear t3_in_after t4_in_after.

eapply unique_thread.
left. apply t3_in_c. left. apply t4_in_c.

subst id3 p4.
assert (p3 = path_cons (path_send n1 id2 p2) p1).
eapply unique_thread. left. apply t3_in_c. right. left.
subst p3. tauto.

subst id3 p4.
assert (p3 = path_cons (path_recv n1 id1 p1) p2).
eapply unique_thread. left. apply t3_in_c. right. right.
subst p3. tauto.

subst id3 p3.
assert (p4 = path_cons (path_send n1 id2 p2) p1).
eapply unique_thread. left. apply t4_in_c. right. left.
subst p4. tauto.

reflexivity.

subst id2 id3 p3 p4.
assert (path_cons (path_send n1 id1 p2) p1 = 
  path_cons (path_recv n1 id1 p1) p2) as false.
eapply unique_thread.
right. left. right. right. inversion false.

subst id3 p3.
assert (p4 = path_cons (path_recv n1 id1 p1) p2).
eapply unique_thread. left. apply t4_in_c. right. right.
subst p4. tauto.

subst id2 id3 p3 p4.
assert (path_cons (path_send n1 id1 p2) p1 = 
  path_cons (path_recv n1 id1 p1) p2) as false.
eapply unique_thread.
right. left. right. right. inversion false.

reflexivity.
Unfocus.

split.

(* Parent thread *)
Focus.
intros t id p t_in_before p_in_dep_t.

inversion t_in_before as [ t_ t_in_c | t_ t_not_in_c ].
subst t_.

(* Let t in c. Then exists p' such that <id, p'> in c_after 
   and p' extends p *)
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id p') /\ extends p' p) as p'.
eapply parent_thread. left. apply t_in_c. apply p_in_dep_t.
destruct p' as [ p' [ t'_in_after p'_extends ] ].

inversion t'_in_after as [ t' t'_in_c | t' [ | ] ]; clear t'_in_after.

(* Case: <id, p'> in c.  Then take p' *)
exists p'. split. left. apply t'_in_c. apply p'_extends.

(* Case: <id, p'> = <id1, Send(<id2, p2>):p1> *)
subst id p'. inversion p'_extends as [ p1' | elt p1_ p_ p1_extends ].

(* Case: p = p' = Send(<id2, p2>):p1 *)
subst p1' p. induction t as [ id_t p_t | hp_t ].
assert (forall p id,
  consistent (thread_trail id p) ->
  In trail (dep_aux_thread id p)
  (thread_trail id1 (path_cons (path_send n1 id2 p2) p1)) ->
  id <> id1 -> id <> id2 ->
  (exists p1'',
    In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
    extends_strict p1'' (path_cons (path_send n1 id2 p2) p1)) \/
  (exists p2'',
    In trail (dep_aux_thread id p) (thread_trail id2 p2'') /\
    extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2))) as helper.
eapply proj1. apply send_helper. 
assert ((exists p1'',
  In trail (dep_aux_thread id_t p_t) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_send n1 id2 p2) p1)) \/
(exists p2'',
  In trail (dep_aux_thread id_t p_t) (thread_trail id2 p2'') /\
  extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2))) as p''.
eapply helper. eapply committable_implies_consistent.
unfold committable. split. esplit. split. apply hp_in_after. auto. 
split. eauto. split. auto. auto. left. auto.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p in_DA | p false ]. 
auto. inversion false. subst. tauto.
unfold not. intros. subst id_t.
assert (p_t = (path_cons (path_send n1 id2 p2) p1)).
eapply unique_thread. left. apply t_in_c. right. left. subst p_t. tauto.
unfold not. intros. subst id_t.
assert (p_t = (path_cons (path_recv n1 id1 p1) p2)).
eapply unique_thread. left. apply t_in_c. right. right. subst p_t. tauto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_send n1 id2 p2) p1) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id1 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_send n1 id2 p2) p1)).
eapply unique_thread. apply in_after. right. left.
subst p'. auto.
assert (~ extends (path_cons (path_send n1 id2 p2) p1) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_recv n1 id1 p1) p2) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id2 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_recv n1 id1 p1) p2)).
eapply unique_thread. apply in_after. right. right.
subst p'. auto.
assert (~ extends (path_cons (path_recv n1 id1 p1) p2) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.

assert (hp = hp_t). apply hp_unique. left. auto. subst hp_t.

assert (forall hp,
  consistent (heap_trail hp) ->
  In trail (dep_aux_heap hp)
  (thread_trail id1 (path_cons (path_send n1 id2 p2) p1)) ->
  (exists p1'',
    In trail (dep_aux_heap hp) (thread_trail id1 p1'') /\
    extends_strict p1'' (path_cons (path_send n1 id2 p2) p1)) \/
  (exists p2'',
    In trail (dep_aux_heap hp) (thread_trail id2 p2'') /\
    extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2))) as helper.
eapply proj2. apply send_helper. 
assert ((exists p1'',
  In trail (dep_aux_heap hp) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_send n1 id2 p2) p1)) \/
(exists p2'',
  In trail (dep_aux_heap hp) (thread_trail id2 p2'') /\
  extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2))) as p''.
eapply helper. eapply committable_implies_consistent.
unfold committable. split. esplit. split. apply hp_in_after. auto. 
split. eauto. split. auto. auto. left. auto.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p in_DA | p false ]. 
auto. inversion false.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_send n1 id2 p2) p1) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id1 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_send n1 id2 p2) p1)).
eapply unique_thread. apply in_after. right. left.
subst p'. auto.
assert (~ extends (path_cons (path_send n1 id2 p2) p1) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_recv n1 id1 p1) p2) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id2 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_recv n1 id1 p1) p2)).
eapply unique_thread. apply in_after. right. right.
subst p'. auto.
assert (~ extends (path_cons (path_recv n1 id1 p1) p2) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.

(* Case: p1 extends p. Take p1. *)
exists p1. split. right. left. auto.

(* Case: <id, p'> = <id2, Recv(<id1, p1>):p2> *)
subst id p'. inversion p'_extends as [ p2' | elt p2_ p_ p2_extends ].

(* Case: p = p' = Recv(<id1, p1>):p2 *)
subst p2' p. induction t as [ id_t p_t | hp_t ].
assert (forall p id,
  consistent (thread_trail id p) ->
  In trail (dep_aux_thread id p)
  (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2)) ->
  id <> id2 -> id <> id1 ->
  (exists p2'',
    In trail (dep_aux_thread id p) (thread_trail id2 p2'') /\
    extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2)) \/
  (exists p1'',
    In trail (dep_aux_thread id p) (thread_trail id1 p1'') /\
    extends_strict p1'' (path_cons (path_send n1 id2 p2) p1))) as helper.
eapply proj1. apply recv_helper. 
assert ((exists p2'',
  In trail (dep_aux_thread id_t p_t) (thread_trail id2 p2'') /\
  extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2))
\/ (exists p1'',
  In trail (dep_aux_thread id_t p_t) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_send n1 id2 p2) p1))) as p''.
eapply helper. eapply committable_implies_consistent.
unfold committable. split. esplit. split. apply hp_in_after. auto. 
split. eauto. split. auto. auto. left. auto.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p in_DA | p false ]. 
auto. inversion false. subst. tauto.
unfold not. intros. subst id_t.
assert (p_t = (path_cons (path_recv n1 id1 p1) p2)).
eapply unique_thread. left. apply t_in_c. right. right. subst p_t. tauto.
unfold not. intros. subst id_t.
assert (p_t = (path_cons (path_send n1 id2 p2) p1)).
eapply unique_thread. left. apply t_in_c. right. left. subst p_t. tauto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_recv n1 id1 p1) p2) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id2 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_recv n1 id1 p1) p2)).
eapply unique_thread. apply in_after. right. right.
subst p'. auto.
assert (~ extends (path_cons (path_recv n1 id1 p1) p2) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_send n1 id2 p2) p1) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id1 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_send n1 id2 p2) p1)).
eapply unique_thread. apply in_after. right. left.
subst p'. auto.
assert (~ extends (path_cons (path_send n1 id2 p2) p1) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.

assert (hp = hp_t). apply hp_unique. left. auto. subst hp_t.

assert (forall hp,
  consistent (heap_trail hp) ->
  In trail (dep_aux_heap hp)
  (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2)) ->
  (exists p2'',
    In trail (dep_aux_heap hp) (thread_trail id2 p2'') /\
    extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2)) \/
  (exists p1'',
    In trail (dep_aux_heap hp) (thread_trail id1 p1'') /\
    extends_strict p1'' (path_cons (path_send n1 id2 p2) p1))) as helper.
eapply proj2. apply recv_helper. 
assert ((exists p2'',
  In trail (dep_aux_heap hp) (thread_trail id2 p2'') /\
  extends_strict p2'' (path_cons (path_recv n1 id1 p1) p2)) \/
(exists p1'',
  In trail (dep_aux_heap hp) (thread_trail id1 p1'') /\
  extends_strict p1'' (path_cons (path_send n1 id2 p2) p1))) as p''.
eapply helper. eapply committable_implies_consistent.
unfold committable. split. esplit. split. apply hp_in_after. auto. 
split. eauto. split. auto. auto. left. auto.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p in_DA | p false ]. 
auto. inversion false.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_recv n1 id1 p1) p2) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id2 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_recv n1 id1 p1) p2)).
eapply unique_thread. apply in_after. right. right.
subst p'. auto.
assert (~ extends (path_cons (path_recv n1 id1 p1) p2) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.
destruct p'' as [ p'' [ in_before strict ] ].
assert (extends (path_cons (path_send n1 id2 p2) p1) p'').
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id1 p') /\ extends p' p'') as p'.
eapply parent_thread. left. apply t_in_c. unfold dep. left. auto.
destruct p' as [ p' [ in_after p'_extends_p'' ] ].
assert (p' = (path_cons (path_send n1 id2 p2) p1)).
eapply unique_thread. apply in_after. right. left.
subst p'. auto.
assert (~ extends (path_cons (path_send n1 id2 p2) p1) p'').
apply extends_strict_implies_reverse_not_extends. auto.
tauto.

(* Case: p1 extends p. Take p1. *)
exists p2. split. right. right. auto.

(* Case: t is not in c *)
inversion t_not_in_c.

(* Case: t is <id1, p1> *)
subst t t_. clear t_not_in_c.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p_ in_DA | p_ p_is_t ].

(* Case: <id, p> in DA(<id1, p1> *)
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id p') /\ extends p' p) as p'.
eapply parent_thread. right. left. unfold dep. simpl. left. left. left. auto.
destruct p' as [ p' [ in_after p'_extends ] ].
inversion in_after as [ t' t'_in_c | t' [ | ] ]; clear in_after.

(* Case: <id, p'> in c *)
exists p'. split. left. auto. auto.

(* Case: <id, p'> = <id1, Send(<id2, p2>):p1> *)
subst id p'.
exists p1. split. right. left.
eapply consistent_implies_dep_extends.
eapply consistent_implies_extends_consistent.
eapply committable_implies_consistent.
unfold committable. apply comm_after.
right. left. apply extends_cons. apply extends_refl.
unfold dep. auto.

(* Case: <id, p'> = <id2, Recv(<id1, p1>):p2> *)
subst id p'. subst p_.
inversion p'_extends as [ p2' | elt p2_ p_ p2_extends ].

(* Case: p = p' = Recv(<id1, p1>):p2 *)
subst p2' p.
assert (exists p1',
  In trail (dep (thread_trail id1 p1)) (thread_trail id1 p1') /\
  extends p1' (path_cons (path_send n1 id2 p2) p1)) as p1'.
apply recv_in_dep_implies_send_in_dep.
unfold dep. left. auto.
destruct p1' as [ p1' [ p1'_in_dep p1'_extends ] ].
assert (extends p1 p1') as p1_extends.
eapply consistent_implies_dep_extends.
eapply consistent_implies_extends_consistent.
eapply committable_implies_consistent.
unfold committable. apply comm_after.
right. left. apply extends_cons. apply extends_refl. auto.
assert (~ extends p1 p1').
assert (extends_strict p1' p1).
eapply extends_cons_implies_extends_strict. apply p1'_extends.
apply extends_strict_implies_reverse_not_extends. auto. tauto.

(* Case: p2 extends p *)
exists p2. split. right. right. auto.

(* Case: p is p1 *)
inversion p_is_t as [ eq ]. subst id p.
exists p1. split. right. left. apply extends_refl.

(* Case: t is <id2, p2> *)
subst t t_. clear t_not_in_c.
unfold dep in p_in_dep_t. inversion p_in_dep_t as [ p_ in_DA | p_ p_is_t ].

(* Case: <id, p> in DA(<id2, p2> *)
assert (exists p',
  In trail (Union trail c
    (Couple trail
      (thread_trail id1 (path_cons (path_send n1 id2 p2) p1))
      (thread_trail id2 (path_cons (path_recv n1 id1 p1) p2))))
  (thread_trail id p') /\ extends p' p) as p'.
eapply parent_thread. right. right. unfold dep. simpl. left. left. left. auto.
destruct p' as [ p' [ in_after p'_extends ] ].
inversion in_after as [ t' t'_in_c | t' [ | ] ]; clear in_after.

(* Case: <id, p'> in c *)
exists p'. split. left. auto. auto.

(* Case: <id, p'> = <id2, Send(<id2, p2>):p1> *)
subst id p'. subst p_.
inversion p'_extends as [ p1' | elt p1_ p_ p1_extends ].

(* Case: p = p' = Recv(<id1, p1>):p2 *)
subst p1' p.
assert (exists p2',
  In trail (dep (thread_trail id2 p2)) (thread_trail id2 p2') /\
  extends p2' (path_cons (path_recv n1 id1 p1) p2)) as p2'.
apply send_in_dep_implies_recv_in_dep.
unfold dep. left. auto.
destruct p2' as [ p2' [ p2'_in_dep p2'_extends ] ].
assert (extends p2 p2') as p2_extends.
eapply consistent_implies_dep_extends.
eapply consistent_implies_extends_consistent.
eapply committable_implies_consistent.
unfold committable. apply comm_after.
right. right. apply extends_cons. apply extends_refl. auto.
assert (~ extends p2 p2').
assert (extends_strict p2' p2).
eapply extends_cons_implies_extends_strict. apply p2'_extends.
apply extends_strict_implies_reverse_not_extends. auto. tauto.

(* Case: p1 extends p *)
exists p1. split. right. left. auto.

(* Case: <id, p'> = <id2, Recv(<id1, p1>):p2> *)
subst id p'.
exists p2. split. right. right.
eapply consistent_implies_dep_extends.
eapply consistent_implies_extends_consistent.
eapply committable_implies_consistent.
unfold committable. apply comm_after.
right. right. apply extends_cons. apply extends_refl.
unfold dep. auto.

(* Case: p is p2 *)
inversion p_is_t. subst id p.
exists p2. split. right. right. apply extends_refl.
Unfocus.

(* Parent heap *)
intros t hp' t_in_before hp'_in_dep_t.
exists hp. split.
left. auto.
destruct t_in_before as [ t t_in_c | t other ].
eapply hp_parent. left. apply t_in_c. auto.
inversion other.
subst t. unfold dep, dep_aux in hp'_in_dep_t.
inversion hp'_in_dep_t as [ | ht false ].
eapply hp_parent. right. left.
unfold dep. left. simpl. left. left. auto.
inversion false.
subst t. unfold dep, dep_aux in hp'_in_dep_t.
inversion hp'_in_dep_t as [ | ht false ].
eapply hp_parent. right. right.
unfold dep. left. simpl. left. left. auto.
inversion false.

Qed.
