Require Import consistent.
Require Import Coq.Logic.Classical.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import send_in_dep_implies_recv_in_dep.
Require Import syntax.
Require Import trails.

Lemma send_helper :
  forall n_s id_s p_s id_r p_r,
    (forall p id,
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      id <> id_s -> id <> id_r ->
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r)))
    /\ (forall hp,
      consistent (heap_trail hp) ->
      In trail (dep_aux_heap hp)
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      (exists p_s'',
        In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
      (exists p_r'',
        In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))).

intros n_s id_s p_s id_r p_r.

apply trail_induction with
(P := fun elt =>
  match elt with
    | path_left => True
    | path_right => True
    | (path_send n id p) => 
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      id <> id_s -> id <> id_r ->
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))
    | (path_recv n id p) =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      id <> id_s -> id <> id_r ->
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))
    | (path_heap n hp) => 
      consistent (heap_trail hp) ->
      In trail (dep_aux_heap hp)
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      (exists p_s'',
        In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
      (exists p_r'',
        In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))
  end)
(P0 := fun elt =>
  match elt with
    | (heap_path_thread n id p) =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      id <> id_s -> id <> id_r ->
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))
  end)
(P1 := fun p =>
  forall id,
    consistent (thread_trail id p) ->
    In trail (dep_aux_thread id p)
    (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
    id <> id_s -> id <> id_r ->
    (exists p_s'',
      In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
      extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
    (exists p_r'',
      In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
      extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r)))
(P2 := fun hp =>
  consistent (heap_trail hp) ->
  In trail (dep_aux_heap hp)
  (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
  (exists p_s'',
    In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
    extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
  (exists p_r'',
    In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
    extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r)));
auto.

intros id consis false. simpl in false. inversion false.

intros elt IH_elt p IH id consis in_DA not_s not_r.

induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left and Right *)
Focus.
simpl. apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
Unfocus. Focus.
simpl. apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
Unfocus.

(* Send *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_s' [ t_s in_DA_t | t_s in_DA_t_x ] | t_s is_t_x ].
assert ((exists p_s'',
  In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_s \/ id_x <> id_s) as cmp_x_s.
apply classic. destruct cmp_x_s. subst id_x.
(* By Lemma B, px \extends Send(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons (path_send n_s id_r p_r) p_s))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Recv(<t,p'>):px \extends(strictly) Send(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_recv n_x id p) p_x)
(path_cons (path_send n_s id_r p_r) p_s)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Recv(<t,p'>):px*)
left. exists (path_cons (path_recv n_x id p) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.
(*   Case tx = tr:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_r.
apply classic. destruct cmp_x_r.
(*   Note that <tr,Recv(<ts,ps>):pr> \in Dep(<ts,Send(<tr,pr>):ps>).
     Hence, <tr,Recv(<ts,ps>):pr> \in Dep(<tx,px>).*)
assert (exists p_r',
  In trail (dep (thread_trail id_x p_x)) 
  (thread_trail id_r p_r')
  /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r))
as p_r'.
apply send_in_dep_implies_recv_in_dep.
unfold dep. left. auto.
destruct p_r' as [ p_r' [ p_r'_in_dep_x p_r'_extends ] ].
(*     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*  By Lemma B, px \extends Recv(<ts,ps>):pr. *)
subst id_r. assert (extends p_x p_r') as p_x_extends_p_r'.
eapply consistent_implies_dep_extends. apply consis_x. auto.
(*     Hence, Recv(<t,p'>):px \extends(strictly) Recv(<ts,ps>):pr.*)
assert (extends_strict (path_cons (path_recv n_x id p) p_x) p_r') as strict.
eapply extends_strict_cons. reflexivity. auto.
(*     Take pr' = Recv(<t,p'>):px. *)
right. exists (path_cons (path_recv n_x id p) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity.
eapply extends_trans. eauto. auto.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_s'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
right. exists p''. split. simpl. left. right. auto. auto.

inversion is_t_x.
Unfocus.

(* Recv *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_s' [ t_s in_DA_t | t_s in_DA_t_x ] | t_s is_t_x ].
assert ((exists p_s'',
  In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_s \/ id_x <> id_s) as cmp_x_s.
apply classic. destruct cmp_x_s. subst id_x.
(* By Lemma B, px \extends Send(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons (path_send n_s id_r p_r) p_s))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Recv(<t,p'>):px \extends(strictly) Send(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_send n_x id p) p_x)
(path_cons (path_send n_s id_r p_r) p_s)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Recv(<t,p'>):px*)
left. exists (path_cons (path_send n_x id p) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.
(*   Case tx = tr:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_r.
apply classic. destruct cmp_x_r.
(*   Note that <tr,Recv(<ts,ps>):pr> \in Dep(<ts,Send(<tr,pr>):ps>).
     Hence, <tr,Recv(<ts,ps>):pr> \in Dep(<tx,px>).*)
assert (exists p_r',
  In trail (dep (thread_trail id_x p_x)) 
  (thread_trail id_r p_r')
  /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r))
as p_r'.
apply send_in_dep_implies_recv_in_dep.
unfold dep. left. auto.
destruct p_r' as [ p_r' [ p_r'_in_dep_x p_r'_extends ] ].
(*     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*  By Lemma B, px \extends Recv(<ts,ps>):pr. *)
subst id_r. assert (extends p_x p_r') as p_x_extends_p_r'.
eapply consistent_implies_dep_extends. apply consis_x. auto.
(*     Hence, Recv(<t,p'>):px \extends(strictly) Recv(<ts,ps>):pr.*)
assert (extends_strict (path_cons (path_send n_x id p) p_x) p_r') as strict.
eapply extends_strict_cons. reflexivity. auto.
(*     Take pr' = Recv(<t,p'>):px. *)
right. exists (path_cons (path_send n_x id p) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity.
eapply extends_trans. eauto. auto.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_s'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
right. exists p''. split. simpl. left. right. auto. auto.

inversion is_t_x. subst id_s id_r p_r p_s.
tauto.
Unfocus.

(* Heap *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_s' [ t_s in_DA_t | t_s in_DA_t_x ] | t_s false ].
assert ((exists p_s'',
  In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (heap_trail hp_x)) as consis_x.
eapply consistent_implies_heap_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply heap_extends_cons. apply heap_extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_s'',
  In trail (dep_aux_heap hp_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_heap hp_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
right. exists p''. split. simpl. left. right. auto. auto.

inversion false.
Unfocus.

(* heap path nil *)
Focus.
intros consis false. simpl in false. inversion false.
Unfocus.

(* heap path *)
intros elt IH_elt hp IH consis in_DA.

simpl in in_DA. induction elt as [ n_x id_x p_x ].
inversion in_DA as [ t_s' [ t_s in_DA_t_x | t_s in_DA_t ] | t_s is_t_x ].

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Send(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_s \/ id_x <> id_s) as cmp_x_s.
apply classic. destruct cmp_x_s. subst id_x.
(* By Lemma B, px \extends Send(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons (path_send n_s id_r p_r) p_s))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Recv(<t,p'>):px \extends(strictly) Send(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_heap n_x hp) p_x)
(path_cons (path_send n_s id_r p_r) p_s)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Recv(<t,p'>):px*)
left. exists (path_cons (path_heap n_x hp) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.
(*   Case tx = tr:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_r.
apply classic. destruct cmp_x_r.
(*   Note that <tr,Recv(<ts,ps>):pr> \in Dep(<ts,Send(<tr,pr>):ps>).
     Hence, <tr,Recv(<ts,ps>):pr> \in Dep(<tx,px>).*)
assert (exists p_r',
  In trail (dep (thread_trail id_x p_x)) 
  (thread_trail id_r p_r')
  /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r))
as p_r'.
apply send_in_dep_implies_recv_in_dep.
unfold dep. left. auto.
destruct p_r' as [ p_r' [ p_r'_in_dep_x p_r'_extends ] ].
(*     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*  By Lemma B, px \extends Recv(<ts,ps>):pr. *)
subst id_r. assert (extends p_x p_r') as p_x_extends_p_r'.
eapply consistent_implies_dep_extends. apply consis_x. auto.
(*     Hence, Recv(<t,p'>):px \extends(strictly) Recv(<ts,ps>):pr.*)
assert (extends_strict (path_cons (path_heap n_x hp) p_x) p_r') as strict.
eapply extends_strict_cons. reflexivity. auto.
(*     Take pr' = Recv(<t,p'>):px. *)
right. exists (path_cons (path_heap n_x hp) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity.
eapply extends_trans. eauto. auto.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Send(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Send(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Recv(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_s'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
right. exists p''. split. simpl. left. left. auto. auto.

assert ((exists p_s'',
  In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_s id_r p_r) p_s)) \/
(exists p_r'',
  In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_s id_s p_s) p_r))) as p''.
apply IH. eapply consistent_implies_heap_extends_consistent.
apply consis. apply heap_extends_cons. apply heap_extends_refl. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. right. auto. auto.

inversion is_t_x. 

Qed.

Lemma recv_helper :
  forall n_r id_r p_r id_s p_s,
    (forall p id,
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      id <> id_r -> id <> id_s ->
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s)))
    /\ (forall hp,
      consistent (heap_trail hp) ->
      In trail (dep_aux_heap hp)
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      (exists p_r'',
        In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
      (exists p_s'',
        In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))).

intros n_r id_r p_r id_s p_s.

apply trail_induction with
(P := fun elt =>
  match elt with
    | path_left => True
    | path_right => True
    | (path_send n id p) =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      id <> id_r -> id <> id_s ->
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))
    | (path_recv n id p) => 
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      id <> id_r -> id <> id_s ->
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))
    | (path_heap n hp) => 
      consistent (heap_trail hp) ->
      In trail (dep_aux_heap hp)
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      (exists p_r'',
        In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
      (exists p_s'',
        In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))
  end)
(P0 := fun elt =>
  match elt with
    | (heap_path_thread n id p) =>
      consistent (thread_trail id p) ->
      In trail (dep_aux_thread id p)
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      id <> id_r -> id <> id_s ->
      (exists p_r'',
        In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
        extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
      (exists p_s'',
        In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
        extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))
  end)
(P1 := fun p =>
  forall id,
    consistent (thread_trail id p) ->
    In trail (dep_aux_thread id p)
    (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
    id <> id_r -> id <> id_s ->
    (exists p_r'',
      In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
      extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
    (exists p_s'',
      In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
      extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s)))
(P2 := fun hp =>
  consistent (heap_trail hp) ->
  In trail (dep_aux_heap hp)
  (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
  (exists p_r'',
    In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
    extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
  (exists p_s'',
    In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
    extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s)));
auto.

intros id consis false. simpl in false. inversion false.

intros elt IH_elt p IH id consis in_DA not_r not_s.

induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left and Right *)
Focus.
simpl. apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
Unfocus. Focus.
simpl. apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
Unfocus.

(* Send *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_r' [ t_r in_DA_t | t_r in_DA_t_x ] | t_r is_t_x ].
assert ((exists p_r'',
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Recv(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_r.
apply classic. destruct cmp_x_r. subst id_x.
(* By Lemma B, px \extends Recv(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons (path_recv n_r id_s p_s) p_r))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Send(<t,p'>):px \extends(strictly) Recv(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_recv n_x id p) p_x)
(path_cons (path_recv n_r id_s p_s) p_r)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Send(<t,p'>):px*)
left. exists (path_cons (path_recv n_x id p) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.
(*   Case tx = tr:*)
assert (id_x = id_s \/ id_x <> id_s) as cmp_x_s.
apply classic. destruct cmp_x_s.
(*   Note that <tr,Send(<ts,ps>):pr> \in Dep(<ts,Recv(<tr,pr>):ps>).
     Hence, <tr,Send(<ts,ps>):pr> \in Dep(<tx,px>).*)
assert (exists p_s',
  In trail (dep (thread_trail id_x p_x)) 
  (thread_trail id_s p_s')
  /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s))
as p_s'.
apply recv_in_dep_implies_send_in_dep.
unfold dep. left. auto.
destruct p_s' as [ p_s' [ p_s'_in_dep_x p_s'_extends ] ].
(*     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*  By Lemma B, px \extends Send(<ts,ps>):pr. *)
subst id_s. assert (extends p_x p_s') as p_x_extends_p_s'.
eapply consistent_implies_dep_extends. apply consis_x. auto.
(*     Hence, Send(<t,p'>):px \extends(strictly) Send(<ts,ps>):pr.*)
assert (extends_strict (path_cons (path_recv n_x id p) p_x) p_s') as strict.
eapply extends_strict_cons. reflexivity. auto.
(*     Take pr' = Send(<t,p'>):px. *)
right. exists (path_cons (path_recv n_x id p) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity.
eapply extends_trans. eauto. auto.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Recv(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Recv(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Send(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_r'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
right. exists p''. split. simpl. left. right. auto. auto.

inversion is_t_x. subst id_r id_s p_s p_r.
tauto.
Unfocus.

(* Recv *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_r' [ t_r in_DA_t | t_r in_DA_t_x ] | t_r is_t_x ].
assert ((exists p_r'',
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Recv(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_r.
apply classic. destruct cmp_x_r. subst id_x.
(* By Lemma B, px \extends Recv(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons (path_recv n_r id_s p_s) p_r))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Send(<t,p'>):px \extends(strictly) Recv(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_send n_x id p) p_x)
(path_cons (path_recv n_r id_s p_s) p_r)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Send(<t,p'>):px*)
left. exists (path_cons (path_send n_x id p) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.
(*   Case tx = tr:*)
assert (id_x = id_s \/ id_x <> id_s) as cmp_x_s.
apply classic. destruct cmp_x_s.
(*   Note that <tr,Send(<ts,ps>):pr> \in Dep(<ts,Recv(<tr,pr>):ps>).
     Hence, <tr,Send(<ts,ps>):pr> \in Dep(<tx,px>).*)
assert (exists p_s',
  In trail (dep (thread_trail id_x p_x)) 
  (thread_trail id_s p_s')
  /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s))
as p_s'.
apply recv_in_dep_implies_send_in_dep.
unfold dep. left. auto.
destruct p_s' as [ p_s' [ p_s'_in_dep_x p_s'_extends ] ].
(*     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*  By Lemma B, px \extends Send(<ts,ps>):pr. *)
subst id_s. assert (extends p_x p_s') as p_x_extends_p_s'.
eapply consistent_implies_dep_extends. apply consis_x. auto.
(*     Hence, Send(<t,p'>):px \extends(strictly) Send(<ts,ps>):pr.*)
assert (extends_strict (path_cons (path_send n_x id p) p_x) p_s') as strict.
eapply extends_strict_cons. reflexivity. auto.
(*     Take pr' = Send(<t,p'>):px. *)
right. exists (path_cons (path_send n_x id p) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity.
eapply extends_trans. eauto. auto.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Recv(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Recv(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Send(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_r'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
right. exists p''. split. simpl. left. right. auto. auto.

inversion is_t_x.
Unfocus.

(* Heap *)
Focus.
simpl in in_DA.
inversion in_DA as [ t_r' [ t_r in_DA_t | t_r in_DA_t_x ] | t_r false ].
assert ((exists p_r'',
  In trail (dep_aux_thread id p) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_thread id p) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH. eapply consistent_implies_extends_consistent.
apply consis. apply extends_cons. apply extends_refl. auto. auto. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. left. auto. auto.

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Recv(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (heap_trail hp_x)) as consis_x.
eapply consistent_implies_heap_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply heap_extends_cons. apply heap_extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Recv(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Recv(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Send(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_r'',
  In trail (dep_aux_heap hp_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_heap hp_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
right. exists p''. split. simpl. left. right. auto. auto.

inversion false.
Unfocus.

(* heap path nil *)
Focus.
intros consis false. simpl in false. inversion false.
Unfocus.

(* heap path *)
intros elt IH_elt hp IH consis in_DA.

simpl in in_DA. induction elt as [ n_x id_x p_x ].
inversion in_DA as [ t_r' [ t_r in_DA_t_x | t_r in_DA_t ] | t_r is_t_x ].

(** Exciting cases! *)
(** From Matthew's email:
Case <ts,Recv(<tr,pr>):ps> \in Dep(<tx,px>).
   Case tx = ts:*)
assert (id_x = id_r \/ id_x <> id_r) as cmp_x_r.
apply classic. destruct cmp_x_r. subst id_x.
(* By Lemma B, px \extends Recv(<tr,pr>):ps.*)
assert (extends_strict p_x (path_cons (path_recv n_r id_s p_s) p_r))
  as p_x_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl. apply in_DA_t_x.
(* Hence, Send(<t,p'>):px \extends(strictly) Recv(<tr,pr>):ps.*)
assert (extends_strict (path_cons (path_heap n_x hp) p_x)
(path_cons (path_recv n_r id_s p_s) p_r)) as p_x'_extends.
eapply extends_strict_cons. reflexivity.
apply extends_strict_implies_extends. apply p_x_extends.
(* Take ps' = Send(<t,p'>):px*)
left. exists (path_cons (path_heap n_x hp) p_x).
split. simpl. right. apply In_singleton. apply p_x'_extends.
(*   Case tx = tr:*)
assert (id_x = id_s \/ id_x <> id_s) as cmp_x_s.
apply classic. destruct cmp_x_s.
(*   Note that <tr,Send(<ts,ps>):pr> \in Dep(<ts,Recv(<tr,pr>):ps>).
     Hence, <tr,Send(<ts,ps>):pr> \in Dep(<tx,px>).*)
assert (exists p_s',
  In trail (dep (thread_trail id_x p_x)) 
  (thread_trail id_s p_s')
  /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s))
as p_s'.
apply recv_in_dep_implies_send_in_dep.
unfold dep. left. auto.
destruct p_s' as [ p_s' [ p_s'_in_dep_x p_s'_extends ] ].
(*     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*  By Lemma B, px \extends Send(<ts,ps>):pr. *)
subst id_s. assert (extends p_x p_s') as p_x_extends_p_s'.
eapply consistent_implies_dep_extends. apply consis_x. auto.
(*     Hence, Send(<t,p'>):px \extends(strictly) Send(<ts,ps>):pr.*)
assert (extends_strict (path_cons (path_heap n_x hp) p_x) p_s') as strict.
eapply extends_strict_cons. reflexivity. auto.
(*     Take pr' = Send(<t,p'>):px. *)
right. exists (path_cons (path_heap n_x hp) p_x).
split. simpl. right. apply In_singleton.
eapply extends_strict_cons. reflexivity.
eapply extends_trans. eauto. auto.

(*   Case tx != ts and tx != tr:
     By Lemma A, Consistent(<t,p>) implies Consistent(<tx,px>).*)
assert (consistent (thread_trail id_x p_x)) as consis_x.
eapply consistent_implies_extends_consistent.
eapply consistent_implies_dep_consistent.
apply consis. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. apply extends_refl.
(*     Then apply induction hypothesis.
     [Note: px structurally smaller than Recv(<tx,px>):p'.]
     There exists ps'' such that <ts,ps''> \in Dep(<tx,px>)
                             and ps'' \extends(strictly) Recv(<tr,pr>):ps
      OR exists pr'' such that <tr,pr''> \in Dep(<tx,px>)
                           and pr'' \extends(strictly) Send(<ts,ps>):pr.
     Take ps' = ps'' OR pr' = pr'' (as appropriate).*)
assert ((exists p_r'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_thread id_x p_x) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH_elt. auto. auto. auto. auto.
destruct p'' as [ p'' | p'' ];
destruct p'' as [ p'' [ p''_in_DA_x strict ] ].
left. exists p''. split. simpl. left. left. auto. auto.
right. exists p''. split. simpl. left. left. auto. auto.

assert ((exists p_r'',
  In trail (dep_aux_heap hp) (thread_trail id_r p_r'') /\
  extends_strict p_r'' (path_cons (path_recv n_r id_s p_s) p_r)) \/
(exists p_s'',
  In trail (dep_aux_heap hp) (thread_trail id_s p_s'') /\
  extends_strict p_s'' (path_cons (path_send n_r id_r p_r) p_s))) as p''.
apply IH. eapply consistent_implies_heap_extends_consistent.
apply consis. apply heap_extends_cons. apply heap_extends_refl. auto.
destruct p'' as [ p'' | p'' ].
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
left. exists p''. split. simpl. left. right. auto. auto.
destruct p'' as [ p'' [ p''_in_DA_t strict ] ].
right. exists p''. split. simpl. left. right. auto. auto.

inversion is_t_x. 

Qed.
