Require Import Coq.Arith.Arith.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import dep_dep_implies_extends.
Require Import trails.

Definition consistent (t:trail) : Prop :=
  (forall id p1 p2,
    In trail (dep t) (thread_trail id p1) ->
    In trail (dep t) (thread_trail id p2) ->
    extends p1 p2 \/ extends p2 p1)
  /\ (forall hp1 hp2,
    In trail (dep t) (heap_trail hp1) ->
    In trail (dep t) (heap_trail hp2) ->
    heap_extends hp1 hp2 \/ heap_extends hp2 hp1).

Lemma consistent_implies_extends_consistent :
  forall id p p',
    consistent (thread_trail id p) ->
    extends p p' ->
    consistent (thread_trail id p').

unfold consistent, dep.
intros id p p' [ consis_thread consis_heap ] p_p'_extends.
split.

intros id1 p1 p2 p1_in_dep p2_in_dep.
assert (Included trail (dep_aux_thread id p') (dep_aux_thread id p))
  as included.
apply extends_implies_dep_aux_thread_included. auto.
unfold Included in included.

inversion p1_in_dep as [ t1 t1_in_DA | t1 t1_is_t' ];
inversion p2_in_dep as [ t2 t2_in_DA | t2 t2_is_t' ].

eapply consis_thread.
left. apply included. eauto.
left. apply included. eauto.

inversion t2_is_t'. subst id1 p2.
assert (extends p1 p \/ extends p p1) as p_extends.
eapply consis_thread. left. apply included. eauto.
right. apply In_singleton.
destruct p_extends as [ p_extends | p_extends ].
left. eapply extends_trans. eauto. auto.
eapply common_parent_implies_extends.
eauto. eauto.

inversion t1_is_t'. subst id1 p1.
assert (extends p2 p \/ extends p p2) as p_extends.
eapply consis_thread. left. apply included. eauto.
right. apply In_singleton.
destruct p_extends as [ p_extends | p_extends ].
right. eapply extends_trans. eauto. auto.
eapply common_parent_implies_extends.
eauto. eauto.

inversion t1_is_t'. inversion t2_is_t'.
subst p1 p2. left. apply extends_refl.

intros hp1 hp2 hp1_in_dep hp2_in_dep.
inversion hp1_in_dep as [ t1 t1_in_DA | t1 false1 ];
inversion hp2_in_dep as [ t2 t2_in_DA | t2 false2 ].
subst t1 t2.
assert (Included trail (dep_aux_thread id p') (dep_aux_thread id p))
  as included.
apply extends_implies_dep_aux_thread_included. auto.
unfold Included in included.
apply consis_heap.
left. auto. left. auto.
inversion false2. inversion false1. inversion false1.

Qed.

Lemma consistent_implies_heap_extends_consistent :
  forall hp hp',
    consistent (heap_trail hp) ->
    heap_extends hp hp' ->
    consistent (heap_trail hp').

unfold consistent, dep.
intros hp hp' [ consis_thread consis_heap ] hp_hp'_extends.

assert (Included trail (dep_aux_heap hp') (dep_aux_heap hp))
  as included.
apply heap_extends_implies_dep_aux_heap_included. auto.
unfold Included in included.

split.

intros id1 p1 p2 p1_in_dep p2_in_dep.

inversion p1_in_dep as [ t1 t1_in_DA | t1 false1 ];
inversion p2_in_dep as [ t2 t2_in_DA | t2 false2 ].

eapply consis_thread.
left. apply included. eauto.
left. apply included. eauto.

inversion false2. inversion false1. inversion false1.

unfold dep_aux.
intros hp1 hp2 hp1_in_dep hp2_in_dep.
inversion hp1_in_dep as [ t1 t1_in_DA | t1 t1_is_t' ];
inversion hp2_in_dep as [ t2 t2_in_DA | t2 t2_is_t' ];
subst t1 t2.

eapply consis_heap.
left. apply included. eauto.
left. apply included. eauto.

inversion t2_is_t'. subst hp2.
assert (heap_extends hp1 hp \/ heap_extends hp hp1) as hp_extends.
eapply consis_heap. left. apply included. eauto.
right. apply In_singleton.
destruct hp_extends as [ hp_extends | hp_extends ].
left. eapply heap_extends_trans. eauto. auto.
eapply common_parent_implies_heap_extends.
eauto. eauto.

inversion t1_is_t'. subst hp1.
assert (heap_extends hp2 hp \/ heap_extends hp hp2) as hp_extends.
eapply consis_heap. left. apply included. eauto.
right. apply In_singleton.
destruct hp_extends as [ hp_extends | hp_extends ].
right. eapply heap_extends_trans. eauto. auto.
eapply common_parent_implies_heap_extends.
eauto. eauto.

inversion t1_is_t'. inversion t2_is_t'.
subst hp1 hp2. left. apply heap_extends_refl.

Qed.

Lemma consistent_implies_dep_consistent : 
  forall t t',
    consistent t ->
    In trail (dep t) t' ->
    consistent t'.

unfold consistent.
intros t1 t2 [ consis_thread consis_heap ] t2_in_dep.
split.
intros id3 p3 p4 t3_in_dep t4_in_dep.
assert (exists p3', 
  In trail (dep t1) (thread_trail id3 p3') /\
  extends p3' p3) as p3'.
eapply dep_dep_implies_extends. apply t2_in_dep. auto.
assert (exists p4', 
  In trail (dep t1) (thread_trail id3 p4') /\
  extends p4' p4) as p4'.
eapply dep_dep_implies_extends. apply t2_in_dep. auto.
destruct p3' as [ p3' [ p3'_in_dep p3'_extends ] ].
destruct p4' as [ p4' [ p4'_in_dep p4'_extends ] ].
assert (extends p3' p4' \/ extends p4' p3') as p3'_p4'_consis.
eapply consis_thread. eauto. auto.
destruct p3'_p4'_consis as [ p3'_extends_p4' | p4'_extends_p3' ].
eapply common_parent_implies_extends.
apply p3'_extends. eapply extends_trans. apply p3'_extends_p4'. auto.
eapply common_parent_implies_extends.
eapply extends_trans. apply p4'_extends_p3'. auto. auto.

intros hp3 hp4 t3_in_dep t4_in_dep.
assert (exists hp3', 
  In trail (dep t1) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'.
eapply dep_dep_implies_heap_extends. apply t2_in_dep. auto.
assert (exists hp4', 
  In trail (dep t1) (heap_trail hp4') /\
  heap_extends hp4' hp4) as hp4'.
eapply dep_dep_implies_heap_extends. apply t2_in_dep. auto.
destruct hp3' as [ hp3' [ hp3'_in_dep hp3'_extends ] ].
destruct hp4' as [ hp4' [ hp4'_in_dep hp4'_extends ] ].
assert (heap_extends hp3' hp4' \/ heap_extends hp4' hp3') as hp3'_hp4'_consis.
eapply consis_heap. eauto. auto.
destruct hp3'_hp4'_consis as [ hp3'_extends_hp4' | hp4'_extends_hp3' ].
eapply common_parent_implies_heap_extends.
apply hp3'_extends. eapply heap_extends_trans. apply hp3'_extends_hp4'. auto.
eapply common_parent_implies_heap_extends.
eapply heap_extends_trans. apply hp4'_extends_hp3'. auto. auto.

Qed.

Lemma consistent_implies_dep_aux_thread_elt_count_gt :
  forall id p p',
    consistent (thread_trail id p) ->
    In trail (dep_aux_thread id p) (thread_trail id p') ->
    elt_count p > elt_count p'.

intros id1 p1 p2 consis in_DA.

induction p1 as [ | elt p1' ].
simpl in in_DA. inversion in_DA.

clear IHp1'.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ];
simpl; simpl in in_DA.

assert (elt_count p1' >= elt_count p2).
assert (trail_elt_count (thread_trail id1 p1') >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. eauto.
simpl in rewriting. auto. auto with arith.

assert (elt_count p1' >= elt_count p2).
assert (trail_elt_count (thread_trail id1 p1') >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. eauto.
simpl in rewriting. auto. auto with arith.

inversion in_DA as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_recv ].
assert (elt_count p1' >= elt_count p2).
assert (trail_elt_count (thread_trail id1 p1') >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. eauto.
simpl in rewriting. auto. auto with arith.
assert (elt_count p_x >= elt_count p2).
assert (trail_elt_count (thread_trail id_x p_x) >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. auto.
simpl in rewriting. auto. rewrite plus_comm. auto with arith.
inversion is_recv. subst p2 id_x.
remember (path_cons (path_send n_x id1 p_x) p1') as send.
remember (path_cons (path_recv n_x id1 p1') p_x) as recv.
assert (extends send recv \/ extends recv send) as false_extends.
unfold consistent in consis. 
destruct consis as [ consis_thread consis_heap ]. eapply consis_thread.
unfold dep. right. apply In_singleton.
unfold dep. subst send. simpl. left. right. subst recv. apply In_singleton.
assert (elt_count send = elt_count recv) as elt_count_eq.
subst send recv. simpl. auto using plus_comm.
destruct false_extends as [ false_extends | false_extends ].
inversion false_extends as [ p p_eq1 p_eq2 
  | elt send' recv_ extends' send_eq ].
subst send recv. inversion p_eq2.
subst recv_ send recv. inversion send_eq. subst elt send'.
assert (elt_count p1' >= 
  elt_count (path_cons (path_recv n_x id1 p1') p_x)) as ge1.
apply extends_implies_elt_count_ge. auto. simpl in ge1.
simpl in elt_count_eq.
assert (S (elt_count p1' + elt_count p_x) > elt_count p1').
auto with arith.
assert (S (elt_count p1' + elt_count p_x) >
  S (elt_count p_x + elt_count p1')) as contra.
auto with arith. tauto.
inversion false_extends as [ p p_eq1 p_eq2 
  | elt recv' send_ extends' recv_eq ].
subst recv send. inversion p_eq2.
subst send_ recv send. inversion recv_eq. subst elt recv'.
assert (elt_count p_x >= 
  elt_count (path_cons (path_send n_x id1 p_x) p1')) as ge1.
apply extends_implies_elt_count_ge. auto. simpl in ge1.
simpl in elt_count_eq.
assert (S (elt_count p1' + elt_count p_x) > elt_count p_x).
auto with arith.
assert (S (elt_count p1' + elt_count p_x) >
  S (elt_count p1' + elt_count p_x)).
eapply gt_le_trans. eauto. unfold ge in ge1. auto.
assert ( ~ S (elt_count p1' + elt_count p_x) > 
  S (elt_count p1' + elt_count p_x)).
apply le_not_gt. auto with arith. tauto.

inversion in_DA as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_send ].
assert (elt_count p1' >= elt_count p2).
assert (trail_elt_count (thread_trail id1 p1') >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. eauto.
simpl in rewriting. auto. auto with arith.
assert (elt_count p_x >= elt_count p2).
assert (trail_elt_count (thread_trail id_x p_x) >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. auto.
simpl in rewriting. auto. rewrite plus_comm. auto with arith.
inversion is_send. subst p2 id_x.
remember (path_cons (path_recv n_x id1 p_x) p1') as recv.
remember (path_cons (path_send n_x id1 p1') p_x) as send.
assert (extends recv send \/ extends send recv) as false_extends.
unfold consistent in consis. 
destruct consis as [ consis_thread consis_heap ]. eapply consis_thread.
unfold dep. right. apply In_singleton.
unfold dep. subst recv. simpl. left. right. subst send. apply In_singleton.
assert (elt_count send = elt_count recv) as elt_count_eq.
subst send recv. simpl. auto using plus_comm.
destruct false_extends as [ false_extends | false_extends ].
inversion false_extends as [ p p_eq1 p_eq2 
  | elt recv' send_ extends' recv_eq ].
subst recv send. inversion p_eq2.
subst send_ recv send. inversion recv_eq. subst elt recv'.
assert (elt_count p1' >= 
  elt_count (path_cons (path_send n_x id1 p1') p_x)) as ge1.
apply extends_implies_elt_count_ge. auto. simpl in ge1.
simpl in elt_count_eq.
assert (S (elt_count p1' + elt_count p_x) > elt_count p1').
auto with arith.
assert (S (elt_count p1' + elt_count p_x) >
  S (elt_count p_x + elt_count p1')) as contra.
auto with arith. tauto.
inversion false_extends as [ p p_eq1 p_eq2 
  | elt send' recv_ extends' send_eq ].
subst send recv. inversion p_eq2.
subst recv_ send recv. inversion send_eq. subst elt send'.
assert (elt_count p_x >= 
  elt_count (path_cons (path_recv n_x id1 p_x) p1')) as ge1.
apply extends_implies_elt_count_ge. auto. simpl in ge1.
simpl in elt_count_eq.
assert (S (elt_count p1' + elt_count p_x) > elt_count p_x).
auto with arith.
assert (S (elt_count p1' + elt_count p_x) >
  S (elt_count p1' + elt_count p_x)).
eapply gt_le_trans. eauto. unfold ge in ge1. auto.
assert ( ~ S (elt_count p1' + elt_count p_x) > 
  S (elt_count p1' + elt_count p_x)).
apply le_not_gt. auto with arith. tauto.

inversion in_DA as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_heap ].
assert (elt_count p1' >= elt_count p2).
assert (trail_elt_count (thread_trail id1 p1') >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. eauto.
simpl in rewriting. auto. auto with arith.
assert (heap_elt_count hp_x >= elt_count p2).
assert (trail_elt_count (heap_trail hp_x) >=
  trail_elt_count (thread_trail id1 p2)) as rewriting.
eapply dep_aux_implies_elt_count_ge. auto.
simpl in rewriting. auto. rewrite plus_comm. auto with arith.
inversion is_heap.

Qed.

Lemma consistent_implies_dep_aux_heap_elt_count_gt :
  forall hp hp',
    consistent (heap_trail hp) ->
    In trail (dep_aux_heap hp) (heap_trail hp') ->
    heap_elt_count hp > heap_elt_count hp'.

intros hp hp' consis in_DA.

induction hp as [ | elt hp ].
inversion in_DA.
induction elt as [ n id p ].
simpl. simpl in in_DA.

inversion in_DA as [ ht' [ ht in_DA_p | ht in_DA_hp ] | ht false ].
assert (trail_elt_count (thread_trail id p) >= 
  trail_elt_count (heap_trail hp')) as rewriting.
eapply dep_aux_implies_elt_count_ge. eauto.
simpl in rewriting. rewrite plus_comm. auto with arith.
assert (heap_elt_count hp > heap_elt_count hp') as rewriting.
apply IHhp. eapply consistent_implies_heap_extends_consistent.
apply consis. apply heap_extends_cons. apply heap_extends_refl. auto.
auto with arith.
inversion false.

Qed.

Lemma consistent_implies_dep_aux_thread_extends_strict :
  forall id p p',
    consistent (thread_trail id p) ->
    In trail (dep_aux_thread id p) (thread_trail id p') ->
    extends_strict p p'.

intros id1 p1 p2 consis in_DA.
assert (extends p1 p2 \/ extends p2 p1) as p1_p2_consis.
unfold consistent in consis. 
destruct consis as [ consis_thread consis_heap ]. eapply consis_thread.
unfold dep. right. apply In_singleton.
unfold dep. left. auto.
assert (elt_count p1 > elt_count p2).
eapply consistent_implies_dep_aux_thread_elt_count_gt.
apply consis. auto.
destruct p1_p2_consis as [ p1_extends | p2_extends ].
inversion p1_extends. subst p1 p2.
assert (~ elt_count p > elt_count p).
apply gt_irrefl. tauto.
eapply extends_strict_cons. reflexivity. auto.
inversion p2_extends as [ p | elt p2' p1_ p2'_extends ]. subst p1 p2.
assert (~ elt_count p > elt_count p).
apply gt_irrefl. tauto.
subst p1_ p2.
assert (elt_count (path_cons elt p2') >= elt_count p1).
apply extends_implies_elt_count_ge. auto.
assert (~ elt_count p1 > elt_count (path_cons elt p2')).
apply le_not_gt. auto with arith. tauto.

Qed.

Lemma consistent_implies_dep_aux_heap_extends_strict :
  forall hp hp',
    consistent (heap_trail hp) ->
    In trail (dep_aux_heap hp) (heap_trail hp') ->
    heap_extends_strict hp hp'.

intros hp1 hp2 consis in_DA.
assert (heap_extends hp1 hp2 \/ heap_extends hp2 hp1) as hp1_hp2_consis.
unfold consistent in consis. 
destruct consis as [ consis_thread consis_heap ]. eapply consis_heap.
unfold dep. right. apply In_singleton.
unfold dep. left. auto.
assert (heap_elt_count hp1 > heap_elt_count hp2).
eapply consistent_implies_dep_aux_heap_elt_count_gt.
apply consis. auto.
destruct hp1_hp2_consis as [ hp1_extends | hp2_extends ].
inversion hp1_extends. subst hp1 hp2.
assert (~ heap_elt_count hp > heap_elt_count hp).
apply gt_irrefl. tauto.
eapply heap_extends_strict_cons. reflexivity. auto.
inversion hp2_extends as [ hp | elt hp2' hp1_ hp2'_extends ]. subst hp1 hp2.
assert (~ heap_elt_count hp > heap_elt_count hp).
apply gt_irrefl. tauto.
subst hp1_ hp2.
assert (heap_elt_count (heap_path_cons elt hp2') >= heap_elt_count hp1).
apply heap_extends_implies_elt_count_ge. auto.
assert (~ heap_elt_count hp1 > heap_elt_count (heap_path_cons elt hp2')).
apply le_not_gt. auto with arith. tauto.

Qed.

Lemma consistent_implies_dep_extends :
  forall id p p',
    consistent (thread_trail id p) ->
    In trail (dep (thread_trail id p)) (thread_trail id p') ->
    extends p p'.

intros id p p' consis in_dep.
unfold dep in in_dep.
inversion in_dep as [ t' t'_in_DA | t' t'_is_t ].
apply extends_strict_implies_extends.
eapply consistent_implies_dep_aux_thread_extends_strict.
apply consis. auto.
inversion t'_is_t. subst p'. apply extends_refl.

Qed.

Lemma consistent_implies_dep_heap_extends :
  forall hp hp',
    consistent (heap_trail hp) ->
    In trail (dep (heap_trail hp)) (heap_trail hp') ->
    heap_extends hp hp'.

intros hp hp' consis in_dep.
unfold dep in in_dep.
inversion in_dep as [ t' t'_in_DA | t' t'_is_t ].
apply heap_extends_strict_implies_extends.
eapply consistent_implies_dep_aux_heap_extends_strict.
apply consis. auto.
inversion t'_is_t. subst hp'. apply heap_extends_refl.

Qed.
