Require Import Coq.Arith.Arith.
Require Import Coq.Sets.Ensembles.
Require Import syntax.
Require Import trails.

Fixpoint dep_aux_thread (id:tid) (p:path) : Ensemble trail :=
  match p with
    | path_nil => Empty_set trail
    | (path_cons elt p') =>
      match elt with
        | path_left => dep_aux_thread id p'
        | path_right => dep_aux_thread id p'
        | (path_send n_x id_r p_r) => 
          (Add trail
            (Union trail 
              (dep_aux_thread id p') 
              (dep_aux_thread id_r p_r))
            (thread_trail id_r (path_cons (path_recv n_x id p') p_r)))
        | (path_recv n_x id_s p_s) =>
          (Add trail
            (Union trail 
              (dep_aux_thread id p') 
              (dep_aux_thread id_s p_s))
            (thread_trail id_s (path_cons (path_send n_x id p') p_s)))
        | (path_heap n_x hp) =>
          (Add trail
            (Union trail 
              (dep_aux_thread id p') 
              (dep_aux_heap hp))
            (heap_trail (heap_path_cons (heap_path_thread n_x id p') hp)))
      end
  end
with dep_aux_heap (hp:heap_path) :=
  match hp with
    | heap_path_nil => Empty_set trail
    | (heap_path_cons elt hp') =>
      match elt with
        | heap_path_thread n_x id p =>
          (Add trail
            (Union trail 
              (dep_aux_thread id p) 
              (dep_aux_heap hp'))
            (thread_trail id (path_cons (path_heap n_x hp') p)))
      end
  end.

Definition dep_aux t : Ensemble trail :=
  match t with 
    | thread_trail id p => dep_aux_thread id p
    | heap_trail hp => dep_aux_heap hp
  end.

Definition dep t : Ensemble trail :=
  Add trail (dep_aux t) t.

Lemma extends_implies_dep_aux_thread_included :
  forall id p p',
    extends p p' ->
    Included trail (dep_aux_thread id p')
    (dep_aux_thread id p).

unfold Included.
intros id p p' ext x in_DA.
induction ext.
auto.
simpl.
induction elt as [ | | | | ].
auto. auto.
left. left. auto.
left. left. auto.
left. left. auto.

Qed.

Lemma heap_extends_implies_dep_aux_heap_included :
  forall hp hp',
    heap_extends hp hp' ->
    Included trail (dep_aux_heap hp')
    (dep_aux_heap hp).

unfold Included.
intros hp hp' ext x in_DA.
induction ext.
auto.
simpl.
induction elt.
left. right. auto.

Qed.

Lemma dep_aux_implies_elt_count_ge :
  forall t1 t2,
    In trail (dep_aux t1) t2 ->
    trail_elt_count t1 >= trail_elt_count t2.

intros t1 t2 in_DA.

cut ((forall p1 id1,
  In trail (dep_aux_thread id1 p1) t2 ->
  elt_count p1 >= trail_elt_count t2)
/\ (forall hp1,
  In trail (dep_aux_heap hp1) t2 ->
  heap_elt_count hp1 >= trail_elt_count t2)).

intros [ H1 H2 ].
induction t1 as [ id p1 | hp1 ];
simpl in in_DA; simpl; eauto.

apply trail_induction with
(P := fun elt =>
  match elt with
    | path_left => True
    | path_right => True
    | path_send n_x id_x p_x => 
      In trail (dep_aux_thread id_x p_x) t2 ->
      elt_count p_x >= trail_elt_count t2
    | path_recv n_x id_x p_x => 
      In trail (dep_aux_thread id_x p_x) t2 ->
      elt_count p_x >= trail_elt_count t2
    | path_heap n_x hp_x => 
      In trail (dep_aux_heap hp_x) t2 ->
      heap_elt_count hp_x >= trail_elt_count t2
  end)
(P0 := fun elt =>
  match elt with
    heap_path_thread n_x id_x p_x =>
    In trail (dep_aux_thread id_x p_x) t2 ->
    elt_count p_x >= trail_elt_count t2
  end)
(P1 := fun p1 =>
  forall id1,
    In trail (dep_aux_thread id1 p1) t2 ->
    elt_count p1 >= trail_elt_count t2)
(P2 := fun hp1 =>
  In trail (dep_aux_heap hp1) t2 ->
  heap_elt_count hp1 >= trail_elt_count t2).

auto. auto. eauto. eauto. auto. eauto.

intros id1 false. simpl in false. inversion false.

intros elt IH_elt p1 IH_path id1 in_DA1.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ];
simpl; simpl in in_DA1.

assert (elt_count p1 >= trail_elt_count t2).
eapply IH_path; eauto. eauto with arith.

assert (elt_count p1 >= trail_elt_count t2).
eapply IH_path; eauto. auto with arith.

destruct in_DA1 as [ t2' [ t2 in_DA1 | t2 in_DAx ] | t2 is_recv ].
assert (elt_count p1 >= trail_elt_count t2).
eapply IH_path; eauto. auto with arith.
unfold ge in IH_elt. unfold ge.
assert (trail_elt_count t2 <= elt_count p_x).
eapply IH_elt; eauto. rewrite <- plus_Sn_m.
rewrite plus_comm. apply le_plus_trans. auto.
inversion is_recv. subst t2.
simpl. rewrite plus_comm. auto with arith.

destruct in_DA1 as [ t2' [ t2 in_DA1 | t2 in_DAx ] | t2 is_send ].
assert (elt_count p1 >= trail_elt_count t2).
eapply IH_path; eauto. auto with arith.
unfold ge in IH_elt. unfold ge.
assert (trail_elt_count t2 <= elt_count p_x).
eapply IH_elt; eauto. rewrite <- plus_Sn_m.
rewrite plus_comm. apply le_plus_trans. auto.
inversion is_send. subst t2.
simpl. rewrite plus_comm. auto with arith.

destruct in_DA1 as [ t2' [ t2 in_DA1 | t2 in_DAx ] | t2 is_heap ].
assert (elt_count p1 >= trail_elt_count t2).
eapply IH_path; eauto. auto with arith.
unfold ge in IH_elt. unfold ge.
assert (trail_elt_count t2 <= heap_elt_count hp_x).
eapply IH_elt; eauto. rewrite <- plus_Sn_m.
rewrite plus_comm. apply le_plus_trans. auto.
inversion is_heap. subst t2.
simpl. rewrite plus_comm. auto with arith.

intros false. simpl in false. inversion false.

intros elt IH_elt hp1 IH_path in_DA1. 
induction elt as [ n_x id_x p_x ]. simpl in in_DA1. simpl.
destruct in_DA1 as [ t2' [ t2 in_DAx | t2 in_DA1 ] | t2 is_x ].
unfold ge in IH_elt. unfold ge.
assert (trail_elt_count t2 <= elt_count p_x).
eapply IH_elt; eauto. rewrite <- plus_Sn_m.
rewrite plus_comm. apply le_plus_trans. auto.
assert (heap_elt_count hp1 >= trail_elt_count t2).
eapply IH_path; eauto. auto with arith.
inversion is_x. subst t2.
simpl. rewrite plus_comm. auto with arith.

Qed.

