Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import trails.

Lemma dep_dep_implies_extends :
  forall t1 t2 id3 p3,
    In trail (dep t1) t2 ->
    In trail (dep t2) (thread_trail id3 p3) ->
    exists p3', 
      In trail (dep t1) (thread_trail id3 p3')
      /\ extends p3' p3.

intros t1 t2 id3 p3 in_dep1 in_dep2.
unfold dep in in_dep1. unfold dep in in_dep2.
remember (thread_trail id3 p3) as t3.
destruct in_dep1 as [ t2 in_DA1 | t2 is_t1 ];
destruct in_dep2 as [ t3 in_DA2 | t3 is_t2 ];
subst t3.

generalize dependent t1.

cut ((forall p1 id1,
  In trail (dep_aux (thread_trail id1 p1)) t2 ->
  exists p3' : path,
    In trail (dep (thread_trail id1 p1)) (thread_trail id3 p3') 
    /\ extends p3' p3)
/\ (forall hp1,
  In trail (dep_aux (heap_trail hp1)) t2 ->
  exists p3' : path,
    In trail (dep (heap_trail hp1)) (thread_trail id3 p3') 
    /\ extends p3' p3)).
intros [ conj1 conj2 ]. induction t1 as [ id1 p1 | hp1 ]; auto.

unfold dep_aux.

apply trail_induction with
(P := fun elt =>
  match elt with
    | path_left => True
    | path_right => True
    | path_send n_x id_x p_x => 
      In trail (dep_aux_thread id_x p_x) t2 ->
      exists p3' : path,
        In trail (dep (thread_trail id_x p_x)) (thread_trail id3 p3') /\
        extends p3' p3
    | path_recv n_x id_x p_x => 
      In trail (dep_aux_thread id_x p_x) t2 ->
      exists p3' : path,
        In trail (dep (thread_trail id_x p_x)) (thread_trail id3 p3') /\
        extends p3' p3
    | path_heap n_x hp_x => 
      In trail (dep_aux_heap hp_x) t2 ->
      exists p3' : path,
        In trail (dep (heap_trail hp_x)) (thread_trail id3 p3') /\
        extends p3' p3
  end)
(P0 := fun elt =>
  match elt with
    | heap_path_thread n_x id_x p_x =>
      In trail (dep_aux_thread id_x p_x) t2 ->
      exists p3' : path,
        In trail (dep (thread_trail id_x p_x)) (thread_trail id3 p3') /\
        extends p3' p3
  end)
(P1 := fun p1 =>
  forall id1,
    In trail (dep_aux_thread id1 p1) t2 ->
    exists p3' : path,
      In trail (dep (thread_trail id1 p1)) (thread_trail id3 p3') /\
      extends p3' p3)
(P2 := fun hp1 =>
  In trail (dep_aux_heap hp1) t2 ->
  exists p3' : path,
    In trail (dep (heap_trail hp1)) (thread_trail id3 p3') /\
    extends p3' p3).

auto. auto. auto. auto. auto. auto.

intros id1 false. simpl in false. inversion false.

intros elt IH_elt p1' IH_p1' id1 in_DA1.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Case: t1 = <id1, Left:p1> *)
simpl in in_DA1.
assert (exists p3' : path,
  In trail (dep (thread_trail id1 p1')) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_dep1' p3'_extends ] ].
unfold dep in in_dep1'.
remember (thread_trail id3 p3') as t3'.
destruct in_dep1' as [ t3' in_DA1' | t3' is_t1' ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. auto. auto.
inversion is_t1'. subst id3 p3'.
exists (path_cons path_left p1').
split. unfold dep. right. apply In_singleton. apply extends_cons. auto.

(* Case: t1 = <id1, Right:p1> *)
simpl in in_DA1.
assert (exists p3' : path,
  In trail (dep (thread_trail id1 p1')) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_dep1' p3'_extends ] ].
unfold dep in in_dep1'.
remember (thread_trail id3 p3') as t3'.
destruct in_dep1' as [ t3' in_DA1' | t3' is_t1' ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. auto. auto.
inversion is_t1'. subst id3 p3'.
exists (path_cons path_right p1').
split. unfold dep. right. apply In_singleton. apply extends_cons. auto.

(* Case: t1 = <id1, Send(<idx, px>):p1> *)
simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_recv ].

assert (exists p3' : path,
  In trail (dep (thread_trail id1 p1')) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_dep1' p3'_extends ] ].
unfold dep in in_dep1'.
remember (thread_trail id3 p3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left.  left. auto. auto.
inversion is_t1'. subst id3 p3'.
exists (path_cons (path_send n_x id_x p_x) p1').
split. unfold dep. right. apply In_singleton. apply extends_cons. auto.

assert (exists p3' : path,
  In trail (dep (thread_trail id_x p_x)) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_depx p3'_extends ] ].
unfold dep in in_depx.
remember (thread_trail id3 p3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left.  right. auto. auto.
inversion is_tx. subst id3 p3'.
exists (path_cons (path_recv n_x id1 p1') p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

inversion is_recv. subst t2.
exists p3. split. unfold dep. simpl.
simpl in in_DA2. remember (thread_trail id3 p3) as t3.
destruct in_DA2 as [ t3 in_DA | ].
destruct in_DA. left. left. right. auto.
left. left. left. auto. right. auto. apply extends_refl.

(* Case: t1 = <id1, Recv(<idx, px>):p1> *)
simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_send ].

assert (exists p3' : path,
  In trail (dep (thread_trail id1 p1')) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_dep1' p3'_extends ] ].
unfold dep in in_dep1'.
remember (thread_trail id3 p3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left.  left. auto. auto.
inversion is_t1'. subst id3 p3'.
exists (path_cons (path_recv n_x id_x p_x) p1').
split. unfold dep. right. apply In_singleton. apply extends_cons. auto.

assert (exists p3' : path,
  In trail (dep (thread_trail id_x p_x)) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_depx p3'_extends ] ].
unfold dep in in_depx.
remember (thread_trail id3 p3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left.  right. auto. auto.
inversion is_tx. subst id3 p3'.
exists (path_cons (path_send n_x id1 p1') p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

inversion is_send. subst t2.
exists p3. split. unfold dep. simpl.
simpl in in_DA2. remember (thread_trail id3 p3) as t3.
destruct in_DA2 as [ t3 in_DA | ].
destruct in_DA. left. left. right. auto.
left. left. left. auto. right. auto. apply extends_refl.

(* Case: t1 = <id1, Heap(hpx):p1> *)
simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_heap ].

assert (exists p3' : path,
  In trail (dep (thread_trail id1 p1')) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_dep1' p3'_extends ] ].
unfold dep in in_dep1'.
remember (thread_trail id3 p3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left.  left. auto. auto.
inversion is_t1'. subst id3 p3'.
exists (path_cons (path_heap n_x hp_x) p1').
split. unfold dep. right. apply In_singleton. apply extends_cons. auto.

assert (exists p3' : path,
  In trail (dep (heap_trail hp_x)) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_depx p3'_extends ] ].
unfold dep in in_depx.
remember (thread_trail id3 p3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left.  right. auto. auto.
inversion is_tx.

inversion is_heap. subst t2.
exists p3. split. unfold dep. simpl.
simpl in in_DA2. remember (thread_trail id3 p3) as t3.
destruct in_DA2 as [ t3 in_DA | ].
destruct in_DA. left. left. left. auto.
left. left. right. auto. right. auto. apply extends_refl.

intros false. simpl in false. inversion false.

intros elt IH_elt p1' IH_p1' in_DA1.
induction elt as [ n_x id_x p_x ].

simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DAx | t2 in_DA1' ] | t2 is_x ].

assert (exists p3' : path,
  In trail (dep (thread_trail id_x p_x)) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_depx p3'_extends ] ].
unfold dep in in_depx.
remember (thread_trail id3 p3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion is_tx. subst id3 p3'.
exists (path_cons (path_heap n_x p1') p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p3' : path,
  In trail (dep (heap_trail p1')) (thread_trail id3 p3') /\
  extends p3' p3) as p3'. auto.
destruct p3' as [ p3' [ in_dep1' p3'_extends ] ].
unfold dep in in_dep1'.
remember (thread_trail id3 p3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists p3'. split. unfold dep. left. simpl. left. right. auto. auto.
inversion is_t1'. 

inversion is_x. subst t2.
exists p3. split. unfold dep. simpl. simpl in in_DA2.
inversion in_DA2 as [ t3' other | t3 false ]. subst t3'.
inversion other as [ t3 | t3 ]; subst t3.
left. left. left. auto. left. left. right. auto.
inversion false. apply extends_refl.

inversion is_t2. subst t2.
exists p3. split. unfold dep. left. auto. apply extends_refl.

inversion is_t1. subst t2.
exists p3. split. unfold dep. left. auto. apply extends_refl.

inversion is_t1. inversion is_t2. subst t1 t2.
exists p3. unfold dep. split. right. apply In_singleton. apply extends_refl.

Qed.

Lemma dep_dep_implies_heap_extends :
  forall t1 t2 hp3,
    In trail (dep t1) t2 ->
    In trail (dep t2) (heap_trail hp3) ->
    exists hp3', 
      In trail (dep t1) (heap_trail hp3')
      /\ heap_extends hp3' hp3.

intros t1 t2 hp3 in_dep1 in_dep2.
unfold dep in in_dep1. unfold dep in in_dep2.
remember (heap_trail hp3) as t3.
destruct in_dep1 as [ t2 in_DA1 | t2 is_t1 ];
destruct in_dep2 as [ t3 in_DA2 | t3 is_t2 ];
subst t3.

generalize dependent t1.

cut ((forall p1 id1,
  In trail (dep_aux (thread_trail id1 p1)) t2 ->
  exists hp3',
    In trail (dep (thread_trail id1 p1)) (heap_trail hp3') 
    /\ heap_extends hp3' hp3)
/\ (forall hp1,
  In trail (dep_aux (heap_trail hp1)) t2 ->
  exists hp3',
    In trail (dep (heap_trail hp1)) (heap_trail hp3') 
    /\ heap_extends hp3' hp3)).
intros [ conj1 conj2 ]. induction t1 as [ id1 p1 | hp1 ]; auto.

unfold dep_aux.

apply trail_induction with
(P := fun elt =>
  match elt with
    | path_left => True
    | path_right => True
    | path_send n_x id_x p_x => 
      In trail (dep_aux_thread id_x p_x) t2 ->
      exists hp3',
        In trail (dep (thread_trail id_x p_x)) (heap_trail hp3') /\
        heap_extends hp3' hp3
    | path_recv n_x id_x p_x => 
      In trail (dep_aux_thread id_x p_x) t2 ->
      exists hp3',
        In trail (dep (thread_trail id_x p_x)) (heap_trail hp3') /\
        heap_extends hp3' hp3
    | path_heap n_x hp_x => 
      In trail (dep_aux_heap hp_x) t2 ->
      exists hp3',
        In trail (dep (heap_trail hp_x)) (heap_trail hp3') /\
        heap_extends hp3' hp3
  end)
(P0 := fun elt =>
  match elt with
    | heap_path_thread n_x id_x p_x =>
      In trail (dep_aux_thread id_x p_x) t2 ->
      exists hp3',
        In trail (dep (thread_trail id_x p_x)) (heap_trail hp3') /\
        heap_extends hp3' hp3
  end)
(P1 := fun p1 =>
  forall id1,
    In trail (dep_aux_thread id1 p1) t2 ->
    exists hp3',
      In trail (dep (thread_trail id1 p1)) (heap_trail hp3') /\
      heap_extends hp3' hp3)
(P2 := fun hp1 =>
  In trail (dep_aux_heap hp1) t2 ->
  exists hp3',
    In trail (dep (heap_trail hp1)) (heap_trail hp3') /\
    heap_extends hp3' hp3).

auto. auto. auto. auto. auto. auto.

intros id1 false. simpl in false. inversion false.

intros elt IH_elt p1' IH_p1' id1 in_DA1.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Case: t1 = <id1, Left:p1> *)
simpl in in_DA1.
assert (exists hp3',
  In trail (dep (thread_trail id1 p1')) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_dep1' hp3'_extends ] ].
unfold dep in in_dep1'.
remember (heap_trail hp3') as t3'.
destruct in_dep1' as [ t3' in_DA1' | t3' is_t1' ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. auto. auto.
inversion is_t1'. 

(* Case: t1 = <id1, Right:p1> *)
simpl in in_DA1.
assert (exists hp3',
  In trail (dep (thread_trail id1 p1')) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_dep1' hp3'_extends ] ].
unfold dep in in_dep1'.
remember (heap_trail hp3') as t3'.
destruct in_dep1' as [ t3' in_DA1' | t3' is_t1' ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. auto. auto.
inversion is_t1'. 

(* Case: t1 = <id1, Send(<idx, px>):p1> *)
simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_recv ].

assert (exists hp3',
  In trail (dep (thread_trail id1 p1')) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_dep1' hp3'_extends ] ].
unfold dep in in_dep1'.
remember (heap_trail hp3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left.  left. auto. auto.
inversion is_t1'. 

assert (exists hp3',
  In trail (dep (thread_trail id_x p_x)) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_depx hp3'_extends ] ].
unfold dep in in_depx.
remember (heap_trail hp3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left.  right. auto. auto.
inversion is_tx. 

inversion is_recv. subst t2.
exists hp3. split. unfold dep. simpl.
simpl in in_DA2. remember (heap_trail hp3) as t3.
destruct in_DA2 as [ t3 in_DA | ].
destruct in_DA. left. left. right. auto.
left. left. left. auto. right. auto. apply heap_extends_refl.

(* Case: t1 = <id1, Recv(<idx, px>):p1> *)
simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_send ].

assert (exists hp3',
  In trail (dep (thread_trail id1 p1')) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_dep1' hp3'_extends ] ].
unfold dep in in_dep1'.
remember (heap_trail hp3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left.  left. auto. auto.
inversion is_t1'. 

assert (exists hp3',
  In trail (dep (thread_trail id_x p_x)) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_depx hp3'_extends ] ].
unfold dep in in_depx.
remember (heap_trail hp3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left.  right. auto. auto.
inversion is_tx.

inversion is_send. subst t2.
exists hp3. split. unfold dep. simpl.
simpl in in_DA2. remember (heap_trail hp3) as t3.
destruct in_DA2 as [ t3 in_DA | ].
destruct in_DA. left. left. right. auto.
left. left. left. auto. right. auto. apply heap_extends_refl.

(* Case: t1 = <id1, Heap(hpx):p1> *)
simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DA1' | t2 in_DAx ] | t2 is_heap ].

assert (exists hp3',
  In trail (dep (thread_trail id1 p1')) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_dep1' hp3'_extends ] ].
unfold dep in in_dep1'.
remember (heap_trail hp3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left.  left. auto. auto.
inversion is_t1'. 

assert (exists hp3',
  In trail (dep (heap_trail hp_x)) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_depx hp3'_extends ] ].
unfold dep in in_depx.
remember (heap_trail hp3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left.  right. auto. auto.
inversion is_tx. subst hp3'.
unfold dep. simpl.
exists (heap_path_cons (heap_path_thread n_x id1 p1') hp_x).
split. left. right. apply In_singleton.
apply heap_extends_cons. auto.

inversion is_heap. subst t2. exists hp3.
split. unfold dep. simpl.
simpl in in_DA2. inversion in_DA2 as [ t3 other | t3 false ].
subst t3. inversion other as [ t3 | t3 ].
left. left. left. auto. left. left. right. auto.
inversion false. apply heap_extends_refl.

intros false. simpl in false. inversion false.

intros elt IH_elt p1' IH_p1' in_DA1.
induction elt as [ n_x id_x p_x ].

simpl in in_DA1.
destruct in_DA1 as [ t2' [ t2 in_DAx | t2 in_DA1' ] | t2 is_x ].

assert (exists hp3',
  In trail (dep (thread_trail id_x p_x)) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_depx hp3'_heap_extends ] ].
unfold dep in in_depx.
remember (heap_trail hp3') as t3'.
destruct in_depx as [ t3' t3'_in_DAx | t3' is_tx ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion is_tx. 

assert (exists hp3',
  In trail (dep (heap_trail p1')) (heap_trail hp3') /\
  heap_extends hp3' hp3) as hp3'. auto.
destruct hp3' as [ hp3' [ in_dep1' hp3'_extends ] ].
unfold dep in in_dep1'.
remember (heap_trail hp3') as t3'.
destruct in_dep1' as [ t3' t3'_in_DA1' | t3' is_t1' ]; subst t3'.
exists hp3'. split. unfold dep. left. simpl. left. right. 
simpl in t3'_in_DA1'. auto. auto.
inversion is_t1'. subst p1'.
unfold dep. simpl.
exists (heap_path_cons (heap_path_thread n_x id_x p_x) hp3').
split. right. apply In_singleton. apply heap_extends_cons. auto.

inversion is_x. subst t2.
simpl in in_DA2. exists hp3. split. unfold dep. simpl.
inversion in_DA2 as [ t3 other | t3 t3_eq ].
subst t3. inversion other as [ t3 | t3 ].
left. left. left. auto. left. left. right. auto.
inversion t3_eq. right. apply In_singleton.
apply heap_extends_refl.

inversion is_t2. subst t2.
exists hp3. split. unfold dep. left. auto. apply heap_extends_refl.

inversion is_t1. subst t2.
exists hp3. split. unfold dep. left. auto. apply heap_extends_refl.

inversion is_t1. inversion is_t2. subst t1 t2.
exists hp3. unfold dep. split. right. apply In_singleton.
apply heap_extends_refl.

Qed.
