Require Import consistent.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import syntax.
Require Import trails.

Lemma heap_in_dep1 :
  forall n_ld id_ld p_ld hp_ld t,
    In trail (dep t) 
    (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
    exists hp_ld',
    In trail (dep t) (heap_trail hp_ld')
    /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld).

intros n_ld id_ld p_ld hp_ld.

cut ((forall p id,
  In trail (dep (thread_trail id p)) 
  (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
  exists hp_ld',
    In trail (dep (thread_trail id p)) (heap_trail hp_ld')
    /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))
/\ (forall hp,
  In trail (dep (heap_trail hp)) 
  (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
  exists hp_ld',
    In trail (dep (heap_trail hp)) (heap_trail hp_ld')
    /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))).

intros [ H1 H2 ].
induction t as [ id p | hp ]; auto.

apply trail_induction with 
(P := fun elt => 
  match elt with
    | path_left => True
    | path_right => True
    | path_send n id p => 
      In trail (dep (thread_trail id p)) 
      (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
      exists hp_ld',
        In trail (dep (thread_trail id p)) (heap_trail hp_ld')
        /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)
    | path_recv n id p => 
      In trail (dep (thread_trail id p)) 
      (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
      exists hp_ld',
        In trail (dep (thread_trail id p)) (heap_trail hp_ld')
        /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)
    | path_heap n hp =>
      In trail (dep (heap_trail hp)) 
      (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
      exists hp_ld',
        In trail (dep (heap_trail hp)) (heap_trail hp_ld')
        /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)
  end)
(P0 := fun elt =>
  match elt with
    | heap_path_thread n id p =>
      In trail (dep (thread_trail id p)) 
      (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
      exists hp_ld',
        In trail (dep (thread_trail id p)) (heap_trail hp_ld')
        /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)
  end)
(P1 := fun p =>
  forall id,
    In trail (dep (thread_trail id p)) 
      (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
      exists hp_ld',
        In trail (dep (thread_trail id p)) (heap_trail hp_ld')
        /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))
(P2 := fun hp =>
  In trail (dep (heap_trail hp)) 
  (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld)) ->
  exists hp_ld',
    In trail (dep (heap_trail hp)) (heap_trail hp_ld')
    /\ heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)).

auto. auto. auto. auto. auto. auto.

intros id false. unfold dep in false.
inversion false as [ t_s false2 | t_s false2 ]; subst t_s.
simpl in false2. inversion false2. inversion false2.

intros elt IH_elt p IH_path id in_dep.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left *)
unfold dep in in_dep.
inversion in_dep as [ t_s t_s_in_DA | t_s false ].
simpl in t_s_in_DA.
assert (exists hp_ld',
  In trail (dep (thread_trail id p)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
eapply IH_path. unfold dep. left. apply t_s_in_DA.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep.
inversion hp_ld'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists hp_ld'. split. unfold dep. left. simpl. auto. auto.
inversion t_r'_is_t'. 
inversion false.

(* Right *)
unfold dep in in_dep.
inversion in_dep as [ t_s t_s_in_DA | t_s false ].
simpl in t_s_in_DA.
assert (exists hp_ld',
  In trail (dep (thread_trail id p)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
eapply IH_path. unfold dep. left. apply t_s_in_DA.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep.
inversion hp_ld'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists hp_ld'. split. unfold dep. left. simpl. auto. auto.
inversion t_r'_is_t'. 
inversion false.

(* Send *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_s other | t_s t_s_is_x ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s false ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_t' | t_s t_s_in_DA_x ].

simpl in t_s_in_DA_t'.
assert (exists hp_ld',
  In trail (dep (thread_trail id p)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t'.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep.
inversion hp_ld'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists hp_ld'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_r'_is_t'.

assert (exists hp_ld',
  In trail (dep (thread_trail id_x p_x)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
apply IH_elt. unfold dep. left. auto.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep_x hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep_x.
inversion hp_ld'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists hp_ld'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_r'_is_x. 
inversion false.

inversion t_s_is_x. 

(* Recv *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_s other | t_s false ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s t_s_is_x ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_t' | t_s t_s_in_DA_x ].

simpl in t_s_in_DA_t'.
assert (exists hp_ld',
  In trail (dep (thread_trail id p)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t'.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep.
inversion hp_ld'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists hp_ld'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_r'_is_t'.

assert (exists hp_ld',
  In trail (dep (thread_trail id_x p_x)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
apply IH_elt. unfold dep. left. auto.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep_x hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep_x.
inversion hp_ld'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists hp_ld'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_r'_is_x. 

inversion t_s_is_x. 

inversion false.

(* Heap *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_s other | t_s false ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s false ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_t | t_s t_s_in_DA_x ].

simpl in t_s_in_DA_t.
assert (exists hp_ld',
  In trail (dep (thread_trail id p)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep.
inversion hp_ld'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists hp_ld'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_r'_is_t'. 

assert (exists hp_ld',
  In trail (dep (heap_trail hp_x)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
apply IH_elt. unfold dep. left. auto.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep_x hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep_x.
inversion hp_ld'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists hp_ld'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_r'_is_x. subst t_r' hp_x.
exists (heap_path_cons (heap_path_thread n_x id p) hp_ld').
split. unfold dep. left. simpl. right. apply In_singleton.
apply heap_extends_cons. auto.

inversion false.

inversion false. subst id_ld n_ld hp_ld p_ld.
exists (heap_path_cons (heap_path_thread n_x id p) hp_x).
split. unfold dep. left. simpl. right. apply In_singleton.
apply heap_extends_refl.

(* heap_path_nil *)
intros false. unfold dep in false.
inversion false as [ t_s false2 | t_s false2 ].
simpl in false2. inversion false2. inversion false2.

(* heap paths *)
intros elt IH_elt hp IH_path in_dep.
induction elt as [ n_x id_x p_x ]. unfold dep in in_dep. simpl in in_dep.

inversion in_dep as [ t_s other | t_s false ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s t_s_is_x ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_x | t_s t_s_in_DA_t ].

assert (exists hp_ld',
  In trail (dep (thread_trail id_x p_x)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
apply IH_elt. unfold dep. left. auto.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep_x hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep_x.
inversion hp_ld'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists hp_ld'. split. unfold dep. simpl. left. left. left. auto. auto.
inversion t_r'_is_x. 

simpl in t_s_in_DA_t.
assert (exists hp_ld',
  In trail (dep (heap_trail hp)) (heap_trail hp_ld')/\
  heap_extends hp_ld' (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t.
destruct hp_ld' as [ hp_ld' [ hp_ld'_in_dep hp_ld'_extends ] ].
unfold dep in hp_ld'_in_dep.
inversion hp_ld'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists hp_ld'. split. unfold dep. left. simpl. left. right. auto. auto.
inversion t_r'_is_t'. subst t_r' hp_ld'.
exists (heap_path_cons (heap_path_thread n_x id_x p_x) hp).
split. unfold dep. right. apply In_singleton.
apply heap_extends_cons. auto.

inversion t_s_is_x. subst id_ld n_ld hp_ld p_ld.
exists (heap_path_cons (heap_path_thread n_x id_x p_x) hp).
split. unfold dep. right. apply In_singleton.
apply heap_extends_refl.

inversion false.

Qed.

Lemma heap_in_dep2 :
  forall n_ld hp_ld id_ld p_ld t,
    In trail (dep t) 
    (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
    exists p_ld',
    In trail (dep t) (thread_trail id_ld p_ld') 
    /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld).

intros n_ld hp_ld id_ld p_ld.

cut ((forall p id,
  In trail (dep (thread_trail id p)) 
  (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
  exists p_ld',
    In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') 
    /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld))
/\ (forall hp,
  In trail (dep (heap_trail hp)) 
  (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
  exists p_ld',
    In trail (dep (heap_trail hp)) (thread_trail id_ld p_ld') 
    /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld))).

intros [ H1 H2 ].
induction t as [ id p | hp ]; auto.

apply trail_induction with 
(P := fun elt => 
  match elt with
    | path_left => True
    | path_right => True
    | path_send n id p => 
      In trail (dep (thread_trail id p)) 
      (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
      exists p_ld',
        In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') 
        /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)
    | path_recv n id p => 
      In trail (dep (thread_trail id p)) 
      (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
      exists p_ld',
        In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') 
        /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)
    | path_heap n hp =>
      In trail (dep (heap_trail hp)) 
      (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
      exists p_ld',
        In trail (dep (heap_trail hp)) (thread_trail id_ld p_ld') 
        /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)
  end)
(P0 := fun elt =>
  match elt with
    | heap_path_thread n id p =>
      In trail (dep (thread_trail id p)) 
      (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
      exists p_ld',
        In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') 
        /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)
  end)
(P1 := fun p =>
  forall id,
    In trail (dep (thread_trail id p)) 
      (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
      exists p_ld',
        In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') 
        /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld))
(P2 := fun hp =>
  In trail (dep (heap_trail hp)) 
  (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) ->
  exists p_ld',
    In trail (dep (heap_trail hp)) (thread_trail id_ld p_ld') 
    /\ extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)).

auto. auto. auto. auto. auto. auto.

intros id false. unfold dep in false.
inversion false as [ t_r false2 | t_r false2 ]; subst t_r.
simpl in false2. inversion false2. inversion false2.

intros elt IH_elt p IH_path id in_dep.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left *)
unfold dep in in_dep.
inversion in_dep as [ t_r t_r_in_DA | t_r false ].
simpl in t_r_in_DA.
assert (exists p_ld',
  In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
eapply IH_path. unfold dep. left. apply t_r_in_DA.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep p_ld'_extends ] ].
unfold dep in p_ld'_in_dep.
inversion p_ld'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_ld'. split. unfold dep. left. simpl. auto. auto.
inversion t_s'_is_t'. subst id_ld p_ld'.
exists (path_cons path_left p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

(* Right *)
unfold dep in in_dep.
inversion in_dep as [ t_r t_r_in_DA | t_r false ].
simpl in t_r_in_DA.
assert (exists p_ld',
  In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
eapply IH_path. unfold dep. left. apply t_r_in_DA.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep p_ld'_extends ] ].
unfold dep in p_ld'_in_dep.
inversion p_ld'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_ld'. split. unfold dep. left. simpl. auto. auto.
inversion t_s'_is_t'. subst id_ld p_ld'.
exists (path_cons path_right p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

(* Send *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_r other | t_r false ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r t_r_is_x ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_t' | t_r t_r_in_DA_x ].

simpl in t_r_in_DA_t'.
assert (exists p_ld',
  In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t'.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep p_ld'_extends ] ].
unfold dep in p_ld'_in_dep.
inversion p_ld'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_ld'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_s'_is_t'. subst id_ld p_ld'.
exists (path_cons (path_send n_x id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_ld',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
apply IH_elt. unfold dep. left. auto.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep_x p_ld'_extends ] ].
unfold dep in p_ld'_in_dep_x.
inversion p_ld'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_ld'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_s'_is_x. subst id_ld p_ld'.
exists (path_cons (path_recv n_x id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

inversion t_r_is_x. 

inversion false.

(* Recv *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_r other | t_r t_r_is_x ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r false ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_t' | t_r t_r_in_DA_x ].

simpl in t_r_in_DA_t'.
assert (exists p_ld',
  In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t'.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep p_ld'_extends ] ].
unfold dep in p_ld'_in_dep.
inversion p_ld'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_ld'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_s'_is_t'. subst id_ld p_ld'.
exists (path_cons (path_recv n_x id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_ld',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
apply IH_elt. unfold dep. left. auto.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep_x p_ld'_extends ] ].
unfold dep in p_ld'_in_dep_x.
inversion p_ld'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_ld'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_s'_is_x. subst id_ld p_ld'.
exists (path_cons (path_send n_x id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

inversion t_r_is_x. 

(* Heap *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_r other | t_r false ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r false ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_t | t_r t_r_in_DA_x ].

simpl in t_r_in_DA_t.
assert (exists p_ld',
  In trail (dep (thread_trail id p)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep p_ld'_extends ] ].
unfold dep in p_ld'_in_dep.
inversion p_ld'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_ld'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_s'_is_t'. subst id_ld p_ld'.
exists (path_cons (path_heap n_x hp_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_ld',
  In trail (dep (heap_trail hp_x)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
apply IH_elt. unfold dep. left. auto.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep_x p_ld'_extends ] ].
unfold dep in p_ld'_in_dep_x.
inversion p_ld'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_ld'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_s'_is_x.

inversion false. subst n_ld id_ld p_ld hp_ld.
exists (path_cons (path_heap n_x hp_x) p).
split. right. apply In_singleton. apply extends_refl.

inversion false.

(* heap_path_nil *)
intros false. unfold dep in false.
inversion false as [ t_r false2 | t_r false2 ].
simpl in false2. inversion false2. inversion false2.

(* heap paths *)
intros elt IH_elt hp IH_path in_dep.
induction elt as [ n_x id_x p_x ]. unfold dep in in_dep. simpl in in_dep.

inversion in_dep as [ t_r other | t_r false ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r t_r_is_x ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_x | t_r t_r_in_DA_t ].

assert (exists p_ld',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
apply IH_elt. unfold dep. left. auto.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep_x p_ld'_extends ] ].
unfold dep in p_ld'_in_dep_x.
inversion p_ld'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_ld'. split. unfold dep. simpl. left. left. left. auto. auto.
inversion t_s'_is_x. subst id_ld p_ld'.
exists (path_cons (path_heap n_x hp) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

simpl in t_r_in_DA_t.
assert (exists p_ld',
  In trail (dep (heap_trail hp)) (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t.
destruct p_ld' as [ p_ld' [ p_ld'_in_dep p_ld'_extends ] ].
unfold dep in p_ld'_in_dep.
inversion p_ld'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_ld'. split. unfold dep. left. simpl. left. right. auto. auto.
inversion t_s'_is_t'.

inversion t_r_is_x. 

inversion false. subst n_ld id_ld p_ld hp_ld.
exists (path_cons (path_heap n_x hp) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_refl.

Qed.
