Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical.
Require Import Coq.Sets.Ensembles.
Require Import syntax.

Definition hl_thread := (prod (prod tid (list ctxt)) exp).

Inductive hl_sync_step :
  heap -> Ensemble hl_thread -> option hl_thread ->
  heap -> Ensemble hl_thread -> option hl_thread -> Prop :=

| hl_sync_step_then_always : 
  forall H S id list_M e E v1 v2,
    get_evt_ctxt e E (thenEvt (alwaysEvt v1) v2) ->
    hl_sync_step H S (Some (id, list_M, e))
    H S (Some (id, list_M, evt_fill_hole E (app v2 v1)))

| hl_sync_step_choose_left :
  forall H S id list_M e E v1 v2,
    get_evt_ctxt e E (chooseEvt v1 v2) ->
    hl_sync_step H S (Some (id, list_M, e))
    H S (Some (id, list_M, evt_fill_hole E v1))

| hl_sync_step_choose_right :
  forall H S id list_M e E v1 v2,
    get_evt_ctxt e E (chooseEvt v1 v2) ->
    hl_sync_step H S (Some (id, list_M, e))
    H S (Some (id, list_M, evt_fill_hole E v2))

| hl_sync_step_begin_lockdown :
  forall H S s,
    ~ In hl_thread S s ->
    hl_sync_step H (Add hl_thread S s) None
    H S (Some s)

| hl_sync_step_end_lockdown :
  forall H S id list_M v,
    ldvalue list_M v ->
    hl_sync_step H S (Some (id, list_M, v))
    H (Add hl_thread S (id, list_M, v)) None

| hl_sync_step_step_lockdown :
  forall H H' S id list_M e e',
    st_step H e H' e' ->
    hl_sync_step H S (Some (id, list_M, e))
    H' S (Some (id, list_M, e'))

| hl_sync_step_send_recv :
  forall H id1 list_M1 e1 E1 (k:channel) v id2 list_M2 e2 E2 S S' s,
    In hl_thread S (id1, list_M1, e1) ->
    In hl_thread S (id2, list_M2, e2) ->
    get_evt_ctxt e1 E1 (sendEvt k v) ->
    get_evt_ctxt e2 E2 (recvEvt k) ->
    S' = (Union hl_thread
      (Setminus hl_thread S 
        (Couple hl_thread (id1, list_M1, e1) (id2, list_M2, e2)))
      (Couple hl_thread 
        (id1, list_M1, evt_fill_hole E1 (alwaysEvt zero))
        (id2, list_M2, evt_fill_hole E2 (alwaysEvt v)))) ->
    hl_sync_step 
    H S s
    H S' s

| hl_sync_step_nested_sync_init :
  forall H S id list_M e M v,
    get_ctxt e M (sync v) ->
    hl_sync_step H S (Some (id, list_M, e))
    H S (Some (id, M::list_M, v))

| hl_sync_step_nested_sync_complete :
  forall H S id M list_M v,
    value v ->
    hl_sync_step H S (Some (id, M::list_M, alwaysEvt v))
    H S (Some (id, list_M, M[v])).

Inductive hl_sync_step_many : 
  heap -> Ensemble hl_thread -> option hl_thread ->
  heap -> Ensemble hl_thread -> option hl_thread -> Prop :=

| hl_sync_step_many_zero :
  forall H S s, hl_sync_step_many H S s H S s

| hl_sync_step_many_nonzero :
  forall H1 S1 s1 H2 S2 s2 H3 S3 s3,
    hl_sync_step H1 S1 s1 H2 S2 s2 ->
    hl_sync_step_many H2 S2 s2 H3 S3 s3 ->
    hl_sync_step_many H1 S1 s1 H3 S3 s3.

Lemma hl_sync_step_many_reversed :
  forall H1 S1 s1 H2 S2 s2 H3 S3 s3,
    hl_sync_step_many H1 S1 s1 H2 S2 s2 ->
    hl_sync_step H2 S2 s2 H3 S3 s3 ->
    hl_sync_step_many H1 S1 s1 H3 S3 s3.

intros H1 S1 s1 H2 S2 s2 H3 S3 s3 step_1_2 step_2_3.
induction step_1_2.

eapply hl_sync_step_many_nonzero. apply step_2_3.
apply hl_sync_step_many_zero.

eapply hl_sync_step_many_nonzero; eauto.

Qed.  

Inductive get_S_hl (T:Ensemble mt_thread) : Ensemble hl_thread :=
| In_S_hl : forall id e M v,
  In mt_thread T (id, e) ->
  get_ctxt e M (sync v) ->
  In hl_thread (get_S_hl T) (id, nil, v).

Inductive get_T'_hl (T:Ensemble mt_thread) 
  (S':Ensemble hl_thread) : Ensemble mt_thread :=
| In_get_T'_hl : forall id e M v v',
  In mt_thread T (id, e) ->
  get_ctxt e M (sync v) ->
  In hl_thread S' (id, nil, (alwaysEvt v')) ->
  In mt_thread (get_T'_hl T S') (id, ctxt_fill_hole M v').

Inductive hl_mt_step :
  heap -> Ensemble mt_thread ->
  heap -> Ensemble mt_thread -> Prop :=

| hl_step_mt : forall H T H' T',
  mt_step H T H' T' ->
  hl_mt_step H T H' T'

| hl_step_sync : forall  H H' T T' S S' T_all T_all' T0,
  (forall id e, In mt_thread T (id, e) -> 
    exists M, exists v, get_ctxt e M (sync v)) ->
  S = get_S_hl T ->
  hl_sync_step_many H S None H' S' None ->
  (forall id M e, In hl_thread S' (id, M, e) ->
    M = nil /\ exists v', value v' /\ e = alwaysEvt v') ->
  T' = get_T'_hl T S' ->
  T_all = (Union mt_thread T T0) ->
  Disjoint mt_thread T T0 ->
  T_all' = (Union mt_thread T' T0) ->
  hl_mt_step H T_all H' T_all'.
  
Inductive hl_mt_step_many : 
  heap -> Ensemble mt_thread ->
  heap -> Ensemble mt_thread -> Prop :=

| hl_mt_step_many_zero :
  forall H T, hl_mt_step_many H T H T

| hl_mt_step_many_nonzero :
  forall H1 T1 H2 T2 H3 T3,
    hl_mt_step H1 T1 H2 T2 ->
    hl_mt_step_many H2 T2 H3 T3 ->
    hl_mt_step_many H1 T1 H3 T3.

Lemma hl_mt_step_many_one : forall H1 T1 H2 T2,
  hl_mt_step H1 T1 H2 T2 ->
  hl_mt_step_many H1 T1 H2 T2.

intros H1 T1 H2 T2 steps.
eapply hl_mt_step_many_nonzero.
eauto. apply hl_mt_step_many_zero.

Qed.

Lemma hl_mt_step_many_many :
  forall H1 T1 H2 T2 H3 T3,
    hl_mt_step_many H1 T1 H2 T2 ->
    hl_mt_step_many H2 T2 H3 T3 ->
    hl_mt_step_many H1 T1 H3 T3.

intros H1 T1 H2 T2 H3 T3 step_1_2 step_2_3.
induction step_1_2. auto.
eapply hl_mt_step_many_nonzero.
eauto. eauto.

Qed.

Definition hl_unique_tids (S:Ensemble hl_thread) : Prop :=
  forall t1 t2 id1 id2 list_M1 list_M2 e1 e2,
    t1 = (id1, list_M1, e1) -> t2 = (id2, list_M2, e2) ->
    In hl_thread S t1 -> In hl_thread S t2 ->
    t1 <> t2 -> id1 <> id2.

Definition ld_unique_tids (S:Ensemble hl_thread)
  (s:option hl_thread) : Prop :=
  hl_unique_tids S /\
  match s with
    | None => True
    | Some (id, list_M, e) => forall id' list_M' e',
      In hl_thread S (id', list_M', e') ->
      id' <> id
  end.
