Require Import committable.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Lists.List.
Require Import Coq.Sets.Ensembles.
Require Import highlevel.
Require Import hl2ll_get_S.
Require Import hl2ll_sync.
Require Import lowlevel.
Require Import preserves_tids.
Require Import syntax.
Require Import trails.
Require Import unique_tids.

Theorem hl2ll : forall H T H' T',
  mt_unique_tids T ->
  hl_mt_step H T H' T' ->
  ll_step_many (unfrozen H) T empty_susp empty_heap empty_search
  (unfrozen H') T' empty_susp empty_heap empty_search.

intros H T H' T' T_unique hl_mt_steps.

inversion hl_mt_steps as [ _H _T _H' _T' mt_steps eqH eqT eqH' eqT' 
  | _H _H' T_sync T_sync' S_hl S_hl' _T _T' T0 T_sync_at_sync S_hl_eq hl_steps 
    S_hl'_complete T_sync'_eq T_eq T_disjoint T'_eq _H_eq _T_eq _H'_eq _T'_eq ].

subst _H _T _H' _T'.
eapply ll_step_many_nonzero.
apply ll_step_mt.
apply mt_steps.
apply ll_step_many_zero.

subst _H _T _H' _T' S_hl T T' T_sync'.

(* first step : SyncInit *)
eapply ll_step_many_nonzero.
eapply ll_step_sync_init; eauto.

(* now take a bunch of synchronizing steps *)
assert (get_S_init T_sync = get_S_sync_init (get_S_hl T_sync)) as S_init_eq.
apply Extensionality_Ensembles. split; intros x x_in_one; inversion x_in_one.
apply In_S_sync_init. apply In_S_hl with (e := e) (M := M); auto.
subst x. inversion H0. eapply In_S_init with (e := e) (M := M); auto.
rewrite -> S_init_eq.
assert (exists thread_trails, exists hp, exists S_H, exists S0, exists S,
  let S_all := Union search_thread S S0 in
    let trails := Add trail thread_trails (heap_trail hp) in
      committable trails /\
      (ll_step_many (frozen H) T0 (get_S_susp_init T_sync) 
        (Singleton search_heap (H, heap_path_nil))
        (get_S_sync_init (get_S_hl T_sync)) (frozen H) T0 (get_S_susp_init T_sync) 
        S_H S_all) /\
      exists S_H0,
        S = get_S_no_ld S_hl' thread_trails /\
        S_H = Add search_heap S_H0 (H', hp) /\
          (forall t, 
            In trail thread_trails t ->
            exists id, exists p, exists list_M, exists e, 
              t = thread_trail id p /\
              In hl_thread S_hl' (id, list_M, e)) /\
          (forall id list_M e,  
            In hl_thread S_hl' (id, list_M, e) ->
            exists p, 
              In trail thread_trails (thread_trail id p))) as ll_sync_steps.
apply hl2ll_sync.
apply hl_steps.
apply get_S_hl_unique_tids.
unfold mt_unique_tids. unfold mt_unique_tids in T_unique.
intros t1 t2 id1 id2 e1 e2 t1_eq t2_eq t1_in_T_sync t2_in_T_sync t1_not_t2.
eapply T_unique. apply t1_eq. apply t2_eq. left. apply t1_in_T_sync. left. apply t2_in_T_sync. apply t1_not_t2.
intros id list_M e in_S_hl.
inversion in_S_hl. subst. split. eapply get_ctxt_sync_value. apply H4. reflexivity.
intros id list_M e in_S_hl'.
assert (list_M = nil /\ (exists v' : exp, value v' /\ e = alwaysEvt v')) as rewriting.
eapply S_hl'_complete. apply in_S_hl'. destruct rewriting as [list_M_is_nil [ v' [ value_v' e_is_aevt_v' ] ] ].
exists v'. auto.

destruct ll_sync_steps as [ thread_trails [ hp [ S_H [ S0 [ S [ comm_trails [ ll_steps 
  [ S_H0 [ S_eq [ S_H_eq [ trails_in_S_hl' S_hl'_in_trails ] ] ] ] ] ] ] ] ] ] ]. 
eapply ll_step_many_reverse.
apply ll_steps.
eapply ll_step_sync_complete.
instantiate (1 := S).
intros id h list_M e p t_in_S.
subst S. inversion t_in_S.
split. reflexivity.
eapply S_hl'_complete.
apply H2.

subst S.
unfold search_unique_tids.
intros id h1 list_M1 e1 p1 h2 list_M2 e2 p2 s1_in_S s2_in_S.
inversion s1_in_S as [ id1 list_M1' e1' p1' ]. subst id1 list_M1' e1' p1'.
inversion s2_in_S as [ id2 list_M2' e2' p2' ]. subst id2 list_M2' e2' p2'.
assert ((id, list_M1, e1) = (id, list_M2, e2)) as hl_eq.
eapply hl_unique_tids_reversed.
apply ld_unique_tids_implies_hl.
eapply hl_sync_step_many_preserves_unique_tids. eauto.
unfold ld_unique_tids. split. unfold hl_unique_tids.
intros t1 t2 id2 id3 list_M3 list_M4 v3 v4 t1_eq t2_eq 
t1_in_S_hl t2_in_S_hl t1_not_t2. subst t1 t2.
inversion t1_in_S_hl as [ id2' e3 M3 v3' ].
inversion t2_in_S_hl as [ id3' e4 M4 v4' ].
unfold mt_unique_tids in T_unique.
eapply T_unique. reflexivity. reflexivity.
left. eauto. left. eauto.
unfold not. intros eq.
inversion eq. subst id3 e4.
assert (M3 = M4 /\ sync v3 = sync v4) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq.
subst M4 v4 v3' v4' list_M3 list_M4 id3'. tauto. auto.
auto. auto.
inversion hl_eq.
unfold committable in comm_trails.
destruct comm_trails as [ heap_unique [ thread_unique other ] ].
assert (p1 = p2). eapply thread_unique.
left. eauto. left. auto. subst p2. reflexivity.

instantiate (1 := get_S_susp_init T_sync).
instantiate (1 := get_T'_hl T_sync S_hl').
apply Extensionality_Ensembles.
split; intros x x_in_one.
inversion x_in_one.
assert (exists p : path,
  In trail thread_trails (thread_trail id p)) as rewriting.
eapply S_hl'_in_trails. apply H2. 
destruct rewriting. 
eapply In_T'_complete.
subst S. apply In_S_no_ld. auto. 
apply H4.
eapply In_S_susp_init.
apply H0. apply H1.
inversion x_in_one. inversion H1. subst. inversion H0. subst. 
eapply In_get_T'_hl. apply H5. apply H7. apply H4.

apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one.
assert (exists list_M', exists e',
  In hl_thread S_hl' (id, list_M', e')
  \/ None = Some (id, list_M', e')) as in_S_hl'.
eapply hl_sync_step_many_preserves_tids.
eauto. left. eapply In_S_hl. eauto. eauto.
destruct in_S_hl' as [ list_M' [ e' [ In_S_hl' | false ] ] ].
assert (exists p, In trail thread_trails (thread_trail id p)) as in_trails.
eapply S_hl'_in_trails. eauto.
destruct in_trails as [ p in_trails ].
eapply In_S_susp_complete. subst x. auto.
subst S. apply In_S_no_ld. eauto. eauto.
inversion false.
inversion x_in_one. auto.
instantiate (1 := Empty_set susp_thread).
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one.
destruct x_in_one. tauto.

instantiate (1 := Empty_set mt_thread).
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one. inversion x_in_one. inversion H0.

apply S_H_eq.
reflexivity.
assert (T0 = Union mt_thread T0 (Empty_set mt_thread)) as T0_eq.
apply Extensionality_Ensembles.
split; intros x x_in_one. left. apply x_in_one.
inversion x_in_one. apply H0. inversion H0.
rewrite <- T0_eq. reflexivity.
instantiate (1 := thread_trails).
subst S. apply Extensionality_Ensembles.
split; intros x x_in_one.
assert (exists id : tid,
  exists p : path,
    exists list_M : list ctxt,
      exists e : exp,
        x = thread_trail id p /\
        In hl_thread S_hl' (id, list_M, e)) as x_is_thread_trail_and_x_in_S_hl'.
apply trails_in_S_hl'. apply x_in_one.
destruct x_is_thread_trail_and_x_in_S_hl' as
  [ id [ p [ list_M [ e [ x_is_thread_trail x_in_S_hl' ] ] ] ] ].
assert (list_M = nil /\ (exists v' : exp, value v' /\ e = alwaysEvt v'))
  as list_M_is_nil.
eapply S_hl'_complete. apply x_in_S_hl'.
destruct list_M_is_nil as 
  [ list_M_is_nil [ v' [ value_v' e_is_aevt_v' ] ] ].
subst x e list_M.
eapply In_thread_trails.
apply In_S_no_ld.
apply x_in_S_hl'.
apply x_in_one.
inversion x_in_one. inversion H0. subst. apply H8.
reflexivity.
apply comm_trails.

Qed.

Corollary hl2ll_many : forall H T H' T',
  mt_unique_tids T ->
  hl_mt_step_many H T H' T' ->
  ll_step_many (unfrozen H) T empty_susp empty_heap empty_search
  (unfrozen H') T' empty_susp empty_heap empty_search.

intros H T H' T' T_unique hl_steps_many.
induction hl_steps_many.

apply ll_step_many_zero.

eapply ll_step_many_many.
apply hl2ll. auto. apply H.
apply IHhl_steps_many.
eapply hl_mt_step_preserves_unique_tids.
apply H. auto.

Qed.

Corollary hl2ll_from_beginning : 
  forall id e H T,
    hl_mt_step_many fresh_heap (Singleton mt_thread (id, e)) H T ->
    ll_step_many 
    (unfrozen fresh_heap) (Singleton mt_thread (id, e)) 
    empty_susp empty_heap empty_search
    (unfrozen H) T empty_susp empty_heap empty_search.

intros id e H T hl_steps_many.
apply hl2ll_many.
unfold mt_unique_tids.
intros t1 t2 id1 id2 e1 e2 t1_eq t2_eq t1_is_id t2_is_id t1_not_t2.
inversion t1_is_id. inversion t2_is_id. subst t1 t2.
inversion H0. inversion H1. subst. subst id2 e2. tauto.
auto.

Qed.
    