Require Import coherent.
Require Import committable.
Require Import committable_after_choose.
Require Import committable_after_send_recv.
Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Sets.Ensembles.
Require Import lowlevel.
Require Import highlevel.
Require Import syntax.
Require Import trails.
Require Import unique_tids.
Require Import hl2ll_get_S.

Lemma hl2ll_ld : 
  forall H' H'' S_hl S_hl' s s' thread_trails hp trails 
    S_H S S0 S_all H T S_susp,
    hl_sync_step H' S_hl (Some s) H'' S_hl' (Some s') ->
    ld_unique_tids S_hl (Some s) ->
    (forall t, In trail thread_trails t ->
      exists id, exists p, exists list_M, exists e, 
        t = thread_trail id p /\ 
        (In hl_thread S_hl (id, list_M, e) \/ 
          s = (id, list_M, e))) ->
    (forall id list_M e,  
      (In hl_thread S_hl (id, list_M, e)
        \/ s = (id, list_M, e)) ->
      exists p, In trail thread_trails (thread_trail id p)) ->
      S = get_S_ld H' S_hl s thread_trails hp ->
      S_all = Union search_thread S S0 ->
      trails = Add trail thread_trails (heap_trail hp) ->
        committable trails ->
        exists thread_trails', exists hp', exists S_H', exists S0',
          let S' := get_S_ld H'' S_hl' s' thread_trails' hp' in
            let S_all' := Union search_thread S' S0' in
              let trails' := Add trail thread_trails' (heap_trail hp') in
                (forall t', 
                  In trail thread_trails' t' ->
                  exists id, exists p', exists list_M', exists e', 
                    t' = thread_trail id p' /\
                    (In hl_thread S_hl' (id, list_M', e')
                      \/ s' = (id, list_M', e'))) /\
                (forall id list_M' e',  
                  (In hl_thread S_hl' (id, list_M', e') 
                    \/ s' = (id, list_M', e')) ->
                  exists p', 
                    In trail thread_trails' (thread_trail id p')) /\
                committable trails' /\
                (ll_step (frozen H) T S_susp S_H S_all
                  (frozen H) T S_susp S_H' S_all').

intros H' H'' S_hl S_hl' s s' thread_trails hp trails S_H S S0 S_all H T 
  S_susp hl_steps hl_before_unique trails_in_hl_before hl_before_in_trails
  S_eq S_all_eq trails_eq comm_before.

subst S S_all trails.

induction s as [ [ id_ld list_M_ld ] e_ld ].

assert (exists! p_ld, In trail thread_trails (thread_trail id_ld p_ld)) 
  as p_ld_exists_unique.
assert (exists p_ld, In trail thread_trails (thread_trail id_ld p_ld)) 
  as p_ld_exists. eapply hl_before_in_trails. right. reflexivity.
destruct p_ld_exists as [ p_ld p_ld_in_thread_trails ].
exists p_ld. split. auto.
intros p_ld' p_ld'_in_thread_trails.
unfold committable in comm_before.
destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
eapply uniq_thread. left. eauto. left. eauto.
destruct p_ld_exists_unique as [ p_ld p_ld_unique ].
unfold unique in p_ld_unique.
destruct p_ld_unique as [ p_ld_in_thread_trails p_ld_unique ].
assert (forall list_M e,
  ~ In hl_thread S_hl (id_ld, list_M, e)) as id_ld_not_in_S_hl'.
intros list_M e.
unfold ld_unique_tids in hl_before_unique.
destruct hl_before_unique as [ S_hl_unique id_ld_not_in_S_hl ].
unfold not. intros in_S_hl. unfold not in id_ld_not_in_S_hl.
eapply id_ld_not_in_S_hl. eauto. reflexivity.

inversion hl_steps as [ 
  _H' _S_hl _id_ld _list_M_ld _e_ld E v1 v2 e_ld_is_then H'_eq S_hl_eq 
  id_ld_eq H''_eq S_hl'_eq s'_eq
  | _H' _S_hl _id_ld _list_M_ld _e_ld E v1 v2 e_ld_is_choose H'_eq S_hl_eq 
    id_ld_eq H'_is_H'' S_hl'_is_S_hl'' s'_eq
  | _H' _S_hl _id_ld _list_M_ld _e_ld E v1 v2 e_ld_is_choose H'_eq S_hl_eq 
    id_ld_eq H'_is_H'' S_hl'_is_S_hl'' s'_eq 
  | | 
  | _H' _H'' _S_hl _id_ld _list_M_ld _e_ld e_ld' st_steps H'_eq S_hl_eq 
    id_ld_eq H''_eq S_hl'_eq s'_eq
  | _H' id_s list_M_s e_s E_s k v id_r list_M_r e_r E_r _S_hl _S_hl' 
    s_opt sender_in_S_hl receiver_in_S_hl e_s_is_send e_r_is_recv eq_S_hl' 
    eq_H' eq_S_hl eq_Some H'_is_H'' eq_S_hl'
  | _H' _S_hl _id_ld _list_M_ld _e_ld M v e_ld_is_sync H'_eq S_hl_eq 
    id_ld_eq H'_is_H'' S_hl_is_S_hl' s'_eq
  | _H' _S_hl _id_ld M list_M_ld' v value_v H'_eq S_hl_eq id_ld_eq H'_is_H'' 
    S_hl_is_S_hl' s'_eq ].

(* ThenAlways *)
subst _H' _S_hl H'' S_hl' s' _id_ld _list_M_ld _e_ld.

exists thread_trails.
exists hp.
exists S_H.
exists S0.

intros S' S_all' trails. subst S' S_all' trails.

(* trails matches S_hl' *)
split.
  intros t t_in_trails.
  assert (exists id, exists p, exists list_M, exists e,
    t = thread_trail id p /\
    (In hl_thread S_hl (id, list_M, e) \/
      (id_ld, list_M_ld, e_ld) = (id, list_M, e))) as ok_before.
  apply trails_in_hl_before. auto.
  destruct ok_before as [ id [ p [ list_M [ e [ is_thread in_hl_or_s ] ] ] ] ].
  exists id. exists p. exists list_M.
  destruct in_hl_or_s as [ in_hl | is_s ].
  exists e. auto.
  inversion is_s. subst. exists (evt_fill_hole E (app v2 v1)). auto.

(* S_hl' matches trails *)
split.
  intros id list_M e in_hl_or_s.
  destruct in_hl_or_s; try inversion H0; subst;
    eapply hl_before_in_trails; eauto.

(* committable *)
split. auto.

(* steps in low-level *)
remember (Union search_thread
  (Intersection search_thread
    (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp)
    (get_S_ld H' S_hl (id_ld, list_M_ld, evt_fill_hole E (app v2 v1))
      thread_trails hp))
  S0) as S''.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x x_in_S0 ].
inversion x_in_S as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, evt_fill_hole E (app v2 v1))
    thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H', hp), list_M_ld, 
    evt_fill_hole E (app v2 v1), p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S' | x x_in_S0 ].
inversion x_in_S' as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

apply ll_step_then_always.
auto.

(* ChooseLeft *)
subst _H' _S_hl H'' S_hl' s' _id_ld _list_M_ld _e_ld.

Inductive get_thread_trails' (thread_trails:Ensemble trail) 
  (id_ld:tid) : Ensemble trail :=

| In_thread_trails'_not_ld : forall id p,
  In trail thread_trails (thread_trail id p) ->
  id <> id_ld ->
  In trail (get_thread_trails' thread_trails id_ld)
  (thread_trail id p)

| In_thread_trails'_ld : forall id p,
  In trail thread_trails (thread_trail id p) ->
  id = id_ld ->
  In trail (get_thread_trails' thread_trails id_ld)
  (thread_trail id (path_cons path_left p)).

exists (get_thread_trails' thread_trails id_ld).
exists hp.
exists S_H.
exists (Add search_thread S0 (id_ld, Some (H', hp), list_M_ld,
  evt_fill_hole E v2, path_cons path_right p_ld)).

intros S' S_all' trails'. subst S' S_all' trails'.

(* trails' matches S_hl' *)
split. intros t t_in_thread_trails.
inversion t_in_thread_trails as [ id p in_thread_trails not_ld is_trail | 
  _id_ld p in_trails id_ld_eq t_eq ].
subst t.
assert (exists id', exists p', exists list_M, exists e,
  thread_trail id p = thread_trail id' p' /\
  (In hl_thread S_hl (id', list_M, e) \/
    (id_ld, list_M_ld, e_ld) = (id', list_M, e))) as ok_before.
apply trails_in_hl_before. auto.
destruct ok_before as [ id' [ p' [ list_M [ e [ is_trail in_before ] ] ] ] ].
inversion is_trail. subst id' p'.
exists id. exists p. exists list_M. exists e.
split. reflexivity.
destruct in_before as [ in_S_hl | false ].
auto. inversion false. subst id. tauto.

subst _id_ld t.
assert (exists id', exists p', exists list_M, exists e,
  thread_trail id_ld p = thread_trail id' p' /\
  (In hl_thread S_hl (id', list_M, e) \/
    (id_ld, list_M_ld, e_ld) = (id', list_M, e))) as ok_before.
apply trails_in_hl_before. auto.
destruct ok_before as [ id' [ p' [ list_M [ e [ is_trail in_before ] ] ] ] ].
inversion is_trail. subst id' p'.
destruct in_before as [ false | is_ld ].
assert False. unfold not in id_ld_not_in_S_hl'. eapply id_ld_not_in_S_hl'.
eauto. tauto.
inversion is_ld. subst list_M e.
exists id_ld. exists (path_cons path_left p). exists list_M_ld.
exists (evt_fill_hole E v1). auto.

(* S_hl in trails' *)
split. intros id list_M e in_hl_or_ld.
destruct in_hl_or_ld as [ in_hl | is_ld ].
assert (exists p, In trail thread_trails (thread_trail id p)) as p_exists.
eauto. destruct p_exists as [ p p_in_trails ].
exists p. apply In_thread_trails'_not_ld. auto.
unfold not. intros is_ld. subst id. 
unfold not in id_ld_not_in_S_hl'. eapply id_ld_not_in_S_hl'. eauto.
exists (path_cons path_left p_ld).
inversion is_ld. subst id list_M e.
apply In_thread_trails'_ld. auto. auto.

(* committable *)
split.
remember (Intersection trail thread_trails
  (get_thread_trails' thread_trails id_ld)) as trails''.
assert (Add trail (get_thread_trails' thread_trails id_ld) (heap_trail hp) =
  Union trail (Add trail trails'' (heap_trail hp)) 
    (Singleton trail (thread_trail id_ld (path_cons path_left p_ld)))) 
as rewriting.
apply Extensionality_Ensembles. subst trails''. split; intros x x_in_one.
destruct x_in_one as [ x x_in_tt' | x x_is_ht ].
inversion x_in_tt' as [ id p in_tt not_ld x_eq | 
  _id_ld p in_tt id_ld_eq x_eq ].
left. left. split. auto. apply In_thread_trails'_not_ld. auto. auto.
subst _id_ld x. right. assert (p_ld = p). apply p_ld_unique; auto.
subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_t'' | x x_is_ld].
destruct x_in_t'' as [ x x_in_int | x x_is_ht ].
destruct x_in_int as [ x x_in_tt x_in_tt' ].
left. auto. right. auto.
inversion x_is_ld. left. apply In_thread_trails'_ld. auto. auto.
rewrite -> rewriting. clear rewriting. subst trails''.

apply committable_after_choose_left.
unfold not. intros ld_in_t'. 
remember (thread_trail id_ld p_ld) as ld.
destruct ld_in_t' as [ ld ld_in_int | ld ld_is_ht ].
destruct ld_in_int as [ ld ld_in_tt ld_in_tt' ].
subst ld. inversion ld_in_tt' as [ | id p ]. tauto.
assert (p_ld = p) as p_ld_is_p. apply p_ld_unique. auto. subst p id.
assert (p_ld <> path_cons path_left p_ld).
apply paths_are_finite. symmetry in H1. tauto.
subst ld. inversion ld_is_ht.

assert (Union trail
  (Add trail
    (Intersection trail thread_trails
      (get_thread_trails' thread_trails id_ld)) 
    (heap_trail hp)) (Singleton trail (thread_trail id_ld p_ld))
  = Add trail thread_trails (heap_trail hp)) as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_tt'' | x x_is_ld ].
destruct x_in_tt'' as [ x x_in_int | x x_is_ht ].
destruct x_in_int as [ x x_in_tt x_in_tt' ].
left. auto. right. auto. inversion x_is_ld. left. auto.
destruct x_in_one as [ x x_in_tt | x x_is_ht ].
assert (x = thread_trail id_ld p_ld \/ 
  x <> thread_trail id_ld p_ld) as classic. apply classic.
destruct classic as [ x_is_ld | x_not_ld ]. subst x. right. apply In_singleton.
left. left. split. auto.
assert (exists id, exists p, exists list_M, exists e,
  x = thread_trail id p /\
  (In hl_thread S_hl (id, list_M, e) \/
    (id_ld, list_M_ld, e_ld) = (id, list_M, e))) as x_is_trail.
apply trails_in_hl_before. auto.
destruct x_is_trail as [ id [ p [ list_M [ e [ x_is_trail other ] ] ] ] ].
subst x. apply In_thread_trails'_not_ld. auto.
unfold not. intros id_is_ld. subst id.
assert (p_ld = p) as false. apply p_ld_unique. auto.
subst p. tauto.
left. right. auto.
rewrite -> rewriting. auto.

(* ll steps *)
remember (Union search_thread
  (Intersection search_thread
    (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp)
    (get_S_ld H' S_hl (id_ld, list_M_ld, evt_fill_hole E v1)
      (get_thread_trails' thread_trails id_ld) hp)) S0) as S''.
assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0 =
  (Add search_thread S''
    (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld))) as rewriting.
subst S''. apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x x_in_S0 ].
inversion x_in_S as [ id list_M e p in_S_hl in_tt x_eq | 
  _id_ld _list_M_ld  _e_ld p eq in_tt x_eq ].
assert (x = (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld) \/
  x <> (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld)) as classic.
apply classic. destruct classic as [ x_is_ld | x_not_ld ].
right. subst x. rewrite -> x_is_ld. apply In_singleton.
left. left.
split; apply In_S_not_ld; auto.
apply In_thread_trails'_not_ld. auto.
unfold not. intros eq. subst id. unfold not in id_ld_not_in_S_hl'.
eapply id_ld_not_in_S_hl'. eauto.
inversion eq. subst _id_ld _list_M_ld _e_ld x.
assert (p_ld = p) as p_ld_eq. apply p_ld_unique. auto. subst p.
right. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ].
left. auto. right. auto.
destruct x_is_ld.
left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, evt_fill_hole E v1)
    (get_thread_trails' thread_trails id_ld) hp)
  (Add search_thread S0
    (id_ld, Some (H', hp), list_M_ld, evt_fill_hole E v2,
      path_cons path_right p_ld)) =
  Union search_thread S''
  (Couple search_thread
    (id_ld, Some (H', hp), list_M_ld, 
      evt_fill_hole E v1, path_cons path_left p_ld)
    (id_ld, Some (H', hp), list_M_ld, 
      evt_fill_hole E v2, path_cons path_right p_ld))) as rewriting.
subst S''. apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x' [ x x_in_S0 | x x_is_right ] ].
inversion x_in_S as [ id list_M e p in_S_hl in_tt' x_eq 
  | _id_ld _list_M_ld _e_ld p eq in_tt' x_eq ].
subst x. left. left. split; apply In_S_not_ld.
auto. inversion in_tt'. auto. subst.
unfold not in id_ld_not_in_S_hl'. assert False. eapply id_ld_not_in_S_hl'.
eauto. inversion H0. auto. auto.
inversion eq. subst _id_ld _list_M_ld _e_ld x.
inversion in_tt'. tauto. subst p.
assert (p_ld = p0). apply p_ld_unique. auto. subst p0.
right. left. left. right. auto.
inversion x_is_right. subst x. right. right.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ].
left. auto. right. left. auto.
destruct x_is_ld.
left. apply In_ld. auto. apply In_thread_trails'_ld. auto. auto.
right. right. apply In_singleton.

rewrite -> rewriting. clear rewriting.
apply ll_step_choose. auto.

(* ChooseRight *)
subst _H' _S_hl H'' S_hl' s' _id_ld _list_M_ld _e_ld.

Inductive get_thread_trails'2 (thread_trails:Ensemble trail) 
  (id_ld:tid) : Ensemble trail :=

| In_thread_trails2'_not_ld : forall id p,
  In trail thread_trails (thread_trail id p) ->
  id <> id_ld ->
  In trail (get_thread_trails'2 thread_trails id_ld)
  (thread_trail id p)

| In_thread_trails2'_ld : forall id p,
  In trail thread_trails (thread_trail id p) ->
  id = id_ld ->
  In trail (get_thread_trails'2 thread_trails id_ld)
  (thread_trail id (path_cons path_right p)).

exists (get_thread_trails'2 thread_trails id_ld).
exists hp.
exists S_H.
exists (Add search_thread S0 (id_ld, Some (H', hp), list_M_ld,
  evt_fill_hole E v1, path_cons path_left p_ld)).

intros S' S_all' trails'. subst S' S_all' trails'.

(* trails' matches S_hl' *)
split. intros t t_in_thread_trails.
inversion t_in_thread_trails as [ id p in_thread_trails not_ld is_trail | 
  _id_ld p in_trails id_ld_eq t_eq ].
subst t.
assert (exists id', exists p', exists list_M, exists e,
  thread_trail id p = thread_trail id' p' /\
  (In hl_thread S_hl (id', list_M, e) \/
    (id_ld, list_M_ld, e_ld) = (id', list_M, e))) as ok_before.
apply trails_in_hl_before. auto.
destruct ok_before as [ id' [ p' [ list_M [ e [ is_trail in_before ] ] ] ] ].
inversion is_trail. subst id' p'.
exists id. exists p. exists list_M. exists e.
split. reflexivity.
destruct in_before as [ in_S_hl | false ].
auto. inversion false. subst id. tauto.

subst _id_ld t.
assert (exists id', exists p', exists list_M, exists e,
  thread_trail id_ld p = thread_trail id' p' /\
  (In hl_thread S_hl (id', list_M, e) \/
    (id_ld, list_M_ld, e_ld) = (id', list_M, e))) as ok_before.
apply trails_in_hl_before. auto.
destruct ok_before as [ id' [ p' [ list_M [ e [ is_trail in_before ] ] ] ] ].
inversion is_trail. subst id' p'.
destruct in_before as [ false | is_ld ].
assert False. unfold not in id_ld_not_in_S_hl'. eapply id_ld_not_in_S_hl'.
eauto. tauto.
inversion is_ld. subst list_M e.
exists id_ld. exists (path_cons path_right p). exists list_M_ld.
exists (evt_fill_hole E v2). auto.

(* S_hl in trails' *)
split. intros id list_M e in_hl_or_ld.
destruct in_hl_or_ld as [ in_hl | is_ld ].
assert (exists p, In trail thread_trails (thread_trail id p)) as p_exists.
eauto. destruct p_exists as [ p p_in_trails ].
exists p. apply In_thread_trails2'_not_ld. auto.
unfold not. intros is_ld. subst id. 
unfold not in id_ld_not_in_S_hl'. eapply id_ld_not_in_S_hl'. eauto.
exists (path_cons path_right p_ld).
inversion is_ld. subst id list_M e.
apply In_thread_trails2'_ld. auto. auto.

(* committable *)
split.
remember (Intersection trail thread_trails
  (get_thread_trails'2 thread_trails id_ld)) as trails''.
assert (Add trail (get_thread_trails'2 thread_trails id_ld) (heap_trail hp) =
  Union trail (Add trail trails'' (heap_trail hp)) 
    (Singleton trail (thread_trail id_ld (path_cons path_right p_ld)))) 
as rewriting.
apply Extensionality_Ensembles. subst trails''. split; intros x x_in_one.
destruct x_in_one as [ x x_in_tt' | x x_is_ht ].
inversion x_in_tt' as [ id p in_tt not_ld x_eq | 
  _id_ld p in_tt id_ld_eq x_eq ].
left. left. split. auto. apply In_thread_trails2'_not_ld. auto. auto.
subst _id_ld x. right. assert (p_ld = p). apply p_ld_unique; auto.
subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_t'' | x x_is_ld].
destruct x_in_t'' as [ x x_in_int | x x_is_ht ].
destruct x_in_int as [ x x_in_tt x_in_tt' ].
left. auto. right. auto.
inversion x_is_ld. left. apply In_thread_trails2'_ld. auto. auto.
rewrite -> rewriting. clear rewriting. subst trails''.

apply committable_after_choose_right.
unfold not. intros ld_in_t'. 
remember (thread_trail id_ld p_ld) as ld.
destruct ld_in_t' as [ ld ld_in_int | ld ld_is_ht ].
destruct ld_in_int as [ ld ld_in_tt ld_in_tt' ].
subst ld. inversion ld_in_tt' as [ | id p ]. tauto.
assert (p_ld = p) as p_ld_is_p. apply p_ld_unique. auto. subst p id.
assert (p_ld <> path_cons path_right p_ld).
apply paths_are_finite. symmetry in H1. tauto.
subst ld. inversion ld_is_ht.

assert (Union trail
  (Add trail
    (Intersection trail thread_trails
      (get_thread_trails'2 thread_trails id_ld)) 
    (heap_trail hp)) (Singleton trail (thread_trail id_ld p_ld))
  = Add trail thread_trails (heap_trail hp)) as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_tt'' | x x_is_ld ].
destruct x_in_tt'' as [ x x_in_int | x x_is_ht ].
destruct x_in_int as [ x x_in_tt x_in_tt' ].
left. auto. right. auto. inversion x_is_ld. left. auto.
destruct x_in_one as [ x x_in_tt | x x_is_ht ].
assert (x = thread_trail id_ld p_ld \/ 
  x <> thread_trail id_ld p_ld) as classic. apply classic.
destruct classic as [ x_is_ld | x_not_ld ]. subst x. right. apply In_singleton.
left. left. split. auto.
assert (exists id, exists p, exists list_M, exists e,
  x = thread_trail id p /\
  (In hl_thread S_hl (id, list_M, e) \/
    (id_ld, list_M_ld, e_ld) = (id, list_M, e))) as x_is_trail.
apply trails_in_hl_before. auto.
destruct x_is_trail as [ id [ p [ list_M [ e [ x_is_trail other ] ] ] ] ].
subst x. apply In_thread_trails2'_not_ld. auto.
unfold not. intros id_is_ld. subst id.
assert (p_ld = p) as false. apply p_ld_unique. auto.
subst p. tauto.
left. right. auto.
rewrite -> rewriting. auto.

(* ll steps *)
remember (Union search_thread
  (Intersection search_thread
    (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp)
    (get_S_ld H' S_hl (id_ld, list_M_ld, evt_fill_hole E v2)
      (get_thread_trails'2 thread_trails id_ld) hp)) S0) as S''.
assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0 =
  (Add search_thread S''
    (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld))) as rewriting.
subst S''. apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x x_in_S0 ].
inversion x_in_S as [ id list_M e p in_S_hl in_tt x_eq | 
  _id_ld _list_M_ld  _e_ld p eq in_tt x_eq ].
assert (x = (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld) \/
  x <> (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld)) as classic.
apply classic. destruct classic as [ x_is_ld | x_not_ld ].
right. subst x. rewrite -> x_is_ld. apply In_singleton.
left. left.
split; apply In_S_not_ld; auto.
apply In_thread_trails2'_not_ld. auto.
unfold not. intros eq. subst id. unfold not in id_ld_not_in_S_hl'.
eapply id_ld_not_in_S_hl'. eauto.
inversion eq. subst _id_ld _list_M_ld _e_ld x.
assert (p_ld = p) as p_ld_eq. apply p_ld_unique. auto. subst p.
right. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ].
left. auto. right. auto.
destruct x_is_ld.
left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, evt_fill_hole E v2)
    (get_thread_trails'2 thread_trails id_ld) hp)
  (Add search_thread S0
    (id_ld, Some (H', hp), list_M_ld, evt_fill_hole E v1,
      path_cons path_left p_ld)) =
  Union search_thread S''
  (Couple search_thread
    (id_ld, Some (H', hp), list_M_ld, 
      evt_fill_hole E v1, path_cons path_left p_ld)
    (id_ld, Some (H', hp), list_M_ld, 
      evt_fill_hole E v2, path_cons path_right p_ld))) as rewriting.
subst S''. apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x' [ x x_in_S0 | x x_is_left ] ].
inversion x_in_S as [ id list_M e p in_S_hl in_tt' x_eq 
  | _id_ld _list_M_ld _e_ld p eq in_tt' x_eq ].
subst x. left. left. split; apply In_S_not_ld.
auto. inversion in_tt'. auto. subst.
unfold not in id_ld_not_in_S_hl'. assert False. eapply id_ld_not_in_S_hl'.
eauto. inversion H0. auto. auto.
inversion eq. subst _id_ld _list_M_ld _e_ld x.
inversion in_tt'. tauto. subst p.
assert (p_ld = p0). apply p_ld_unique. auto. subst p0.
right. right. left. right. auto.
inversion x_is_left. subst x. right. left.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ].
left. auto. right. left. auto.
destruct x_is_ld.
right. right. apply In_singleton.
left. apply In_ld. auto. apply In_thread_trails2'_ld. auto. auto.

rewrite -> rewriting. clear rewriting.
apply ll_step_choose. auto.

(* StepLockdown *)
subst _H' _H'' _S_hl _id_ld _list_M_ld _e_ld S_hl' s'.

exists thread_trails.
exists hp.
exists S_H.
exists S0.

intros S' S_all' trails'. subst S' S_all' trails'.

(* trails matches S_hl' *)
split.
  intros t t_in_trails.
  assert (exists id, exists p, exists list_M, exists e,
    t = thread_trail id p /\
    (In hl_thread S_hl (id, list_M, e) \/
      (id_ld, list_M_ld, e_ld) = (id, list_M, e))) as ok_before.
  apply trails_in_hl_before. auto.
  destruct ok_before as [ id [ p [ list_M [ e [ is_thread in_hl_or_s ] ] ] ] ].
  exists id. exists p. exists list_M.
  destruct in_hl_or_s as [ in_hl | is_s ].
  exists e. auto.
  inversion is_s. subst. exists e_ld'. auto.

(* S_hl' matches trails *)
split.
  intros id list_M e in_hl_or_s.
  destruct in_hl_or_s; try inversion H0; subst;
    eapply hl_before_in_trails; eauto.

(* committable *)
split. auto.

(* steps in low-level *)
remember (Union search_thread
  (Intersection search_thread
    (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp)
    (get_S_ld H'' S_hl (id_ld, list_M_ld, e_ld')
      thread_trails hp))
  S0) as S''.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x x_in_S0 ].
inversion x_in_S as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

assert (Union search_thread
  (get_S_ld H'' S_hl (id_ld, list_M_ld, e_ld')
    thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H'', hp), list_M_ld, 
    e_ld', p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S' | x x_in_S0 ].
inversion x_in_S' as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

apply ll_step_lockdown.
auto.

(* SendRecv *)
Focus.

subst _H' _S_hl H'' _S_hl'.

assert (exists p_s, In trail thread_trails (thread_trail id_s p_s)) 
  as p_s_exists. eauto.
assert (exists p_r, In trail thread_trails (thread_trail id_r p_r)) 
  as p_r_exists. eauto.
destruct p_s_exists as [ p_s p_s_in_thread_trails ].
destruct p_r_exists as [ p_r p_r_in_thread_trails ].

inversion hl_before_unique as [ S_hl_unique ld_unique ].

assert (id_s <> id_r) as id_s_not_id_r.
unfold not. intros id_s_is_id_r. subst id_r.
assert ((id_s, list_M_s, e_s) = (id_s, list_M_r, e_r)) as e_s_is_e_r.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
apply NNPP. unfold not at 1. intros e_s_not_e_r.
eapply S_hl_unique. reflexivity. reflexivity.
apply sender_in_S_hl. apply receiver_in_S_hl. auto. reflexivity.
inversion e_s_is_e_r. subst e_r.
assert (sendEvt k v = recvEvt k) as false.
eapply proj2. eapply get_evt_ctxt_unique. apply e_s_is_send. apply e_r_is_recv.
inversion false.

remember (get_fresh_nonce S_H (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0)) as n_sr.

Inductive get_thread_trails_sr_ld (thread_trails:Ensemble trail)
  (n_sr : nonce) (id_s : tid) (p_s : path) 
  (id_r : tid) (p_r : path) : Ensemble trail :=
| In_thread_trails_sr_ld_sender : 
  In trail (get_thread_trails_sr_ld thread_trails n_sr id_s p_s id_r p_r)
  (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s))
| In_thread_trails_sr_ld_receiver : 
  In trail (get_thread_trails_sr_ld thread_trails n_sr id_s p_s id_r p_r)
  (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))
| In_thread_trails_sr_ld_other : forall id p,
  id <> id_s -> id <> id_r ->
  In trail thread_trails (thread_trail id p) ->
  In trail (get_thread_trails_sr_ld thread_trails n_sr id_s p_s id_r p_r)
  (thread_trail id p).

exists (get_thread_trails_sr_ld thread_trails n_sr id_s p_s id_r p_r).

exists hp. exists S_H.

exists (Union search_thread S0
  (Couple search_thread
    (id_s, None, list_M_s, e_s, p_s)
    (id_r, None, list_M_r, e_r, p_r))).

intros S' S_all' trails'. subst S' S_all' trails'.
split. Focus 2. split. Focus 2. split.

remember (Add trail (Setminus trail thread_trails
  (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r)))
(heap_trail hp)) as c.
remember (get_thread_trails_sr_ld thread_trails n_sr id_s p_s id_r p_r) 
as thread_trails'.
remember (Add trail (Setminus trail thread_trails'
  (Couple trail 
    (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s))
    (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))))
(heap_trail hp)) as c'.
assert ((Add trail thread_trails (heap_trail hp)) =
  (Union trail c 
    (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r)))) as before.
apply Extensionality_Ensembles.
unfold Same_set. split.
unfold Included. intros t t_in_c1.
destruct t_in_c1 as [ t t_is_thread | t t_is_heap ].
subst c.
assert 
  (In trail (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r)) t
    \/
    ~ In trail (Couple trail (thread_trail id_s p_s) 
      (thread_trail id_r p_r)) t) as t_is_comm_or_not. 
apply classic. destruct t_is_comm_or_not as [ t_is_comm | t_not_comm ].
right. apply t_is_comm. left. left. unfold Setminus. split; auto.
left. subst c. right. auto.
unfold Included. intros t t_in_c2.
destruct t_in_c2 as [ t t_not_comm | t t_is_comm ].
subst c. destruct t_not_comm as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_thread_trails t_not_comm ].
left. auto. right. auto.
destruct t_is_comm. left. auto. left. auto.

assert ((Add trail thread_trails' (heap_trail hp)) =
  (Union trail c' 
    (Couple trail (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s)) 
      (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))))) as after.
apply Extensionality_Ensembles.
unfold Same_set. split.
unfold Included. intros t t_in_c1'.
destruct t_in_c1' as [ t t_is_thread | t t_is_heap ].
subst c'.
assert 
  (In trail (Couple trail 
    (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s)) 
    (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))) t
    \/
    ~ In trail (Couple trail 
      (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s)) 
      (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))) t)
    as t_is_comm_or_not.
apply classic. destruct t_is_comm_or_not as [ t_is_comm | t_not_comm ].
right. apply t_is_comm. left. left. unfold Setminus. split; auto.
left. subst c'. right. auto.
unfold Included. intros t t_in_c2'.
destruct t_in_c2' as [ t t_not_comm | t t_is_comm ].
subst c'. destruct t_not_comm as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_thread_trails' t_not_comm ].
left. auto. right. auto.
left. subst thread_trails'.
destruct t_is_comm. 
apply In_thread_trails_sr_ld_sender.
apply In_thread_trails_sr_ld_receiver.
rewrite -> after. apply committable_after_send_recv.
subst c'. unfold not. intros t_s_in_after.
remember (thread_trail id_s p_s) as t_s.
destruct t_s_in_after as [ t_s t_s_is_thread | t_s t_s_is_heap ].
destruct t_s_is_thread as [ t_s_in_thread_trails' t_s_not_comm ].
unfold In in t_s_in_thread_trails'. subst thread_trails'.
inversion t_s_in_thread_trails'.
subst t_s. inversion H0.
inversion H1. assert (p_s <> path_cons (path_send n_sr id_r p_r) p_s).
apply paths_are_finite. symmetry in H4. tauto.
subst t_s. inversion H0.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
eapply S_hl_unique. reflexivity. reflexivity.
apply sender_in_S_hl. apply receiver_in_S_hl.
unfold not. intros t_s_is_t_r.
inversion t_s_is_t_r. assert (E_s = E_r /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique. apply e_s_is_send. subst e_r. apply e_r_is_recv.
destruct H1. inversion H3. inversion H7. symmetry.
inversion H1. reflexivity.
subst t_s. inversion H4. tauto.
subst t_s. inversion t_s_is_heap.
subst c'. unfold not. intros t_r_in_after.
remember (thread_trail id_r p_r) as t_r.
destruct t_r_in_after as [ t_r t_r_is_thread | t_r t_r_is_heap ].
destruct t_r_is_thread as [ t_r_in_thread_trails' t_r_not_comm ].
unfold In in t_r_in_thread_trails'. subst thread_trails'.
inversion t_r_in_thread_trails'.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
eapply S_hl_unique. reflexivity. reflexivity.
apply receiver_in_S_hl. apply sender_in_S_hl.
unfold not. intros t_r_is_t_s.
inversion t_r_is_t_s. assert (E_s = E_r /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique. apply e_s_is_send. subst e_r. apply e_r_is_recv.
destruct H1. inversion H5. 
inversion H2. inversion H7.
subst t_r. inversion H1. reflexivity.
subst t_r. inversion H1.
assert (p_r <> path_cons (path_recv n_sr id_s p_s) p_r).
apply paths_are_finite. symmetry in H3. tauto.
subst t_r. inversion H4. tauto.
subst t_r. inversion t_r_is_heap.

unfold not. intros id_s_is_id_r.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
eapply S_hl_unique. reflexivity. reflexivity.
apply sender_in_S_hl. apply receiver_in_S_hl.
unfold not. intros t_s_is_t_r. inversion t_s_is_t_r.
assert (E_s = E_r /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique. apply e_s_is_send. subst e_r. apply e_r_is_recv.
destruct H0. inversion H4. 
inversion H1. inversion H6.
apply id_s_is_id_r.

assert (c = c') as c_is_c'.
apply Extensionality_Ensembles.
unfold Same_set. split.
subst c c'. unfold Included. intros t t_in_c.
destruct t_in_c as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_trails t_not_comm ].
left. split.
subst thread_trails'.
assert (exists id : tid, exists p : path, 
  exists list_M : list ctxt, exists e : exp,
  t = thread_trail id p /\ (In hl_thread S_hl (id, list_M, e) \/
  (id_ld, list_M_ld, e_ld) = (id, list_M, e))). auto.
destruct H1 as [ id [ p [ list_M [ e 
  [ t_is_id [ id_in_S_hl | id_is_ld ] ] ] ] ] ].
subst t.
assert ((id = id_s \/ id = id_r) \/ ~(id = id_s \/ id = id_r)).
apply classic. destruct H1. destruct H1.
subst id. assert (p = p_s).
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
eapply uniq_thread. left. apply t_in_trails. left. auto.
subst p.
unfold not in t_not_comm. destruct t_not_comm. left.
subst id. assert (p = p_r).
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
eapply uniq_thread. left. apply t_in_trails. left. auto.
subst p.
unfold not in t_not_comm. destruct t_not_comm. right.
assert (id <> id_s /\ id <> id_r).
apply not_or_and. auto. destruct H2.
apply In_thread_trails_sr_ld_other. auto. auto. auto.
subst t. eapply In_thread_trails_sr_ld_other.
inversion id_is_ld. subst id list_M e.
assert (id_s <> id_ld). eapply ld_unique. eauto.
unfold not. intros. subst id_s. tauto.
inversion id_is_ld. subst id list_M e.
assert (id_r <> id_ld). eapply ld_unique. eauto.
unfold not. intros. subst id_r. tauto.
inversion id_is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto.
subst p. auto.
unfold not. intros t_is_comm'. destruct t_is_comm'.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_s = (path_cons (path_send n_sr id_r p_r) p_s)).
eapply uniq_thread. left. apply p_s_in_thread_trails. left. auto.
assert (p_s <> path_cons (path_send n_sr id_r p_r) p_s).
apply paths_are_finite. tauto.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_r = (path_cons (path_recv n_sr id_s p_s) p_r)).
eapply uniq_thread. left. apply p_r_in_thread_trails. left. auto.
assert (p_r <> path_cons (path_recv n_sr id_s p_s) p_r).
apply paths_are_finite. tauto.

right. auto.

subst c c'. unfold Included. intros t t_in_c'.
destruct t_in_c' as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_trails t_not_comm ].
left. unfold Setminus. split. subst thread_trails'.
inversion t_in_trails.
unfold not in t_not_comm. destruct t_not_comm.
subst t. left.
unfold not in t_not_comm. destruct t_not_comm.
subst t. right. auto.
unfold not. intros t_is_comm.
destruct t_is_comm.
subst thread_trails'. inversion t_in_trails.
assert (p_s <> path_cons (path_send n_sr id_r p_r) p_s).
apply paths_are_finite. symmetry in H2. tauto.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_s = p_r).
eapply uniq_thread. left. apply p_s_in_thread_trails. left. subst id_r. auto.
subst p_s. 
assert (p_r <> path_cons (path_recv n_sr id_s p_r) p_r).
apply paths_are_finite. symmetry in H3. tauto.
tauto.
subst thread_trails'. inversion t_in_trails.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_s = p_r).
eapply uniq_thread. left. apply p_s_in_thread_trails. left. subst id_r. auto.
subst p_s.
assert (p_r <> path_cons (path_send n_sr id_r p_r) p_r).
apply paths_are_finite. symmetry in H3. tauto.
assert (p_r <> path_cons (path_recv n_sr id_s p_s) p_r).
apply paths_are_finite. symmetry in H2. tauto.
tauto.

right. auto.

subst c'. subst c. rewrite <- c_is_c'. rewrite <- before. apply comm_before.

Unfocus.

Focus 3.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0
  =
  Union search_thread
  (Union search_thread S0
    (Setminus search_thread
      (Union search_thread
        (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0)
      (Couple search_thread (id_s, None, list_M_s, e_s, p_s) 
        (id_r, None, list_M_r, e_r, p_r))))
  (Couple search_thread (id_s, None, list_M_s, e_s, p_s) 
    (id_r, None, list_M_r, e_r, p_r)))
as before.

apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_set | x x_in_S0 ].
inversion x_in_set as [ id list_M e p in_S_hl in_trails |
  id list_M e p s_eq in_trails ]; subst x.
assert (id = id_s \/ id <> id_s) as [ | id_not_id_s ].
apply classic. intros; subst id.
assert ((id_s, list_M_s, e_s) = (id_s, list_M, e)) as is_sender.
eapply hl_unique_tids_reversed. apply S_hl_unique. auto. auto.
inversion is_sender. subst list_M e.
right. assert (p_s = p).
destruct comm_before as [ unique_heap [ unique_thread other ] ].
eapply unique_thread. left. eauto. left. auto.
subst p. left.
assert (id = id_r \/ id <> id_r) as [ | id_not_id_r ].
apply classic. intros; subst id.
assert ((id_r, list_M_r, e_r) = (id_r, list_M, e)) as is_receiver.
eapply hl_unique_tids_reversed. apply S_hl_unique. auto. auto.
inversion is_receiver. subst list_M e.
right. assert (p_r = p).
destruct comm_before as [ unique_heap [ unique_thread other ] ].
eapply unique_thread. left. eauto. left. auto.
subst p. right.
left. right. split.
left. apply In_S_not_ld. auto. auto.
intros false. inversion false as [ false2 | false2 ].
subst id. tauto. subst id. tauto.
inversion s_eq. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p.
left. right. split. left.
apply In_ld. reflexivity. auto.
intros false. inversion false as [ | ].
left. left. auto.

destruct x_in_one as [ x other | x is_comm ].
destruct other as [ x in_S0 | x other ].
right. auto.
destruct other as [ other not_comm ].
auto.
destruct is_comm as [ | ].
left. apply In_S_not_ld. auto. auto.
left. apply In_S_not_ld. auto. auto.

assert ((Union search_thread
  (get_S_ld H' S_hl' (id_ld, list_M_ld, e_ld)
    (get_thread_trails_sr_ld thread_trails n_sr id_s p_s id_r p_r) hp)
  (Union search_thread S0
    (Couple search_thread (id_s, None, list_M_s, e_s, p_s)
      (id_r, None, list_M_r, e_r, p_r))))
=
(Union search_thread 
  (Union search_thread S0
    (Setminus search_thread
      (Union search_thread
        (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0)
      (Couple search_thread (id_s, None, list_M_s, e_s, p_s) 
        (id_r, None, list_M_r, e_r, p_r))))
  (Union search_thread
    (Couple search_thread (id_s, None, list_M_s, e_s, p_s) (id_r, None, list_M_r, e_r, p_r))
    (Couple search_thread 
      (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero), 
        path_cons (path_send n_sr id_r p_r) p_s) 
      (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v), 
        path_cons (path_recv n_sr id_s p_s) p_r))))) as after.

apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_one | x x_in_one ].
inversion x_in_one as [ id list_M e p in_S_hl in_trails | 
  id list_M e p s_eq in_trails ].
inversion in_trails as [ is_sender | is_receiver | 
  id' p' not_sender not_receiver in_trails' ].
subst id p.
subst S_hl'.
inversion in_S_hl as [ s [ s_in_S_hl s_not_comm ] | s [ | ] ].
assert ((id_s, list_M, e) = (id_s, list_M_s, e_s)) as eq.
eapply hl_unique_tids_reversed. eauto. auto. auto.
inversion eq. subst list_M e.
destruct s_not_comm. left.
subst list_M e. right. right. left.
subst id_r. tauto.
subst id p. subst S_hl'.
inversion in_S_hl as [ s [ s_in_S_hl s_not_comm ] | s [ | ] ].
assert ((id_r, list_M, e) = (id_r, list_M_r, e_r)) as eq.
eapply hl_unique_tids_reversed. eauto. auto. auto.
inversion eq. subst list_M e.
destruct s_not_comm. right.
subst id_r. tauto.
subst list_M e. right. right. right.
subst id' p'. subst S_hl'.
inversion in_S_hl as [ s [ s_in_S_hl s_not_comm ] | s [ | ] ].
left. right. split. left. apply In_S_not_ld. auto. auto.
intros false. inversion false as [ | ].
subst id list_M e p. destruct s_not_comm. left.
subst id list_M e p. destruct s_not_comm. right.
subst id list_M e. tauto.
subst id. tauto.
inversion s_eq. subst id list_M e.
inversion in_trails as [ is_sender | is_receiver | 
  id' p' not_sender not_receiver in_trails' ].
assert (id_s <> id_ld). eapply ld_unique. eauto. tauto.
assert (id_r <> id_ld). eapply ld_unique. eauto. tauto.
subst id' p'. assert (p_ld = p). apply p_ld_unique. auto.
subst p_ld. left. right. split.
left. apply In_ld. reflexivity. auto.
intros false. inversion false as [ | ]; tauto.
destruct x_in_one as [ x in_S0 | x [ | ] ].
left. left. auto. right. left. left. right. left. right.

destruct x_in_one as [ x x_in_one | x x_in_one ].
destruct x_in_one as [ x in_S0 | x other ].
right. left. auto.
destruct other as [ other not_comm ].
destruct other as [ x other | x in_S0 ].
inversion other as [ id list_M e p in_S_hl in_trails | 
  id list_M e p s_eq in_trails ].
left. apply In_S_not_ld. subst S_hl'. left. split. auto.
intros false. inversion false as [ | ]; destruct not_comm.
subst x id list_M e.
assert (p = p_s). unfold committable in comm_before.
destruct comm_before as [ unique_heap [ unique_thread parent ] ].
eapply unique_thread. left. eauto. left. auto.
subst p. left.
subst x id list_M e.
assert (p = p_r). unfold committable in comm_before.
destruct comm_before as [ unique_heap [ unique_thread parent ] ].
eapply unique_thread. left. eauto. left. auto.
subst p. right.
apply In_thread_trails_sr_ld_other.
unfold not. intros false. subst id.
assert ((id_s, list_M, e) = (id_s, list_M_s, e_s)) as eq.
eapply hl_unique_tids_reversed. eauto. auto. auto.
inversion eq. subst list_M e. destruct not_comm. subst x.
assert (p = p_s). unfold committable in comm_before.
destruct comm_before as [ unique_heap [ unique_thread parent ] ].
eapply unique_thread. left. eauto. left. auto.
subst p. left.
unfold not. intros false. subst id.
assert ((id_r, list_M, e) = (id_r, list_M_r, e_r)) as eq.
eapply hl_unique_tids_reversed. eauto. auto. auto.
inversion eq. subst list_M e. destruct not_comm. subst x.
assert (p = p_r). unfold committable in comm_before.
destruct comm_before as [ unique_heap [ unique_thread parent ] ].
eapply unique_thread. left. eauto. left. auto.
subst p. right. auto.
inversion s_eq. subst x id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p.
left. apply In_ld. auto.
apply In_thread_trails_sr_ld_other.
intros false. subst id_s.
unfold not in id_ld_not_in_S_hl'. eapply id_ld_not_in_S_hl'. eauto.
intros false. subst id_r.
unfold not in id_ld_not_in_S_hl'. eapply id_ld_not_in_S_hl'. eauto.
auto.
right. left. auto.
destruct x_in_one as [ x x_in_one | x x_in_one ].
inversion x_in_one as [ | ]; subst x.
right. right. left. right. right. right.
inversion x_in_one as [ | ]; subst x.
left. apply In_S_not_ld. subst S_hl'. right. left.
apply In_thread_trails_sr_ld_sender.
left. apply In_S_not_ld. subst S_hl'. right. right.
apply In_thread_trails_sr_ld_receiver.

rewrite before. rewrite after.
eapply ll_step_send_recv.
subst n_sr. rewrite <- before. apply get_fresh_nonce_correct.
eauto. eauto.
eapply committable_implies_coherent.
eauto. auto. left. auto. left. auto.

Unfocus.

Focus.

intros t' t'_in_thread_trails'.
inversion t'_in_thread_trails' as [ is_sender | is_receiver | 
  id p not_sender not_receiver in_trails' ].
exists id_s. exists (path_cons (path_send n_sr id_r p_r) p_s).
exists list_M_s. exists (evt_fill_hole E_s (alwaysEvt zero)).
split. reflexivity.
left. subst S_hl'. right. left.
exists id_r. exists (path_cons (path_recv n_sr id_s p_s) p_r).
exists list_M_r. exists (evt_fill_hole E_r (alwaysEvt v)).
split. reflexivity.
left. subst S_hl'. right. right.
subst t'.
assert (exists id' : tid, exists p' : path, 
  exists list_M' : list ctxt, exists e' : exp,
  thread_trail id p = thread_trail id' p' /\ 
  (In hl_thread S_hl (id', list_M', e') \/
  (id_ld, list_M_ld, e_ld) = (id', list_M', e')))
as [id' [p' [list_M' [e' [eq in_S_hl ]]]]].
apply trails_in_hl_before. auto.
exists id. exists p. exists list_M'. exists e'. 
split. reflexivity.
destruct in_S_hl.
subst S_hl'. left. left. split.
inversion eq. auto. unfold not. intros false.
inversion eq. subst id' p'.
remember (id, list_M', e') as t'.
inversion false.
subst t'. inversion H2. symmetry in H4. tauto.
subst t'. inversion H2. symmetry in H4. tauto.
right. inversion H1. subst id' list_M' e'.
inversion eq. reflexivity.

Unfocus.

Focus.
intros id list_M' e' e'_in_S_hl'.
subst S_hl'.
remember (id, list_M', e') as t'.
destruct e'_in_S_hl' as [ e'_in_S_hl' | is_ld ].
destruct e'_in_S_hl' as [ t' t'_in_S_hl | t' t'_is_comm' ].
destruct t'_in_S_hl as [ t'_in_S_hl t'_not_comm ].
assert (exists p : path, In trail thread_trails (thread_trail id p))
as [ p p_in_trails ].
eapply hl_before_in_trails. left. subst t'. apply t'_in_S_hl.
exists p. apply In_thread_trails_sr_ld_other.
unfold not. intros id_is_id_s. subst id. subst t'.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique. eapply S_hl_unique.
reflexivity. reflexivity. apply t'_in_S_hl. apply sender_in_S_hl.
unfold not. intros false.
destruct t'_not_comm. rewrite -> false. left. reflexivity.
unfold not. intros id_is_id_r. subst id. subst t'.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique. eapply S_hl_unique.
reflexivity. reflexivity. apply t'_in_S_hl. apply receiver_in_S_hl.
unfold not. intros false.
destruct t'_not_comm. rewrite -> false. right. reflexivity.
auto.
destruct t'_is_comm'.
inversion Heqt'. subst id list_M' e'.
exists (path_cons (path_send n_sr id_r p_r) p_s).
apply In_thread_trails_sr_ld_sender.
inversion Heqt'. subst id list_M' e'.
exists (path_cons (path_recv n_sr id_s p_s) p_r).
apply In_thread_trails_sr_ld_receiver.
subst t'. inversion Heqt'. subst id list_M' e'.
exists p_ld. eapply In_thread_trails_sr_ld_other.
intros false. subst id_s.
assert (id_ld <> id_ld). eapply ld_unique. eauto. tauto.
intros false. subst id_r.
assert (id_ld <> id_ld). eapply ld_unique. eauto. tauto.
auto.
Unfocus.

(* NestedSyncInit *)
subst _H' _S_hl _id_ld _list_M_ld _e_ld H'' S_hl' s'.

exists thread_trails.
exists hp.
exists S_H.
exists S0.

intros S' S_all' trails. subst S' S_all' trails.

(* trails matches S_hl' *)
split.
  intros t t_in_trails.
  assert (exists id, exists p, exists list_M, exists e,
    t = thread_trail id p /\
    (In hl_thread S_hl (id, list_M, e) \/
      (id_ld, list_M_ld, e_ld) = (id, list_M, e))) as ok_before.
  apply trails_in_hl_before. auto.
  destruct ok_before as [ id [ p [ list_M [ e [ is_thread in_hl_or_s ] ] ] ] ].
  exists id. exists p.
  destruct in_hl_or_s as [ in_hl | is_s ].
  exists list_M. exists e. auto.
  inversion is_s. subst. exists (M :: list_M). exists v. auto.

(* S_hl' matches trails *)
split.
  intros id list_M e in_hl_or_s.
  destruct in_hl_or_s; try inversion H0; subst;
    eapply hl_before_in_trails; eauto.

(* committable *)
split. auto.

(* steps in low-level *)
remember (Union search_thread
  (Intersection search_thread
    (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp)
    (get_S_ld H' S_hl (id_ld, M :: list_M_ld, v)
      thread_trails hp))
  S0) as S''.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld, e_ld) thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H', hp), list_M_ld, e_ld, p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x x_in_S0 ].
inversion x_in_S as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, M :: list_M_ld, v)
    thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H', hp), M :: list_M_ld, v, p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S' | x x_in_S0 ].
inversion x_in_S' as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

apply ll_step_nested_sync_init.
auto.

(* NestedSyncComplete *)
subst _H' _S_hl _id_ld list_M_ld H'' S_hl' s' e_ld.

exists thread_trails.
exists hp.
exists S_H.
exists S0.

intros S' S_all' trails. subst S' S_all' trails.

(* trails matches S_hl' *)
split.
  intros t t_in_trails.
  assert (exists id, exists p, exists list_M, exists e,
    t = thread_trail id p /\
    (In hl_thread S_hl (id, list_M, e) \/
      (id_ld, M::list_M_ld', alwaysEvt v) = (id, list_M, e))) as ok_before.
  apply trails_in_hl_before. auto.
  destruct ok_before as [ id [ p [ list_M [ e [ is_thread in_hl_or_s ] ] ] ] ].
  exists id. exists p.
  destruct in_hl_or_s as [ in_hl | is_s ].
  exists list_M. exists e. auto.
  inversion is_s. subst. exists list_M_ld'. exists (M[v]). auto.

(* S_hl' matches trails *)
split.
  intros id list_M e in_hl_or_s.
  destruct in_hl_or_s; try inversion H0; subst;
    eapply hl_before_in_trails; eauto.

(* committable *)
split. auto.

(* steps in low-level *)
remember (Union search_thread
  (Intersection search_thread
    (get_S_ld H' S_hl (id_ld, M::list_M_ld', alwaysEvt v) thread_trails hp)
    (get_S_ld H' S_hl (id_ld, list_M_ld', M[v])
      thread_trails hp))
  S0) as S''.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, M::list_M_ld', alwaysEvt v) thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H', hp), M::list_M_ld', 
    alwaysEvt v, p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S | x x_in_S0 ].
inversion x_in_S as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

assert (Union search_thread
  (get_S_ld H' S_hl (id_ld, list_M_ld', M[v])
    thread_trails hp) S0
  = Add search_thread S'' (id_ld, Some (H', hp), list_M_ld', M[v], p_ld))
as rewriting.
subst S''. apply Extensionality_Ensembles. split ; intros x x_in_one.
destruct x_in_one as [ x x_in_S' | x x_in_S0 ].
inversion x_in_S' as [ id list_M e p in_S_hl in_trails x_eq | 
  id list_M e p is_ld in_trails x_eq ].
left. left. split. apply In_S_not_ld. auto. auto.
apply In_S_not_ld. auto. auto.
right. inversion is_ld. subst id list_M e.
assert (p_ld = p). apply p_ld_unique. auto. subst p. apply In_singleton.
left. right. auto.
destruct x_in_one as [ x x_in_S'' | x x_is_ld ].
destruct x_in_S'' as [ x x_in_int | x x_in_S0 ].
destruct x_in_int as [ x x_in_S x_in_S' ]. left. auto.
right. auto.
inversion x_is_ld. left. apply In_ld. reflexivity. auto.
rewrite -> rewriting. clear rewriting.

apply ll_step_nested_sync_complete.
auto.

Qed.
