Require Import coherent.
Require Import committable.
Require Import committable_after_send_recv.
Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Sets.Ensembles.
Require Import lowlevel.
Require Import highlevel.
Require Import syntax.
Require Import trails.
Require Import unique_tids.
Require Import hl2ll_get_S.

Lemma hl2ll_no_ld : 
  forall H' H'' S_hl S_hl' thread_trails hp trails 
    S_H0 S_H S S0 S_all H T S_susp,
    hl_sync_step H' S_hl None H'' S_hl' None ->
    hl_unique_tids S_hl ->
    (forall t, In trail thread_trails t ->
      exists id, exists p, exists list_M, exists e, 
        t = thread_trail id p /\ In hl_thread S_hl (id, list_M, e)) ->
    (forall id list_M e,  In hl_thread S_hl (id, list_M, e) ->
      exists p, In trail thread_trails (thread_trail id p)) ->
    S_H = Add search_heap S_H0 (H', hp) ->
    S = get_S_no_ld S_hl thread_trails ->
    S_all = Union search_thread S S0 ->
    trails = Add trail thread_trails (heap_trail hp) ->
      committable trails ->
      exists thread_trails', exists hp', exists S_H0', exists S0',
        let S' := get_S_no_ld S_hl' thread_trails' in
          let S_all' := Union search_thread S' S0' in
            let S_H' := Add search_heap S_H0' (H'', hp') in
              let trails' := Add trail thread_trails' (heap_trail hp') in
                (forall t', 
                  In trail thread_trails' t' ->
                  exists id, exists p', exists list_M', exists e', 
                    t' = thread_trail id p' /\ In hl_thread S_hl' (id, list_M', e')) /\
                (forall id list_M' e',  
                  In hl_thread S_hl' (id, list_M', e') ->
                  exists p', 
                    In trail thread_trails' (thread_trail id p')) /\
                committable trails' /\
                (ll_step (frozen H) T S_susp S_H S_all
                  (frozen H) T S_susp S_H' S_all').

intros H' H'' S_hl S_hl' thread_trails hp trails S_H0 S_H
  S S0 S_all H T S_susp hl_steps S_hl_unique trails_in_S_hl 
  S_hl_in_trails eq_S_H eq_S eq_S_all eq_trails comm_before.

subst S_H S S_all trails.

inversion hl_steps as [ | | | | | | 
  _H' id_s list_M_s e_s E_s k v id_r list_M_r e_r E_r _S_hl _S_hl' 
  s_opt sender_in_S_hl receiver_in_S_hl e_s_is_send e_r_is_recv eq_S_hl' 
  eq_H' eq_S_hl eq_None H'_is_H'' eq_S_hl' | | ].

subst _H' _S_hl H'' _S_hl' s_opt.

assert (exists p_s, In trail thread_trails (thread_trail id_s p_s)) 
  as p_s_exists. eauto.
assert (exists p_r, In trail thread_trails (thread_trail id_r p_r)) 
  as p_r_exists. eauto.
destruct p_s_exists as [ p_s p_s_in_thread_trails ].
destruct p_r_exists as [ p_r p_r_in_thread_trails ].

assert (id_s <> id_r) as id_s_not_id_r.
unfold not. intros id_s_is_id_r. subst id_r.
assert ((id_s, list_M_s, e_s) = (id_s, list_M_r, e_r)) as e_s_is_e_r.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
apply NNPP. unfold not at 1. intros e_s_not_e_r.
eapply S_hl_unique. reflexivity. reflexivity.
apply sender_in_S_hl. apply receiver_in_S_hl. auto. reflexivity.
inversion e_s_is_e_r. subst e_r.
assert (sendEvt k v = recvEvt k) as false.
eapply proj2. eapply get_evt_ctxt_unique. apply e_s_is_send. apply e_r_is_recv.
inversion false.

remember (get_fresh_nonce (Add search_heap S_H0 (H', hp)) 
  (Union search_thread (get_S_no_ld S_hl thread_trails) S0)) as n_sr.

Inductive get_thread_trails' (thread_trails:Ensemble trail) (n_sr:nonce)
  (id_s : tid) (p_s : path) (id_r : tid) (p_r : path) : Ensemble trail :=
| In_thread_trails'_sender : 
  In trail (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r)
  (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s))
| In_thread_trails'_receiver : 
  In trail (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r)
  (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))
| In_thread_trails'_other : forall id p,
  id <> id_s -> id <> id_r ->
  In trail thread_trails (thread_trail id p) ->
  In trail (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r)
  (thread_trail id p).

exists (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r).

exists hp. exists S_H0.

exists (Union search_thread S0
  (Couple search_thread
    (id_s, None, list_M_s, e_s, p_s)
    (id_r, None, list_M_r, e_r, p_r))).

intros S' S_all' S_H' trails'. subst S' S_all' S_H' trails'.
split. Focus 2. split. Focus 2. split.

Focus.
remember (Add trail (Setminus trail thread_trails
  (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r)))
(heap_trail hp)) as c.
remember (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r) 
as thread_trails'.
remember (Add trail (Setminus trail thread_trails'
  (Couple trail 
    (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s))
    (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))))
(heap_trail hp)) as c'.
assert ((Add trail thread_trails (heap_trail hp)) =
  (Union trail c 
    (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r)))) as before.
apply Extensionality_Ensembles.
unfold Same_set. split.
unfold Included. intros t t_in_c1.
destruct t_in_c1 as [ t t_is_thread | t t_is_heap ].
subst c.
assert 
  (In trail (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r)) t
    \/
    ~ In trail (Couple trail (thread_trail id_s p_s) 
      (thread_trail id_r p_r)) t) as t_is_comm_or_not. 
apply classic. destruct t_is_comm_or_not as [ t_is_comm | t_not_comm ].
right. apply t_is_comm. left. left. unfold Setminus. split; auto.
left. subst c. right. auto.
unfold Included. intros t t_in_c2.
destruct t_in_c2 as [ t t_not_comm | t t_is_comm ].
subst c. destruct t_not_comm as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_thread_trails t_not_comm ].
left. auto. right. auto.
destruct t_is_comm. left. auto. left. auto.

assert ((Add trail thread_trails' (heap_trail hp)) =
  (Union trail c' 
    (Couple trail (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s)) 
      (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))))) as after.
apply Extensionality_Ensembles.
unfold Same_set. split.
unfold Included. intros t t_in_c1'.
destruct t_in_c1' as [ t t_is_thread | t t_is_heap ].
subst c'.
assert 
  (In trail (Couple trail 
    (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s)) 
    (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))) t
    \/
    ~ In trail (Couple trail 
      (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s)) 
      (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))) t)
    as t_is_comm_or_not.
apply classic. destruct t_is_comm_or_not as [ t_is_comm | t_not_comm ].
right. apply t_is_comm. left. left. unfold Setminus. split; auto.
left. subst c'. right. auto.
unfold Included. intros t t_in_c2'.
destruct t_in_c2' as [ t t_not_comm | t t_is_comm ].
subst c'. destruct t_not_comm as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_thread_trails' t_not_comm ].
left. auto. right. auto.
left. subst thread_trails'.
destruct t_is_comm. 
apply In_thread_trails'_sender.
apply In_thread_trails'_receiver.
rewrite -> after. apply committable_after_send_recv.
subst c'. unfold not. intros t_s_in_after.
remember (thread_trail id_s p_s) as t_s.
destruct t_s_in_after as [ t_s t_s_is_thread | t_s t_s_is_heap ].
destruct t_s_is_thread as [ t_s_in_thread_trails' t_s_not_comm ].
unfold In in t_s_in_thread_trails'. subst thread_trails'.
inversion t_s_in_thread_trails'.
subst t_s. inversion H0. 
assert (p_s <> path_cons (path_send n_sr id_r p_r) p_s).
apply paths_are_finite. symmetry in H2. tauto.
subst t_s. inversion H0.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
eapply S_hl_unique. reflexivity. reflexivity.
apply sender_in_S_hl. apply receiver_in_S_hl.
unfold not. intros t_s_is_t_r.
inversion t_s_is_t_r. assert (E_s = E_r /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique. apply e_s_is_send. subst e_r. apply e_r_is_recv.
destruct H1. inversion H7. symmetry. apply H2.
subst t_s. inversion H3. tauto.
subst t_s. inversion t_s_is_heap.
subst c'. unfold not. intros t_r_in_after.
remember (thread_trail id_r p_r) as t_r.
destruct t_r_in_after as [ t_r t_r_is_thread | t_r t_r_is_heap ].
destruct t_r_is_thread as [ t_r_in_thread_trails' t_r_not_comm ].
unfold In in t_r_in_thread_trails'. subst thread_trails'.
inversion t_r_in_thread_trails'.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
eapply S_hl_unique. reflexivity. reflexivity.
apply receiver_in_S_hl. apply sender_in_S_hl.
unfold not. intros t_r_is_t_s.
inversion t_r_is_t_s. assert (E_s = E_r /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique. apply e_s_is_send. subst e_r. apply e_r_is_recv.
destruct H1. inversion H5. subst t_r.  inversion H0. reflexivity.
subst t_r. inversion H0. 
assert (p_r <> path_cons (path_recv n_sr id_s p_s) p_r).
apply paths_are_finite. symmetry in H2. auto.
subst t_r. inversion H3. tauto.
subst t_r. inversion t_r_is_heap.

unfold not. intros id_s_is_id_r.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
eapply S_hl_unique. reflexivity. reflexivity.
apply sender_in_S_hl. apply receiver_in_S_hl.
unfold not. intros t_s_is_t_r. inversion t_s_is_t_r.
assert (E_s = E_r /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique. apply e_s_is_send. subst e_r. apply e_r_is_recv.
destruct H0. inversion H4. apply id_s_is_id_r.

assert (c = c') as c_is_c'.
apply Extensionality_Ensembles.
unfold Same_set. split.
subst c c'. unfold Included. intros t t_in_c.
destruct t_in_c as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_trails t_not_comm ].
left. split.
subst thread_trails'.
assert (exists id : tid, exists p : path, exists list_M : list ctxt, exists e : exp,
  t = thread_trail id p /\ In hl_thread S_hl (id, list_M, e)). auto.
destruct H0 as [ id [ p [ list_M [ e [ t_is_id id_in_S_hl ] ] ] ] ].
subst t.
assert ((id = id_s \/ id = id_r) \/ ~(id = id_s \/ id = id_r)).
apply classic. destruct H0. destruct H0.
subst id. assert (p = p_s).
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
eapply uniq_thread. left. apply t_in_trails. left. auto.
subst p.
unfold not in t_not_comm. destruct t_not_comm. left.
subst id. assert (p = p_r).
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
eapply uniq_thread. left. apply t_in_trails. left. auto.
subst p.
unfold not in t_not_comm. destruct t_not_comm. right.
assert (id <> id_s /\ id <> id_r).
apply not_or_and. auto. destruct H1.
apply In_thread_trails'_other. auto. auto. auto.
unfold not. intros t_is_comm'. destruct t_is_comm'.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_s = (path_cons (path_send n_sr id_r p_r) p_s)).
eapply uniq_thread. left. apply p_s_in_thread_trails. left. auto.
assert (p_s <> path_cons (path_send n_sr id_r p_r) p_s).
apply paths_are_finite. tauto.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_r = (path_cons (path_recv n_sr id_s p_s) p_r)).
eapply uniq_thread. left. apply p_r_in_thread_trails. left. auto.
assert (p_r <> path_cons (path_recv n_sr id_s p_s) p_r).
apply paths_are_finite. tauto.

right. auto.

subst c c'. unfold Included. intros t t_in_c'.
destruct t_in_c' as [ t t_is_thread | t t_is_heap ].
destruct t_is_thread as [ t_in_trails t_not_comm ].
left. unfold Setminus. split. subst thread_trails'.
inversion t_in_trails.
unfold not in t_not_comm. destruct t_not_comm.
subst t. left.
unfold not in t_not_comm. destruct t_not_comm.
subst t. right. apply H2.
unfold not. intros t_is_comm.
destruct t_is_comm.
subst thread_trails'. inversion t_in_trails.
assert (p_s <> path_cons (path_send n_sr id_r p_r) p_s).
apply paths_are_finite. symmetry in H1. tauto.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_s = p_r).
eapply uniq_thread. left. apply p_s_in_thread_trails. left. subst id_r. auto.
subst p_s. 
assert (p_r <> path_cons (path_recv n_sr id_s p_r) p_r).
apply paths_are_finite. symmetry in H2. tauto.
tauto.
subst thread_trails'. inversion t_in_trails.
unfold committable in comm_before. destruct comm_before as [ uniq_heap [ uniq_thread other ] ].
assert (p_s = p_r).
eapply uniq_thread. left. apply p_s_in_thread_trails. left. subst id_r. auto.
subst p_s.
assert (p_r <> path_cons (path_send n_sr id_r p_r) p_r).
apply paths_are_finite. symmetry in H2. tauto.
assert (p_r <> path_cons (path_recv n_sr id_s p_s) p_r).
apply paths_are_finite. symmetry in H1. tauto.
tauto.

right. auto.

subst c'. subst c. rewrite <- c_is_c'. rewrite <- before. apply comm_before.

Unfocus.

Focus 3.

assert (Union search_thread (get_S_no_ld S_hl thread_trails) S0
  =
  Union search_thread
  (Union search_thread S0
    (Setminus search_thread
      (get_S_no_ld S_hl thread_trails)
      (Couple search_thread (id_s, None, list_M_s, e_s, p_s) 
        (id_r, None, list_M_r, e_r, p_r))))
  (Couple search_thread (id_s, None, list_M_s, e_s, p_s) 
    (id_r, None, list_M_r, e_r, p_r)))
as before.

Focus 2.

assert ((Union search_thread
  (get_S_no_ld S_hl'
    (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r))
  (Union search_thread S0
    (Couple search_thread (id_s, None, list_M_s, e_s, p_s)
      (id_r, None, list_M_r, e_r, p_r))))
=
(Union search_thread 
  (Union search_thread S0
    (Setminus search_thread
      (get_S_no_ld S_hl' (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r))
      (Couple search_thread 
        (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero), 
          path_cons (path_send n_sr id_r p_r) p_s) 
        (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v), 
          path_cons (path_recv n_sr id_s p_s) p_r))))
  (Union search_thread
    (Couple search_thread (id_s, None, list_M_s, e_s, p_s) (id_r, None, list_M_r, e_r, p_r))
    (Couple search_thread 
      (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero), 
        path_cons (path_send n_sr id_r p_r) p_s) 
      (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v), 
        path_cons (path_recv n_sr id_s p_s) p_r))))) as after.

Focus 2.

assert
  (Union search_thread S0
    (Setminus search_thread
      (get_S_no_ld S_hl thread_trails)
      (Couple search_thread (id_s, None, list_M_s, e_s, p_s) 
        (id_r, None, list_M_r, e_r, p_r)))
    = Union search_thread S0
    (Setminus search_thread
      (get_S_no_ld S_hl' (get_thread_trails' thread_trails n_sr id_s p_s id_r p_r))
      (Couple search_thread 
        (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero), 
          path_cons (path_send n_sr id_r p_r) p_s) 
        (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v), 
          path_cons (path_recv n_sr id_s p_s) p_r)))) as eq.

Focus 2.

rewrite -> before.
rewrite -> after.
rewrite <- eq.
eapply ll_step_send_recv.
subst n_sr. rewrite before. apply get_fresh_nonce_correct.
apply e_s_is_send.
apply e_r_is_recv.
eapply committable_implies_coherent.
apply comm_before.
apply id_s_not_id_r.
left. auto. left. auto.
Unfocus.

Focus 3.
apply Extensionality_Ensembles.
split.

intros t t_in_small.
assert (In search_thread
  (Couple search_thread (id_s, None, list_M_s, e_s, p_s)
    (id_r, None, list_M_r, e_r, p_r)) t
  \/ ~ In search_thread
  (Couple search_thread (id_s, None, list_M_s, e_s, p_s)
    (id_r, None, list_M_r, e_r, p_r)) t) as t_comm.
apply classic.
destruct t_comm as [ t_is_comm | t_not_comm ].
right. auto.
destruct t_in_small as [ t t_in_S | t t_in_S0 ].
left. right. split; auto.
left. left. auto.
intros t t_in_big.
destruct t_in_big as [ t t_in_left | t t_is_comm ].
destruct t_in_left as [ t t_in_S0 | t t_in_S ].
right. auto.
destruct t_in_S. left. auto.
left. destruct t_is_comm;
apply In_S_no_ld; auto; auto.
Unfocus.

Focus 3.
apply Extensionality_Ensembles.
split.

intros t t_in_small.
destruct t_in_small as [ t t_in_S' | t other ].
assert (In search_thread (Couple search_thread
  (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
    path_cons (path_send n_sr id_r p_r) p_s)
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) t
\/ ~ In search_thread (Couple search_thread
  (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
    path_cons (path_send n_sr id_r p_r) p_s)
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) t) as t_comm'.
apply classic.
destruct t_comm' as [ t_is_comm' | t_not_comm' ].
right. right. auto.
left. right. split. auto. auto.
destruct other as [ t t_in_S0 | t t_is_comm ].
left. left. auto.
right. left. auto.

intros t t_in_big.
destruct t_in_big as [ t t_in_left | t t_in_right ].
destruct t_in_left as [ t t_in_S0 | t t_in_S' ].
right; left; auto. destruct t_in_S'. left; auto.
destruct t_in_right as [ t t_is_comm | t t_is_comm' ].
right; right; auto.
left. destruct t_is_comm'.
eapply In_S_no_ld. subst S_hl'. right. left.
eapply In_thread_trails'_sender.
eapply In_S_no_ld. subst S_hl'. right. right.
eapply In_thread_trails'_receiver.
Unfocus.

Focus 3.
apply Extensionality_Ensembles.
split.

intros t t_in_version1.
destruct t_in_version1 as [ t t_in_S0 | t t_in_S ].
left; auto.
destruct t_in_S as [ t_in_S t_not_comm ].
induction t as [ [ [ [ id h ] list_M ] e ] p ].
inversion t_in_S. subst id0 h list_M0 e0 p0.
assert (id <> id_s /\ id <> id_r) as id_not_comm.
apply not_or_and. unfold not. intros t_is_comm.
destruct t_is_comm. subst id.
unfold committable in comm_before.
destruct comm_before as [ uniq_heap other ].
destruct other as [ uniq_thread other ].
assert (p = p_s). eapply uniq_thread.
left. apply H6. left. apply p_s_in_thread_trails. subst p.
unfold not in t_not_comm. apply t_not_comm.
assert ((id_s, list_M, e) = (id_s, list_M_s, e_s)) as t_is_s.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique. apply NNPP. unfold not at 1. intros t_not_s.
eapply S_hl_unique. reflexivity. reflexivity. apply H2. apply sender_in_S_hl. auto. reflexivity.
inversion t_is_s. subst. left.
subst id.
unfold committable in comm_before.
destruct comm_before as [ uniq_heap other ].
destruct other as [ uniq_thread other ].
assert (p = p_r). eapply uniq_thread.
left. apply H6. left. apply p_r_in_thread_trails. subst p.
unfold not in t_not_comm. apply t_not_comm.
assert ((id_r, list_M, e) = (id_r, list_M_r, e_r)) as t_is_r.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique. apply NNPP. unfold not at 1. intros t_not_r.
eapply S_hl_unique. reflexivity. reflexivity. apply H2. apply receiver_in_S_hl. auto. reflexivity.
inversion t_is_r. subst. right.
destruct id_not_comm as [ id_not_id_s id_not_id_r ].

clear before after.
right. split.
eapply In_S_no_ld. subst S_hl'.
left. split. auto. unfold not. intros false.
remember (id, list_M, e) as t.
destruct false. inversion Heqt. symmetry in H1. tauto.
inversion Heqt. symmetry in H1. tauto.
apply In_thread_trails'_other. auto. auto. auto.
unfold not. intros false.
remember (id, (None:option search_heap), list_M, e, p) as t.
inversion false. symmetry in H1. tauto.
symmetry in H1. tauto.

unfold Included. intros t t_in_big.
destruct t_in_big as [ t t_in_S0 | t t_in_S' ].
left. auto.
destruct t_in_S' as [ t_in_S' t_not_comm' ].
inversion t_in_S'. subst t.
assert (id <> id_s /\ id <> id_r) as id_not_comm.
apply not_or_and. unfold not. intros id_is_comm.
destruct id_is_comm as [ id_is_id_s | id_is_id_r ].
assert (p = path_cons (path_send n_sr id_r p_r) p_s).
subst id. inversion H1. reflexivity. symmetry in H3. tauto. tauto.
subst id. assert ((id_s, list_M, e) = (id_s, list_M_s, evt_fill_hole E_s (alwaysEvt zero))).
assert (hl_unique_tids S_hl') as S_hl'_unique.
assert (ld_unique_tids S_hl' None).
eapply hl_sync_step_preserves_unique_tids.
apply hl_steps. unfold ld_unique_tids. split. auto. auto.
unfold ld_unique_tids in H3. eapply proj1. apply H3.
unfold hl_unique_tids in S_hl'_unique. unfold not at 2 in S_hl'_unique.
apply NNPP. unfold not. intros t_is_s'. eapply S_hl'_unique.
reflexivity. reflexivity. apply H0.
subst S_hl'. right. left. unfold not. auto. reflexivity.
inversion H3. subst.
destruct t_not_comm'. left.
assert (p = path_cons (path_recv n_sr id_s p_s) p_r).
subst id. inversion H1. tauto. reflexivity. tauto.
subst id. assert ((id_r, list_M, e) = (id_r, list_M_r, evt_fill_hole E_r (alwaysEvt v))).
assert (hl_unique_tids S_hl') as S_hl'_unique.
assert (ld_unique_tids S_hl' None).
eapply hl_sync_step_preserves_unique_tids.
apply hl_steps. unfold ld_unique_tids. split. auto. auto.
unfold ld_unique_tids in H3. eapply proj1. apply H3.
unfold hl_unique_tids in S_hl'_unique. unfold not at 2 in S_hl'_unique.
apply NNPP. unfold not. intros t_is_r'. eapply S_hl'_unique.
reflexivity. reflexivity. apply H0.
subst S_hl'. right. right. unfold not. auto. reflexivity.
inversion H3. subst.
destruct t_not_comm'. right.
destruct id_not_comm as [ id_not_id_s id_not_id_r ].
right. split. eapply In_S_no_ld.
inversion t_in_S'. subst.
clear before after.
remember (id, list_M, e) as t.
destruct H4. destruct H2. auto. destruct H2.
inversion Heqt. symmetry in H3. tauto.
inversion Heqt. symmetry in H3. tauto.
remember (thread_trail id p) as tr.
inversion H1. subst tr. inversion H2. symmetry in H4. tauto.
subst tr. inversion H2. symmetry in H4. tauto. auto.
unfold not. intros false. remember (id, None:option search_heap, list_M, e, p) as t.
inversion false; subst.
inversion H2. symmetry in H4. tauto.
inversion H2. symmetry in H4. tauto.
Unfocus.

Focus.
intros t' t'_in_thread_trails'.
inversion t'_in_thread_trails'.
exists id_s. exists (path_cons (path_send n_sr id_r p_r) p_s).
exists list_M_s. exists (evt_fill_hole E_s (alwaysEvt zero)).
split. reflexivity.
subst S_hl'. right. left.
exists id_r. exists (path_cons (path_recv n_sr id_s p_s) p_r).
exists list_M_r. exists (evt_fill_hole E_r (alwaysEvt v)).
split. reflexivity.
subst S_hl'. right. right.
subst t'.
assert (exists id' : tid, exists p' : path, exists list_M' : list ctxt, exists e' : exp,
  thread_trail id p = thread_trail id' p' /\ In hl_thread S_hl (id', list_M', e')).
apply trails_in_S_hl. auto. destruct H3 as [id' [p' [list_M' [e' [eq in_S_hl ]]]]].
exists id. exists p. exists list_M'. exists e'. 
split. reflexivity.
subst S_hl'. left. split.
inversion eq. auto. unfold not. intros false.
inversion eq. subst id' p'.
remember (id, list_M', e') as t'.
inversion false.
subst t'. inversion H3. symmetry in H5. tauto.
subst t'. inversion H3. symmetry in H5. tauto.
Unfocus.

intros id list_M' e' e'_in_S_hl'.
subst S_hl'.
remember (id, list_M', e') as t'.
destruct e'_in_S_hl' as [ t' t'_in_S_hl | t' t'_is_comm' ].
destruct t'_in_S_hl as [ t'_in_S_hl t'_not_comm ].
assert (exists p : path, In trail thread_trails (thread_trail id p)).
eapply S_hl_in_trails. subst t'. apply t'_in_S_hl.
destruct H0 as [ p p_in_trails ].
exists p.
apply In_thread_trails'_other.
unfold not. intros id_is_id_s. subst id. subst t'.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique. eapply S_hl_unique.
reflexivity. reflexivity. apply t'_in_S_hl. apply sender_in_S_hl.
unfold not. intros false.
destruct t'_not_comm. rewrite -> false. left. reflexivity.
unfold not. intros id_is_id_r. subst id. subst t'.
unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique. eapply S_hl_unique.
reflexivity. reflexivity. apply t'_in_S_hl. apply receiver_in_S_hl.
unfold not. intros false.
destruct t'_not_comm. rewrite -> false. right. reflexivity.
auto.
destruct t'_is_comm'.
inversion Heqt'. subst id list_M' e'.
exists (path_cons (path_send n_sr id_r p_r) p_s).
apply In_thread_trails'_sender.
inversion Heqt'. subst id list_M' e'.
exists (path_cons (path_recv n_sr id_s p_s) p_r).
apply In_thread_trails'_receiver.

Qed.