Require Import committable.
Require Import dep.
Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Sets.Ensembles.
Require Import highlevel.
Require Import hl2ll_get_S.
Require Import hl2ll_many.
Require Import lowlevel.
Require Import syntax.
Require Import trails.

Corollary hl2ll_sync : forall H' S_hl_init H'' S_hl_complete H T S_susp,
  hl_sync_step_many H' S_hl_init None H'' S_hl_complete None ->
  hl_unique_tids S_hl_init ->
  (forall id list_M e,
    In hl_thread S_hl_init (id, list_M, e) ->
    value e /\ list_M = nil) ->
  (forall id list_M' e',
    In hl_thread S_hl_complete (id, list_M', e') ->
    exists v',
      value v' /\ e' = alwaysEvt v' /\ list_M' = nil) ->
  let S_init := get_S_sync_init S_hl_init in
    let S_H_init := Singleton search_heap (H', heap_path_nil) in
      exists thread_trails, exists hp, exists S_H, exists S0, exists S,
        let S_all := Union search_thread S S0 in
          let trails := Add trail thread_trails (heap_trail hp) in
            committable trails /\
            (ll_step_many (frozen H) T S_susp S_H_init S_init
              (frozen H) T S_susp S_H S_all) /\
            exists S_H0,
              S = get_S_no_ld S_hl_complete thread_trails /\
              S_H = Add search_heap S_H0 (H'', hp) /\
                (forall t, 
                  In trail thread_trails t ->
                  exists id, exists p, exists list_M, exists e, 
                    t = thread_trail id p /\
                    In hl_thread S_hl_complete (id, list_M, e)) /\
                (forall id list_M e,  
                  In hl_thread S_hl_complete (id, list_M, e) ->
                  exists p, 
                    In trail thread_trails (thread_trail id p)).

intros H' S_hl_init H'' S_hl_complete H T S_susp hl_steps S_hl_init_unique
  S_hl_init_values S_hl_complete_values S_init S_H_init.

remember (None:option hl_thread) as s.

Inductive get_thread_trails_init (S_hl_init:Ensemble hl_thread) 
  : Ensemble trail :=
| In_thread_trails_init : forall id v,
  In hl_thread S_hl_init (id, nil, v) ->
  In trail (get_thread_trails_init S_hl_init)
  (thread_trail id (path_nil)).

remember (get_thread_trails_init S_hl_init) as thread_trails_init.
remember (Add trail thread_trails_init 
  (heap_trail (heap_path_nil))) as trails_init.

assert (exists thread_trails, exists hp, exists S_H, exists S0, exists S,
  let S_all := Union search_thread S S0 in
    let trails := Add trail thread_trails (heap_trail hp) in
      committable trails /\
      (ll_step_many (frozen H) T S_susp S_H_init S_init
        (frozen H) T S_susp S_H S_all) /\
      (( (* no ld *)
        s = None /\
        exists S_H0,
          S = get_S_no_ld S_hl_complete thread_trails /\
          S_H = Add search_heap S_H0 (H'', hp) /\
            (forall t, 
              In trail thread_trails t ->
              exists id, exists p, exists list_M, exists e, 
                t = thread_trail id p /\
                In hl_thread S_hl_complete (id, list_M, e)) /\
            (forall id list_M e,  
              In hl_thread S_hl_complete (id, list_M, e) ->
              exists p, 
                In trail thread_trails (thread_trail id p)))
      \/ ( (* ld *)
        exists id_ld, exists list_M_ld, exists e_ld,
          s = Some (id_ld, list_M_ld, e_ld) /\
          S = get_S_ld H'' S_hl_complete 
          (id_ld, list_M_ld, e_ld) thread_trails hp /\
          (forall t, 
            In trail thread_trails t ->
            exists id, exists p, exists list_M, exists e, 
              t = thread_trail id p /\
              (In hl_thread S_hl_complete (id, list_M, e)
                \/ (id_ld, list_M_ld, e_ld) = (id, list_M, e))) /\
          (forall id list_M e,  
            (In hl_thread S_hl_complete (id, list_M, e) 
              \/ (id_ld, list_M_ld, e_ld) = (id, list_M, e)) ->
            exists p, 
              In trail thread_trails (thread_trail id p))))) as many.

eapply hl2ll_many.
apply hl_steps.
subst s. unfold ld_unique_tids. split. apply S_hl_init_unique. auto.
apply Heqtrails_init.

(* committability of initial trails *)
subst thread_trails_init trails_init.
unfold committable.
repeat split.
unfold unique. exists (heap_path_nil).
split. right. apply In_singleton.
intros hp' hp'_in_trails.
remember (heap_trail hp') as ht'.
destruct hp'_in_trails.
subst x.
inversion H0.
inversion H0. subst x. inversion H1. reflexivity.
intros id p1 p2 t1_in_trails t2_in_trails.
inversion t1_in_trails; inversion t2_in_trails.
inversion H0. inversion H2. reflexivity.
inversion H2. inversion H0. inversion H0.
intros t id p t_in_trails p_in_dep_t.
exists (path_nil).
remember (thread_trail id p) as t'.
unfold dep in p_in_dep_t.
inversion t_in_trails. inversion H0.
subst t x t'.
simpl in p_in_dep_t.
inversion p_in_dep_t. inversion H1. subst.
split. left. auto. 
inversion H1. subst id p. auto.
inversion H1. subst id p. apply extends_refl.
inversion H1.
inversion H0. subst x t t'.
simpl in p_in_dep_t.
inversion p_in_dep_t. inversion H1. inversion H1.
intros t p t_in_trails p_in_dep_t.
exists (heap_path_nil).
inversion t_in_trails.
inversion H0.
unfold dep in p_in_dep_t.
subst t x.
simpl in p_in_dep_t.
inversion p_in_dep_t.
inversion H1. inversion H1.
inversion H0. subst x t.
unfold dep in p_in_dep_t.
simpl in p_in_dep_t.
inversion p_in_dep_t. inversion H1. 
inversion H1. subst.
split. right. apply In_singleton. apply heap_extends_refl.

assert (S_init = get_S_no_ld S_hl_init thread_trails_init) as S_init_eq.
apply Extensionality_Ensembles.
subst S_init.
split.
unfold Included. intros x x_in_get_init.
inversion x_in_get_init.
unfold In. apply In_S_no_ld. apply H0.
subst thread_trails_init. eapply In_thread_trails_init.
apply H0.
unfold Included. intros x x_in_get_no_ld.
inversion x_in_get_no_ld.
assert (list_M = nil) as list_M_is_nil.
eapply proj2.
eapply S_hl_init_values.
apply H0.
assert (p = path_nil) as p_is_nil.
subst thread_trails_init.
inversion H1.
reflexivity.
subst list_M p.
eapply In_S_sync_init.
apply H0.
instantiate (2 := S_init). 
instantiate (1 := Empty_set search_thread).
apply Extensionality_Ensembles.
split. unfold Included. intros x x_in_S_init. left. auto.
unfold Included. intros x x_in_union. inversion x_in_union. auto. inversion H0.

left. repeat split.
apply Heqs.
intros t t_in_thread_trails_init.
subst thread_trails_init.
inversion t_in_thread_trails_init. subst t.
repeat esplit. apply H0.
intros id list_M e t_in_S_hl_init.
subst thread_trails_init.
exists (path_nil).
eapply In_thread_trails_init. 
assert (list_M = nil) as list_M_is_nil.
eapply proj2. eapply S_hl_init_values.
apply t_in_S_hl_init.
subst list_M. apply t_in_S_hl_init.
instantiate (1 := Empty_set search_heap).
subst S_H_init.
apply Extensionality_Ensembles.
split; unfold Included; intros x x_in_one.
inversion x_in_one. subst x. right. apply In_singleton.
inversion x_in_one. inversion H0. auto.
subst S_init. apply Extensionality_Ensembles.
split; unfold Included; intros x x_in_one; inversion x_in_one.
eapply In_S_no_ld. apply H0. subst thread_trails_init. eapply In_thread_trails_init. apply H0.
assert (list_M = nil) as list_M_is_nil.
eapply proj2. eapply S_hl_init_values. apply H0. subst list_M.
subst thread_trails_init. inversion H1. subst p.
eapply In_S_sync_init. subst id0. auto.
destruct many as [ thread_trails [ hp [ S_H [ S0 [ S [ comm_after [ ll_steps ld_or_not_ld ] ] ] ] ] ] ].
destruct ld_or_not_ld as [ ld | not_ld ].
destruct ld as [ s_is_None [ S_H0 [ S_eq [ S_H_eq [ trails_in_complete complete_in_trails ] ] ] ] ].
exists thread_trails.
exists hp.
exists S_H.
exists S0.
exists S.
intros S_all trails.
split. subst trails. apply comm_after.
split. subst S_all. apply ll_steps.
exists S_H0.
split. apply S_eq.
split. apply S_H_eq.
split. apply trails_in_complete.
apply complete_in_trails.
destruct not_ld as [ id [ list [ e [ false [ other ] ] ] ] ]. 
subst s. inversion false.

Qed.
