Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Lists.List.
Require Import Coq.Sets.Ensembles.
Require Import highlevel.
Require Import ll_invariants.
Require Import ll2hl_get_S.
Require Import lowlevel.
Require Import sets.
Require Import syncsim.
Require Import syntax.
Require Import trails.

Definition translate_heap (h:global_heap) : heap :=
  match h with
    | (frozen H) => H
    | (unfrozen H) => H
  end.

Definition translate_threads (T:Ensemble mt_thread) 
  (S_susp:Ensemble susp_thread) : Ensemble mt_thread :=
  Union mt_thread T (translate_susp S_susp).

Theorem ll2hl : forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2,
  susp_get_ctxt S_susp1 ->
  T_S_susp_unique_tids T1 S_susp1 ->
  sync_sim h1 S_susp1 S_H1 S1 ->
  ll_step h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
  hl_mt_step_many (translate_heap h1) (translate_threads T1 S_susp1)
  (translate_heap h2) (translate_threads T2 S_susp2).

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2
  susp_get_ctxt_before T_S_susp_unique_before sync_sim_before ll_steps.

unfold translate_heap, translate_threads.

inversion ll_steps as [
  H1 _T1 H2 _T2 mt_steps h1_eq T1_eq S_susp1_eq S_H1_eq S1_eq 
  h2_eq T2_eq S_susp2_eq S_H2_eq S2_eq
  | H T_sync _T2 _T1 _S_susp2 _S2 _S_H2 T_sync_at_sync T1_eq
    disjoint_T_sync_T2 S_susp2_eq S2_eq S_H2_eq h1_eq T1_e1 S_susp1_eq 
    S_H1_eq S1_eq h2_eq T2_eq S_susp2_eq2 S_H2_eq2 S2_eq2
  | | | | | | | | 
  | H S_complete _T1 T_susp T_complete _T2 S_susp_complete 
    S_susp_incomplete _S_susp1 H' hp S_H0 _S_H1 thread_trails
    trails S0 _S1 S_complete_complete S_unique T_complete_eq 
    S_susp_complete_eq S_susp_incomplete_eq T_susp_eq2 S_H1_eq 
    S1_eq T2_eq thread_trails_eq trails_eq comm h1_eq T1_eq 
    S_susp1_eq2 S_H1_eq2 S1_eq2 h2_eq T2_eq2 S_susp2_eq S_H2_eq S2_eq
];
try (subst; apply hl_mt_step_many_zero).

(* StepTopLevel *)

Focus.

subst h1 _T1 h2 _T2 S_susp1 S_H1 S1 S_susp2 S_H2 S2.

unfold translate_heap, translate_threads, empty_susp.

rewrite -> translate_susp_empty.
rewrite -> union_empty. rewrite -> union_empty.
eapply hl_mt_step_many_nonzero.
apply hl_step_mt. eauto.
apply hl_mt_step_many_zero.

Unfocus.

(* SyncInit *)

Focus.

subst _T1 _T2 _S_susp2 _S_H2 _S2 T1 S_susp2 S2 S_H2 h1 S_susp1 S_H1 S1 h2.
unfold empty_susp.

rewrite -> translate_susp_empty. rewrite -> union_empty.

assert (translate_susp (get_S_susp_init T_sync) = T_sync) as T_sync_eq.
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id M v in_S_susp x_eq ].
inversion in_S_susp as [ _id e _M _v in_T_sync e_is_sync id_eq ]. 
subst _id _M _v x.
replace (M[sync v]) with e. auto.
symmetry. apply fill_hole_same. auto.
induction x as [ id e ].
assert (exists M, exists v, get_ctxt e M (sync v)) as e_at_sync by eauto.
destruct e_at_sync as [ M [ v e_at_sync ] ].
replace e with (M[sync v]).
apply In_translate_susp. eapply In_S_susp_init. eauto. auto.
apply fill_hole_same. auto.

rewrite T_sync_eq. rewrite union_swap.
apply hl_mt_step_many_zero.

Unfocus.

(* SyncComplete *)

subst T_complete T_susp S_H1 S1 T2 thread_trails trails h1 _T1 
  _S_susp1 _S_H1 _S1 h2 _T2 S_susp2 S_H2 S2.

assert (S_susp1 = Union susp_thread S_susp_complete S_susp_incomplete)
  as S_susp1_eq.
apply Extensionality_Ensembles. subst S_susp_complete S_susp_incomplete.
split; intros x x_in_one.
assert (In susp_thread (get_S_susp_complete S_susp1 S_complete) x \/
  ~ In susp_thread (get_S_susp_complete S_susp1 S_complete) x) as classic.
apply classic. destruct classic as [ in_incomplete | in_incomplete ].
left. auto. right. split. auto. auto.
destruct x_in_one as [ x in_complete | x in_incomplete ].
inversion in_complete. auto. inversion in_incomplete. auto.
assert (Disjoint susp_thread S_susp_complete S_susp_incomplete)
  as disjoint_susp.
subst S_susp_complete S_susp_incomplete.
apply disjoint_setminus.
subst S_susp1.

unfold empty_susp. rewrite translate_susp_empty.
rewrite union_empty.
apply hl_mt_step_many_one.
eapply hl_step_sync.

(* S_susp_complete, translated, at sync *)
instantiate (1 := translate_susp S_susp_complete).
intros id e in_T_init.
inversion in_T_init as [ _id M v in_s_susp_complete id_eq ].
exists M. exists v.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before.
eauto. left. eauto.

reflexivity.

instantiate (1 := get_S_hl_S S_complete).

unfold sync_sim in sync_sim_before.
destruct sync_sim_before as [ sync_sim_no_ld_before sync_sim_ld_before ].
eapply sync_sim_no_ld_before.
instantiate (1 := S_complete).
unfold Included. apply Union_introl.
auto.
intros id h list_M e p in_S_complete.
assert (h = None /\ list_M = nil /\ 
  exists v', value v' /\ e = alwaysEvt v') as h_is_none.
eapply S_complete_complete. apply in_S_complete.
destruct h_is_none as [ h_is_none other ]. apply h_is_none.
right. apply In_singleton.
reflexivity.
instantiate (1 := S_susp_complete).
rewrite S_susp_complete_eq.
apply Extensionality_Ensembles. split; intros x x_in_one.
induction x as [ [ id M ] v ].
inversion x_in_one as [ _id _M _v h list_M e' p in_S_susp in_S_complete ].
eapply In_S_susp_complete.
left. eapply In_S_susp_complete. auto. apply in_S_complete.
apply in_S_complete.
inversion x_in_one as [ id M v h list_M e' p in_S_susp in_S_complete ].
inversion in_S_susp as [ _x in_S_susp_complete | _x in_incomplete ].
auto. 

eapply In_S_susp_complete.
right. auto. eauto.

reflexivity. reflexivity. auto.

intros id list_M e in_S_hl'.
inversion in_S_hl' as 
  [ _id h _list_M _e p in_S_complete ].
subst _id _list_M _e.
eapply proj2. eauto.

reflexivity.

instantiate (1 := Union mt_thread T1 (translate_susp S_susp_incomplete)).
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_in_T1 | x x_in_T2 ].
right. left. auto.
inversion x_in_T2 as [ id M v in_S_susp ]. 
inversion in_S_susp;
auto using Union_introl, Union_intror, In_translate_susp.
destruct x_in_one as [ x in_T 
  | x' [ x in_T | x in_T ] ];
try inversion in_T;
auto using Union_introl, Union_intror, In_translate_susp.

apply Disjoint_intro. unfold not. intros x in_int.
destruct in_int as [ x in_complete in_T1_or_incomplete ].
inversion in_complete as [ id M v ].
destruct in_T1_or_incomplete as [ x x_in_T1 | x x_in_incomplete ]; subst x.
assert (~ (In mt_thread T1 (id, M[sync v]) /\ 
  In susp_thread (Union susp_thread S_susp_complete S_susp_incomplete)
  (id, M, v)))
  as contra.
unfold T_S_susp_unique_tids in T_S_susp_unique_before.
destruct T_S_susp_unique_before as 
  [ T1_unique [ S_susp_unique T1_S_susp_disjoint ] ].
unfold T_S_susp_disjoint_tids in T1_S_susp_disjoint.
apply T1_S_susp_disjoint.
destruct contra. split; auto using Union_introl.
inversion disjoint_susp as [ not_in_int ].
unfold not in not_in_int. eapply not_in_int.
split. eauto. 
inversion x_in_incomplete as [ id' M' v' in_incomplete [ id_eq e_eq] ].
subst id'.
assert (M = M' /\ sync v = sync v') as M_v_eq.
eapply get_ctxt_unique.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before; eauto using Union_introl.
rewrite <- e_eq.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before; eauto using Union_intror.
destruct M_v_eq as [ M_eq sync_v_eq ]. inversion sync_v_eq. subst M' v'.
auto.

assert (get_T'_complete S_complete S_susp_complete =
  get_T'_hl (translate_susp S_susp_complete) (get_S_hl_S S_complete))
as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id M v v' p ].
eapply In_get_T'_hl. apply In_translate_susp. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. eauto. left. eauto.
eapply In_S_hl_S. eauto.
inversion x_in_one as 
  [ id e M v v' in_T_complete_init e_at_sync in_S_hl_complete ].
inversion in_T_complete_init as [ _id M1 v1 in_susp_complete1 ]. 
subst _id e.
assert (M = M1 /\ sync v = sync v1) as M_sync_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before.
eauto. left. eauto.
destruct M_sync_v_eq as [ M_eq sync_v_eq ]. inversion sync_v_eq. subst M1 v1.
inversion in_S_hl_complete as [ _id h list_M e' p in_S_complete ]. 
subst _id list_M e'.
assert (h = None /\ (nil:list ctxt) = nil /\
  (exists v''' : exp, value v''' /\ alwaysEvt v' = alwaysEvt v''')) as complete.
eapply S_complete_complete. eauto.
destruct complete as [ h_eq other ]. subst h.
eapply In_T'_complete. eauto. eauto.
rewrite rewriting. clear rewriting. reflexivity.

Qed.

Corollary ll2hl_many : forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2,
  susp_get_ctxt S_susp1 ->
  T_S_susp_unique_tids T1 S_susp1 ->
  sync_sim h1 S_susp1 S_H1 S1 ->
  ll_step_many h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
  hl_mt_step_many (translate_heap h1) (translate_threads T1 S_susp1)
  (translate_heap h2) (translate_threads T2 S_susp2).

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2
  susp_get_ctxt_before T_S_susp_unique_before sync_sim_before ll_steps.

induction ll_steps.

apply hl_mt_step_many_zero.

eapply hl_mt_step_many_many.
eapply ll2hl. auto. auto. eauto. eauto.
apply IHll_steps.
eapply ll_step_preserves_susp_get_ctxt. eauto. auto.
eapply ll_step_preserves_T_S_susp_unique_tids. eauto. auto.
eapply ll_step_preserves_sync_sim. eauto. eauto. eauto. eauto.

Qed.

Corollary ll2hl_from_beginning :
  forall id e H T,
    ll_step_many (unfrozen fresh_heap) (Singleton mt_thread (id, e)) 
    empty_susp empty_heap empty_search
    (unfrozen H) T empty_susp empty_heap empty_search ->
    hl_mt_step_many fresh_heap (Singleton mt_thread (id, e)) H T.

intros id e H T ll_steps.

assert (fresh_heap = translate_heap (unfrozen fresh_heap)) as rewrite_H.
unfold translate_heap. auto.
assert (Singleton mt_thread (id, e) = 
  translate_threads (Singleton mt_thread (id, e)) empty_susp)
as rewrite_T.
unfold translate_threads. unfold empty_susp.
rewrite translate_susp_empty. rewrite union_empty. auto.
assert (H = translate_heap (unfrozen H)) as rewrite_H'.
unfold translate_heap. auto.
assert (T = translate_threads T empty_susp) as rewrite_T'.
unfold translate_threads. unfold empty_susp. 
rewrite translate_susp_empty. rewrite union_empty. auto.
rewrite rewrite_H. rewrite rewrite_T. rewrite rewrite_H'. rewrite rewrite_T'.

eapply ll2hl_many.

unfold susp_get_ctxt.
intros id1 M1 v1 false. unfold empty_susp in false. inversion false.

split.
unfold mt_unique_tids.
intros t1 t2 id1 id2 e1 e2 t1_eq t2_eq t1_is_id t2_is_id t1_not_t2.
inversion t1_is_id as [ eq1 ]. inversion t2_is_id as [ eq2 ]. 
subst t1 t2. inversion eq1. inversion eq2. subst id1 id2 e1 e2.
tauto.

split. unfold susp_unique_tids, empty_susp.
intros id1 M1 M2 v1 v2 false. inversion false.

unfold T_S_susp_disjoint_tids, empty_susp.
intros id1 e1 M1 v1. apply or_not_and. right.
unfold not. intros false. inversion false.

unfold sync_sim. auto.

eauto.

Qed.
