Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Sets.Ensembles.
Require Import highlevel.
Require Import ll2hl_get_S.
Require Import lowlevel.
Require Import syntax.
Require Import trails.

Definition susp_get_ctxt (S_susp:Ensemble susp_thread) : Prop :=
  forall id M v,
    In susp_thread S_susp (id, M, v) ->
    get_ctxt (M[sync v]) M (sync v).

Definition T_S_susp_disjoint_tids (T:Ensemble mt_thread)
  (S_susp:Ensemble susp_thread) : Prop :=
  forall id e M v,
    ~ (In mt_thread T (id, e) /\ 
      In susp_thread S_susp (id, M, v)).

Definition susp_unique_tids (S_susp:Ensemble susp_thread) : Prop :=
  forall id M1 M2 v1 v2,
    In susp_thread S_susp (id, M1, v1) ->
    In susp_thread S_susp (id, M2, v2) ->
    (id, M1, v1) = (id, M2, v2).

Definition T_S_susp_unique_tids (T:Ensemble mt_thread)
  (S_susp:Ensemble susp_thread) : Prop :=
  mt_unique_tids T
  /\ susp_unique_tids S_susp
  /\ T_S_susp_disjoint_tids T S_susp.

Lemma ll_step_preserves_susp_get_ctxt :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2,
    ll_step h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    susp_get_ctxt S_susp1 ->
    susp_get_ctxt S_susp2.

unfold susp_get_ctxt.
intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2
  ll_steps susp_get_ctxt_before id M v in_S_susp2.

inversion ll_steps;
try (subst S_susp2; inversion in_S_susp2);
try (subst S_susp2; eapply susp_get_ctxt_before; eauto).

assert (M[sync v] = e).
apply fill_hole_same. auto.
subst e. auto.

Qed.

Lemma ll_step_many_preserves_susp_get_ctxt :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2,
    ll_step_many h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    susp_get_ctxt S_susp1 ->
    susp_get_ctxt S_susp2.

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2
  ll_steps susp_get_ctxt_before.

induction ll_steps.
auto. eauto using ll_step_preserves_susp_get_ctxt.

Qed.

Lemma ll_step_preserves_T_S_susp_unique_tids :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2,
    ll_step h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    T_S_susp_unique_tids T1 S_susp1 ->
    T_S_susp_unique_tids T2 S_susp2.

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2
  ll_steps unique_before.

destruct unique_before as [ T_unique [ S_susp_unique disjoint ] ].

inversion ll_steps; subst; repeat split; auto.

(* StepTopLevel *)

(** T unique **)
eapply mt_step_preserves_unique_tids. eauto. auto.

(** S_susp unique **)

(** disjoint **)
unfold T_S_susp_disjoint_tids. intros.
apply or_not_and. right. unfold not. intros false.
inversion false.

(* SyncInit *)

(** T unique **)
unfold mt_unique_tids. unfold mt_unique_tids in T_unique.
intros. eapply T_unique; eauto using Union_intror.

(** S_susp unique **)
unfold susp_unique_tids.
intros id M1 M2 v1 v2 t1_in_T t2_in_T.
inversion t1_in_T as [ id1 e1 ]. inversion t2_in_T as [ id2 e2 ].
subst id1 id2.
assert (e1 = e2). assert ((id, e1) = (id, e2)) as eq.
apply NNPP. unfold not at 1. intros neq.
unfold mt_unique_tids in T_unique. unfold not at 2 in T_unique.
eapply T_unique. reflexivity. reflexivity.
instantiate (1 := e1). left. eauto.
instantiate (1 := e2). left. eauto.
auto. reflexivity. inversion eq. reflexivity. subst e2.
assert (M1 = M2 /\ sync v1 = sync v2) as M_sync_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_sync_v_eq as [ M_eq v_eq ]. inversion v_eq.
subst. reflexivity.

(** disjoint **)
unfold T_S_susp_disjoint_tids. unfold not.
intros id e M v [ in_T2 in_S_susp ].
inversion in_S_susp as [ _id e' ].
unfold mt_unique_tids in T_unique. unfold not at 2 in T_unique.
eapply T_unique. reflexivity. reflexivity.
right. instantiate (1 := e). eauto.
left. instantiate (1 := e'). eauto.
cut (Disjoint mt_thread T T2).
intros T_disjoint. inversion T_disjoint as [not_in_int].
unfold not. unfold not in not_in_int.
intros eq. inversion eq. subst e'.
eapply not_in_int. split. eauto. auto. auto. reflexivity.

(* SyncComplete *)

(** T' unique **)
Focus.
unfold mt_unique_tids. unfold not at 2.
intros t1 t2 id1 id2 e1 e2 eq1 eq2 t1_in_T t2_in_T neq id_eq.
subst id2.
destruct t1_in_T as [ t1 t1_in_complete | 
  t1' [ t1 t1_in_T1 | t1 t1_in_incomplete ] ];
destruct t2_in_T as [ t2 t2_in_complete | 
  t2' [ t2 t2_in_T1 | t2 t2_in_incomplete ] ];
subst t1 t2.

(*** complete/complete *)
inversion t1_in_complete as [ id1' M1 v1 v1' p1 t1_in_S t1_in_susp_complete ].
inversion t1_in_susp_complete.
inversion t2_in_complete as [ id2' M2 v2 v2' p2 t2_in_S t2_in_susp_complete ].
inversion t2_in_susp_complete.
unfold susp_unique_tids in S_susp_unique. subst.
assert ((id1, M1, v1) = (id1, M2, v2)) as M_v_eq.
apply S_susp_unique. auto. auto.
inversion M_v_eq. subst M2 v2.
assert (v1' = v2').
assert ((id1, None:local_heap, nil:list ctxt, alwaysEvt v1', p1) = 
  (id1, None, nil, alwaysEvt v2', p2)) as eq.
cut (search_unique_tids S). intros S_unique.
unfold search_unique_tids in S_unique.
apply S_unique. auto. auto. auto.
inversion eq. auto. subst v2'. tauto.

(*** complete/T1 *)
unfold T_S_susp_disjoint_tids, not in disjoint.
inversion t1_in_complete as [ id1' M1 v1 v1' p1 in_S in_S_susp_complete ].
inversion in_S_susp_complete.
eapply disjoint. split.
eauto. eauto.

(*** complete/incomplete *)
inversion t1_in_complete as [ id1' M1 v1 v1' p1 t1_in_S t1_in_S_susp_complete ].
subst id1' e1.
inversion t1_in_S_susp_complete as 
  [ id1' M1' v1'' h1 list_M1 e1' p1' t1_in_S_susp1 t1_in_S_again ]. 
subst id1' M1' v1''.
inversion t2_in_incomplete as [ id2' M2 v2 t2_in_S_susp_incomplete ].
subst id2' e2.
inversion t2_in_S_susp_incomplete as 
  [ t2_in_S_susp1 t2_not_in_S_susp_complete ].
unfold susp_unique_tids in S_susp_unique.
assert ((id1, M1, v1) = (id1, M2, v2)) as M_V_eq.
apply S_susp_unique. auto. auto.
inversion M_V_eq. subst M2 v2.
destruct t2_not_in_S_susp_complete.
eapply In_S_susp_complete. auto. eauto.

(*** T1/complete *)
unfold T_S_susp_disjoint_tids, not in disjoint.
inversion t2_in_complete as [ id2' M2 v2 v2' p2 in_S in_S_susp_complete ].
inversion in_S_susp_complete.
eapply disjoint. split.
eauto. eauto.

(*** T1/T1 *)
unfold mt_unique_tids in T_unique.
unfold not at 2 in T_unique.
eapply T_unique. reflexivity. reflexivity.
apply t1_in_T1. apply t2_in_T1. auto. reflexivity.

(*** T1/incomplete *)
inversion t2_in_incomplete as [ id2' M2 v2 t2_in_S_susp_incomplete ].
subst id2' e2.
inversion t2_in_S_susp_incomplete as 
  [ t2_in_S_susp1 t2_not_in_S_susp_complete ].
unfold T_S_susp_disjoint_tids in disjoint.
unfold not in disjoint. eapply disjoint.
split. eauto. eauto.

(*** incomplete/complete *)
inversion t2_in_complete as [ id2' M2 v2 v2' p2 t2_in_S t2_in_S_susp_complete ].
subst id2' e2.
inversion t2_in_S_susp_complete as 
  [ id2' M2' v2'' h2 list_M2 e2' p2' t2_in_S_susp1 t2_in_S_again ]. 
subst id2' M2' v2''.
inversion t1_in_incomplete as [ id1' M1 v1 t1_in_S_susp_incomplete ].
subst id1' e1.
inversion t1_in_S_susp_incomplete as 
  [ t1_in_S_susp1 t1_not_in_S_susp_complete ].
unfold susp_unique_tids in S_susp_unique.
assert ((id1, M1, v1) = (id1, M2, v2)) as M_V_eq.
apply S_susp_unique. auto. auto.
inversion M_V_eq. subst M2 v2.
destruct t1_not_in_S_susp_complete.
eapply In_S_susp_complete. auto. eauto.

(*** incomplete/T1 *)
inversion t1_in_incomplete as [ id1' M1 v1 t1_in_S_susp_incomplete ].
subst id1' e1.
inversion t1_in_S_susp_incomplete as 
  [ t1_in_S_susp1 t1_not_in_S_susp_complete ].
unfold T_S_susp_disjoint_tids in disjoint.
unfold not in disjoint. eapply disjoint.
split. eauto. eauto.

(*** incomplete/incomplete *)
inversion t1_in_incomplete as [ id1' M1 v1 t1_in_S_susp_incomplete ].
subst id1' e1.
inversion t1_in_S_susp_incomplete as 
  [ t1_in_S_susp1 t1_not_in_S_susp_complete ].
inversion t2_in_incomplete as [ id2' M2 v2 t2_in_S_susp_incomplete ].
subst id2' e2.
inversion t2_in_S_susp_incomplete as 
  [ t2_in_S_susp1 t2_not_in_S_susp_complete ].
assert ((id1, M1, v1) = (id1, M2, v2)) as M_v_eq.
unfold susp_unique_tids in S_susp_unique.
apply S_susp_unique. auto. auto.
inversion M_v_eq. subst M2 v2. tauto.

Unfocus.

unfold susp_unique_tids, empty_susp.
intros id M1 M2 v1 v2 false. inversion false.

unfold T_S_susp_disjoint_tids, empty_susp.
intros id e M v. apply or_not_and.
right. unfold not. intros false. inversion false.

Qed.

Lemma ll_step_many_preserves_T_S_susp_unique_tids :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2,
    ll_step_many h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    T_S_susp_unique_tids T1 S_susp1 ->
    T_S_susp_unique_tids T2 S_susp2.

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2
  ll_steps T_S_susp_unique_before.

induction ll_steps.
auto. eauto using ll_step_preserves_T_S_susp_unique_tids.

Qed.