Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Sets.Ensembles.
Require Import highlevel.
Require Import syntax.

Lemma hl_sync_step_preserves_tids_no_ld : 
  forall H S_hl H' S_hl' id list_M e,
    hl_sync_step H S_hl None H' S_hl' None ->
    In hl_thread S_hl (id, list_M, e) ->
    exists list_M, exists e, 
      In hl_thread S_hl' (id, list_M, e).

intros H S_hl H' S_hl' id list_M e steps in_S_hl.
inversion steps. subst.
assert ((id = id1 \/ id = id2) \/ ~(id = id1 \/ id = id2)) as classic.
apply classic.
destruct classic. destruct H. subst id.
esplit. esplit. right. left.
subst id. esplit. esplit. right. right.
exists list_M. exists e. left.
split. apply in_S_hl. unfold not. unfold not in H. intros false. apply H.
inversion false. left. auto. right. auto.

Qed.

Lemma hl_sync_step_preserves_tids_begin_ld : 
  forall H S_hl H' S_hl' s id list_M e,
    hl_sync_step H S_hl None H' S_hl' (Some s) ->
    In hl_thread S_hl (id, list_M, e) ->
    exists list_M', exists e', 
      In hl_thread S_hl' (id, list_M', e') 
      \/ s = (id, list_M', e').

intros H S_hl H' S_hl' s id list_M e steps in_S_hl.
inversion steps. subst.
inversion in_S_hl. esplit. esplit. left. apply H.
inversion H. esplit. esplit. right. reflexivity.

Qed.

Lemma hl_sync_step_preserves_tids_end_ld : 
  forall H S_hl s H' S_hl' id list_M e,
    hl_sync_step H S_hl (Some s) H' S_hl' None ->
    In hl_thread S_hl (id, list_M, e) 
    \/ s = (id, list_M, e) ->
    exists list_M', exists e', 
      In hl_thread S_hl' (id, list_M', e').

intros H S_hl s H' S_hl' id list_M e steps in_S_hl_or_s.
inversion steps. subst.
inversion in_S_hl_or_s. esplit. esplit. left. apply H.
inversion H. esplit. esplit. right. apply In_singleton.

Qed.

Lemma hl_sync_step_preserves_tids_ld :
  forall H S_hl s H' S_hl' s' id list_M e,
    hl_sync_step H S_hl (Some s) H' S_hl' (Some s') ->
    In hl_thread S_hl (id, list_M, e)
    \/ s = (id, list_M, e) ->
    exists list_M', exists e',
      In hl_thread S_hl' (id, list_M', e')
      \/ s' = (id, list_M', e').

intros H S_hl s H' S_hl' s' id list_M e steps in_S_hl_or_s.

inversion steps; 
try solve [ inversion in_S_hl_or_s; subst;
esplit; esplit; try inversion H8; subst; eauto ].

subst.
assert ((id = id1 \/ id = id2) \/ ~(id = id1 \/ id = id2)) as classic.
apply classic.
destruct classic as [ | neq ]. destruct H. subst id.

esplit. esplit. left. right. left.
subst id. esplit. esplit. left. right. right. 
inversion in_S_hl_or_s.

unfold not in neq.
esplit. esplit. left. left. split. eauto.
intros false. inversion false. subst id1.  tauto.
subst id2. tauto.

esplit. esplit. right. eauto.

Qed.

Lemma hl_sync_step_preserves_tids :
  forall H S_hl s H' S_hl' s' id list_M e,
    hl_sync_step H S_hl s H' S_hl' s' ->
    In hl_thread S_hl (id, list_M, e)
    \/ s = Some (id, list_M, e) ->
    exists list_M', exists e',
      In hl_thread S_hl' (id, list_M', e')
      \/ s' = Some (id, list_M', e').

intros H S_hl s H' S_hl' s' id list_M e steps in_S_hl_or_s.
induction s; induction s'.

assert (exists list_M' : List.list ctxt,
  exists e' : exp,
    In hl_thread S_hl' (id, list_M', e') \/
    a0 = (id, list_M', e')).
eapply hl_sync_step_preserves_tids_ld. apply steps.
destruct in_S_hl_or_s. eauto. inversion H0. auto.
destruct H0 as [ list_M' [ e' [ in_S_hl' | is_s' ] ] ]; subst; eauto.

assert (exists list_M' : List.list ctxt,
  exists e' : exp,
    In hl_thread S_hl' (id, list_M', e')).
eapply hl_sync_step_preserves_tids_end_ld. apply steps.
destruct in_S_hl_or_s; try inversion H0; eauto.
destruct H0 as [ list_M' [ e' in_S_hl' ] ]; subst; eauto.

assert (exists list_M' : List.list ctxt,
  exists e' : exp,
    In hl_thread S_hl' (id, list_M', e') \/
    a = (id, list_M', e')).
eapply hl_sync_step_preserves_tids_begin_ld. apply steps.
destruct in_S_hl_or_s; try inversion H0; eauto.
destruct H0 as [ list_M' [ e' [ in_S_hl' | is_s' ] ] ]; subst; eauto.

destruct in_S_hl_or_s; try inversion H0.
assert (exists list_M' : List.list ctxt,
  exists e' : exp,
    In hl_thread S_hl' (id, list_M', e')).
eapply hl_sync_step_preserves_tids_no_ld. apply steps. eauto.
destruct H1 as [ list_M' [ e' in_S_hl' ] ]; subst; eauto.

Qed.

Lemma hl_sync_step_many_preserves_tids :
  forall H S_hl s H' S_hl' s' id list_M e,
    hl_sync_step_many H S_hl s H' S_hl' s' ->
    In hl_thread S_hl (id, list_M, e)
    \/ s = Some (id, list_M, e) ->
    exists list_M', exists e',
      In hl_thread S_hl' (id, list_M', e')
      \/ s' = Some (id, list_M', e').

intros H S_hl s H' S_hl' s' id list_M e steps in_S_hl_or_s.
generalize dependent e. generalize dependent list_M. induction steps.

eauto.

intros list_M e in_S1_or_s1.
assert (exists list_M2, exists e2,
  In hl_thread S2 (id, list_M2, e2) \/ 
  s2 = Some (id, list_M2, e2)).
eapply hl_sync_step_preserves_tids; eauto.
destruct H0 as [ list_M2 [ e2 true ] ].
eapply IHsteps. 
apply true.

Qed.
