Require Import consistent.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import syntax.
Require Import trails.

Lemma send_in_dep_implies_recv_in_dep :
  forall n_s id_s p_s id_r p_r t,
    In trail (dep t) 
    (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
    exists p_r',
    In trail (dep t) (thread_trail id_r p_r') 
    /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r).

intros n_s id_s p_s id_r p_r.

cut ((forall p id,
  In trail (dep (thread_trail id p)) 
  (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
  exists p_r',
    In trail (dep (thread_trail id p)) (thread_trail id_r p_r') 
    /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r))
/\ (forall hp,
  In trail (dep (heap_trail hp)) 
  (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
  exists p_r',
    In trail (dep (heap_trail hp)) (thread_trail id_r p_r') 
    /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r))).

intros [ H1 H2 ].
induction t as [ id p | hp ]; auto.

apply trail_induction with 
(P := fun elt => 
  match elt with
    | path_left => True
    | path_right => True
    | path_send n id p => 
      In trail (dep (thread_trail id p)) 
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      exists p_r',
        In trail (dep (thread_trail id p)) (thread_trail id_r p_r') 
        /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)
    | path_recv n id p => 
      In trail (dep (thread_trail id p)) 
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      exists p_r',
        In trail (dep (thread_trail id p)) (thread_trail id_r p_r') 
        /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)
    | path_heap n hp =>
      In trail (dep (heap_trail hp)) 
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      exists p_r',
        In trail (dep (heap_trail hp)) (thread_trail id_r p_r') 
        /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)
  end)
(P0 := fun elt =>
  match elt with
    | heap_path_thread n id p =>
      In trail (dep (thread_trail id p)) 
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      exists p_r',
        In trail (dep (thread_trail id p)) (thread_trail id_r p_r') 
        /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)
  end)
(P1 := fun p =>
  forall id,
    In trail (dep (thread_trail id p)) 
      (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
      exists p_r',
        In trail (dep (thread_trail id p)) (thread_trail id_r p_r') 
        /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r))
(P2 := fun hp =>
  In trail (dep (heap_trail hp)) 
  (thread_trail id_s (path_cons (path_send n_s id_r p_r) p_s)) ->
  exists p_r',
    In trail (dep (heap_trail hp)) (thread_trail id_r p_r') 
    /\ extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)).

auto. auto. auto. auto. auto. auto.

intros id false. unfold dep in false.
inversion false as [ t_s false2 | t_s false2 ]; subst t_s.
simpl in false2. inversion false2. inversion false2.

intros elt IH_elt p IH_path id in_dep.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left *)
unfold dep in in_dep.
inversion in_dep as [ t_s t_s_in_DA | t_s false ].
simpl in t_s_in_DA.
assert (exists p_r',
  In trail (dep (thread_trail id p)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
eapply IH_path. unfold dep. left. apply t_s_in_DA.
destruct p_r' as [ p_r' [ p_r'_in_dep p_r'_extends ] ].
unfold dep in p_r'_in_dep.
inversion p_r'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists p_r'. split. unfold dep. left. simpl. auto. auto.
inversion t_r'_is_t'. subst id_r p_r'.
exists (path_cons path_left p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

(* Right *)
unfold dep in in_dep.
inversion in_dep as [ t_s t_s_in_DA | t_s false ].
simpl in t_s_in_DA.
assert (exists p_r',
  In trail (dep (thread_trail id p)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
eapply IH_path. unfold dep. left. apply t_s_in_DA.
destruct p_r' as [ p_r' [ p_r'_in_dep p_r'_extends ] ].
unfold dep in p_r'_in_dep.
inversion p_r'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists p_r'. split. unfold dep. left. simpl. auto. auto.
inversion t_r'_is_t'. subst id_r p_r'.
exists (path_cons path_right p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

(* Send *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_s other | t_s t_s_is_x ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s false ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_t' | t_s t_s_in_DA_x ].

simpl in t_s_in_DA_t'.
assert (exists p_r',
  In trail (dep (thread_trail id p)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t'.
destruct p_r' as [ p_r' [ p_r'_in_dep p_r'_extends ] ].
unfold dep in p_r'_in_dep.
inversion p_r'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists p_r'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_r'_is_t'. subst id_r p_r'.
exists (path_cons (path_send n_x id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_r',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
apply IH_elt. unfold dep. left. auto.
destruct p_r' as [ p_r' [ p_r'_in_dep_x p_r'_extends ] ].
unfold dep in p_r'_in_dep_x.
inversion p_r'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists p_r'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_r'_is_x. subst id_r p_r'.
exists (path_cons (path_recv n_x id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

inversion t_s_is_x. subst id_s id_r p_r p_s.
exists (path_cons (path_recv n_s id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_refl.

(* Recv *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_s other | t_s false ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s t_s_is_x ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_t' | t_s t_s_in_DA_x ].

simpl in t_s_in_DA_t'.
assert (exists p_r',
  In trail (dep (thread_trail id p)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t'.
destruct p_r' as [ p_r' [ p_r'_in_dep p_r'_extends ] ].
unfold dep in p_r'_in_dep.
inversion p_r'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists p_r'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_r'_is_t'. subst id_r p_r'.
exists (path_cons (path_recv n_x id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_r',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
apply IH_elt. unfold dep. left. auto.
destruct p_r' as [ p_r' [ p_r'_in_dep_x p_r'_extends ] ].
unfold dep in p_r'_in_dep_x.
inversion p_r'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists p_r'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_r'_is_x. subst id_r p_r'.
exists (path_cons (path_send n_x id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

inversion t_s_is_x. subst id_s id_r p_r p_s.
exists (path_cons (path_recv n_s id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_refl.

inversion false.

(* Heap *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_s other | t_s false ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s false ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_t | t_s t_s_in_DA_x ].

simpl in t_s_in_DA_t.
assert (exists p_r',
  In trail (dep (thread_trail id p)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t.
destruct p_r' as [ p_r' [ p_r'_in_dep p_r'_extends ] ].
unfold dep in p_r'_in_dep.
inversion p_r'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists p_r'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_r'_is_t'. subst id_r p_r'.
exists (path_cons (path_heap n_x hp_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_r',
  In trail (dep (heap_trail hp_x)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
apply IH_elt. unfold dep. left. auto.
destruct p_r' as [ p_r' [ p_r'_in_dep_x p_r'_extends ] ].
unfold dep in p_r'_in_dep_x.
inversion p_r'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists p_r'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_r'_is_x.

inversion false. inversion false.

(* heap_path_nil *)
intros false. unfold dep in false.
inversion false as [ t_s false2 | t_s false2 ].
simpl in false2. inversion false2. inversion false2.

(* heap paths *)
intros elt IH_elt hp IH_path in_dep.
induction elt as [ n_x id_x p_x ]. unfold dep in in_dep. simpl in in_dep.

inversion in_dep as [ t_s other | t_s false ].
subst t_s. inversion other as [ t_s t_s_in_DA | t_s t_s_is_x ].
subst t_s. inversion t_s_in_DA as [ t_s t_s_in_DA_x | t_s t_s_in_DA_t ].

assert (exists p_r',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
apply IH_elt. unfold dep. left. auto.
destruct p_r' as [ p_r' [ p_r'_in_dep_x p_r'_extends ] ].
unfold dep in p_r'_in_dep_x.
inversion p_r'_in_dep_x as [ t_r' t_r'_in_DA | t_r' t_r'_is_x ].
exists p_r'. split. unfold dep. simpl. left. left. left. auto. auto.
inversion t_r'_is_x. subst id_r p_r'.
exists (path_cons (path_heap n_x hp) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

simpl in t_s_in_DA_t.
assert (exists p_r',
  In trail (dep (heap_trail hp)) (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_s id_s p_s) p_r)) as p_r'.
eapply IH_path. unfold dep. left. apply t_s_in_DA_t.
destruct p_r' as [ p_r' [ p_r'_in_dep p_r'_extends ] ].
unfold dep in p_r'_in_dep.
inversion p_r'_in_dep as [ t_r' t_r'_in_DA | t_r' t_r'_is_t' ].
exists p_r'. split. unfold dep. left. simpl. left. right. auto. auto.
inversion t_r'_is_t'.

inversion t_s_is_x. 

inversion false.

Qed.

Lemma recv_in_dep_implies_send_in_dep :
  forall n_r id_r p_r id_s p_s t,
    In trail (dep t) 
    (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
    exists p_s',
    In trail (dep t) (thread_trail id_s p_s') 
    /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s).

intros n_r id_r p_r id_s p_s.

cut ((forall p id,
  In trail (dep (thread_trail id p)) 
  (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
  exists p_s',
    In trail (dep (thread_trail id p)) (thread_trail id_s p_s') 
    /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s))
/\ (forall hp,
  In trail (dep (heap_trail hp)) 
  (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
  exists p_s',
    In trail (dep (heap_trail hp)) (thread_trail id_s p_s') 
    /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s))).

intros [ H1 H2 ].
induction t as [ id p | hp ]; auto.

apply trail_induction with 
(P := fun elt => 
  match elt with
    | path_left => True
    | path_right => True
    | path_send n id p => 
      In trail (dep (thread_trail id p)) 
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      exists p_s',
        In trail (dep (thread_trail id p)) (thread_trail id_s p_s') 
        /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s)
    | path_recv n id p => 
      In trail (dep (thread_trail id p)) 
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      exists p_s',
        In trail (dep (thread_trail id p)) (thread_trail id_s p_s') 
        /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s)
    | path_heap n hp =>
      In trail (dep (heap_trail hp)) 
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      exists p_s',
        In trail (dep (heap_trail hp)) (thread_trail id_s p_s') 
        /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s)
  end)
(P0 := fun elt =>
  match elt with
    | heap_path_thread n id p =>
      In trail (dep (thread_trail id p)) 
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      exists p_s',
        In trail (dep (thread_trail id p)) (thread_trail id_s p_s') 
        /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s)
  end)
(P1 := fun p =>
  forall id,
    In trail (dep (thread_trail id p)) 
      (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
      exists p_s',
        In trail (dep (thread_trail id p)) (thread_trail id_s p_s') 
        /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s))
(P2 := fun hp =>
  In trail (dep (heap_trail hp)) 
  (thread_trail id_r (path_cons (path_recv n_r id_s p_s) p_r)) ->
  exists p_s',
    In trail (dep (heap_trail hp)) (thread_trail id_s p_s') 
    /\ extends p_s' (path_cons (path_send n_r id_r p_r) p_s)).

auto. auto. auto. auto. auto. auto.

intros id false. unfold dep in false.
inversion false as [ t_r false2 | t_r false2 ]; subst t_r.
simpl in false2. inversion false2. inversion false2.

intros elt IH_elt p IH_path id in_dep.
induction elt as [ | | n_x id_x p_x | n_x id_x p_x | n_x hp_x ].

(* Left *)
unfold dep in in_dep.
inversion in_dep as [ t_r t_r_in_DA | t_r false ].
simpl in t_r_in_DA.
assert (exists p_s',
  In trail (dep (thread_trail id p)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
eapply IH_path. unfold dep. left. apply t_r_in_DA.
destruct p_s' as [ p_s' [ p_s'_in_dep p_s'_extends ] ].
unfold dep in p_s'_in_dep.
inversion p_s'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_s'. split. unfold dep. left. simpl. auto. auto.
inversion t_s'_is_t'. subst id_s p_s'.
exists (path_cons path_left p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

(* Right *)
unfold dep in in_dep.
inversion in_dep as [ t_r t_r_in_DA | t_r false ].
simpl in t_r_in_DA.
assert (exists p_s',
  In trail (dep (thread_trail id p)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
eapply IH_path. unfold dep. left. apply t_r_in_DA.
destruct p_s' as [ p_s' [ p_s'_in_dep p_s'_extends ] ].
unfold dep in p_s'_in_dep.
inversion p_s'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_s'. split. unfold dep. left. simpl. auto. auto.
inversion t_s'_is_t'. subst id_s p_s'.
exists (path_cons path_right p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

(* Send *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_r other | t_r false ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r t_r_is_x ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_t' | t_r t_r_in_DA_x ].

simpl in t_r_in_DA_t'.
assert (exists p_s',
  In trail (dep (thread_trail id p)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t'.
destruct p_s' as [ p_s' [ p_s'_in_dep p_s'_extends ] ].
unfold dep in p_s'_in_dep.
inversion p_s'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_s'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_s'_is_t'. subst id_s p_s'.
exists (path_cons (path_send n_x id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_s',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
apply IH_elt. unfold dep. left. auto.
destruct p_s' as [ p_s' [ p_s'_in_dep_x p_s'_extends ] ].
unfold dep in p_s'_in_dep_x.
inversion p_s'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_s'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_s'_is_x. subst id_s p_s'.
exists (path_cons (path_recv n_x id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

inversion t_r_is_x. subst id_r id_s p_s p_r.
exists (path_cons (path_send n_r id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_refl.

inversion false.

(* Recv *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_r other | t_r t_r_is_x ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r false ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_t' | t_r t_r_in_DA_x ].

simpl in t_r_in_DA_t'.
assert (exists p_s',
  In trail (dep (thread_trail id p)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t'.
destruct p_s' as [ p_s' [ p_s'_in_dep p_s'_extends ] ].
unfold dep in p_s'_in_dep.
inversion p_s'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_s'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_s'_is_t'. subst id_s p_s'.
exists (path_cons (path_recv n_x id_x p_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_s',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
apply IH_elt. unfold dep. left. auto.
destruct p_s' as [ p_s' [ p_s'_in_dep_x p_s'_extends ] ].
unfold dep in p_s'_in_dep_x.
inversion p_s'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_s'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_s'_is_x. subst id_s p_s'.
exists (path_cons (path_send n_x id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.
inversion false.

inversion t_r_is_x. subst id_r id_s p_s p_r.
exists (path_cons (path_send n_r id p) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_refl.

(* Heap *)
unfold dep in in_dep. simpl in in_dep.
inversion in_dep as [ t_r other | t_r false ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r false ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_t | t_r t_r_in_DA_x ].

simpl in t_r_in_DA_t.
assert (exists p_s',
  In trail (dep (thread_trail id p)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t.
destruct p_s' as [ p_s' [ p_s'_in_dep p_s'_extends ] ].
unfold dep in p_s'_in_dep.
inversion p_s'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_s'. split. unfold dep. left. simpl. left. left. auto. auto.
inversion t_s'_is_t'. subst id_s p_s'.
exists (path_cons (path_heap n_x hp_x) p).
split. unfold dep. right. apply In_singleton.
apply extends_cons. auto.

assert (exists p_s',
  In trail (dep (heap_trail hp_x)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
apply IH_elt. unfold dep. left. auto.
destruct p_s' as [ p_s' [ p_s'_in_dep_x p_s'_extends ] ].
unfold dep in p_s'_in_dep_x.
inversion p_s'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_s'. split. unfold dep. simpl. left. left. right. auto. auto.
inversion t_s'_is_x.

inversion false. inversion false.

(* heap_path_nil *)
intros false. unfold dep in false.
inversion false as [ t_r false2 | t_r false2 ].
simpl in false2. inversion false2. inversion false2.

(* heap paths *)
intros elt IH_elt hp IH_path in_dep.
induction elt as [ n_x id_x p_x ]. unfold dep in in_dep. simpl in in_dep.

inversion in_dep as [ t_r other | t_r false ].
subst t_r. inversion other as [ t_r t_r_in_DA | t_r t_r_is_x ].
subst t_r. inversion t_r_in_DA as [ t_r t_r_in_DA_x | t_r t_r_in_DA_t ].

assert (exists p_s',
  In trail (dep (thread_trail id_x p_x)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
apply IH_elt. unfold dep. left. auto.
destruct p_s' as [ p_s' [ p_s'_in_dep_x p_s'_extends ] ].
unfold dep in p_s'_in_dep_x.
inversion p_s'_in_dep_x as [ t_s' t_s'_in_DA | t_s' t_s'_is_x ].
exists p_s'. split. unfold dep. simpl. left. left. left. auto. auto.
inversion t_s'_is_x. subst id_s p_s'.
exists (path_cons (path_heap n_x hp) p_x).
split. unfold dep. simpl. left. right. apply In_singleton.
apply extends_cons. auto.

simpl in t_r_in_DA_t.
assert (exists p_s',
  In trail (dep (heap_trail hp)) (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_r id_r p_r) p_s)) as p_s'.
eapply IH_path. unfold dep. left. apply t_r_in_DA_t.
destruct p_s' as [ p_s' [ p_s'_in_dep p_s'_extends ] ].
unfold dep in p_s'_in_dep.
inversion p_s'_in_dep as [ t_s' t_s'_in_DA | t_s' t_s'_is_t' ].
exists p_s'. split. unfold dep. left. simpl. left. right. auto. auto.
inversion t_s'_is_t'.

inversion t_r_is_x. 

inversion false.

Qed.
