Require Import coherent.
Require Import committable.
Require Import committable_before_choose_left.
Require Import committable_before_choose_right.
Require Import committable_before_end_lockdown.
Require Import committable_before_send_recv.
Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical.
Require Import Coq.Sets.Ensembles.
Require Import dep.
Require Import highlevel.
Require Import ll2hl_get_S.
Require Import ll_invariants.
Require Import lowlevel.
Require Import nonces.
Require Import sets.
Require Import syntax.
Require Import trails.
Require Import unique_tids.

Definition sync_sim (h:global_heap) (S_susp:Ensemble susp_thread)
  (S_H:Ensemble search_heap) (S:Ensemble search_thread) : Prop :=
  match h with
    | unfrozen H =>
      S_susp = empty_susp
      /\ S_H = empty_heap
      /\ S = empty_search
    | frozen H =>
      (forall S_complete H' hp trails S_susp_complete 
        S_hl_init S_hl_complete,
        Included search_thread S_complete S ->
        search_unique_tids S_complete ->
        (forall id h list_M e p,
          In search_thread S_complete (id, h, list_M, e, p) ->
          h = None) ->
        In search_heap S_H (H', hp) ->
        trails = Add trail 
          (get_thread_trails S_complete) (heap_trail hp) ->
          S_susp_complete = get_S_susp_complete S_susp S_complete ->
          S_hl_init = get_S_hl (translate_susp S_susp_complete) ->
          S_hl_complete = get_S_hl_S S_complete ->
          committable trails ->
          hl_sync_step_many H S_hl_init None H' S_hl_complete None)
      /\ 
      (forall S_complete ld id H' hp list_M e p trails S_susp_complete
        S_hl_init S_hl_complete,
        Included search_thread S_complete S ->
        search_unique_tids S_complete ->
        (forall id h list_M e p,
          In search_thread S_complete (id, h, list_M, e, p) ->
          h = None) ->
        ld = (id, Some (H', hp), list_M, e, p) ->
        In search_thread S ld ->
        (forall h' list_M' e' p',
          ~ In search_thread S_complete (id, h', list_M', e', p')) ->
        trails = (Union trail 
          (get_thread_trails S_complete) 
          (Couple trail (thread_trail id p) (heap_trail hp))) ->
        S_susp_complete = (get_S_susp_complete S_susp 
          (Add search_thread S_complete ld)) ->
        S_hl_init = get_S_hl (translate_susp S_susp_complete) ->
        S_hl_complete = get_S_hl_S S_complete ->
        committable trails ->
        (hl_sync_step_many H S_hl_init None 
          H' S_hl_complete (Some (id, list_M, e))))
  end.

Lemma ll_step_preserves_sync_sim :
  forall h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2,
    susp_get_ctxt S_susp1 ->
    T_S_susp_unique_tids T1 S_susp1 ->
    sync_sim h1 S_susp1 S_H1 S1 ->
    ll_step h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2 ->
    sync_sim h2 S_susp2 S_H2 S2.

intros h1 T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2
  susp_get_ctxt_before T_S_susp_unique_before sync_sim_before ll_steps.

(* some invariants proved in ll_invariants 
   -the decomposition of each thread held in S_susp is correct
   -S_susp has unique tids
   *)
assert (susp_get_ctxt S_susp2) as susp_get_ctxt_after.
eapply ll_step_preserves_susp_get_ctxt. eauto. auto.
assert (T_S_susp_unique_tids T2 S_susp2) as T_S_susp_unique_after.
eapply ll_step_preserves_T_S_susp_unique_tids. eauto. auto.
destruct T_S_susp_unique_before as 
  [ T1_unique [ S_susp1_unique T1_S_susp1_disjoint ] ].
destruct T_S_susp_unique_after as 
  [ T2_unique [ S_susp2_unique T2_S_susp2_disjoint ] ].

(* which low-level step did we take? *)
inversion ll_steps as [
  (* mt step *)
  H1 _T1 H2 _T2 mt_steps h1_eq T1_eq S_susp1_eq S_H1_eq S1_eq 
  h2_eq T2_eq S_susp2_eq S_H2_eq S2_eq
  | (* SyncInit *)
    H T_sync _T2 _T1 _S_susp2 _S2 _S_H2 T_sync_at_sync T1_eq
    disjoint_T_sync_T2 S_susp2_eq S2_eq S_H2_eq h1_eq T1_e1 S_susp1_eq 
    S_H1_eq S1_eq h2_eq T2_eq S_susp2_eq2 S_H2_eq2 S2_eq2
  | (* BeginLockdown *)
    H id p list_M e H' hp _T1 _S_susp1 _S_H1 _S1 hp_in_S_H1 in_S1 
    coherent_p_hp
  | (* EndLockdown *)
    H id_ld p_ld list_M_ld e_ld H_ld hp_ld T S_susp S_H1_ S n_ld n_ld_fresh 
    ldv T1_ T1_eq S_susp1_eq S_H1_eq S1_eq h2_eq T2_eq S_susp2_eq 
    S_H2_eq S2_eq
  | (* SendRecv *)
    H T S_susp S_H S id_s list_M_s e_s E_s k v p_s id_r list_M_r e_r
    E_r p_r n_sr n_sr_fresh s_is_send r_is_recv coher nothing T1_eq
    S_susp1_eq S_H1_eq S1_eq h2_eq T2_eq S_susp2_eq S_H2_eq S2_eq
  | (* Choose *)
    H T S_susp S_H S id s_H list_M e E v1 v2 p e_is_choose
  | (* ThenAlways *)
    H T S_susp S_H S id r list_M e E v1 v2 p e_is_then
  | (* StepLockdown *)
    H T S_susp S_H S id H' H'' hp list_M e e' p st_steps
  | (* NestedSyncInit *)
    H T S_susp S_H S id r list_M e M v p e_is_sync
  | (* NestedSyncComplete *)
    H T S_susp S_H S id r list_M M v p v_value
  | (* SyncComplete *)
    H S_complete _T1 T_susp T_complete _T2 S_susp_complete 
    S_susp_incomplete _S_susp1 H' hp S_H0 _S_H1 thread_trails
    trails S0 _S1 S_complete_complete S_unique T_complete_eq 
    S_susp_complete_eq S_susp_incomplete_eq T_susp_eq2 S_H1_eq 
    S1_eq T2_eq thread_trails_eq trails_eq comm h1_eq T1_eq 
    S_susp1_eq2 S_H1_eq2 S1_eq2 h2_eq T2_eq2 S_susp2_eq S_H2_eq S2_eq
]; 
subst h1;
unfold sync_sim in sync_sim_before;
try destruct sync_sim_before as [ sync_sim_no_ld_before sync_sim_ld_before ].

(* StepTopLevel *)
Focus.
unfold sync_sim. auto.
Unfocus.

(* SyncInit *)
Focus.
subst T1 S_susp2 S2 S_H2 _T1 S_susp1 S_H1 S1 h2 _T2 _S_susp2 _S_H2 _S2.
unfold sync_sim.

split.

(** collections of search threads with no lockdown thread *)
intros S_complete H' hp trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S S_complete_unique no_ld hp_is_nil trails_eq 
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.

(***  so our heap thread must be the initial heap *)
inversion hp_is_nil. subst H' hp.
subst S_susp_complete S_hl_init S_hl_complete trails.

(*** we haven't taken any steps yet, so show that the initial set 
   of high-level threads is the same as the final set *)
assert (get_S_hl (translate_susp
  (get_S_susp_complete (get_S_susp_init T_sync) S_complete)) =
get_S_hl_S S_complete) as rewriting.
apply Extensionality_Ensembles. split; intros x in_one.
inversion in_one as [ id e M v in_translate_susp e_at_sync ].
inversion in_translate_susp as [ id' M' v' in_S_susp_complete ]. subst id' e.
inversion in_S_susp_complete as 
  [ id' M'' v'' h list_M v''' p in_S_susp in_S_complete ]. subst id' M'' v''.
assert (In search_thread (get_S_init T_sync) (id, h, list_M, v''', p))
  as in_S_init. unfold Included in S_complete_in_S.
apply S_complete_in_S. auto.
inversion in_S_init as [ id' e' M'' v'''' in_T_sync e_at_sync_again ].
subst id' h list_M v'''' p.
inversion in_S_susp as [ id' e M''' v'' in_T_sync_again e_at_sync_yet_again ].
assert (e = e'). assert ((id, e) = (id, e')) as eq.
eapply mt_unique_tids_reversed. eauto.
left. auto. left. auto. inversion eq. auto.
subst id' M''' v'' e'.
assert (get_ctxt (M' [sync v']) M' (sync v')) as get_ctxt_M'_v'.
unfold susp_get_ctxt in susp_get_ctxt_after.
eapply susp_get_ctxt_after. eauto.
assert (M' = M /\ sync v' = sync v) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v'.
assert (M = M'' /\ sync v = sync v''') as M_v_eq.
eapply get_ctxt_unique. apply e_at_sync_yet_again. auto.
destruct M_v_eq as [ M_eq v'_eq ]. inversion v'_eq. subst M'' v'''.
eapply In_S_hl_S. eauto.

inversion in_one as [ id h list_M v p in_S_complete ].
assert (In search_thread (get_S_init T_sync) (id, h, list_M, v, p)) as in_S1.
unfold Included in S_complete_in_S. apply S_complete_in_S. auto.
inversion in_S1 as [ id' e M v' in_T_sync e_at_sync_v ].
subst id' h list_M v' p.
eapply In_S_hl. apply In_translate_susp. eapply In_S_susp_complete.
eapply In_S_susp_init. apply in_T_sync. apply e_at_sync_v.
apply in_S_complete.
unfold susp_get_ctxt in susp_get_ctxt_after.
eapply susp_get_ctxt_after.
eapply In_S_susp_init. apply in_T_sync. apply e_at_sync_v.

(*** take zero steps *)
rewrite rewriting. apply hl_sync_step_many_zero.

(** collections of search threads with a lockdown thread (vacuous) *)
intros S_complete ld id H' hp list_M e p trails S_susp_complete
  S_hl_init S_hl_complete S_complete_in_S2 S_complete_unique
  S_complete_no_ld ld_eq ld_in_S2 id_not_in_S_complete trails_eq
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.
subst ld. inversion ld_in_S2.
Unfocus.

(* BeginLockdown *)
Focus.
subst _T1 _S_susp1 _S_H1 _S1 h2 T2 S_susp2 S_H2 S2.
unfold sync_sim.

split.

(** no lockdown thread *)
intros S_complete H'' hp' trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique S_complete_no_ld hp'_in_S_H1 trails_eq 
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.
subst trails S_susp_complete S_hl_init S_hl_complete.
eapply sync_sim_no_ld_before.
instantiate (1 := S_complete).
intros t in_S_complete.
assert (In search_thread (Add search_thread S1
  (id, Some (H', hp), list_M, e, p)) t) as in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
destruct in_S2 as [ t t_in_S1 | t false ]. auto.
inversion false. subst t.
assert (Some (H', hp) = None) as contra.
eapply S_complete_no_ld. apply in_S_complete. inversion contra.
auto. auto. apply hp'_in_S_H1. reflexivity.
reflexivity. reflexivity. reflexivity. auto.

(** lockdown thread *)
intros S_complete ld id' H'' hp' list_M' e' p' trails S_susp_complete
  S_hl_init S_hl_complete S_complete_in_S2 S_complete_unique S_complete_no_ld
  ld_eq ld_in_S2 ld_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst ld trails S_susp_complete S_hl_init S_hl_complete.

(*** the non-lockdown threads were all in the soup before stepping *)
assert (Included search_thread S_complete S1) as S_complete_in_S1.
intros t t_in_S_complete.
assert (In search_thread (Add search_thread S1
  (id, Some (H', hp), list_M, e, p)) t) as t_in_S2.
apply S_complete_in_S2. auto.
destruct t_in_S2 as [ t t_in_S1 | t false ].
auto. inversion false. subst t.
assert (Some (H', hp) = None) as contra.
eapply S_complete_no_ld. apply t_in_S_complete. inversion contra.
clear S_complete_in_S2.

(*** either the ld thread was in the previous soup, or it's the one
   just created - previous soup case first *)
inversion ld_in_S2 as [ t t_in_S1 | t t_is_ld ].
eapply sync_sim_ld_before.
apply S_complete_in_S1.
auto. auto. reflexivity. apply t_in_S1. auto.
reflexivity. reflexivity. auto. auto. auto.

(*** the exciting case - our lockdown thread is the one that just began
   lockdown mode *)
inversion t_is_ld. subst id' H'' hp' list_M' e' p'.
eapply hl_sync_step_many_reversed.

(**** first take a bunch of steps to just before the BeginLockdown*)
eapply sync_sim_no_ld_before.

(**** the committable set is the same but with the pre-lockdown thread *)
instantiate (1 := Add search_thread S_complete (id, None, list_M, e, p)).
intros x x_in_S_complete'.
inversion x_in_S_complete' as [ x' in_S_complete | x' is_ld ].
unfold Included in S_complete_in_S1. auto.
inversion is_ld. auto.
unfold search_unique_tids.
intros id1 h1 list_M1 e1 p1 h2 list_M2 e2 p2
  t1_in_S_complete' t2_in_S_complete'.
inversion t1_in_S_complete' as [ t1' t1_in_S_complete | t1' t1_is_ld ];
inversion t2_in_S_complete' as [ t2' t2_in_S_complete | t2' t2_is_ld ].
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
inversion t2_is_ld. subst id1.
assert False as false.
unfold not in ld_not_in_S_complete. eapply ld_not_in_S_complete.
eauto. inversion false.
inversion t1_is_ld. subst id1.
assert False as false.
unfold not in ld_not_in_S_complete. eapply ld_not_in_S_complete.
eauto. inversion false.
inversion t1_is_ld. subst id1 h1 list_M1 e1 p1.
inversion t2_is_ld. reflexivity.
intros id1 h1 list_M1 e1 p1 in_S_complete'.
inversion in_S_complete' as [ x x_in_S_complete | x x_is_ld ].
eapply S_complete_no_ld. eauto.
inversion x_is_ld. reflexivity.
apply hp_in_S_H1.
reflexivity. reflexivity.
assert (get_S_susp_complete S_susp1 (Add search_thread S_complete
  (id, Some (H', hp), list_M, e, p)) =
get_S_susp_complete S_susp1
(Add search_thread S_complete (id, None, list_M, e, p))) as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id1 M1 v1 h1 list_M1 e1 p1 in_S_susp1 in_S_complete' ].
inversion in_S_complete' as [ x' x'_in_S_complete | x' x'_is_ld ].
eapply In_S_susp_complete. auto. left. eauto.
inversion x'_is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_susp_complete. auto. right. apply In_singleton.
inversion x_in_one as [ id1 M1 v1 h1 list_M1 e1 p1 in_S_susp1 in_S_complete ].
inversion in_S_complete as [ x' x'_in_S_complete | x' x'_is_ld ].
eapply In_S_susp_complete. auto. left. eauto.
inversion x'_is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_susp_complete. auto. right. apply In_singleton.
rewrite rewriting. reflexivity.

reflexivity.

assert (get_thread_trails
  (Add search_thread S_complete (id, None, list_M, e, p)) =
  Add trail (get_thread_trails S_complete) (thread_trail id p))
as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id' h' list_M' e' p' in_S_complete' ].
inversion in_S_complete' as [ x' x'_in_S_complete | x' x'_is_ld ].
left. eapply In_thread_trails. apply x'_in_S_complete.
inversion x'_is_ld. subst id' p'. right. apply In_singleton.
destruct x_in_one as [ x x_in_S_complete | x x_is_ld ].
inversion x_in_S_complete as [ id' h' list_M' e' p' in_S_complete_again ].
eapply In_thread_trails. left. apply in_S_complete_again.
inversion x_is_ld. eapply In_thread_trails. right. apply In_singleton.
rewrite -> rewriting. clear rewriting.
assert (Add trail (Add trail (get_thread_trails S_complete) 
  (thread_trail id p)) (heap_trail hp) =
Union trail (get_thread_trails S_complete)
(Couple trail 
  (thread_trail id p)
  (heap_trail hp))) as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_is_thread | x x_is_heap ].
destruct x_is_thread as [ x x_not_ld | x x_is_ld ].
left. auto.
inversion x_is_ld. right. left. 
inversion x_is_heap. right. right.
destruct x_in_one as [ x x_in_S_complete | x x_is_ld_or_heap ].
left. left. auto.
destruct x_is_ld_or_heap. left. right. apply In_singleton. right.
apply In_singleton.
rewrite rewriting. clear rewriting.
auto.

(*** then take a single BeginLockdown step *)
assert (get_S_hl_S (Add search_thread S_complete (id, None, list_M, e, p)) =
  Add hl_thread (get_S_hl_S S_complete) (id, list_M, e)) as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id' h' list_M' e' p' in_S_complete' ].
inversion in_S_complete' as [ x' x'_in_S_complete | x' x'_is_ld ].
left. eapply In_S_hl_S. eauto.
inversion x'_is_ld. right. apply In_singleton.
inversion x_in_one as [ x' x'_in_S_complete | x' x'_is_ld ].
inversion x'_in_S_complete as [ id' h' list_M' e' p' in_S_complete' ].
eapply In_S_hl_S. left. eauto.
inversion x'_is_ld. eapply In_S_hl_S. right. apply In_singleton.
rewrite rewriting.
apply hl_sync_step_begin_lockdown.
intros ld_in_S_hl_complete.
inversion ld_in_S_hl_complete as [ id' h' list_M' e' p' in_S_complete ].
unfold not in ld_not_in_S_complete. eapply ld_not_in_S_complete. eauto.
Unfocus.

(* EndLockdown *)
Focus.
subst T1 S_susp1 S_H1_ S1 h2 T2 S_susp2 S_H2 S2.
unfold sync_sim. split.

(** no lockdown thread *)
intros S_complete H' hp trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique S_complete_no_ld hp_in_S_H
  trails_eq S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm.
subst trails S_susp_complete S_hl_init S_hl_complete.

(*** is ld thread in the committable set? *)
assert (In search_thread S_complete 
  (id_ld, None, list_M_ld, e_ld,
    path_cons (path_heap n_ld hp_ld) p_ld) \/
~ In search_thread S_complete 
(id_ld, None, list_M_ld, e_ld,
  path_cons (path_heap n_ld hp_ld) p_ld)) as classic.
apply classic.
destruct classic as [ ld_in_complete | ld_not_in_complete ].
assert ((H', hp) = (H_ld,
  heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld) \/
(H', hp) <> (H_ld,
  heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))
as classic.
apply classic.
destruct classic as [ h_ld_is_heap | h_ld_not_heap ].

(* ld thread and heap both in complete - the "interesting" case *)
inversion h_ld_is_heap. subst H' hp.
eapply hl_sync_step_many_reversed
  with (H2 := H_ld) (s2 := Some (id_ld, list_M_ld, e_ld)).
eapply sync_sim_ld_before.

(** a billion things to prove *)

(** first: show that the "before" set of committable search threads
   is in the "before" set of all search threads *)
instantiate (1 := Subtract search_thread
      S_complete
      (id_ld, None, list_M_ld, e_ld,
        path_cons (path_heap n_ld hp_ld) p_ld)).
unfold Included. intros x x_in_one.
destruct x_in_one as [ x_in_complete x_not_ld ].
assert (In search_thread
  (Add search_thread S
    (id_ld, None, list_M_ld, e_ld,
      path_cons (path_heap n_ld hp_ld) p_ld)) x) as x_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
inversion x_in_S2 as [ | x' x_is_ld ]. left. auto.
tauto.

(** show that the before set has unique tids *)
unfold search_unique_tids.
intros id h1 list_M1 e1 p1 h2 list_M2 e2 p2 t1_in_before t2_in_before.
inversion t1_in_before as [ t1_in_S_complete t1_not_ld ].
inversion t2_in_before as [ t2_in_S_complete t2_not_ld ].
apply S_complete_unique; auto.

(** show that the before set has no ld thread *)
intros id h list_M e p t_in_before.
inversion t_in_before as [ t_in_complete t_not_ld ].
eapply S_complete_no_ld.
apply t_in_complete.

(** lockdown thread *)
reflexivity.

(** lockdown thread exists *)
right. apply In_singleton.

(** ld id not in complete *)
unfold not. intros h' list_M' e' path' in_before.
inversion in_before as [ in_complete not_comm ].
assert ((id_ld, h', list_M', e', path') = 
  (id_ld, None, list_M_ld, e_ld,
    path_cons (path_heap n_ld hp_ld) p_ld)) as false.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
destruct not_comm. rewrite false. apply In_singleton.

(** set of "before" trails *)
reflexivity.

(** set of suspended threads matching the committable set *)
reflexivity.

(** checking initial set of high-level threads is the same *)
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id e M v_init in_translate_susp ].
inversion in_translate_susp as [ id_ M' v_init' in_susp_complete ]. subst id_.
inversion in_susp_complete as 
  [ id_ M_ v_init'_ h list_M e_ll p in_susp in_complete ]. 
subst id_ M_ v_init'_.
assert (id = id_ld \/ id <> id_ld) as classic.
apply classic. destruct classic.
subst id.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. split. apply in_complete.
unfold not. intros false. inversion false; subst; tauto.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.

inversion x_in_one as [ id e M v_init in_translate_susp ].
inversion in_translate_susp as [ id_ M' v_init' in_susp_complete ]. subst id_.
inversion in_susp_complete as 
  [ id_ M_ v_init'_ h list_M e_ll p in_susp in_complete ]. 
subst id_ M_ v_init'_.
inversion in_complete as [ t [ in_S_complete not_comm ] | t t_is_ld ].
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. eauto.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
inversion t_is_ld. subst id h list_M e_ll p.
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. apply ld_in_complete.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.

(** the "before" set of high-level threads *)
reflexivity.

(** committable before *)
eapply committable_before_end_lockdown.
unfold not. intros in_trails.
inversion in_trails as [ id h list_M e p in_complete ].
inversion in_complete as [ false not_comm ].
instantiate (1 := n_ld) in false.
assert ((id_ld, h, list_M, e, path_cons (path_heap n_ld hp_ld) p_ld) =
  (id_ld, None, list_M_ld, e_ld, path_cons (path_heap n_ld hp_ld) p_ld))
as rewriting.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique; auto.
destruct not_comm. rewrite -> rewriting. apply In_singleton.
unfold not. intros in_trails.
inversion in_trails as [ id h list_M e p in_complete ].

assert (Union trail
  (get_thread_trails
    (Subtract search_thread S_complete
      (id_ld, None, list_M_ld, e_ld,
        path_cons (path_heap n_ld hp_ld) p_ld)))
  (Couple trail
    (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld))
    (heap_trail
      (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))) =
  (Add trail (get_thread_trails S_complete)
    (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))))
as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_not_ld_or_heap | x x_is_ld_or_heap ].
inversion x_not_ld_or_heap as [ id h list_M e p [ in_complete not_ld ] ].
left. eapply In_thread_trails. apply in_complete.
inversion x_is_ld_or_heap.
left. eapply In_thread_trails. eauto.
right. apply In_singleton.
destruct x_in_one as [ x x_is_thread | x x_is_heap ]. 
inversion x_is_thread as [ id h list_M e p in_complete ].
assert (id = id_ld \/ id <> id_ld) as classic. 
apply classic. destruct classic.
subst id.
assert ((id_ld, h, list_M, e, p) = (id_ld, None, list_M_ld, e_ld,
  path_cons (path_heap n_ld hp_ld) p_ld)) as rewriting.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
inversion rewriting. right. left.
left. eapply In_thread_trails. split. eauto.
unfold not. intros false. inversion false. subst id. tauto.
inversion x_is_heap. right. right.
rewrite -> rewriting. auto.

assert (Add hl_thread 
  (get_S_hl_S
    (Subtract search_thread S_complete
      (id_ld, None, list_M_ld, e_ld,
        path_cons (path_heap n_ld hp_ld) p_ld)))
  (id_ld, list_M_ld, e_ld) =
  get_S_hl_S S_complete) as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_not_ld | x x_is_ld ].
inversion x_not_ld as [ id h list_M e p in_complete ].
destruct in_complete as [ in_complete not_ld ].
eapply In_S_hl_S. apply in_complete.
inversion x_is_ld. eapply In_S_hl_S. eauto.
inversion x_in_one as [ id h list_M e p in_complete ].
assert (id = id_ld \/ id <> id_ld) as classic.
apply classic. destruct classic.
subst id.
assert ((id_ld, h, list_M, e, p) = (id_ld, None, list_M_ld, e_ld,
  path_cons (path_heap n_ld hp_ld) p_ld)) as rewriting.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
inversion rewriting. right. apply In_singleton.
left. eapply In_S_hl_S. split. eauto.
unfold not. intros false. inversion false. subst id. tauto.
rewrite <- rewriting. clear rewriting.

eapply hl_sync_step_end_lockdown.
auto.

(* heap not the lockdown heap *)
assert (heap_extends hp 
  (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_extends_ld.
unfold committable in comm.
destruct comm as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
assert (exists hp',
  (Add trail (get_thread_trails S_complete) (heap_trail hp)) (heap_trail hp')
  /\ heap_extends hp'
  (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp'.
apply parent_heap with 
  (t := (thread_trail id_ld (path_cons (path_heap n_ld hp_ld) p_ld))).
left. 
apply In_thread_trails with (h := (None:option search_heap))
  (list_M := list_M_ld) (e := e_ld).
apply ld_in_complete.
unfold dep, dep_aux. simpl. left. right. apply In_singleton.
destruct hp' as [ hp' [ hp'_in_trails hp'_extends ] ].
assert (hp = hp'). 
unfold unique in unique_heap.
destruct unique_heap as [ hp'' [ hp''_in_trails hp''_unique ] ].
assert (hp'' = hp). apply hp''_unique. right. apply In_singleton.
assert (hp'' = hp'). apply hp''_unique. auto.
subst hp' hp''. reflexivity.
subst hp'. auto.
inversion hp_in_S_H as [ ht in_S_H1 | ht false ].
assert (In nonce (nonces_heap_path hp) n_ld) as n_ld_in_nonces_hp.
assert (Included nonce (nonces_heap_path 
  (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) 
  (nonces_heap_path hp)) as nonces_included.
apply heap_extends_implies_nonces_included. auto.
unfold Included in nonces_included. apply nonces_included.
simpl. right. apply In_singleton.
assert (n_ld <> n_ld) as false.
unfold fresh_nonce in n_ld_fresh.
destruct n_ld_fresh as [ fresh_S_H fresh_S ].
apply fresh_S_H with (r := (H', hp)).
auto. auto. tauto.
inversion false. subst H' hp. tauto.

(* is the heap the ld heap? *)
assert ((H', hp) = (H_ld,
  heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld) \/
(H', hp) <> (H_ld,
  heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))
as classic.
apply classic.
destruct classic as [ h_ld_is_heap | h_ld_not_heap ].

(** ld thread not in set *)
inversion h_ld_is_heap. subst H' hp.
assert (exists p_ld',
  In trail (Add trail (get_thread_trails S_complete)
    (heap_trail (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)))
  (thread_trail id_ld p_ld') /\
  extends p_ld' (path_cons (path_heap n_ld hp_ld) p_ld)) as p_ld'.
unfold committable in comm.
destruct comm as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
eapply parent_thread. right. apply In_singleton.
unfold dep, dep_aux. simpl. left. right. apply In_singleton.
destruct p_ld' as [ p_ld' [ p_ld'_in_trails p_ld'_extends ] ].
assert (In nonce (nonces_path p_ld') n_ld) as n_ld_in_nonces_p_ld'.
assert (Included nonce
  (nonces_path (path_cons (path_heap n_ld hp_ld) p_ld))
  (nonces_path p_ld')) as included.
apply extends_implies_nonces_included. auto.
unfold Included in included. apply included.
simpl. right. apply In_singleton.
inversion p_ld'_in_trails as [ t_ld' t_ld'_in_complete | t_ld' false ].
inversion t_ld'_in_complete as [ id h list_M e p in_complete ].
assert (In search_thread (Add search_thread S
  (id_ld, None, list_M_ld, e_ld,
    path_cons (path_heap n_ld hp_ld) p_ld))
(id_ld, h, list_M, e, p_ld')) as in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
inversion in_S2 as [ t in_S | t is_ld ].
assert (n_ld <> n_ld) as false.
unfold fresh_nonce in n_ld_fresh.
destruct n_ld_fresh as [ n_ld_fresh_S_H n_ld_fresh_S ].
eapply n_ld_fresh_S. left. apply in_S. simpl.
induction h as [ nothing1 | ].
induction nothing1. left. auto. auto. tauto.
inversion is_ld. subst h list_M e p_ld'. subst p. tauto.
inversion false.

(* neither lockdown thread nor heap appears *)
eapply sync_sim_no_ld_before.
instantiate (1 := S_complete).
intros x x_in_complete.
assert (In search_thread (Add search_thread S
  (id_ld, None, list_M_ld, e_ld, path_cons (path_heap n_ld hp_ld) p_ld)) x)
as in_S2. unfold Included in S_complete_in_S2.
apply S_complete_in_S2. auto.
destruct in_S2 as [ x in_S | x false ].
left. auto. inversion false. subst x. tauto.

auto. auto.

instantiate (1 := hp).
inversion hp_in_S_H as [ ht in_S_H1 | ht false ].
auto. inversion false. subst H' hp. tauto.

reflexivity. reflexivity. reflexivity. reflexivity.
auto.

(** lockdown thread! *)
intros S_complete ld1 id_ld1 H'_ld1 hp_ld1 list_M_ld1 e_ld1 p_ld1 
  trails S_susp_complete S_hl_init S_hl_complete S_complete_in_S2
  S_complete_unique S_complete_no_ld ld1_eq ld1_in_S2 
  id_ld1_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst trails S_susp_complete S_hl_init S_hl_complete ld1.
inversion ld1_in_S2 as [ ld1 ld1_in_S | ld1 false ].

assert (~ In search_thread S_complete
  (id_ld, None, list_M_ld, e_ld, path_cons (path_heap n_ld hp_ld) p_ld))
as ld_not_in_complete.
unfold not. intros ld_in_complete.
unfold committable in comm_trails.
destruct comm_trails as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
unfold unique in unique_heap. 
destruct unique_heap as [ hp [ hp_in_trails hp_unique ] ].
assert (heap_extends hp_ld1
  (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld)) as hp_ld1_extends.
assert (exists hp'',
  In trail (Union trail (get_thread_trails S_complete)
    (Couple trail (thread_trail id_ld1 p_ld1) (heap_trail hp_ld1)))
  (heap_trail hp'')
  /\ heap_extends hp'' 
  (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))
as [ hp'' [ hp''_in_trails hp''_extends ] ].
eapply parent_heap. left. eapply In_thread_trails. apply ld_in_complete.
unfold dep, dep_aux. simpl. left. right. apply In_singleton.
assert (hp = hp_ld1). eapply hp_unique. right. right.
assert (hp = hp''). eapply hp_unique. auto.
subst hp hp''. auto.
assert (n_ld <> n_ld) as false.
unfold fresh_nonce in n_ld_fresh. 
destruct n_ld_fresh as [ fresh_S_H fresh_S ].
eapply fresh_S. left. apply ld1_in_S.
simpl. right.
assert (Included nonce 
  (nonces_heap_path (heap_path_cons (heap_path_thread n_ld id_ld p_ld) hp_ld))
  (nonces_heap_path hp_ld1)) as included.
apply heap_extends_implies_nonces_included. auto.
unfold Included in included. apply included.
simpl. right. apply In_singleton.
tauto.

eapply sync_sim_ld_before.
instantiate (1 := S_complete).
intros x x_in_one. 
unfold Included in S_complete_in_S2.
assert (In search_thread (Add search_thread S
  (id_ld, None, list_M_ld, e_ld, path_cons (path_heap n_ld hp_ld) p_ld)) x)
as in_S2. apply S_complete_in_S2. auto.
destruct in_S2 as [ x in_S | x false ].
left. auto. inversion false. subst x. tauto.

auto. auto. reflexivity.

left. apply ld1_in_S. auto. reflexivity. reflexivity.
reflexivity. reflexivity. auto.

inversion false.
Unfocus.

(* SendRecv *)
Focus.
subst T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2.
assert (id_s <> id_r) as id_s_not_id_r.
unfold coherent in coher.
assert (id_s <> id_r /\
  (forall p : path,
  In trail (dep (thread_trail id_r p_r))
  (thread_trail id_s p) -> extends p_s p) /\
(forall p : path,
  In trail (dep (thread_trail id_s p_s))
  (thread_trail id_r p) -> extends p_r p) /\
(forall (id : tid) (p1 p2 : path),
  In trail (dep (thread_trail id_s p_s))
  (thread_trail id p1) ->
  In trail (dep (thread_trail id_r p_r))
  (thread_trail id p2) ->
  extends p1 p2 \/ extends p2 p1) /\
(forall hp1 hp2 : heap_path,
  In trail (dep (thread_trail id_s p_s))
  (heap_trail hp1) ->
  In trail (dep (thread_trail id_r p_r))
  (heap_trail hp2) ->
  heap_extends hp1 hp2 \/ heap_extends hp2 hp1)) as contra.
apply coher. reflexivity. reflexivity. reflexivity. reflexivity.
destruct contra. auto. 
unfold sync_sim. split.

(** no lockdown thread *)
intros S_complete H' hp trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique S_complete_no_ld hp_in_S_H
  trails_eq S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm.
subst trails S_susp_complete S_hl_init S_hl_complete.

(*** is the sender in the committable set? *)
assert (In search_thread S_complete 
  (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
    path_cons (path_send n_sr id_r p_r) p_s) \/
~ In search_thread S_complete 
(id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
  path_cons (path_send n_sr id_r p_r) p_s)) as classic.
apply classic.
destruct classic as [ sender_in_complete | sender_not_in_complete ].
assert (In search_thread S_complete 
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r) \/
~ In search_thread S_complete 
(id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) as classic.
apply classic.
destruct classic as [ receiver_in_complete | receiver_not_in_complete ].

(* sender and receiver both in complete - the "interesting" case *)
eapply hl_sync_step_many_reversed.
eapply sync_sim_no_ld_before.

(** a billion things to prove *)

(** first: show that the "before" set of committable search threads
   is in the "before" set of all search threads *)
instantiate (1 := Union search_thread
    (Setminus search_thread
      S_complete
      (Couple search_thread
        (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r)))
    (Couple search_thread
      (id_s, None, list_M_s, e_s, p_s)
      (id_r, None, list_M_r, e_r, p_r))).
unfold Included. intros x x_in_one.
destruct x_in_one as [ x x_not_comm | x x_is_comm ].
destruct x_not_comm as [ x_in_complete x_not_comm ].
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r)))) x) as x_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
inversion x_in_S2 as [ | x' x_is_comm ]. left. auto.
destruct x_is_comm as [ x x_is_comm_before | x false ].
right. auto. tauto. right. auto.

(** show that the before set has unique tids *)
unfold search_unique_tids.
intros id h1 list_M1 e1 p1 h2 list_M2 e2 p2 t1_in_before t2_in_before.
inversion t1_in_before as [ t1 [ t1_in_S_complete t1_not_comm ] | t1 [ | ] ];
inversion t2_in_before as [ t2 [ t2_in_S_complete t2_not_comm ] | t2 [ | ] ].
apply S_complete_unique; auto.
subst id h2 list_M2 e2 p2.
assert ((id_s, h1, list_M1, e1, p1) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero),
  path_cons (path_send n_sr id_r p_r) p_s)) as false.
apply S_complete_unique; auto. rewrite -> false in t1_not_comm.
assert False. unfold not in t1_not_comm. apply t1_not_comm. left. tauto.
subst id h2 list_M2 e2 p2.
assert ((id_r, h1, list_M1, e1, p1) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as false.
apply S_complete_unique; auto. rewrite -> false in t1_not_comm.
assert False. unfold not in t1_not_comm. apply t1_not_comm. right. tauto.
subst id h1 list_M1 e1 p1.
assert ((id_s, h2, list_M2, e2, p2) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero),
  path_cons (path_send n_sr id_r p_r) p_s)) as false.
apply S_complete_unique; auto. rewrite -> false in t2_not_comm.
assert False. unfold not in t2_not_comm. apply t2_not_comm. left. tauto.
auto.
subst id_r id_s. tauto.
subst id h1 list_M1 e1 p1.
assert ((id_r, h2, list_M2, e2, p2) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as false.
apply S_complete_unique; auto. rewrite -> false in t2_not_comm.
assert False. unfold not in t2_not_comm. apply t2_not_comm. right. tauto.
subst id_r id_s. tauto.
auto.

(** show that the before set has no ld thread *)
intros id h list_M e p t_in_before.
inversion t_in_before as [ t [ t_in_complete ] | t [ | ] ].
intros. subst. eapply S_complete_no_ld.
apply t_in_complete. auto. auto.

(** the search heap still exists *)
eauto.

(** set of "before" trails *)
reflexivity.

(** set of suspended threads matching the committable set *)
reflexivity.

(** checking initial set of high-level threads is the same *)
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id e M v_init in_translate_susp ].
inversion in_translate_susp as [ id_ M' v_init' in_susp_complete ]. subst id_.
inversion in_susp_complete as 
  [ id_ M_ v_init'_ h list_M e_ll p in_susp in_complete ]. 
subst id_ M_ v_init'_.
assert (id = id_s \/ id <> id_s) as classic.
apply classic. destruct classic.
subst id.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. left. 
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
assert (id = id_r \/ id <> id_r) as classic.
apply classic. destruct classic.
subst id.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. right. 
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. split. apply in_complete.
unfold not. intros false. inversion false; subst; tauto.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.

inversion x_in_one as [ id e M v_init in_translate_susp ].
inversion in_translate_susp as [ id_ M' v_init' in_susp_complete ]. subst id_.
inversion in_susp_complete as 
  [ id_ M_ v_init'_ h list_M e_ll p in_susp in_complete ]. 
subst id_ M_ v_init'_.
inversion in_complete as [ t [ in_S_complete not_comm ] | t [ | ] ].
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. eauto.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
subst id h list_M e_ll p.
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. apply sender_in_complete.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
subst id h list_M e_ll p.
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. apply receiver_in_complete.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.

(** the "before" set of high-level threads *)
reflexivity.

(** committable before *)
assert (Add trail
  (get_thread_trails
    (Union search_thread
      (Setminus search_thread S_complete
        (Couple search_thread
          (id_s, None, list_M_s,
            evt_fill_hole E_s (alwaysEvt zero),
            path_cons (path_send n_sr id_r p_r) p_s)
          (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
            path_cons (path_recv n_sr id_s p_s) p_r)))
      (Couple search_thread (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r)))) 
  (heap_trail hp) =
  (Union trail
    (Add trail
      (get_thread_trails
        (Setminus search_thread S_complete
          (Couple search_thread
            (id_s, None, list_M_s,
              evt_fill_hole E_s (alwaysEvt zero),
              path_cons (path_send n_sr id_r p_r) p_s)
            (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
              path_cons (path_recv n_sr id_s p_s) p_r))))
      (heap_trail hp))
    (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r))))
as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_is_thread | x x_is_heap ].
inversion x_is_thread as [ id h list_M e p in_S_complete_before ].
inversion in_S_complete_before as [ t [ t_in_complete t_not_comm ] | t [ | ] ].
left. left. eapply In_thread_trails.
split. apply t_in_complete. auto.
subst id h list_M e p. right. left.
subst id h list_M e p. right. right.
inversion x_is_heap. left. right. apply In_singleton.
destruct x_in_one as [ x x_not_comm | x x_is_comm ].
destruct x_not_comm as [ x x_is_thread | x x_is_heap ].
inversion x_is_thread as [ id h list_M e p in_S_complete_before ].
left. eapply In_thread_trails. left. eauto.
inversion x_is_heap. right. apply In_singleton.
left. destruct x_is_comm.
eapply In_thread_trails. right. left.
eapply In_thread_trails. right. right.

rewrite -> rewriting. clear rewriting.
eapply committable_before_send_recv with (n1 := n_sr).
unfold not. intros false.
inversion false as [ t_s' false2 | t_s' false2 ].
inversion false2 as [ id_s_ h list_M e p_s' [ in_complete not_comm ] ].
assert ((id_s, h, list_M, e, path_cons (path_send n_sr id_r p_r) p_s) =
  (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
    path_cons (path_send n_sr id_r p_r) p_s)) as false3.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
inversion false3. subst h list_M e.
unfold not in not_comm. apply not_comm. left.
inversion false2.
unfold not. intros false.
inversion false as [ t_s' false2 | t_s' false2 ].
inversion false2 as [ id_r_ h list_M e p_r' [ in_complete not_comm ] ].
assert ((id_r, h, list_M, e, path_cons (path_recv n_sr id_s p_s) p_r) =
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) as false3.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
inversion false3. subst h list_M e.
unfold not in not_comm. apply not_comm. right.
inversion false2.

assert (Union trail
  (Add trail
    (get_thread_trails
      (Setminus search_thread S_complete
        (Couple search_thread
          (id_s, None, list_M_s,
            evt_fill_hole E_s (alwaysEvt zero),
            path_cons (path_send n_sr id_r p_r) p_s)
          (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
            path_cons (path_recv n_sr id_s p_s) p_r))))
    (heap_trail hp))
  (Couple trail
    (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s))
    (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))) =
  (Add trail (get_thread_trails S_complete)
    (heap_trail hp))) as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_not_comm | x [ | ] ].
destruct x_not_comm as [ x x_is_thread | x x_is_heap ].
inversion x_is_thread as [ id h list_M e p [ in_complete not_comm ] ].
left. eapply In_thread_trails. apply in_complete.
right. auto.
left. eapply In_thread_trails. apply sender_in_complete.
left. eapply In_thread_trails. apply receiver_in_complete.
destruct x_in_one as [ x x_in_complete | x x_is_heap ].
inversion x_in_complete as [ id h list_M e p in_complete ].
assert (id = id_s \/ id <> id_s) as classic. apply classic. 
destruct classic as [ | id_not_s ].
assert ((id, h, list_M, e, p) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero), 
  path_cons (path_send n_sr id_r p_r) p_s)) as is_sender.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_sender. right. left.
assert (id = id_r \/ id <> id_r) as classic. apply classic. 
destruct classic as [ | id_not_r ].
assert ((id, h, list_M, e, p) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as is_receiver.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_receiver. right. right.
left. left. eapply In_thread_trails. split.
apply in_complete. unfold not. intros false.
inversion false. subst. tauto. subst. tauto.
left. right. auto.
rewrite -> rewriting. auto.

eapply hl_sync_step_send_recv.
eapply In_S_hl_S. right. left.
eapply In_S_hl_S. right. right.
eauto. eauto.

apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id h list_M e p in_S_complete ].
assert (id = id_s \/ id <> id_s) as classic. apply classic. 
destruct classic as [ | id_not_s ].
assert ((id, h, list_M, e, p) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero), 
  path_cons (path_send n_sr id_r p_r) p_s)) as is_sender.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_sender. right. left.
assert (id = id_r \/ id <> id_r) as classic. apply classic. 
destruct classic as [ | id_not_r ].
assert ((id, h, list_M, e, p) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as is_receiver.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_receiver. right. right.
left. split. eapply In_S_hl_S. left. split. eauto.
unfold not. intros false. inversion false; subst; tauto.
unfold not. intros false. inversion false; subst; tauto.
destruct x_in_one as [ x x_not_comm | x x_is_comm ].
destruct x_not_comm as [ x_in_complete x_not_comm ].
inversion x_in_complete as [ id h list_M e p in_complete ].
subst x. inversion in_complete as [ x in_complete2 | x false ].
inversion in_complete2.
eapply In_S_hl_S. eauto.
inversion false as [ | ].
subst. destruct x_not_comm. left.
subst. destruct x_not_comm. right.
inversion x_is_comm as [ | ].
eapply In_S_hl_S. eauto.
eapply In_S_hl_S. eauto.

(* receiver not in complete *)
assert (exists p_r',
  In trail (Add trail (get_thread_trails S_complete) (heap_trail hp))
  (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_sr id_s p_s) p_r)) as p_r'.
unfold committable in comm.
destruct comm as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
eapply parent_thread.
left. eapply In_thread_trails. apply sender_in_complete.
unfold dep, dep_aux. simpl. left. right. apply In_singleton.
destruct p_r' as [ p_r' [ p_r'_in_trails p_r'_extends ] ].
assert (In nonce (nonces_path p_r') n_sr) as n_sr_in_nonces_p_r'.
assert (Included nonce (nonces_path (path_cons (path_recv n_sr id_s p_s) p_r)) 
  (nonces_path p_r')) as nonces_included.
apply extends_implies_nonces_included. auto.
unfold Included in nonces_included. apply nonces_included.
simpl. right. apply In_singleton.
assert (n_sr <> n_sr) as false.
unfold fresh_nonce in n_sr_fresh.
destruct n_sr_fresh as [ fresh_S_H fresh_S ].
inversion p_r'_in_trails as [ tr_r' tr_r'_in_S_complete | tr_r' false ].
inversion tr_r'_in_S_complete as 
  [ id_r_ h_r' M_r' e_r' p_r'_ t_r'_in_S_complete ]. subst id_r_ p_r'_.
remember (id_r, h_r', M_r', e_r', p_r') as t_r'.
eapply fresh_S with (s := t_r').
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r))))
    t_r') as t_r'_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
destruct t_r'_in_S2 as [ t_r' t_r'_in_S | t_r' other ].
left. apply t_r'_in_S.
destruct other as [ t_r' t_r'_in_before | t_r' t_r'_in_after ].
right. auto.
destruct t_r'_in_after as [ | ].
inversion Heqt_r'. subst id_r.
assert (id_s <> id_s) as contra.
unfold coherent in coher.
assert (id_s <> id_s /\
  (forall p : path,
  In trail (dep (thread_trail id_s p_r))
  (thread_trail id_s p) -> extends p_s p) /\
(forall p : path,
  In trail (dep (thread_trail id_s p_s))
  (thread_trail id_s p) -> extends p_r p) /\
(forall (id : tid) (p1 p2 : path),
  In trail (dep (thread_trail id_s p_s))
  (thread_trail id p1) ->
  In trail (dep (thread_trail id_s p_r))
  (thread_trail id p2) ->
  extends p1 p2 \/ extends p2 p1) /\
(forall hp1 hp2 : heap_path,
  In trail (dep (thread_trail id_s p_s))
  (heap_trail hp1) ->
  In trail (dep (thread_trail id_s p_r))
  (heap_trail hp2) ->
  heap_extends hp1 hp2 \/ heap_extends hp2 hp1)) as contra.
apply coher. reflexivity. reflexivity. reflexivity. reflexivity.
destruct contra. auto. tauto.
inversion Heqt_r'. subst. tauto.
subst t_r'. simpl. induction h_r' as [ r' | ].
induction r'. left. auto. auto.
inversion false. tauto.

assert (In search_thread S_complete 
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r) \/
~ In search_thread S_complete 
(id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) as classic.
apply classic.
destruct classic as [ receiver_in_complete | receiver_not_in_complete ].

(* sender not in complete *)
assert (exists p_s',
  In trail (Add trail (get_thread_trails S_complete) (heap_trail hp))
  (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_sr id_r p_r) p_s)) as p_s'.
unfold committable in comm.
destruct comm as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
eapply parent_thread.
left. eapply In_thread_trails. apply receiver_in_complete.
unfold dep, dep_aux. simpl. left. right. apply In_singleton.
destruct p_s' as [ p_s' [ p_s'_in_trails p_s'_extends ] ].
assert (In nonce (nonces_path p_s') n_sr) as n_sr_in_nonces_p_s'.
assert (Included nonce (nonces_path (path_cons (path_send n_sr id_r p_r) p_s)) 
  (nonces_path p_s')) as nonces_included.
apply extends_implies_nonces_included. auto.
unfold Included in nonces_included. apply nonces_included.
simpl. right. apply In_singleton.
assert (n_sr <> n_sr) as false.
unfold fresh_nonce in n_sr_fresh.
destruct n_sr_fresh as [ fresh_S_H fresh_S ].
inversion p_s'_in_trails as [ tr_s' tr_s'_in_S_complete | tr_s' false ].
inversion tr_s'_in_S_complete as 
  [ id_s_ h_s' M_s' e_s' p_s'_ t_s'_in_S_complete ]. subst id_s_ p_s'_.
remember (id_s, h_s', M_s', e_s', p_s') as t_s'.
eapply fresh_S with (s := t_s').
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r))))
    t_s') as t_s'_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
destruct t_s'_in_S2 as [ t_s' t_s'_in_S | t_s' other ].
left. apply t_s'_in_S.
destruct other as [ t_s' t_s'_in_before | t_s' t_s'_in_after ].
right. auto.
destruct t_s'_in_after as [ | ].
inversion Heqt_s'. subst. tauto.
inversion Heqt_s'. subst id_s.
assert (id_r <> id_r) as contra.
unfold coherent in coher.
assert (id_r <> id_r /\
  (forall p : path,
  In trail (dep (thread_trail id_r p_r))
  (thread_trail id_r p) -> extends p_s p) /\
(forall p : path,
  In trail (dep (thread_trail id_r p_s))
  (thread_trail id_r p) -> extends p_r p) /\
(forall (id : tid) (p1 p2 : path),
  In trail (dep (thread_trail id_r p_s))
  (thread_trail id p1) ->
  In trail (dep (thread_trail id_r p_r))
  (thread_trail id p2) ->
  extends p1 p2 \/ extends p2 p1) /\
(forall hp1 hp2 : heap_path,
  In trail (dep (thread_trail id_r p_s))
  (heap_trail hp1) ->
  In trail (dep (thread_trail id_r p_r))
  (heap_trail hp2) ->
  heap_extends hp1 hp2 \/ heap_extends hp2 hp1)) as contra.
apply coher. reflexivity. reflexivity. reflexivity. reflexivity.
destruct contra. auto. tauto.
subst t_s'. simpl. induction h_s' as [ s' | ].
induction s'. left. auto. auto.
inversion false. tauto.

(* neither in committable set *)
eapply sync_sim_no_ld_before.
instantiate (1 := S_complete).
intros x x_in_complete.
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r)))) x) as x_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
destruct x_in_S2 as [ | x other ].
left. auto.
destruct other as [ | x [ | ] ].
right. auto. tauto. tauto.
auto.

auto. eauto. reflexivity. reflexivity. 
reflexivity. reflexivity. auto.

(** OK! now onto the other half.  lockdown thread! *)
intros S_complete ld id_ld H'_ld hp_ld list_M_ld e_ld p_ld trails
  S_susp_complete S_hl_init S_hl_complete S_complete_in_S2
  S_complete_unique S_complete_no_ld ld_eq ld_in_S2 
  id_ld_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst trails S_susp_complete S_hl_init S_hl_complete.

assert (In search_thread S_complete 
  (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
    path_cons (path_send n_sr id_r p_r) p_s) \/
~ In search_thread S_complete 
(id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
  path_cons (path_send n_sr id_r p_r) p_s)) as classic.
apply classic.
destruct classic as [ sender_in_complete | sender_not_in_complete ].
assert (In search_thread S_complete 
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r) \/
~ In search_thread S_complete 
(id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) as classic.
apply classic.
destruct classic as [ receiver_in_complete | receiver_not_in_complete ].

(* sender and receiver both in complete - the "interesting" case *)
eapply hl_sync_step_many_reversed with
  (H2 := H'_ld) (s2 := (Some (id_ld, list_M_ld, e_ld))).
eapply sync_sim_ld_before.

(** a billion things to prove *)

(** first: show that the "before" set of committable search threads
   is in the "before" set of all search threads *)
instantiate (1 := Union search_thread
    (Setminus search_thread
      S_complete
      (Couple search_thread
        (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r)))
    (Couple search_thread
      (id_s, None, list_M_s, e_s, p_s)
      (id_r, None, list_M_r, e_r, p_r))).
unfold Included. intros x x_in_one.
destruct x_in_one as [ x x_not_comm | x x_is_comm ].
destruct x_not_comm as [ x_in_complete x_not_comm ].
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r)))) x) as x_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
inversion x_in_S2 as [ | x' x_is_comm ]. left. auto.
destruct x_is_comm as [ x x_is_comm_before | x false ].
right. auto. tauto. right. auto.

(** show that the before set has unique tids *)
unfold search_unique_tids.
intros id h1 list_M1 e1 p1 h2 list_M2 e2 p2 t1_in_before t2_in_before.
inversion t1_in_before as [ t1 [ t1_in_S_complete t1_not_comm ] | t1 [ | ] ];
inversion t2_in_before as [ t2 [ t2_in_S_complete t2_not_comm ] | t2 [ | ] ].
apply S_complete_unique; auto.
subst id h2 list_M2 e2 p2.
assert ((id_s, h1, list_M1, e1, p1) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero),
  path_cons (path_send n_sr id_r p_r) p_s)) as false.
apply S_complete_unique; auto. rewrite -> false in t1_not_comm.
assert False. unfold not in t1_not_comm. apply t1_not_comm. left. tauto.
subst id h2 list_M2 e2 p2.
assert ((id_r, h1, list_M1, e1, p1) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as false.
apply S_complete_unique; auto. rewrite -> false in t1_not_comm.
assert False. unfold not in t1_not_comm. apply t1_not_comm. right. tauto.
subst id h1 list_M1 e1 p1.
assert ((id_s, h2, list_M2, e2, p2) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero),
  path_cons (path_send n_sr id_r p_r) p_s)) as false.
apply S_complete_unique; auto. rewrite -> false in t2_not_comm.
assert False. unfold not in t2_not_comm. apply t2_not_comm. left. tauto.
auto.
subst id_r id_s. tauto.
subst id h1 list_M1 e1 p1.
assert ((id_r, h2, list_M2, e2, p2) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as false.
apply S_complete_unique; auto. rewrite -> false in t2_not_comm.
assert False. unfold not in t2_not_comm. apply t2_not_comm. right. tauto.
subst id_r id_s. tauto.
auto.

(** show that the before set has no ld thread *)
intros id h list_M e p t_in_before.
inversion t_in_before as [ t [ t_in_complete ] | t [ | ] ].
intros. subst. eapply S_complete_no_ld.
apply t_in_complete. auto. auto.

(** the lockdown thread exists *)
reflexivity.

(** the lockdown thread is in the pool *)
destruct ld_in_S2 as [ ld ld_in_S | ld false ].
subst ld. left. apply ld_in_S.
destruct false as [ ld false | ld false ];
destruct false; inversion ld_eq.

(** lockdown thread not in S_complete *)
unfold not. intros h' list_M' e' p' ld_in_S_complete_before.
remember (id_ld, h', list_M', e', p') as t_ld'.
destruct ld_in_S_complete_before as 
  [ t_ld' [ t_ld'_in_complete t_ld'_not_comm ] | t_ld' t_ld'_is_comm ].
subst t_ld'. unfold not in id_ld_not_in_S_complete.
eapply id_ld_not_in_S_complete. apply t_ld'_in_complete.
inversion t_ld'_is_comm as [ eq | eq ]. subst t_ld'. inversion eq.
subst id_ld. unfold not in id_ld_not_in_S_complete.
eapply id_ld_not_in_S_complete. apply sender_in_complete.
subst t_ld'. inversion eq.
subst id_ld. unfold not in id_ld_not_in_S_complete.
eapply id_ld_not_in_S_complete. apply receiver_in_complete.

(** set of "before" trails *)
reflexivity.

(** set of suspended threads matching the committable set *)
reflexivity.

(** checking initial set of high-level threads is the same *)
apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id e M v_init in_translate_susp ].
inversion in_translate_susp as [ id_ M' v_init' in_susp_complete ]. subst id_.
inversion in_susp_complete as 
  [ id_ M_ v_init'_ h list_M e_ll p in_susp in_complete ]. 
subst id_ M_ v_init'_.
subst ld. inversion in_complete as [ t t_in_complete | t t_is_ld ].
assert (id = id_s \/ id <> id_s) as classic.
apply classic. destruct classic.
subst id.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. right. left. 
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
assert (id = id_r \/ id <> id_r) as classic.
apply classic. destruct classic.
subst id.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. right. right. 
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. left. split. apply t_in_complete.
unfold not. intros false. inversion false; subst; tauto.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
inversion t_is_ld. subst id h list_M e_ll p.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.

inversion x_in_one as [ id e M v_init in_translate_susp ].
inversion in_translate_susp as [ id_ M' v_init' in_susp_complete ]. subst id_.
inversion in_susp_complete as 
  [ id_ M_ v_init'_ h list_M e_ll p in_susp in_complete ]. 
inversion in_complete as [ t t_in_complete | t t_is_ld ].
subst id_ M_ v_init'_. subst t.
inversion t_in_complete as [ t [ in_S_complete not_comm ] | t [ | ] ].
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. left. eauto.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
subst id h list_M e_ll p.
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. left. apply sender_in_complete.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
subst id h list_M e_ll p.
eapply In_S_hl. eapply In_translate_susp.
eapply In_S_susp_complete. apply in_susp. left. apply receiver_in_complete.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.
inversion t_is_ld. subst id h list_M e_ll p.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. subst ld. subst id_. apply In_singleton.
subst e. assert (M = M' /\ (sync v_init = sync v_init')) as M_v_eq.
eapply get_ctxt_unique. eauto.
unfold susp_get_ctxt in susp_get_ctxt_before.
eapply susp_get_ctxt_before. apply in_susp.
destruct M_v_eq as [ M_eq v_eq ]. inversion v_eq. subst M' v_init'.
eauto.

(** the "before" set of high-level threads *)
reflexivity.

(** committable before *)
assert (Union trail
  (get_thread_trails
    (Union search_thread
      (Setminus search_thread S_complete
        (Couple search_thread
          (id_s, None, list_M_s,
            evt_fill_hole E_s (alwaysEvt zero),
            path_cons (path_send n_sr id_r p_r) p_s)
          (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
            path_cons (path_recv n_sr id_s p_s) p_r)))
      (Couple search_thread (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r)))) 
  (Couple trail
    (thread_trail id_ld p_ld) (heap_trail hp_ld)) =
  (Union trail
    (Union trail
      (get_thread_trails
        (Setminus search_thread S_complete
          (Couple search_thread
            (id_s, None, list_M_s,
              evt_fill_hole E_s (alwaysEvt zero),
              path_cons (path_send n_sr id_r p_r) p_s)
            (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
              path_cons (path_recv n_sr id_s p_s) p_r))))
      (Couple trail (thread_trail id_ld p_ld) (heap_trail hp_ld)))
    (Couple trail (thread_trail id_s p_s) (thread_trail id_r p_r))))
as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_is_thread | x x_is_ld_or_heap ].
inversion x_is_thread as [ id h list_M e p in_S_complete_before ].
inversion in_S_complete_before as [ t [ t_in_complete t_not_comm ] | t [ | ] ].
left. left. eapply In_thread_trails.
split. apply t_in_complete. auto.
subst id h list_M e p. right. left.
subst id h list_M e p. right. right.
inversion x_is_ld_or_heap; subst x.
left. right. left. 
left. right. right.
destruct x_in_one as [ x x_not_comm | x x_is_comm ].
destruct x_not_comm as [ x x_is_thread | x x_is_ld_or_heap ].
inversion x_is_thread as [ id h list_M e p in_S_complete_before ].
left. eapply In_thread_trails. left. eauto.
inversion x_is_ld_or_heap. right. left. right. right.
left. destruct x_is_comm.
eapply In_thread_trails. right. left.
eapply In_thread_trails. right. right.

rewrite -> rewriting. clear rewriting.
eapply committable_before_send_recv with (n1 := n_sr).
unfold not. intros false.
inversion false as [ t_s' false2 | t_s' false2 ].
inversion false2 as [ id_s_ h list_M e p_s' [ in_complete not_comm ] ].
assert ((id_s, h, list_M, e, path_cons (path_send n_sr id_r p_r) p_s) =
  (id_s, None, list_M_s, evt_fill_hole E_s (alwaysEvt zero),
    path_cons (path_send n_sr id_r p_r) p_s)) as false3.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
inversion false3. subst h list_M e.
unfold not in not_comm. apply not_comm. left.
inversion false2. subst id_ld.
unfold not in id_ld_not_in_S_complete.
eapply id_ld_not_in_S_complete. apply sender_in_complete.
unfold not. intros false.
inversion false as [ t_s' false2 | t_s' false2 ].
inversion false2 as [ id_r_ h list_M e p_r' [ in_complete not_comm ] ].
assert ((id_r, h, list_M, e, path_cons (path_recv n_sr id_s p_s) p_r) =
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) as false3.
unfold search_unique_tids in S_complete_unique.
apply S_complete_unique. auto. auto.
inversion false3. subst h list_M e.
unfold not in not_comm. apply not_comm. right.
inversion false2.
unfold not in id_ld_not_in_S_complete.
eapply id_ld_not_in_S_complete. subst id_ld. apply receiver_in_complete.

assert (Union trail
  (Union trail
    (get_thread_trails
      (Setminus search_thread S_complete
        (Couple search_thread
          (id_s, None, list_M_s,
            evt_fill_hole E_s (alwaysEvt zero),
            path_cons (path_send n_sr id_r p_r) p_s)
          (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
            path_cons (path_recv n_sr id_s p_s) p_r))))
    (Couple trail (thread_trail id_ld p_ld) (heap_trail hp_ld)))
  (Couple trail
    (thread_trail id_s (path_cons (path_send n_sr id_r p_r) p_s))
    (thread_trail id_r (path_cons (path_recv n_sr id_s p_s) p_r))) =
  (Union trail (get_thread_trails S_complete)
    (Couple trail (thread_trail id_ld p_ld) (heap_trail hp_ld))))
as rewriting.
apply Extensionality_Ensembles. split; intros x x_in_one.
destruct x_in_one as [ x x_not_comm | x [ | ] ].
destruct x_not_comm as [ x x_is_thread | x x_is_ld_or_heap ].
inversion x_is_thread as [ id h list_M e p [ in_complete not_comm ] ].
left. eapply In_thread_trails. apply in_complete.
right. auto.
left. eapply In_thread_trails. apply sender_in_complete.
left. eapply In_thread_trails. apply receiver_in_complete.
destruct x_in_one as [ x x_in_complete | x x_is_ld_or_heap ].
inversion x_in_complete as [ id h list_M e p in_complete ].
assert (id = id_s \/ id <> id_s) as classic. apply classic. 
destruct classic as [ | id_not_s ].
assert ((id, h, list_M, e, p) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero), 
  path_cons (path_send n_sr id_r p_r) p_s)) as is_sender.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_sender. right. left.
assert (id = id_r \/ id <> id_r) as classic. apply classic. 
destruct classic as [ | id_not_r ].
assert ((id, h, list_M, e, p) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as is_receiver.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_receiver. right. right.
left. left. eapply In_thread_trails. split.
apply in_complete. unfold not. intros false.
inversion false. subst. tauto. subst. tauto.
left. right. auto.
rewrite -> rewriting. auto.

eapply hl_sync_step_send_recv.
eapply In_S_hl_S. right. left.
eapply In_S_hl_S. right. right.
eauto. eauto.

apply Extensionality_Ensembles. split; intros x x_in_one.
inversion x_in_one as [ id h list_M e p in_S_complete ].
assert (id = id_s \/ id <> id_s) as classic. apply classic. 
destruct classic as [ | id_not_s ].
assert ((id, h, list_M, e, p) = (id_s, None, list_M_s,
  evt_fill_hole E_s (alwaysEvt zero), 
  path_cons (path_send n_sr id_r p_r) p_s)) as is_sender.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_sender. right. left.
assert (id = id_r \/ id <> id_r) as classic. apply classic. 
destruct classic as [ | id_not_r ].
assert ((id, h, list_M, e, p) = (id_r, None, list_M_r, 
  evt_fill_hole E_r (alwaysEvt v),
  path_cons (path_recv n_sr id_s p_s) p_r)) as is_receiver.
unfold search_unique_tids in S_complete_unique.
subst id. apply S_complete_unique; auto.
inversion is_receiver. right. right.
left. split. eapply In_S_hl_S. left. split. eauto.
unfold not. intros false. inversion false; subst; tauto.
unfold not. intros false. inversion false; subst; tauto.
destruct x_in_one as [ x x_not_comm | x x_is_comm ].
destruct x_not_comm as [ x_in_complete x_not_comm ].
inversion x_in_complete as [ id h list_M e p in_complete ].
subst x. inversion in_complete as [ x in_complete2 | x false ].
inversion in_complete2.
eapply In_S_hl_S. eauto.
inversion false as [ | ].
subst. destruct x_not_comm. left.
subst. destruct x_not_comm. right.
inversion x_is_comm as [ | ].
eapply In_S_hl_S. eauto.
eapply In_S_hl_S. eauto.

(** receiver not in complete *)
assert (exists p_r',
  In trail (Union trail (get_thread_trails S_complete) 
    (Couple trail (thread_trail id_ld p_ld) (heap_trail hp_ld)))
  (thread_trail id_r p_r') /\
  extends p_r' (path_cons (path_recv n_sr id_s p_s) p_r)) as p_r'.
unfold committable in comm_trails.
destruct comm_trails as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
eapply parent_thread.
left. eapply In_thread_trails. apply sender_in_complete.
unfold dep, dep_aux. simpl. left. right. apply In_singleton.
destruct p_r' as [ p_r' [ p_r'_in_trails p_r'_extends ] ].
assert (In nonce (nonces_path p_r') n_sr) as n_sr_in_nonces_p_r'.
assert (Included nonce (nonces_path (path_cons (path_recv n_sr id_s p_s) p_r)) 
  (nonces_path p_r')) as nonces_included.
apply extends_implies_nonces_included. auto.
unfold Included in nonces_included. apply nonces_included.
simpl. right. apply In_singleton.
assert (n_sr <> n_sr) as false.
unfold fresh_nonce in n_sr_fresh.
destruct n_sr_fresh as [ fresh_S_H fresh_S ].
inversion p_r'_in_trails as [ tr_r' tr_r'_in_S_complete | tr_r' false ].
inversion tr_r'_in_S_complete as 
  [ id_r_ h_r' M_r' e_r' p_r'_ t_r'_in_S_complete ]. subst id_r_ p_r'_.
remember (id_r, h_r', M_r', e_r', p_r') as t_r'.
eapply fresh_S with (s := t_r').
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r))))
    t_r') as t_r'_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
destruct t_r'_in_S2 as [ t_r' t_r'_in_S | t_r' other ].
left. apply t_r'_in_S.
destruct other as [ t_r' t_r'_in_before | t_r' t_r'_in_after ].
right. auto.
destruct t_r'_in_after as [ | ].
inversion Heqt_r'. subst id_r.
assert (id_s <> id_s) as contra.
unfold coherent in coher.
assert (id_s <> id_s /\
  (forall p : path,
  In trail (dep (thread_trail id_s p_r))
  (thread_trail id_s p) -> extends p_s p) /\
(forall p : path,
  In trail (dep (thread_trail id_s p_s))
  (thread_trail id_s p) -> extends p_r p) /\
(forall (id : tid) (p1 p2 : path),
  In trail (dep (thread_trail id_s p_s))
  (thread_trail id p1) ->
  In trail (dep (thread_trail id_s p_r))
  (thread_trail id p2) ->
  extends p1 p2 \/ extends p2 p1) /\
(forall hp1 hp2 : heap_path,
  In trail (dep (thread_trail id_s p_s))
  (heap_trail hp1) ->
  In trail (dep (thread_trail id_s p_r))
  (heap_trail hp2) ->
  heap_extends hp1 hp2 \/ heap_extends hp2 hp1)) as contra.
apply coher. reflexivity. reflexivity. reflexivity. reflexivity.
destruct contra. auto. tauto.
inversion Heqt_r'. subst. tauto.
subst t_r'. simpl. induction h_r' as [ r' | ].
induction r'. left. auto. auto.
inversion false.
subst id_ld p_r' tr_r'. 
eapply fresh_S. left. subst ld.
inversion ld_in_S2 as [ ld ld_in_S | ld false2 ].
apply ld_in_S.
inversion false2 as [ ld' [ | ] | ld' [ | ] ].
simpl. left. auto. tauto.

assert (In search_thread S_complete 
  (id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r) \/
~ In search_thread S_complete 
(id_r, None, list_M_r, evt_fill_hole E_r (alwaysEvt v),
    path_cons (path_recv n_sr id_s p_s) p_r)) as classic.
apply classic.
destruct classic as [ receiver_in_complete | receiver_not_in_complete ].

(* sender not in complete *)
assert (exists p_s',
  In trail (Union trail (get_thread_trails S_complete) 
    (Couple trail (thread_trail id_ld p_ld) (heap_trail hp_ld)))
  (thread_trail id_s p_s') /\
  extends p_s' (path_cons (path_send n_sr id_r p_r) p_s)) as p_s'.
unfold committable in comm_trails.
destruct comm_trails as 
  [ unique_heap [ unique_thread [ parent_thread parent_heap ] ] ].
eapply parent_thread.
left. eapply In_thread_trails. apply receiver_in_complete.
unfold dep, dep_aux. simpl. left. right. apply In_singleton.
destruct p_s' as [ p_s' [ p_s'_in_trails p_s'_extends ] ].
assert (In nonce (nonces_path p_s') n_sr) as n_sr_in_nonces_p_rs.
assert (Included nonce (nonces_path (path_cons (path_send n_sr id_r p_r) p_s)) 
  (nonces_path p_s')) as nonces_included.
apply extends_implies_nonces_included. auto.
unfold Included in nonces_included. apply nonces_included.
simpl. right. apply In_singleton.
assert (n_sr <> n_sr) as false.
unfold fresh_nonce in n_sr_fresh.
destruct n_sr_fresh as [ fresh_S_H fresh_S ].
inversion p_s'_in_trails as [ tr_s' tr_s'_in_S_complete | tr_s' false ].
inversion tr_s'_in_S_complete as 
  [ id_s_ h_s' M_s' e_s' p_s'_ t_s'_in_S_complete ]. subst id_s_ p_s'_.
remember (id_s, h_s', M_s', e_s', p_s') as t_s'.
eapply fresh_S with (s := t_s').
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r))))
    t_s') as t_s'_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
destruct t_s'_in_S2 as [ t_s' t_s'_in_S | t_s' other ].
left. apply t_s'_in_S.
destruct other as [ t_s' t_s'_in_before | t_s' t_s'_in_after ].
right. auto.
destruct t_s'_in_after as [ | ].
inversion Heqt_s'. subst. tauto.
inversion Heqt_s'. subst id_s.
assert (id_r <> id_r) as contra.
unfold coherent in coher.
assert (id_r <> id_r /\
  (forall p : path,
  In trail (dep (thread_trail id_r p_r))
  (thread_trail id_r p) -> extends p_s p) /\
(forall p : path,
  In trail (dep (thread_trail id_r p_s))
  (thread_trail id_r p) -> extends p_r p) /\
(forall (id : tid) (p1 p2 : path),
  In trail (dep (thread_trail id_r p_s))
  (thread_trail id p1) ->
  In trail (dep (thread_trail id_r p_r))
  (thread_trail id p2) ->
  extends p1 p2 \/ extends p2 p1) /\
(forall hp1 hp2 : heap_path,
  In trail (dep (thread_trail id_r p_s))
  (heap_trail hp1) ->
  In trail (dep (thread_trail id_r p_r))
  (heap_trail hp2) ->
  heap_extends hp1 hp2 \/ heap_extends hp2 hp1)) as contra.
apply coher. reflexivity. reflexivity. reflexivity. reflexivity.
destruct contra. auto. tauto.
subst t_s'. simpl. induction h_s' as [ r' | ].
induction r'. left. auto. auto.
inversion false.
subst id_ld p_s' tr_s'. 
eapply fresh_S. left. subst ld.
inversion ld_in_S2 as [ ld ld_in_S | ld false2 ].
apply ld_in_S.
inversion false2 as [ ld' [ | ] | ld' [ | ] ].
simpl. left. auto. tauto.

(* neither in committable set *)
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
intros x x_in_complete.
assert (In search_thread
  (Union search_thread S
    (Union search_thread
      (Couple search_thread
        (id_s, None, list_M_s, e_s, p_s)
        (id_r, None, list_M_r, e_r, p_r))
      (Couple search_thread
        (id_s, None, list_M_s,
          evt_fill_hole E_s (alwaysEvt zero),
          path_cons (path_send n_sr id_r p_r) p_s)
        (id_r, None, list_M_r,
          evt_fill_hole E_r (alwaysEvt v),
          path_cons (path_recv n_sr id_s p_s) p_r)))) x) as x_in_S2.
unfold Included in S_complete_in_S2. apply S_complete_in_S2. auto.
destruct x_in_S2 as [ | x other ].
left. auto.
destruct other as [ | x [ | ] ].
right. auto. tauto. tauto.
auto.

auto. eauto. 
inversion ld_in_S2 as [ ld' ld_in_S | ld' false ].
left. auto.
subst ld.
inversion false as [ ld false2 | ld false2 ];
subst ld; inversion false2.
auto.
reflexivity. reflexivity. 
reflexivity. reflexivity. auto.
Unfocus.

(* Choose *)
Focus.
subst T1 S_susp1 S_H1 S1 h2 T2 S_susp2 S_H2 S2.
unfold sync_sim.
split.

(* no lockdown thread *)
intros S_complete H' hp trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique no_ld hp_is_nil trails_eq 
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread (Union search_thread S
  (Couple search_thread 
    (id, Some s_H, list_M, evt_fill_hole E v1, path_cons path_left p)
  (id, Some s_H, list_M, evt_fill_hole E v2, path_cons path_right p))) x)
as x_in_S2.
unfold Included in S_complete_in_S2. auto.
destruct x_in_S2 as [ x x_in_S | x [ | ] ].
auto.
assert (Some s_H = None) as false. eapply no_ld. eauto. inversion false.
assert (Some s_H = None) as false. eapply no_ld. eauto. inversion false.

eapply sync_sim_no_ld_before.
left. unfold Included in S_complete_in_S. apply S_complete_in_S. eauto.
auto. auto. eauto. reflexivity. reflexivity.
subst S_hl_init. subst S_susp_complete. reflexivity.
subst S_hl_complete. reflexivity.
subst trails. auto.

(* lockdown thread *)
intros S_complete ld id' H'' hp' list_M' e' p' trails S_susp_complete
  S_hl_init S_hl_complete S_complete_in_S2 S_complete_unique S_complete_no_ld
  ld_eq ld_in_S2 ld_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst ld trails S_susp_complete S_hl_init S_hl_complete.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread (Union search_thread S
  (Couple search_thread 
    (id, Some s_H, list_M, evt_fill_hole E v1, path_cons path_left p)
  (id, Some s_H, list_M, evt_fill_hole E v2, path_cons path_right p))) x)
as x_in_S2.
unfold Included in S_complete_in_S2. auto.
destruct x_in_S2 as [ x x_in_S | x [ | ] ].
auto.
assert (Some s_H = None) as false. eapply S_complete_no_ld. 
eauto. inversion false.
assert (Some s_H = None) as false. eapply S_complete_no_ld. 
eauto. inversion false.

(* which is the lockdown thread - choose left, choose right,
   or something else entirely? *)
remember (id', Some (H'', hp'), list_M', e', p') as ld.
destruct ld_in_S2 as [ ld ld_in_S | ld [ | ] ]. subst ld.
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity. left.
apply ld_in_S. auto. reflexivity. reflexivity.
reflexivity. reflexivity. auto.

(* choose left *)
inversion Heqld. subst id' s_H list_M' e' p'.
eapply hl_sync_step_many_reversed.
instantiate (3 := H'').
instantiate (1 := Some (id, list_M, e)).
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity.
right. apply In_singleton. auto.
reflexivity. reflexivity.

apply Extensionality_Ensembles. 
split; intros x x_in_one;
inversion x_in_one as 
  [ id1 e_init1 M1 v_init1 in_translate_susp e_init1_is_sync ];
inversion in_translate_susp as [ id_ M2 v_init2 in_susp_complete ];
subst e_init1;
inversion in_susp_complete as 
  [ id__ M2_ v_init2_ h1 list_M1 e1 p1 in_susp in_complete_or_is_ld ];
subst id_ id__ M2_ v_init2_.
assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

reflexivity.

rewrite <- add_add_to_union_couple.
apply committable_before_choose_left.
unfold not. intros in_complete_or_heap.
inversion in_complete_or_heap as [ x in_complete | x false ].
inversion in_complete. unfold not in ld_not_in_S_complete.
eapply ld_not_in_S_complete. eauto. inversion false.
rewrite -> add_add_to_union_couple. auto.

eapply hl_sync_step_choose_left.
apply e_is_choose.

(* choose right *)
inversion Heqld. subst id' s_H list_M' e' p'.
eapply hl_sync_step_many_reversed.
instantiate (3 := H'').
instantiate (1 := Some (id, list_M, e)).
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity.
right. apply In_singleton. auto.
reflexivity. reflexivity.

apply Extensionality_Ensembles. 
split; intros x x_in_one;
inversion x_in_one as 
  [ id1 e_init1 M1 v_init1 in_translate_susp e_init1_is_sync ];
inversion in_translate_susp as [ id_ M2 v_init2 in_susp_complete ];
subst e_init1;
inversion in_susp_complete as 
  [ id__ M2_ v_init2_ h1 list_M1 e1 p1 in_susp in_complete_or_is_ld ];
subst id_ id__ M2_ v_init2_.
assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

reflexivity.

rewrite <- add_add_to_union_couple.
apply committable_before_choose_right.
unfold not. intros in_complete_or_heap.
inversion in_complete_or_heap as [ x in_complete | x false ].
inversion in_complete. unfold not in ld_not_in_S_complete.
eapply ld_not_in_S_complete. eauto. inversion false.
rewrite -> add_add_to_union_couple. auto.

eapply hl_sync_step_choose_right.
apply e_is_choose.
Unfocus.

(* ThenAlways *)
Focus.
remember (id, Some r, list_M, evt_fill_hole E (app v2 v1), p) as ld2.

unfold sync_sim.
split.

(* no lockdown thread *)
intros S_complete H' hp trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique no_ld hp_is_nil trails_eq 
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread S2 x) as x_in_S2.
unfold Included in S_complete_in_S2. subst S2. auto.
subst S2. destruct x_in_S2 as [ x x_in_S | x x_is_ld ].
auto.
inversion x_is_ld. subst x ld2.
assert (Some r = None) as false. eapply no_ld. eauto. inversion false.

eapply sync_sim_no_ld_before.
subst S1. left. unfold Included in S_complete_in_S. 
apply S_complete_in_S. eauto.
auto. auto. subst S_H2. eauto. reflexivity. reflexivity.
subst S_hl_init. subst S_susp_complete. subst S_susp2. reflexivity.
subst S_hl_complete. reflexivity.
subst trails. auto.

(* lockdown thread *)
intros S_complete ld' id' H'' hp' list_M' e' p' trails S_susp_complete
  S_hl_init S_hl_complete S_complete_in_S2 S_complete_unique S_complete_no_ld
  ld_eq ld_in_S2 ld_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst trails S_susp_complete S_hl_init S_hl_complete.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread (Add search_thread S ld2) x) as x_in_S2.
unfold Included in S_complete_in_S2. auto.
destruct x_in_S2 as [ x x_in_S | x x_is_ld2 ].
auto.
inversion x_is_ld2. subst x ld2.
assert (Some r = None) as false.
eapply S_complete_no_ld. eauto.
inversion false.

(* which is the lockdown thread - ld2, or something else entirely? *)
destruct ld_in_S2 as [ ld' ld'_in_S | ld' ld'_is_ld2 ]. subst ld'.
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity. subst S1. left.
apply ld'_in_S. auto. reflexivity. reflexivity.
subst S_susp2. reflexivity. reflexivity. auto.

(* lockdown thread is ld2 *)
inversion ld'_is_ld2 as [ ld'_eq ].
eapply hl_sync_step_many_reversed.
instantiate (3 := H'').
instantiate (1 := Some (id, list_M, e)).
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity.
subst S1. right. 
subst ld' ld2. inversion ld'_eq. subst id'.
apply In_singleton.
subst ld' ld2. inversion ld'_eq. subst id'.
auto.
reflexivity. reflexivity. 
subst S_susp2 S_susp1. subst ld' ld2. inversion ld'_eq. subst.


apply Extensionality_Ensembles. 
split; intros x x_in_one;
inversion x_in_one as 
  [ id1 e_init1 M1 v_init1 in_translate_susp e_init1_is_sync ];
inversion in_translate_susp as [ id_ M2 v_init2 in_susp_complete ];
subst e_init1;
inversion in_susp_complete as 
  [ id__ M2_ v_init2_ h1 list_M1 e1 p1 in_susp in_complete_or_is_ld ];
subst id_ id__ M2_ v_init2_.
assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

reflexivity.

subst ld' ld2. inversion ld'_eq. subst.
auto.

subst ld' ld2. inversion ld'_eq. subst.
eapply hl_sync_step_then_always. auto.
Unfocus.

(* StepLockdown *)
Focus.
remember (id, Some (H'', hp), list_M, e', p) as ld2.

unfold sync_sim.
split.

(* no lockdown thread *)
intros S_complete H''' hp' trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique no_ld hp_is_nil trails_eq 
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread S2 x) as x_in_S2.
unfold Included in S_complete_in_S2. subst S2. auto.
subst S2. destruct x_in_S2 as [ x x_in_S | x x_is_ld ].
auto.
inversion x_is_ld. subst x ld2.
assert (Some (H'', hp) = None) as false. eapply no_ld. eauto. inversion false.

eapply sync_sim_no_ld_before.
subst S1. left. unfold Included in S_complete_in_S. 
apply S_complete_in_S. eauto.
auto. auto. subst S_H2. eauto. reflexivity. reflexivity.
subst S_hl_init. subst S_susp_complete. subst S_susp2. reflexivity.
subst S_hl_complete. reflexivity.
subst trails. auto.

(* lockdown thread *)
intros S_complete ld3 id3 H3' hp3 list_M3 e3 p3 trails S_susp_complete
  S_hl_init S_hl_complete S_complete_in_S2 S_complete_unique S_complete_no_ld
  ld_eq ld_in_S2 ld_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst trails S_susp_complete S_hl_init S_hl_complete.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread (Add search_thread S ld2) x) as x_in_S2.
unfold Included in S_complete_in_S2. auto.
destruct x_in_S2 as [ x x_in_S | x x_is_ld2 ].
auto.
inversion x_is_ld2. subst x ld2.
assert (Some (H'', hp) = None) as false.
eapply S_complete_no_ld. eauto.
inversion false.

(* which is the lockdown thread - ld2, or something else entirely? *)
destruct ld_in_S2 as [ ld' ld'_in_S | ld' ld'_is_ld2 ]. subst ld'.
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity. subst S1. left.
apply ld'_in_S. auto. reflexivity. reflexivity.
subst S_susp2. reflexivity. reflexivity. auto.

(* lockdown thread is ld2 *)
inversion ld'_is_ld2 as [ ld'_eq ].
eapply hl_sync_step_many_reversed.
instantiate (3 := H').
instantiate (1 := Some (id, list_M, e)).
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity.
subst S1. right. 
subst ld' ld2. inversion ld'_eq. subst id3.
apply In_singleton.
subst ld' ld2. inversion ld'_eq. subst id3.
auto.
reflexivity. reflexivity. 
subst S_susp2 S_susp1. subst ld' ld2. inversion ld'_eq. subst.

apply Extensionality_Ensembles. 
split; intros x x_in_one;
inversion x_in_one as 
  [ id1 e_init1 M1 v_init1 in_translate_susp e_init1_is_sync ];
inversion in_translate_susp as [ id_ M2 v_init2 in_susp_complete ];
subst e_init1;
inversion in_susp_complete as 
  [ id__ M2_ v_init2_ h1 list_M1 e1 p1 in_susp in_complete_or_is_ld ];
subst id_ id__ M2_ v_init2_.
assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

reflexivity.

subst ld' ld2. inversion ld'_eq. subst.
auto.

subst ld' ld2. inversion ld'_eq. subst.
eapply hl_sync_step_step_lockdown. auto.
Unfocus.

(* NestedSyncInit *)
Focus.
remember (id, Some r, M :: list_M, v, p) as ld2.

unfold sync_sim.
split.

(* no lockdown thread *)
intros S_complete H' hp trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique no_ld hp_is_nil trails_eq 
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread S2 x) as x_in_S2.
unfold Included in S_complete_in_S2. subst S2. auto.
subst S2. destruct x_in_S2 as [ x x_in_S | x x_is_ld ].
auto.
inversion x_is_ld. subst x ld2.
assert (Some r = None) as false. eapply no_ld. eauto. inversion false.

eapply sync_sim_no_ld_before.
subst S1. left. unfold Included in S_complete_in_S. 
apply S_complete_in_S. eauto.
auto. auto. subst S_H2. eauto. reflexivity. reflexivity.
subst S_hl_init. subst S_susp_complete. subst S_susp2. reflexivity.
subst S_hl_complete. reflexivity.
subst trails. auto.

(* lockdown thread *)
intros S_complete ld' id' H'' hp' list_M' e' p' trails S_susp_complete
  S_hl_init S_hl_complete S_complete_in_S2 S_complete_unique S_complete_no_ld
  ld_eq ld_in_S2 ld_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst trails S_susp_complete S_hl_init S_hl_complete.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread (Add search_thread S ld2) x) as x_in_S2.
unfold Included in S_complete_in_S2. auto.
destruct x_in_S2 as [ x x_in_S | x x_is_ld2 ].
auto.
inversion x_is_ld2. subst x ld2.
assert (Some r = None) as false.
eapply S_complete_no_ld. eauto.
inversion false.

(* which is the lockdown thread - ld2, or something else entirely? *)
destruct ld_in_S2 as [ ld' ld'_in_S | ld' ld'_is_ld2 ]. subst ld'.
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity. subst S1. left.
apply ld'_in_S. auto. reflexivity. reflexivity.
subst S_susp2. reflexivity. reflexivity. auto.

(* lockdown thread is ld2 *)
inversion ld'_is_ld2 as [ ld'_eq ].
eapply hl_sync_step_many_reversed.
instantiate (3 := H'').
instantiate (1 := Some (id, list_M, e)).
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity.
subst S1. right. 
subst ld' ld2. inversion ld'_eq. subst id'.
apply In_singleton.
subst ld' ld2. inversion ld'_eq. subst id'.
auto.
reflexivity. reflexivity. 
subst S_susp2 S_susp1. subst ld' ld2. inversion ld'_eq. subst.


apply Extensionality_Ensembles. 
split; intros x x_in_one;
inversion x_in_one as 
  [ id1 e_init1 M1 v_init1 in_translate_susp e_init1_is_sync ];
inversion in_translate_susp as [ id_ M2 v_init2 in_susp_complete ];
subst e_init1;
inversion in_susp_complete as 
  [ id__ M2_ v_init2_ h1 list_M1 e1 p1 in_susp in_complete_or_is_ld ];
subst id_ id__ M2_ v_init2_.
assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

reflexivity.

subst ld' ld2. inversion ld'_eq. subst.
auto.

subst ld' ld2. inversion ld'_eq. subst.
eapply hl_sync_step_nested_sync_init. auto.
Unfocus.

(* NestedSyncComplete *)
Focus.
remember (id, Some r, list_M, M [v], p) as ld2.

unfold sync_sim.
split.

(* no lockdown thread *)
intros S_complete H' hp trails S_susp_complete S_hl_init S_hl_complete
  S_complete_in_S2 S_complete_unique no_ld hp_is_nil trails_eq 
  S_susp_complete_eq S_hl_init_eq S_hl_complete_eq comm_trails.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread S2 x) as x_in_S2.
unfold Included in S_complete_in_S2. subst S2. auto.
subst S2. destruct x_in_S2 as [ x x_in_S | x x_is_ld ].
auto.
inversion x_is_ld. subst x ld2.
assert (Some r = None) as false. eapply no_ld. eauto. inversion false.

eapply sync_sim_no_ld_before.
subst S1. left. unfold Included in S_complete_in_S. 
apply S_complete_in_S. eauto.
auto. auto. subst S_H2. eauto. reflexivity. reflexivity.
subst S_hl_init. subst S_susp_complete. subst S_susp2. reflexivity.
subst S_hl_complete. reflexivity.
subst trails. auto.

(* lockdown thread *)
intros S_complete ld' id' H'' hp' list_M' e' p' trails S_susp_complete
  S_hl_init S_hl_complete S_complete_in_S2 S_complete_unique S_complete_no_ld
  ld_eq ld_in_S2 ld_not_in_S_complete trails_eq S_susp_complete_eq
  S_hl_init_eq S_hl_complete_eq comm_trails.
subst trails S_susp_complete S_hl_init S_hl_complete.

assert (Included search_thread S_complete S) as S_complete_in_S.
intros x x_in_complete.
assert (In search_thread (Add search_thread S ld2) x) as x_in_S2.
unfold Included in S_complete_in_S2. auto.
destruct x_in_S2 as [ x x_in_S | x x_is_ld2 ].
auto.
inversion x_is_ld2. subst x ld2.
assert (Some r = None) as false.
eapply S_complete_no_ld. eauto.
inversion false.

(* which is the lockdown thread - ld2, or something else entirely? *)
destruct ld_in_S2 as [ ld' ld'_in_S | ld' ld'_is_ld2 ]. subst ld'.
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity. subst S1. left.
apply ld'_in_S. auto. reflexivity. reflexivity.
subst S_susp2. reflexivity. reflexivity. auto.

(* lockdown thread is ld2 *)
inversion ld'_is_ld2 as [ ld'_eq ].
eapply hl_sync_step_many_reversed.
instantiate (3 := H'').
instantiate (1 := Some (id, M :: list_M, alwaysEvt v)).
eapply sync_sim_ld_before.
instantiate (1 := S_complete).
subst S1. left. unfold Included in S_complete_in_S. auto.
auto. auto. reflexivity.
subst S1. right. 
subst ld' ld2. inversion ld'_eq. subst id'.
apply In_singleton.
subst ld' ld2. inversion ld'_eq. subst id'.
auto.
reflexivity. reflexivity. 
subst S_susp2 S_susp1. subst ld' ld2. inversion ld'_eq. subst.


apply Extensionality_Ensembles. 
split; intros x x_in_one;
inversion x_in_one as 
  [ id1 e_init1 M1 v_init1 in_translate_susp e_init1_is_sync ];
inversion in_translate_susp as [ id_ M2 v_init2 in_susp_complete ];
subst e_init1;
inversion in_susp_complete as 
  [ id__ M2_ v_init2_ h1 list_M1 e1 p1 in_susp in_complete_or_is_ld ];
subst id_ id__ M2_ v_init2_.
assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

assert (get_ctxt (ctxt_fill_hole M2 (sync v_init2)) M2 (sync v_init2)).
unfold susp_get_ctxt in susp_get_ctxt_after. eapply susp_get_ctxt_after.
apply in_susp.
assert (M1 = M2 /\ sync v_init1 = sync v_init2) as M_v_eq.
eapply get_ctxt_unique. eauto. auto.
destruct M_v_eq as [ M_eq v_eq ]. subst M2. inversion v_eq. subst v_init2.

subst x.
inversion in_complete_or_is_ld as [ x in_complete | x is_ld ].

eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. left. apply in_complete. eauto.
inversion is_ld. subst id1 h1 list_M1 e1 p1.
eapply In_S_hl. eapply In_translate_susp. eapply In_S_susp_complete.
apply in_susp. right. apply In_singleton. eauto.

reflexivity.

subst ld' ld2. inversion ld'_eq. subst.
auto.

subst ld' ld2. inversion ld'_eq. subst.
eapply hl_sync_step_nested_sync_complete. auto.
Unfocus.

(* SyncComplete *)
Focus.
unfold sync_sim. auto.
Unfocus.

Qed.
