Require Import Coq.Lists.List.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Sets.Ensembles.
Require Import highlevel.
Require Import syntax.

Lemma hl_sync_step_preserves_unique_tids :
  forall H S s H' S' s',
    hl_sync_step H S s H' S' s' ->
    ld_unique_tids S s ->
    ld_unique_tids S' s'.

intros H S s H' S' s' steps unique_before.
induction s as [ t | ]; 
destruct unique_before as [ S_unique_tids t_is_unique ];
induction s' as [ t' | ];
inversion steps; subst; unfold ld_unique_tids; split; auto.

unfold hl_unique_tids.
intros t1 t2 id3 id4 list_M3 list_M4 e3 e4 t1_is_id3 t2_is_id4
  t1_in_S' t2_in_S' t1_not_t2.
unfold hl_unique_tids in S_unique_tids.
destruct t1_in_S' as [ t1 t1_not_comm | t1 t1_is_sender_or_receiver ];
destruct t2_in_S' as [ t2 t2_not_comm | t2 t2_is_sender_or_receiver ].

eapply S_unique_tids. apply t1_is_id3. apply t2_is_id4.
destruct t1_not_comm. auto.
destruct t2_not_comm. auto. auto.

destruct t2_is_sender_or_receiver.

inversion t2_is_id4. subst id4 list_M4 e4.
destruct t1_not_comm.
eapply S_unique_tids.
apply t1_is_id3. reflexivity. auto. eauto.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. left.

inversion t2_is_id4. subst id4 list_M4 e4.
destruct t1_not_comm.
eapply S_unique_tids.
apply t1_is_id3. reflexivity. auto. eauto.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. right.

destruct t1_is_sender_or_receiver.
destruct t2_not_comm.
inversion t1_is_id3. subst id3 list_M3 e3.
eapply S_unique_tids.
reflexivity. apply t2_is_id4. eauto. auto.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. left.

inversion t1_is_id3. subst id3.
destruct t2_not_comm.
eapply S_unique_tids.
reflexivity. apply t2_is_id4. eauto. auto.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. right.

unfold not. intros id3_is_id4. subst id4.
destruct t1_is_sender_or_receiver; 
destruct t2_is_sender_or_receiver;
inversion t1_is_id3; inversion t2_is_id4; subst;
eauto.

assert ((id3, list_M3, e1) = (id3, list_M4, e2)) as t1_is_t2.
apply NNPP. unfold not at 1. intros e1_not_e2.
assert (id3 <> id3) as false.
eapply S_unique_tids. reflexivity. reflexivity.
apply H2. apply H3. apply e1_not_e2. tauto.
inversion t1_is_t2. subst list_M4 e2.
assert (E1 = E2 /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique.
eauto. auto.
destruct H. inversion H0.

assert ((id3, list_M4, e1) = (id3, list_M3, e2)) as t1_is_t2.
apply NNPP. unfold not at 1. intros e1_not_e2.
assert (id3 <> id3) as false.
eapply S_unique_tids. reflexivity. reflexivity.
apply H2. apply H3. apply e1_not_e2. tauto.
inversion t1_is_t2. subst list_M4 e2.
assert (E1 = E2 /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique.
eauto. eauto.
destruct H. inversion H0.

induction t' as [ [ id list_M ] e ].
intros id' list_M' e' in_S.
assert (id' = id1 \/ id' <> id1) as classic.
apply classic. destruct classic. subst id'.
eapply t_is_unique. eauto.
assert (id' = id2 \/ id' <> id2) as classic.
apply classic. destruct classic. subst id'.
eapply t_is_unique. eauto.
eapply t_is_unique.
inversion in_S as [ t' t'_in_S | t' false ].
destruct t'_in_S. eauto.
inversion false. subst id'. tauto.
inversion false. subst id'. tauto.

subst id'. tauto.

unfold hl_unique_tids.
intros t1 t2 id1 id2 list_M1 list_M2 e1 e2 t1_is_id1 t2_is_id2 
  t1_in_S' t2_in_S' t1_not_t2.
unfold hl_unique_tids in S_unique_tids.
inversion t1_in_S'; inversion t2_in_S'; subst.
eauto.
inversion H1; subst. eauto. inversion H. subst.
assert (id2 <> id1) as swapped. eauto. auto.
inversion H. subst id1 list_M1 e1.
inversion H1. subst id2 list_M2 e2.
assert ((id, list_M, v) = (id, list_M, v)) as equality.
reflexivity.
unfold not in t1_not_t2.
assert False as false. auto. inversion false.

unfold hl_unique_tids in S_unique_tids.
unfold hl_unique_tids.
intros t1 t2 id1 id2 list_M1 list_M2 e1 e2 t1_is_id1 t2_is_id2
  t1_in_S' t2_in_S' t1_not_t2.
eapply S_unique_tids. apply t1_is_id1. apply t2_is_id2.
left. auto. left. auto. auto.

unfold hl_unique_tids in S_unique_tids.
destruct t'. destruct p. intros id' list_M' e' id'_in_S'.
eapply S_unique_tids.
reflexivity. reflexivity. left. apply id'_in_S'. right.
apply In_singleton.
unfold not. intros false. rewrite <- false in H4. unfold not in H4. apply H4.
apply id'_in_S'.

unfold hl_unique_tids.
intros t1 t2 id3 id4 list_M3 list_M4 e3 e4 t1_is_id3 t2_is_id4
  t1_in_S' t2_in_S' t1_not_t2.
unfold hl_unique_tids in S_unique_tids.
destruct t1_in_S' as [ t1 t1_not_comm | t1 t1_is_sender_or_receiver ];
destruct t2_in_S' as [ t2 t2_not_comm | t2 t2_is_sender_or_receiver ].

eapply S_unique_tids. apply t1_is_id3. apply t2_is_id4.
destruct t1_not_comm. auto.
destruct t2_not_comm. auto. auto.

destruct t2_is_sender_or_receiver.

inversion t2_is_id4. subst id4 list_M4 e4.
destruct t1_not_comm.
eapply S_unique_tids.
apply t1_is_id3. reflexivity. auto. apply H1.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. left.

inversion t2_is_id4. subst id4 list_M4 e4.
destruct t1_not_comm.
eapply S_unique_tids.
apply t1_is_id3. reflexivity. auto. apply H2.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. right.

destruct t1_is_sender_or_receiver.
destruct t2_not_comm.
inversion t1_is_id3. subst id3 list_M3 e3.
eapply S_unique_tids.
reflexivity. apply t2_is_id4. apply H1. auto.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. left.

inversion t1_is_id3. subst id3.
destruct t2_not_comm.
eapply S_unique_tids.
reflexivity. apply t2_is_id4. apply H2. auto.
unfold not in H0.
unfold not.
intros false. apply H0. rewrite <- false. right.

unfold not. intros id3_is_id4. subst id4.
destruct t1_is_sender_or_receiver; 
destruct t2_is_sender_or_receiver;
inversion t1_is_id3; inversion t2_is_id4; subst;
eauto.

assert ((id3, list_M3, e1) = (id3, list_M4, e2)) as t1_is_t2.
apply NNPP. unfold not at 1. intros e1_not_e2.
assert (id3 <> id3) as false.
eapply S_unique_tids. reflexivity. reflexivity.
apply H1. apply H2. apply e1_not_e2. tauto.
inversion t1_is_t2. subst list_M4 e2.
assert (E1 = E2 /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique.
apply H3. apply H4.
destruct H. inversion H0.

assert ((id3, list_M4, e1) = (id3, list_M3, e2)) as t1_is_t2.
apply NNPP. unfold not at 1. intros e1_not_e2.
assert (id3 <> id3) as false.
eapply S_unique_tids. reflexivity. reflexivity.
apply H1. apply H2. apply e1_not_e2. tauto.
inversion t1_is_t2. subst list_M4 e2.
assert (E1 = E2 /\ sendEvt k v = recvEvt k).
eapply get_evt_ctxt_unique.
apply H3. apply H4.
destruct H. inversion H0.

Qed.

Lemma get_S_hl_unique_tids : forall T,
  mt_unique_tids T ->
  hl_unique_tids (get_S_hl T).

intros T T_unique.
unfold hl_unique_tids.
intros t1 t2 id1 id2 list_M1 list_M2 e1 e2 t1_is_id1 t2_is_id2 t1_in_S t2_in_S t1_not_t2.
inversion t1_in_S. inversion t2_in_S. subst.
inversion H4. inversion H1. subst.
assert (id1 <> id2 \/ e1 <> e2) as id1_not_id2_or_e1_not_e2.
assert (~ (id1 = id2 /\ e1 = e2)) as not_id1_is_id2_and_e1_is_e2.
unfold not. intros [ id1_is_id2 e1_is_e2 ].
unfold not in t1_not_t2. apply t1_not_t2. subst. reflexivity.
apply not_and_or. auto.
destruct id1_not_id2_or_e1_not_e2 as [ id1_not_id2 | e1_not_e2 ].
apply id1_not_id2.
assert (e <> e0) as e_not_e0.
unfold not. intros e_is_e0. subst e0.
assert (M = M0 /\ sync e1 = sync e2) as ctxt_uniq.
eapply get_ctxt_unique.
apply H0. apply H3.
destruct ctxt_uniq as [ M_is_M0 e1_is_e2 ].
inversion e1_is_e2. tauto.
unfold mt_unique_tids in T_unique.
eapply T_unique. reflexivity. reflexivity. apply H. apply H2.
unfold not. intros t1_is_t2. inversion t1_is_t2. tauto.

Qed.

Lemma hl_sync_step_many_preserves_unique_tids :
  forall H S s H' S' s',
    hl_sync_step_many H S s H' S' s' ->
    ld_unique_tids S s ->
    ld_unique_tids S' s'.

intros H S s H' S' s' steps S_unique.
induction steps.
apply S_unique.
apply IHsteps.
eapply hl_sync_step_preserves_unique_tids.
apply H. apply S_unique.

Qed.

Lemma hl_mt_step_preserves_unique_tids :
  forall H T H' T',
    hl_mt_step H T H' T' ->
    mt_unique_tids T ->
    mt_unique_tids T'.

intros H T_all H' T_all' steps T_all_unique.

inversion steps as [ H_ T_all_ H'_ T_all'_ mt_steps eqH eqT_all eqH' eqT_all' 
  | H_ H'_ T T' S S' T_all_ T_all'_ T0 T_at_sync S_is_get_S_hl_T S_steps_to_S' 
    S'_is_complete T'_is_get_T'_hl_T_S' eqH disjoint_T_T0 eqT_all eqH' eqT_all' ];
subst H_ T_all_ H'_ T_all'_.

eapply mt_step_preserves_unique_tids.
apply mt_steps. apply T_all_unique.

assert (mt_unique_tids T) as T_unique.
unfold mt_unique_tids.
intros t1 t2 id1 id2 e1 e2 t1_is_id1 t2_is_id2 
  t1_in_T t2_in_T t1_not_t2.
unfold mt_unique_tids in T_all_unique. subst T_all.
eapply T_all_unique.
apply t1_is_id1. apply t2_is_id2. left. auto. left. auto. auto.

assert (mt_unique_tids T') as T'_unique.

unfold mt_unique_tids.
intros t1' t2' id1 id2 e1' e2' t1'_is_id1 t2'_is_id2 
  t1'_in_T' t2'_in_T' t1'_not_t2'.
subst t1' t2' S T'.
inversion t1'_in_T' as
  [ id1_ e1 M1 v1 v1' t1_in_T get_ctxt_e1_is_M1_sync_v1 s1'_in_S' eqid1 ].
inversion t2'_in_T' as 
  [ id2_ e2 M2 v2 v2' t2_in_T get_ctxt_e2_is_M2_sync_v2 s2'_in_S' eqid2 ].
subst id1_ e1' id2_ e2'.

unfold not in t1'_not_t2'.
assert (~ (id1 = id2 /\ M1 [v1'] = M2 [v2'])) as not_both_equal.
unfold not. intros both_equal. 
destruct both_equal as [ id1_is_id2 e1'_is_e2' ]. apply t1'_not_t2'.
subst id2. rewrite -> e1'_is_e2'. reflexivity.
assert (~ (id1 = id2) \/ ~ (M1 [v1'] = M2 [v2'])) as ids_or_e's_not_equal.
apply not_and_or. apply not_both_equal.
destruct ids_or_e's_not_equal as [ id1_not_id2 | e1'_not_e2' ]. 
apply id1_not_id2.
assert (~ (M1 = M2 /\ v1' = v2')) as not_both_equal_e'.
unfold not. intros both_equal_e'. destruct both_equal_e' as [M1_is_M2 v1'_is_v2'].
unfold not in e1'_not_e2'. apply e1'_not_e2'. subst M2 v2'. reflexivity.
assert (M1 <> M2 \/ v1' <> v2') as Ms_or_v's_not_equal.
apply not_and_or. apply not_both_equal_e'.
destruct Ms_or_v's_not_equal as [M1_not_M2 | v1'_not_v2'].
assert (e1 = e2 -> M1 = M2 /\ sync v1 = sync v2) as to_be_reversed.
intros e1_is_e2. eapply get_ctxt_unique.
apply get_ctxt_e1_is_M1_sync_v1. subst e2. apply get_ctxt_e2_is_M2_sync_v2.
assert (e1 <> e2) as e1_not_e2.
unfold not. intros e1_is_e2.
assert (M1 = M2 /\ sync v1 = sync v2) as false. auto.
destruct false as [false1 false2]. tauto.
unfold mt_unique_tids in T_unique.
eapply T_unique. reflexivity. reflexivity. apply t1_in_T. apply t2_in_T.
unfold not. intros t1_is_t2. unfold not in e1_not_e2. 
apply e1_not_e2. inversion t1_is_t2. reflexivity.
assert (hl_unique_tids S') as S'_unique.
assert (ld_unique_tids S' None) as S'_ld_unique.
eapply hl_sync_step_many_preserves_unique_tids.
apply S_steps_to_S'.
unfold ld_unique_tids. split.
apply get_S_hl_unique_tids. apply T_unique. auto.
unfold ld_unique_tids in S'_ld_unique.
destruct S'_ld_unique. auto.
unfold hl_unique_tids in S'_unique.
eapply S'_unique.
reflexivity. reflexivity. apply s1'_in_S'. apply s2'_in_S'.
unfold not. intros s1'_is_s2'.
inversion s1'_is_s2'. tauto.

assert (mt_unique_tids T0) as T0_unique.
unfold mt_unique_tids.
intros t1 t2 id1 id2 e1 e2 t1_is_id1 t2_is_id2 
  t1_in_T0 t2_in_T0 t1_not_t2.
unfold mt_unique_tids in T_all_unique. subst T_all.
eapply T_all_unique.
apply t1_is_id1. apply t2_is_id2. right. auto. right. auto. auto.

unfold mt_unique_tids.
intros t1' t2' id1 id2 e1' e2' t1'_is_id1 t2'_is_id2
  t1'_in_T_all' t2'_in_T_all' t1'_not_t2'.
subst T_all'.
destruct t1'_in_T_all' as [ t1' t1'_in_T' | t1' t1'_in_T0 ];
destruct t2'_in_T_all' as [ t2' t2'_in_T' | t2' t2'_in_T0 ];
subst t1' t2'.
unfold mt_unique_tids in T'_unique. eapply T'_unique.
reflexivity. reflexivity. apply t1'_in_T'. apply t2'_in_T'. apply t1'_not_t2'.
subst T'. inversion t1'_in_T'. subst.
unfold mt_unique_tids in T_all_unique.
eapply T_all_unique. reflexivity. reflexivity. left. apply H2. right. apply t2'_in_T0.
unfold not. intros t1'_is_t2'.
inversion disjoint_T_T0.
inversion t1'_is_t2'. subst id1 e2'.
assert (In mt_thread (Intersection mt_thread T T0) (id2, e)).
apply Intersection_intro.
auto. auto.
assert (~ In mt_thread (Intersection mt_thread T T0) (id2, e)). auto. tauto.

subst T'. inversion t2'_in_T'. subst.
unfold mt_unique_tids in T_all_unique.
eapply T_all_unique. reflexivity. reflexivity. right. apply t1'_in_T0. left. apply H2.
unfold not. intros t1'_is_t2'.
inversion disjoint_T_T0.
inversion t1'_is_t2'. subst id1 e1'.
assert (In mt_thread (Intersection mt_thread T T0) (id2, e)).
apply Intersection_intro.
auto. auto.
assert (~ In mt_thread (Intersection mt_thread T T0) (id2, e)). auto. tauto.

unfold mt_unique_tids in T0_unique.
eapply T0_unique. reflexivity. reflexivity. apply t1'_in_T0. apply t2'_in_T0. apply t1'_not_t2'.

Qed.

Lemma hl_unique_tids_reversed :
  forall S_hl id list_M1 e1 list_M2 e2,
  hl_unique_tids S_hl ->
  In hl_thread S_hl (id, list_M1, e1) ->
  In hl_thread S_hl (id, list_M2, e2) ->
  (id, list_M1, e1) = (id, list_M2, e2).

intros S_hl id list_M1 e1 list_M2 e2 S_hl_unique s1_in_S_hl s2_in_S_hl.

unfold hl_unique_tids in S_hl_unique.
unfold not at 2 in S_hl_unique.
apply NNPP. unfold not at 1. intros neq.
eapply S_hl_unique. reflexivity.  reflexivity. 
apply s1_in_S_hl. apply s2_in_S_hl. auto. reflexivity.

Qed.

Lemma ld_unique_tids_implies_hl :
  forall S_hl,
    ld_unique_tids S_hl None ->
    hl_unique_tids S_hl.

intros S_hl ld_unique.
unfold ld_unique_tids in ld_unique.
destruct ld_unique. auto.

Qed.

Lemma mt_unique_tids_reversed :
  forall T id e1 e2,
  mt_unique_tids T ->
  In mt_thread T (id, e1) ->
  In mt_thread T (id, e2) ->
  (id, e1) = (id, e2).

intros T id e1 e2 unique e1_in_T e2_in_T.
unfold mt_unique_tids in unique.
apply NNPP. unfold not at 1. intros neq.
unfold not at 2 in unique.
eapply unique. reflexivity. reflexivity.
apply e1_in_T. apply e2_in_T. auto. reflexivity.

Qed.